//___________________________________________________________________
//
//									Class AVIGen
//___________________________________________________________________

class AVIGen  
{
public:
	wxString filename;	// name of output file
	int rate;	// Frame rate 
	BITMAPINFOHEADER bih;	// structure contains information for a single stream
	int frame;	// frame counter
	wxString aviError;	// last error string

	AVIGen();
	AVIGen(charP filename_, wxWindow *window, int rate_);
	AVIGen(charP filename_, LPBITMAPINFOHEADER lpbih, int rate_);

	~AVIGen();

	int InitEngine(bool newFile=true, int frame_=0);

	// The data pointed by bmBits has to be compatible 
	// with the bitmap description of the movie.
	int AddFrame(charP pixels);

	void ReleaseEngine();

	void SetBitmapHeader(wxWindow *window);
	void SetBitmapHeader(LPBITMAPINFOHEADER lpbih);

	LPBITMAPINFOHEADER GetBitmapHeader()							
	{	return &bih;
	}

	void SetFileName(wxString filename_)
	{	filename=filename_; 
		MakeExtAvi();
	}

	// Sets FrameRate (should equal or greater than one)
	void SetRate(int rate_)										
	{	rate=rate_;
	}

	wxString GetLastErrorMessage() const								
	{	return aviError;
	}

private:
	void MakeExtAvi();

	PAVIFILE pAVIFile;		// file interface pointer
	PAVISTREAM pStream;	// Address of the stream interface
	PAVISTREAM pStreamCompressed;	// Address of the compressed video stream
};

/* A simple class to create AVI video stream
============================================
Usage
-----
	Step 1 : Declare an AVIGenerator object
	Step 2 : Set Bitmap by calling SetBitmapHeader functions + other parameters
	Step 3 : Initialize engine by calling InitEngine
	Step 4 : Send each frames to engine with function AddFrame
	Step 5 : Close engine by calling ReleaseEngine

Demo Code
---------
		AVIGen aviGen;
		BYTE* bmBits;

		// set characteristics
		aviGen.SetRate(20);								// set 20fps
		aviGen.SetBitmapHeader(GetActiveView());	// give info about bitmap

		aviGen.InitEngine();
		..... // Draw code, bmBits is the buffer containing the frame
		aviGen.AddFrame(bmBits);
		.....
		aviGen.ReleaseEngine();

Update history
--------------
	22-10-2002 Minor changes in constructors.

	author : Jonathan de Halleux, dehalleux@auto.ucl.ac.be (2001)
*/
