//___________________________________________________________________
//
//                          Class  CheckBox
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/checkbox.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>

#include "Main.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "Color.h"
#include "Float.h"
#include "Point.h"
#include "Error.h"
#include "Eval.h"
#include "Stack.h"

#include "Event.h"
#include "CheckBox.h"
#include "Frame.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"
#include "Console.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(CheckBox, wxCheckBox)
	EVT_CHECKBOX	(-1, CheckBox::OnClick	)
END_EVENT_TABLE()

CheckBox::CheckBox(CheckBoxNodeP checkBoxNode_):
	checkBoxNode(checkBoxNode_)
{
	wxWindow *parent=console;
	if (checkBoxNode->parentFrame)
		parent=checkBoxNode->parentFrame->frame;
	else if (checkBoxNode->parentGraph)
		parent=checkBoxNode->parentGraph->graph;
	Create(parent, -1, 
			wxEmptyString, 
			wxDefaultPosition, 
			wxDefaultSize, 
			0, 
			wxDefaultValidator, 
			"CheckBox");
}

CheckBox::CheckBox(CheckBoxNodeP checkBoxNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxString &label,
			PairP onClick_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
wxCheckBox(parent, id, label, 
			pos, size, style, 
			validator, name),
onClick(onClick_),
checkBoxNode(checkBoxNode_)
{
	checkBoxNode->checkBox=this;
}

CheckBox::~CheckBox()
{
	if (checkBoxNode)
		checkBoxNode->checkBox=0;
}

void CheckBox::OnClick(wxCommandEvent &event)
{
	Event::fifo << new EventCheckBox(this);
}

wxString CheckBox::wxstring()
{
	wxString str;
	str << "CheckBox [" << GetLabel() << "] " 
		<< '[' << GetName() << "] ";
	if (onClick)
		str << onClick->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " 
		<< GetValue();
	return str;
}

wxDataInputStream &CheckBox::read(wxDataInputStream &istr) 
{	
	wxString label, name;
	wxPoint pos;
	wxSize size;
	int style, state;
	
	istr >> label >> name >> (int&)onClick
		>> pos.x >> pos.y >> size.x >> size.y >> style >> state;

	SetLabel(label);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetWindowStyle(style);
	SetValue(state!=0);
	return istr;
}
wxDataOutputStream &CheckBox::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetLabel() << GetName() 
		<< GC_::ptrToInt->find(onClick)
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y
		<< int(GetWindowStyle()) << int(GetValue());
	return ostr;
}

PairP CheckBoxNode::onClick()
{
	if (checkBox)
		return checkBox->onClick;
	return 0;
}

void CheckBoxNode::setonClick(NodeP nd) 
{
	if (checkBox)
	{	checkBox->onClick = PairP(nd);  check_oldyoung(nd);
	}
}

void CheckBoxNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void CheckBoxNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

CheckBoxNode::CheckBoxNode(): 
	Node(CHECKBOX), checkBox(0), 
	parentFrame(0),
	parentGraph(0)
{}

CheckBoxNode::CheckBoxNode(FrameNodeP parentFrame_,
									GraphNodeP parentGraph_): 
	Node(CHECKBOX), checkBox(0), 
	parentFrame(parentFrame_),
	parentGraph(parentGraph_)
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void CheckBoxNode::free() 
{	
	if (checkBox) 
	{	checkBox->checkBoxNode=0;
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		checkBox->Destroy(); 
		EventPP out=Event::fifo.out;
		EventP ev;
		while (Event::fifo)
		{	Event::fifo >> ev;
			switch (ev->type)
			{
			case Event::tCheckBox:
				if (((EventCheckBox*)ev)->checkBox==checkBox)
					((EventCheckBox*)ev)->checkBox=0;
				break;
			default:
				;
			}
		}
		Event::fifo.out=out;
		checkBox=0;
	}
}
void CheckBoxNode::inc_gc() {fifo << onClick();}
int CheckBoxNode::youngs() {return gmark(onClick());}
int CheckBoxNode::youngsPtr() {return gmarkPtr(onClick());}
void CheckBoxNode::check_oldyoungs() {check_oldyoung(onClick());}

wxDataInputStream &CheckBoxNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	FrameNodeP parentFrameNode;
		GraphNodeP parentGraphNode;
		istr >> (int&)parentFrameNode >> (int&)parentGraphNode;
		checkBox=new CheckBox(CheckBoxNodeP(this));
		checkBox->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}

wxDataOutputStream &CheckBoxNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (checkBox)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame)
			<< GC_::ptrToInt->find(parentGraph);
		checkBox->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void CheckBoxNode::setPtrs()
{
	parentFrame=FrameNodeP((FrameNode*)GC_::intToPtr->find(int(parentFrame)));
	parentGraph=GraphNodeP((GraphNode*)GC_::intToPtr->find(int(parentGraph)));
	if (checkBox)
	{	setonClick(PairP((Pair*)GC_::intToPtr->find(int(checkBox->onClick))));
	}
}

void CheckBoxNode::recreate()
{
	wxWindow *p=0;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	CheckBoxP c=checkBox, 
		c2=new CheckBox(CheckBoxNodeP(this),
				p, -1, c->GetLabel(),
				c->onClick, c->GetPosition(), c->GetSize(), c->GetWindowStyle(), 
				wxDefaultValidator, c->GetName());
	c2->SetValue(c->GetValue());
	checkBox=c2;
	setonClick(c->onClick);
	delete c;
	checkBox=c2;
}

StringP CheckBoxNode::string()
{
	if (checkBox)
		return new String(checkBox->wxstring());
	return _Null_Word->name();
}

wxString CheckBoxNode::wxstring()
{
	if (checkBox)
		return checkBox->wxstring();
	return wxEmptyString;
}

NodeP lCheckBox(PairPP arg)
{
	wxWindow *parent=graph;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=theApp->mainFrame->graphNode;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}
	wxString label=wxEmptyString;
	{	StringP l=arg[2]->a_string(2);
		if throwing return Unbound;
		label=l->ptr();
	}
	PairP onClick=0;
	if (exist(3))
	{	PairP oc=arg[3]->a_runnable(3);
		if throwing return Unbound;
		onClick=oc;
	}
	long style = 0;
	if (exist(4))
	{	int s=arg[4]->a_int(4);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(5))
	{	PairP p=arg[5]->a_int_vector(5);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(6))
	{	PairP s=arg[6]->a_int_vector(6);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	wxWindowID id=-1;
	if (exist(7))
	{	int i=arg[7]->a_int(7);
		if throwing return Unbound;
		id=i;	
	}
	wxString name=_T("CheckBox");
	if (exist(8))
	{	StringP n=arg[8]->a_string(8);
		if throwing return Unbound;
		name=n->ptr();
	}
	CheckBoxNodeP checkBoxNode=new CheckBoxNode(
		parentFrameNode,

		parentGraphNode);
	new CheckBox(checkBoxNode,
			parent, id, label, 
			onClick, pos, size, style, wxDefaultValidator, name);
	return checkBoxNode;
}

NodeP lCheckBoxOnClick(PairPP arg)
{
	CheckBoxNodeP checkBox=arg[1]->a_checkBox(1);
	if throwing return Unbound;
	PairP function=arg[2]->a_runnable(2);
	if throwing return Unbound;
	checkBox->setonClick(function);
	return Unbound;
}

NodeP lCheckBoxDestroy(PairPP arg)
{
	CheckBoxNodeP checkBox=arg[1]->a_checkBox(1);
	if throwing return Unbound;
	{	FrameNodeP p=checkBox->parentFrame;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(checkBox));
			p->check_oldyoung(p->children);
		}
	}
	{	GraphNodeP p=checkBox->parentGraph;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(checkBox));
			p->check_oldyoung(p->children);
		}
	}
	checkBox->checkBox->Destroy();
	return Unbound;
}

NodeP lCheckBoxSet(PairPP arg)
{
	CheckBoxNodeP checkBox=arg[1]->a_checkBox(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	checkBox->checkBox->SetValue(state!=0);
	return Unbound;
}

NodeP lCheckBoxValue(PairPP arg)
{
	CheckBoxNodeP checkBox=arg[1]->a_checkBox(1);
	if throwing return Unbound;
	bool state=checkBox->checkBox->GetValue();
	return Node::boolnode(state);
}

NodeP lCheckBoxEnable(PairPP arg)
{
	CheckBoxNodeP checkBox=arg[1]->a_checkBox(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	checkBox->checkBox->Enable(state!=0);
	return Unbound;
}
