//___________________________________________________________________
//
//                           Class ComboBox
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/textctrl.h>
#include <wx/stattext.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/listbox.h>
#include <wx/combobox.h>

#include "Main.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "Float.h"
#include "Point.h"
#include "Error.h"
#include "Eval.h"
#include "Stack.h"

#include "Event.h"
#include "ComboBox.h"
#include "Frame.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"
#include "Color.h"
#include "Graph.h"
#include "Console.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

//___________________________________________________________________

class MyComboBox : public wxComboBox
{
public:
	MyComboBox(wxWindow* parent, wxWindowID id, 
		const wxString &value,
		const wxArrayString& choices, 
		const wxPoint& pos=wxDefaultPosition, 
		const wxSize& size=wxDefaultSize, 
		long style = 0, 
		const wxValidator& validator = wxDefaultValidator, 
		const wxString& name = "ComboBox");

	// MSW sends extra kill focus event
	virtual ~MyComboBox()
	{
		if (parent) 
			parent->comboBox = 0;
		parent=0;
	}

	ComboBox *parent;

	void OnChar   (wxKeyEvent &event);	// pass chars to ComboBox
	void OnKeyDown(wxKeyEvent &event);
	void OnKeyUp  (wxKeyEvent &event);
	void OnSelect(wxCommandEvent &event);
	void OnChange(wxCommandEvent &event);
	void OnEnter (wxCommandEvent &event);
	void Set(wxArrayString &choices);

private:
	DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(MyComboBox,wxComboBox)
	EVT_CHAR		(MyComboBox::OnChar		)
	EVT_KEY_DOWN(MyComboBox::OnKeyDown	)
	EVT_KEY_UP	(MyComboBox::OnKeyUp		)
	EVT_COMBOBOX(-1,	MyComboBox::OnSelect	)
	EVT_TEXT		(-1,	MyComboBox::OnChange	)
	EVT_TEXT_ENTER(-1,	MyComboBox::OnEnter	)
END_EVENT_TABLE()

MyComboBox::MyComboBox(wxWindow *parent_, wxWindowID id,
							  const wxString &value,
							  const wxArrayString &choices_,
							  const wxPoint &pos, const wxSize &size,
							  long style,
							  const wxValidator& validator,
							  const wxString &name):
wxComboBox(parent_, id, value, pos, size, choices_, style, validator, name)
{
	parent = (ComboBox*)parent_;
}

void MyComboBox::OnChar(wxKeyEvent &event)
{
	if (parent) 
		parent->OnChar(event);
	event.Skip();
}

void MyComboBox::OnKeyDown(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyDown(event);
	event.Skip();
}

void MyComboBox::OnKeyUp(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyUp(event);
	event.Skip();
}

void MyComboBox::OnSelect(wxCommandEvent &event)
{
	if (parent) 
		parent->OnSelect(event);
}

void MyComboBox::OnChange(wxCommandEvent &event)
{
	if (parent) 
		parent->OnChange(event);
}

void MyComboBox::OnEnter(wxCommandEvent &event)
{
	if (parent) 
		parent->OnEnter(event);
}

void MyComboBox::Set(wxArrayString &choices)
{
	Clear();
	int nr=choices.GetCount();
	for (int i=0; i<nr; i++)
		Append(choices[i]);
}
//___________________________________________________________________

BEGIN_EVENT_TABLE(ComboBox, wxControl)
	EVT_CHAR				(		ComboBox::OnChar		)
	EVT_KEY_DOWN		(		ComboBox::OnKeyDown	)
	EVT_KEY_UP			(		ComboBox::OnKeyUp		)
	EVT_CHOICE			(-1,	ComboBox::OnSelect	)
	EVT_TEXT				(-1,	ComboBox::OnChange	)
	EVT_TEXT_ENTER		(-1,	ComboBox::OnEnter		)
END_EVENT_TABLE()

ComboBox::ComboBox(ComboBoxNodeP comboBoxNode_, int style):
label(0), comboBox(0), bestSize(-999, -999),
choices(0),
onSelect(0), onChange(0), onEnter(0),
onChar(0), onKeyUp(0), onKeyDown(0), 
comboBoxNode(comboBoxNode_)
{
	wxWindow *parent=console;
	if (comboBoxNode->parentFrame)
		parent=comboBoxNode->parentFrame->frame;
	else if (comboBoxNode->parentGraph)
		parent=comboBoxNode->parentGraph->graph;
	create(parent, -1, wxEmptyString, wxArrayString(), 
			wxDefaultPosition, wxDefaultSize, style,
			wxDefaultValidator, "ComboBox");
}

ComboBox::ComboBox(ComboBoxNodeP comboBoxNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxString &value_,
			const wxArrayString &choices_, 
			PairP onSelect_,
			PairP onChange_,
			PairP onEnter_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
label(0), comboBox(0), bestSize(-999, -999), 
value(value_), choices(choices_),
onSelect(onSelect_), onChange(onChange_), onEnter(onEnter_),
onChar(0), onKeyUp(0), onKeyDown(0), 
comboBoxNode(comboBoxNode_)
{
	bestSize = wxSize(-999, -999);
	comboBoxNode->comboBox=this;
	comboBoxNode->style=style;
	create(parent, id, value, choices, pos, size, style, validator, name);
}

bool ComboBox::create(wxWindow* parent, wxWindowID id, 
			const wxString &value_,
			const wxArrayString &choices_, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name)
{
	if (!wxControl::Create(parent, id, pos, size, style|wxNO_BORDER,
								wxDefaultValidator, name))
	  return false;

	wxControl::SetLabel(name);
	wxControl::SetBackgroundColour(parent->GetBackgroundColour());
	wxControl::SetForegroundColour(parent->GetForegroundColour());

	int width = size.GetWidth(), height = size.GetHeight();

	wxSize best_size(DoGetBestSize());
	if (width  == -1) width  = best_size.GetWidth();
	if (height == -1) height = best_size.GetHeight();

	if (name != wxEmptyString)
		label = new wxStaticText(this, id, name, 
			wxPoint(0, 0), wxDefaultSize, 0, "staticText");
	else
		label=0;

	if (label)
		comboBox = new MyComboBox(this, id, value_, choices_,
			wxPoint(0,0),
			wxSize(width, height-label->GetSize().GetHeight()),
			style, validator, name);
	else
		comboBox = new MyComboBox(this, id, value_, choices_,
			wxPoint(0,0),
			wxSize(width, height), style, validator, name);

	value=wxEmptyString;
	choices.Clear();

	DoSetSize(pos.x, pos.y, width, height);
	SetBestFittingSize(wxSize(width, height));
	return true;
}

void ComboBox::DoSetSize(int x, int y, int width, int height, int sizeFlags)
{
	wxSize bestSize(DoGetBestSize());
	if (width < 0)  width  = bestSize.GetWidth();
	if (height < 0) height = bestSize.GetHeight();

	wxWindow::DoSetSize(x, y, width, height, sizeFlags);

	int labelwidth  = 0;
	int labelheight = 0;
	if (label)
		label->GetSize(&labelwidth, &labelheight);

	if (comboBox)   
		comboBox->SetSize(0, labelheight, 
			width, height-labelheight-2);
}

wxSize ComboBox::DoGetBestSize()
{
	if (bestSize.x == -999)
	{	wxComboBox *t=new wxComboBox((wxWindow*)this, wxID_ANY, 
			wxEmptyString, wxPoint(0,0), wxDefaultSize, 
			choices, GetWindowStyle());
		bestSize=t->GetBestSize();
		delete t;
		if (GetLabel()!=wxEmptyString)
		{	wxStaticText label((wxWindow*)this, wxID_ANY, GetLabel());
			wxSize ls=label.GetBestSize();
			bestSize.y+=ls.y;
			if (ls.x > bestSize.x)
				bestSize.x=ls.x;
		}
	}
	return bestSize;
}

ComboBox::~ComboBox()
{
	if (comboBoxNode)
		comboBoxNode->comboBox=0;
	if (label)
	{	delete label;
		label=0;
	}
	if (comboBox)
	{	delete comboBox;
		comboBox=0;
	}

	EventPP out=Event::fifo.out;
	EventP ev;
	while (Event::fifo)
	{	Event::fifo >> ev;
		switch (ev->type)
		{
		case Event::ComboBoxSelect:
		case Event::ComboBoxChange:
		case Event::ComboBoxEnter:
		case Event::ComboBoxChar:
		case Event::ComboBoxKeyDown:
		case Event::ComboBoxKeyUp:
			if (((EventComboBox*)ev)->comboBox==this)
				((EventComboBox*)ev)->comboBox=0;
			break;
		default:
			;
		}
	}
	Event::fifo.out=out;
}

void ComboBox::Set(wxArrayString &choices)
{
	if (comboBox)
		comboBox->Set(choices);
}

void ComboBox::SetValue(const wxString &value)
{
	if (comboBox)
		comboBox->SetValue(value);
}

wxString ComboBox::GetValue() const
{
	if (comboBox)
		return comboBox->GetValue();
	return wxEmptyString;
}

#define LB_ON(NAME)																\
void ComboBox::On##NAME(wxKeyEvent &event)									\
{																						\
	Event::fifo << new EventComboBox##NAME(this, event.GetKeyCode());	\
	event.Skip();																	\
}

LB_ON(Char)
LB_ON(KeyDown)
LB_ON(KeyUp)

void ComboBox::OnSelect(wxCommandEvent &event)
{
	Event::fifo << new EventComboBoxSelect(this);
}

void ComboBox::OnChange(wxCommandEvent &event)
{
	Event::fifo << new EventComboBoxChange(this);
}

void ComboBox::OnEnter(wxCommandEvent &event)
{
	Event::fifo << new EventComboBoxEnter(this);
}

wxString ComboBox::wxstring()
{
	wxString str;
	str << "ComboBox " 
		<< '[' << GetName() << "] ";
	if (onSelect)
		str << onSelect->string()->ptr() << ' ';
	else
		str << "[] ";
	if (onChange)
		str << onChange->string()->ptr() << ' ';
	else
		str << "[] ";
	if (onEnter)
		str << onEnter->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

wxDataInputStream &ComboBox::read(wxDataInputStream &istr) 
{	
	wxPoint pos;
	wxSize size;
	wxString name, value, fontName;
	int fontSize, fontStyle, fontWeight;
	unsigned char fr,fg,fb, br,bg,bb;
	
	istr >> name >> value
		>> (int&)onSelect 
		>> (int&)onChange 
		>> (int&)onEnter 
		>> (int&)onChar
		>> (int&)onKeyDown
		>> (int&)onKeyUp
		>> pos.x >> pos.y >> size.x >> size.y
		>> fr >> fg >> fb 
		>> br >> bg >> bb 
		>> fontSize >> fontName >> fontStyle >> fontWeight;

	int nr;
	istr >> nr;
	choices.Clear();
	for (int i=0; i<nr; i++)
	{	wxString item;
		istr >> item;
		choices.Add(item);
	}
	comboBox->Set(choices);

	istr >> nr;
	comboBox->SetSelection(nr);
	SetName(name);
	SetValue(value);
	SetPosition(pos);
	SetSize(size);
	SetForegroundColour(wxColour(fr,fg,fb));
	SetBackgroundColour(wxColour(br,bg,bb));
	wxFont f=GetFont();
	f.SetPointSize(fontSize);
	f.SetFaceName(fontName);
	f.SetStyle(fontStyle);
	f.SetWeight(fontWeight);
	SetFont(f);
	return istr;
}
wxDataOutputStream &ComboBox::write(wxDataOutputStream &ostr) 
{
//	flush();
	ostr
		<< GetName() 
		<< GetValue()
		<< GC_::ptrToInt->find(onSelect)
		<< GC_::ptrToInt->find(onChange)
		<< GC_::ptrToInt->find(onEnter )
		<< GC_::ptrToInt->find(onChar)
		<< GC_::ptrToInt->find(onKeyDown)
		<< GC_::ptrToInt->find(onKeyUp)
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y;

	{	wxColour 
			f=comboBox->GetForegroundColour(), 
			b=comboBox->GetBackgroundColour();
		ostr
			<< f.Red() << f.Green() << f.Blue() 
			<< b.Red() << b.Green() << b.Blue();
	}
	wxFont f=comboBox->GetFont();
	ostr 
		<< f.GetPointSize() << f.GetFaceName() 
		<< f.GetStyle() << f.GetWeight();

	int nr=comboBox->GetCount();
	ostr << nr;
	for (int i=0; i<nr; i++)
		ostr << comboBox->GetString(i);

	wxArrayInt sel;
	nr=comboBox->GetSelection();
	ostr << nr;
	return ostr;
}

bool ComboBox::SetBackgroundColour(const wxColour &c)
{
	if (comboBox)
		return comboBox->SetBackgroundColour(c);
	return false;
}

bool ComboBox::SetForegroundColour(const wxColour &c)
{
	if (comboBox)
		return comboBox->SetForegroundColour(c);
	return false;
}

bool ComboBox::SetFont(const wxFont &f)
{
	if (comboBox)
		return comboBox->SetFont(f);
	return false;
}

//___________________________________________________________________

#define NODE_ON(x)							\
PairP ComboBoxNode::on##x()					\
{													\
	if (comboBox)								\
		return comboBox->on##x;				\
	return 0;									\
}													\
													\
void ComboBoxNode::seton##x(NodeP nd)	\
{													\
	if (comboBox)								\
	{	comboBox->on##x = PairP(nd);		\
		check_oldyoung(nd);					\
	}												\
}

NODE_ON(Char   )
NODE_ON(KeyDown)
NODE_ON(KeyUp  )
NODE_ON(Select )
NODE_ON(Change )
NODE_ON(Enter  )

void ComboBoxNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void ComboBoxNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

ComboBoxNode::ComboBoxNode(): 
	Node(COMBOBOX), comboBox(0), 
	parentFrame(0), 
	parentGraph(0) 
{}

ComboBoxNode::ComboBoxNode(FrameNodeP parentFrame_, 
											GraphNodeP parentGraph_): 
	Node(COMBOBOX), comboBox(0), 
	parentFrame(parentFrame_),
	parentGraph(parentGraph_) 
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void ComboBoxNode::free() 
{	
	if (comboBox) 
	{	comboBox->comboBoxNode=0;
		comboBox->Destroy(); 
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		EventPP out=Event::fifo.out;
		EventP ev;
		while (Event::fifo)
		{	Event::fifo >> ev;
			switch (ev->type)
			{
			case Event::ComboBoxSelect:
			case Event::ComboBoxChange:
			case Event::ComboBoxEnter:
			case Event::ComboBoxChar:
			case Event::ComboBoxKeyDown:
			case Event::ComboBoxKeyUp:
				if (((EventComboBox*)ev)->comboBox==comboBox)
					((EventComboBox*)ev)->comboBox=0;
				break;
			default:
				;
			}
		}
		Event::fifo.out=out;
		comboBox=0;
	}
}
void ComboBoxNode::inc_gc() 
{	fifo << onSelect() << onChange() << onEnter()
		 << onChar() << onKeyDown() << onKeyUp();
}
int ComboBoxNode::youngs() 
{	return gmark(onSelect())+gmark(onChange())+gmark(onEnter())
		+gmark(onChar())+gmark(onKeyDown())+gmark(onKeyUp());
}
int ComboBoxNode::youngsPtr() 
{	return gmarkPtr(onSelect())+gmarkPtr(onChange())+gmarkPtr(onEnter())
		+gmarkPtr(onChar())+gmarkPtr(onKeyDown())+gmarkPtr(onKeyUp());
}
void ComboBoxNode::check_oldyoungs() 
{	check_oldyoung(onSelect());
	check_oldyoung(onChange());
	check_oldyoung(onEnter());
	check_oldyoung(onChar());
	check_oldyoung(onKeyDown());
	check_oldyoung(onKeyUp());
}

wxDataInputStream &ComboBoxNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	FrameNodeP parentFrameNode;
		GraphNodeP parentGraphNode;
		istr >> (int&)parentFrameNode >> (int&)parentGraphNode >> style;
		comboBox=new ComboBox(ComboBoxNodeP(this), style);
		comboBox->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}
wxDataOutputStream &ComboBoxNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (comboBox)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame) 
			<< GC_::ptrToInt->find(parentGraph)
			<< style;
		comboBox->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void ComboBoxNode::setPtrs()
{
	parentFrame=(FrameNode*)GC_::intToPtr->find(int(parentFrame));
	parentGraph=(GraphNode*)GC_::intToPtr->find(int(parentGraph));
	if (comboBox)
	{	ComboBoxP t=comboBox;
		setonSelect ((Pair*)GC_::intToPtr->find(int(t->onSelect )));
		setonChange ((Pair*)GC_::intToPtr->find(int(t->onChange )));
		setonEnter  ((Pair*)GC_::intToPtr->find(int(t->onEnter  )));
		setonChar   ((Pair*)GC_::intToPtr->find(int(t->onChar   )));
		setonKeyDown((Pair*)GC_::intToPtr->find(int(t->onKeyDown)));
		setonKeyUp  ((Pair*)GC_::intToPtr->find(int(t->onKeyUp  )));
	}
}

void ComboBoxNode::recreate()
{
	wxWindow *p=0;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	ComboBoxP t=comboBox, t2=0;
	wxPoint pos=t->GetPosition();
	t2=new ComboBox(ComboBoxNodeP(this),
			p, -1, t->value, t->choices, 
			t->onSelect, t->onChange, t->onEnter, 
			pos, t->GetSize(), 
			style, wxDefaultValidator, t->GetName());
	t2->SetValue(t->GetValue());
	comboBox=t2;
	wxArrayInt sel;
	int nr=t->comboBox->GetSelection();
	t2->comboBox->SetSelection(nr);
	setonSelect (t->onSelect );
	setonChange (t->onChange );
	setonEnter  (t->onEnter  );
	setonChar   (t->onChar   );
	setonKeyDown(t->onKeyDown);
	setonKeyUp  (t->onKeyUp  );
	t2->comboBox->SetForegroundColour(t->comboBox->GetForegroundColour());
	t2->comboBox->SetBackgroundColour(t->comboBox->GetBackgroundColour());
	wxFont f=t->comboBox->GetFont();
	t2->comboBox->SetFont(f);
	delete t;
	comboBox=t2;
}

StringP ComboBoxNode::string()
{
	if (comboBox)
		return new String(comboBox->wxstring());
	return _Null_Word->name();
}

wxString ComboBoxNode::wxstring()
{
	if (comboBox)
		return comboBox->wxstring();
	return wxEmptyString;
}

NodeP lComboBox(PairPP arg)
{
	wxWindowID id=-1;
	wxWindow *parent=graph;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=theApp->mainFrame->graphNode;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}
	wxString name=_T("ComboBox");
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	wxString value;
	{	StringP v=arg[3]->a_string(3);
		if throwing return Unbound;
		value=v->ptr();	
	}
	wxArrayString choices;
	{	PairP list=arg[4]->a_list(4);
		if throwing return Unbound;
		while (list)
		{	StringP s=list->first()->a_string(4); 
			if throwing return Unbound;
			choices.Add(s->ptr());
			next(list);
		}
	}
	PairP onSelect=0;
	if (exist(5))
	{	PairP os=arg[5]->a_runnable(5);
		if throwing return Unbound;
		onSelect=os;
	}
	PairP onChange=0;
	if (exist(6))
	{	PairP oc=arg[6]->a_runnable(6);
		if throwing return Unbound;
		onChange=oc;
	}
	PairP onEnter=0;
	if (exist(7))
	{	PairP oe=arg[7]->a_runnable(7);
		if throwing return Unbound;
		onEnter=oe;
	}
	long style = 0;
	if (exist(8))
	{	int s=arg[8]->a_int(8);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(9))
	{	PairP p=arg[9]->a_int_vector(9);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(10))
	{	PairP s=arg[10]->a_int_vector(10);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	ComboBoxNodeP comboBoxNode=new ComboBoxNode(
		parentFrameNode, 
		parentGraphNode);
	new ComboBox(comboBoxNode,
			parent, id, value, choices, 
			onSelect, onChange, onEnter,  
			pos, size, style, wxDefaultValidator, name);
	return comboBoxNode;
}

NodeP lComboBoxSelection(PairPP arg)
{
	ComboBox *comboBox=0;
	if (exist(1))
	{	ComboBoxNodeP comboBoxNode=arg[1]->a_comboBox(1);
		if throwing return Unbound;
		comboBox=comboBoxNode->comboBox;
	}
	else
	{	comboBox=Event::staticComboBox;
	}
	if (!comboBox) return 0;
	wxArrayInt sel;
	int nr=comboBox->comboBox->GetSelection();
	return new Int(nr);
}

NodeP lComboBoxSetSelection(PairPP arg)
{
	ComboBox *comboBox=0;
	if (arg[1])
	{	ComboBoxNodeP comboBoxNode=arg[1]->a_comboBox(1);
		if throwing return Unbound;
		comboBox=comboBoxNode->comboBox;
	}
	else
	{	comboBox=Event::staticComboBox;
	}
	if (!comboBox) return 0;
	int sel=arg[2]->a_int(2);
	if throwing return Unbound;
	int nr=comboBox->comboBox->GetCount();
	if (sel < 0 || sel >= nr)
	{	error(BAD_DATA, arg[2]);
		return Unbound;
	}
	comboBox->comboBox->SetSelection(sel);
	return Unbound;
}

NodeP lComboBoxSetChoices(PairPP arg)
{
	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);
	if throwing return Unbound;
	wxArrayString choices;
	PairP list=arg[2]->a_list(2);
	if throwing return Unbound;
	while (list)
	{	StringP s=list->first()->a_string(2);
		if throwing return Unbound;
		choices.Add(s->ptr());
		next(list);
	}
	comboBox->comboBox->comboBox->Set(choices);
	return Unbound;
}

NodeP lComboBoxAppend(PairPP arg)
{
	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);
	if throwing return Unbound;
	StringP value=arg[2]->a_string(2);
	if throwing return Unbound;
	comboBox->comboBox->comboBox->Append(value->ptr());
	return Unbound;
}

NodeP lComboBoxSetItem(PairPP arg)
{
	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);
	if throwing return Unbound;
	int n=arg[2]->a_int(2);
	if throwing return Unbound;
	if (n < 0 || n >= comboBox->comboBox->comboBox->GetCount())
	{	error(BAD_DATA, arg[2]);
		return Unbound;
	}
	StringP str=arg[3]->a_string(3);
	if throwing return Unbound;
	comboBox->comboBox->comboBox->SetString(n, str->ptr());
	return Unbound;
}

NodeP lComboBoxRemoveItem(PairPP arg)
{
	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);
	if throwing return Unbound;
	int n=arg[2]->a_int(2);
	if throwing return Unbound;
	if (n < 0 || n >= comboBox->comboBox->comboBox->GetCount())
	{	error(BAD_DATA, arg[2]);
		return Unbound;
	}
	comboBox->comboBox->comboBox->Delete(n);
	return Unbound;
}

NodeP lComboBoxCount(PairPP arg)
{
	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);
	if throwing return Unbound;
	int count=comboBox->comboBox->comboBox->GetCount();
	return new Int(count);
}

#define _TC_ON(NAME)									\
NodeP lComboBoxOn##NAME(PairPP arg)			\
{															\
	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);\
	if throwing return Unbound;					\
	PairP function=arg[2]->a_runnable(2);		\
	if throwing return Unbound;					\
	comboBox->seton##NAME(function);			\
	return Unbound;									\
}

_TC_ON(Select)
_TC_ON(Change)
_TC_ON(Enter)
_TC_ON(Char)
_TC_ON(KeyDown)
_TC_ON(KeyUp)

NodeP lComboBoxValue(PairPP arg)
{
	if (exist(1))
	{	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);
		if throwing return Unbound;
		return new String(comboBox->comboBox->GetValue());
	}
	else
		return new String(Event::comboBoxValue);
}

NodeP lComboBoxSetValue(PairPP arg)
{
	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);
	if throwing return Unbound;
	StringP value=arg[2]->a_string(2);

	if throwing return Unbound;
	comboBox->comboBox->SetValue(value->ptr());
	return Unbound;
}

NodeP lComboBoxSetBackgroundColor(PairPP arg)
{
	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
	comboBox->comboBox->SetBackgroundColour(wc);
	return Unbound;
}

NodeP lComboBoxSetColor(PairPP arg)
{
	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
	comboBox->comboBox->SetForegroundColour(wc);
	return Unbound;
}

NodeP lComboBoxSetFontSize(PairPP arg)
{
	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=comboBox->comboBox->GetFont();
	f.SetPointSize(s);
	comboBox->comboBox->SetFont(f);
	return Unbound;
}

NodeP lComboBoxSetFontName(PairPP arg)
{
	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);
	if throwing return Unbound;
	StringP s=arg[2]->a_string(2);
	if throwing return Unbound;
	wxFont f=comboBox->comboBox->GetFont();
	f.SetFaceName(s->ptr());
	comboBox->comboBox->SetFont(f);
	return Unbound;
}

NodeP lComboBoxSetFontStyle(PairPP arg)
{
	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=comboBox->comboBox->GetFont();
	f.SetStyle(s);
	comboBox->comboBox->SetFont(f);
	return Unbound;
}

NodeP lComboBoxSetFontWeight(PairPP arg)
{
	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=comboBox->comboBox->GetFont();
	f.SetWeight(s);
	comboBox->comboBox->SetFont(f);
	return Unbound;
}

NodeP lComboBoxEnable(PairPP arg)
{
	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	comboBox->comboBox->Enable(state!=0);
	return Unbound;
}

NodeP lComboBoxDestroy(PairPP arg)
{
	ComboBoxNodeP comboBox=arg[1]->a_comboBox(1);
	if throwing return Unbound;
	{	FrameNodeP p=comboBox->parentFrame;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(comboBox));
			p->check_oldyoung(p->children);
		}
	}
	{	GraphNodeP p=comboBox->parentGraph;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(comboBox));
			p->check_oldyoung(p->children);
		}
	}
	comboBox->comboBox->Destroy();
	return Unbound;
}
