#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/app.h>
#include <wx/listctrl.h>

#include <stdio.h>

#include "Main.h"

#include "File.h"
#include "Pair.h"
#include "Procedur.h"
#include "Object.h"
#include "Quote.h"
#include "Colon.h"
#include "Tree.h"
#include "Int.h"
#include "Array.h"
#include "String.h"

#include "Error.h"
#include "Eval.h"
#include "Stack.h"
#include "Init.h"

#include "Calls.h"
#include "Vars.h"
#include "LogoApp.h"

#include "EdDefsExt.h"
#include "Edit.h"
#include "EdFrame.h"
#include "Mem.h"

extern Function llocal;

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

#ifdef WIN32
	#pragma inline_depth(12)
	#pragma inline_recursion(on)
#endif

bool Evaluator::dont_fix_ift=false;

int Evaluator::ift_iff_flag=-1,
	 Evaluator:: trace_level= 0,	// indentation level when tracing
	 Evaluator::repcount,
	 Evaluator::user_repcount,
	 Evaluator::ifori,
	 Evaluator::iforend,
	 Evaluator::iforstep;

double Evaluator::fori,
	Evaluator::forend,
	Evaluator::forstep;

 TailCall Evaluator:: tailcall = in_sequence;
ValStatus Evaluator::val_status= dont_care;

NodeP
	Evaluator::last_call=0,  // the last proc called
	didnt_output_name=0;  // the name of the proc that didn't OP 

PairP	output_node=0;  // the output of the current function 

NodeP			fun=0;  // current function name 
ObjectP	  ufun=0;  // current user-defined function name
LineP this_line=0;  // the current instruction line 


LineP   Evaluator::last_line=0;  // the line that called this one 
PairP   Evaluator::loopcond=0;	// the loop condition list
ObjectP Evaluator::last_ufun=0,  // the function that called this one 
		  Evaluator::forname=0;

int qmList=-1;  // question mark list 

ErrorStructP Evaluator::didnt_get_output=0;	// the name of the proc that wanted the OP 

int Evaluator::var=0;	// frame pointer into varStack

NodePStack             argStack,		// argument stack
            Evaluator::stack;			// eval register stack 
   IntStack Evaluator::intStack;		// integer stack
 FloatStack Evaluator::floatStack;	// float stack
  ToDoStack Evaluator::toDoStack;	// instruction stack 
   VarStack Evaluator::varStack;		// stack for storing global objects while local vars are bound
   IntStack Evaluator::varPStack;	// stack for pointers into varStack
   IntStack Evaluator::argPStack;	// stack for pointers into argStack
 ErrorStack Evaluator::errorStack;	// error stack

#define create_name(x) #x,

const char Evaluator::commandName[][end+1] =
{	Eval_Labels(create_name)
	"null"
};

#ifdef PROC_EVAL	// for debugging & profiling
	#ifdef _DEBUG		// for use with stack frames
		#define Goto(f)	{register Do fun=__##f.instruction; __asm{mov ebx,fun __asm mov ecx,this __asm mov esp,ebp __asm pop ebp __asm jmp ebx} return 0;}
	#else				// for use without stack frames
		#define Goto(f)	{return *intP(&__##f.instruction);}
	#endif
	#define TO(x)	}	Evaluator::_##x()	{
#else	// for the optimized release version; the fastest Goto is a goto !-)
	#define Goto(f)  goto f;
	#define TO(x)	}	x:	{
#endif

extern NodeP _callcode(PairPP) {return Unbound;}

//___________________________________________________________________
//
//		An explicit control evaluator, according to SICP, section 5.2.
//			<nodeList> is a flat list of expressions to evaluate.
//			<what> is the EvalCommand to begin with.
//			<Return value>  depends on what.
//___________________________________________________________________

Evaluator::Evaluator(PairP nodeList_, ToDo what)
{
	eval(nodeList_, what);
}

NodeP Evaluator::eval(PairP nodeList_, ToDo what)

#ifndef PROC_EVAL
{
#endif
				{	oldtailcall=tailcall;
					tracing=1;
					old_ift_iff=ift_iff_flag;

					val=catch_tag=0;
					exp=unev= arg= parm=formals=0;
					proc=0;
					asp=argStack;
					vsp=0;
					var=varStack;

					varPStack << var;
					stack << fun << ufun << this_line;
					toDoStack << __end;

					nodeList=nodeList_;
#ifdef PROC_EVAL
 					ToDo command = what;
					while(command.instruction)  
						command.i = (this->*command.instruction)();
					return val;
#else
					toDoStack << what;
					Goto( fetch_command )
#endif

TO( begin_line )
					toDoStack << __end_line;	Goto( begin_sequence )
TO( end_line )
					if (val != Unbound && !throwing) error(DK_WHAT, val);
					val=0;
					Goto( fetch_command )

TO( tail_dispatch )
					tailcall = tailcall_tail;
					Goto( dispatch )
TO( dispatch )
					switch (exp->type)
					{
					case LINE:
					case PAIR:	// procedure application
									{	ObjectP obj=ObjectP(exp->first());

										if (tailcall==tailcall_tail)
										{	if (!obj->bit.permanent 
											&&   obj->bit.proc_macro 
											||   obj==_Goto
											||   obj==_For)	// this is not very nice
																	// but I don't know how to avoid it
											{	// tail call to user-defined macro must be treated as non-tail
												// because the expression returned by the macro
												// remains to be evaluated in the caller's context
												unev=0; Goto( non_tail_eval )
											}
										}
										fun=obj;
									}
									if (exp->tail()) Goto( proc_with_args )
									asp=argStack;	
									// Evaluate an application of a procedure with no arguments. 
									Goto( apply_dispatch )	// apply the procedure 

					case COLON:	val = ColonP(exp)->value();	// variable
									if (val==Unbound && !throwing) 
										error(NO_VALUE, ColonP(exp)->colon());
									break;

					case QUOTE:	val = QuoteP(exp)->quote();	break; // quoted literal
					case ARRAY:	//val = ArrayP(exp)->deepCopy();break; // array must be copied
									val = new Array(ArrayP(exp));	break; 
					default:		val=exp;		// self-evaluating
					}
					Goto( fetch_command )

TO( proc_with_args )	// Evaluate an application of a procedure with arguments. 
					unev=exp->tail();
					asp=argStack;
					intStack.push(val_status, tailcall, ift_iff_flag);
					varPStack << var;
					stack << didnt_output_name;
					Goto( arg_loop )
TO( arg_loop )
	 arg_loop_again:
					if (!unev) Goto( args_done )
					exp = unev->first();
					if (!exp) {exp=_Not_Enough_node; goto arg_is_function;}
//if (!is_valid_pointer(exp)) {lpause(0); return Unbound;}
					switch (exp->type)
					{
					case LINE:
					case PAIR:{	ObjectP obj=ObjectP(exp->first());
									// Fix a colonless var if no proc or prim exists. 
									if (obj && obj->proc_or_prim()!=Unbound)
										goto arg_is_function;
									exp = new Colon(obj);
									unev->setfirst( exp );	// now we have a COLON
								 }
					case COLON:	val = ColonP(exp)->value();		// variable
									if (val==Unbound && !throwing)	
										error(NO_VALUE, ColonP(exp)->colon());
									break;

					case QUOTE:	val = QuoteP(exp)->quote();	break; // quoted literal
					case ARRAY:	//val = ArrayP(exp)->deepCopy();break; // array must be copied
									val = new Array(ArrayP(exp));	break; 
					default:		val=exp;		// self-evaluating
					}
					argStack << val;
					next(unev);
					goto arg_loop_again;
	 arg_is_function:
					if (exp==_Not_Enough_node && !throwing) 
					{	error(NOT_ENOUGH, 0); 
						argStack.pop_till(asp);
						Goto( args_done )
					}
					argPStack << asp;
					stack.push(unev,proc, fun,ufun,this_line, last_ufun,last_line);
					errorStack.push(ObjectP(exp->first()), ufun, this_line);
					didnt_output_name=0;

					var=varStack;
					tailcall = tailcall_arg;
					val_status=value_required;

					toDoStack << __accumulate_arg; Goto( dispatch )   // evaluate the current argument 
TO( accumulate_arg )	// Put the evaluated argument on the argStack.
					varStack.pop_till(var);
					last_call=fun;
					errorStack.pop();
					stack.pop(unev,proc, fun,ufun,this_line, last_ufun,last_line);
					argPStack >> asp;

					if (val==Unbound && !throwing) 
					{	error(DIDNT_OUTPUT, 0);
						Goto( fetch_command )
					}
					if (!throwing)
						argStack << val;
					next(unev);
					Goto( arg_loop )
TO( args_done )
					stack >> didnt_output_name;
					varPStack >> var;
					intStack.pop(val_status, tailcall, ift_iff_flag);

					last_call=fun;
					if (throwing) 
					{	val=Unbound;
						Goto( fetch_command )
					}
					Goto( apply_dispatch )

TO( apply_dispatch )	// Load in the procedure's definition and decide 
							// whether it's a compound procedure or a primitive procedure.
					if (!fun)
					{	c_out << "No way! This is bogus!"; 
						// better than a direct crash, but no error msg yet.
						lpause(argStack);
						Goto( fetch_command )
					}
					else
					{	proc=ObjectP(fun)->proc_or_prim();
						if (proc==Unbound && ObjectP(fun)->var()!=Unbound)
						{	// fixing colonless var before output
							val=ObjectP(fun)->var();	
							Goto( fetch_command );
						}
					}
					if ( ObjectP(fun)->bit.proc_macro )
					{	intStack << val_status << tailcall;
						val_status=value_required;
						toDoStack << __macro_return;
					}
					if (proc->is.primitive)		// primitive_apply
					{	if (!ObjectP(fun)->bit.proc_traced) 
						{	val = PrimitiveP(proc)->call(
								(PairPP)&argStack.base[asp]);	// call the primitive
							argStack.pop_till(asp);
						}else
						{	tracing = ObjectP(fun)->bit.proc_traced;
							for (int i=0; i < trace_level; i++) 
								c_out << ' ';
							c_out << "( " << fun << ' ';
							if (asp!=argStack)
								for (PairPP arg=argStack;  
								arg!=(PairPP)&argStack.base[asp];  arg--)
									c_out << (*arg)->maybe_quote() << ' ';
							c_out << ')' << endL;
						
							val = PrimitiveP(proc)->call(
								(PairPP)&argStack.base[asp]);	// call the primitive
							argStack.pop_till(asp);

							if (!throwing) 
							{	for (int i=0;  i < trace_level;  i++) 
									c_out << ' ';
								c_out << fun;
								if (val==Unbound)
									c_out << " stops" << endL;
								else 
									c_out << " outputs " << val->maybe_quote() << endL;
							}
						}
					}
					else if (proc->type==PROC_)
						Goto( compound_apply )
					else 
					{	//if (ufun) 
						//	untreeify_proc(ufun);
						if (!throwing) val=error(DK_HOW, fun);
					}
					Goto( fetch_command )

TO( fetch_command )
					ToDo __command;
					toDoStack >> __command; 
#ifdef PROC_EVAL
					Goto( command )
#else
	#define CREATE_CASE(x)	case x:	goto x;

					switch (__command.mnemonic)
					{	Eval_Labels(CREATE_CASE)
					}
#endif

TO( compound_apply )
					File::checkBreak();
					tracing = ObjectP(fun)->bit.proc_traced;
					if (tracing) 
					{	for (int i=0; i < trace_level; i++) 
							writer << ' ';
						trace_level++;
						writer << "( " << fun << ' ';
					}
					Goto( lambda_apply )
TO( lambda_apply )					// Bind the actuals to the formals 
					vsp=varStack;	// remember where we came in 
					argPStack << asp;
					formals=ProcedureP(proc)->formals();
					Goto( begin_formal_loop )
TO( begin_formal_loop )
					if (formals)
					{	parm=formals->first();
						if (type_of(parm)==INT_)	// default # args 
							Goto( after_formals )

						if (asp != argStack)
						{	arg = PairP(argStack.base[++asp]);
							if (tracing) 
								writer << arg->maybe_quote() << ' ';
						}else
							arg = PairP(Unbound);

						switch (type_of(parm))
						{
						case OBJECT: 
							{	ObjectP parmobj = ObjectP(parm);
								if (parmobj->not_local(vsp)) 
									varStack << parmobj;
								parmobj->tell_shadow();
								parmobj->setvar( arg );
								break;

							}
						case LINE:
						case PAIR:		// parm is optional or rest 
							{	ObjectP parm1 = ObjectP(parm->first());

								if (parm1->not_local(vsp)) 
									varStack << parm1;
								parm1->tell_shadow();

								if (!parm->tail())	// parm is rest 
								{
									parm1->setvar( argStack.to_list(asp-1) );	// SLOW 8-(
									if (tracing && asp!=argStack) 
									{	PairPP arg=(PairPP)&argStack.base[asp];
										while (arg!=argStack) 
											writer << (*++arg)->maybe_quote() << ' ';
									}
									Goto( after_formals )
								}

								if (arg==Unbound)		// use default 
								{
									varPStack << var << vsp;
									stack.push( 
										didnt_output_name, proc,  
										fun, ufun, this_line,  
										last_ufun, last_line,
										formals);

									argPStack << asp;
									intStack << ift_iff_flag << val_status;

									var=varStack;
									tailcall = tailcall_arg;
									val_status=value_required;

									nodeList=parm->tail();
									if (!throwing)
										nodeList->make_tree();
									else
										nodeList=0;

									if (nodeList && nodeList->type==TREE) 
									{	unev = TreeP(nodeList)->tree();
										val=Unbound;
										toDoStack << __set_args_continue;
										Goto( sequence )
									}else
										val=Unbound;
									Goto( set_args_continue )
								}
								Goto( set_args_continue2 )
							}
						default: ;
						}
						Goto( end_formal_loop )
					}
					Goto( after_formals )
TO( set_args_continue )
					varPStack >> vsp;
					argPStack >> asp;
					stack >> formals;
					parm=formals->first();
					varStack.pop_till(var);

					intStack >> val_status >> ift_iff_flag;
					stack.pop( 
						didnt_output_name, proc,  
						fun, ufun, this_line,  
						last_ufun, last_line);
					varPStack >> var;

					arg=PairP(val);
					Goto( set_args_continue2 )
TO( set_args_continue2 )
					ObjectP(parm->first())->setvar( arg );
					Goto( end_formal_loop )
TO( end_formal_loop )
					next(formals);
					Goto( begin_formal_loop )
TO( after_formals )
					argPStack >> asp;
					argStack.pop_till(asp);

					if (checkThrowing()) 
					{	val=Unbound;
						Goto( fetch_command )
					}
					vsp=0;
					tracing= !fun->is.list;
					if (tracing && ObjectP(fun)->bit.proc_traced) 
					{	if (!throwing) 
							writer << ')';
						writer << endL;
						stack << fun;
						toDoStack << __compound_apply_continue;
					}
					val=Unbound;
					last_ufun=ufun;

			//		if (!fun->is.list)	
					// commented out to make apply in procedure text form work
						ufun=ObjectP(fun);

					last_line=this_line;
					this_line=0;
					proc = fun->is.list ?   PairP(fun)->anonymous_function()
											  : ObjectP(fun)->proc();

					nodeList=ProcedureP(proc)->bodylist();		// get the body ... 
					nodeList->make_tree_from_body();

					if (!nodeList || nodeList->type!=TREE 
					|| !TreeP(nodeList)->tree()) //Kludge to make empty to def work
						Goto( fetch_command )

					unev = TreeP(nodeList)->tree();
					if (!throwing) 
						stoppingFlag=RUN;

					output_node=PairP(Unbound);


					switch (val_status)
					{	case value_required:				val_status=output_ok; break;
						case value_maybe_ok_in_macro:	val_status=dont_care; break;
						default:								val_status=no_value_allowed;
					}
					Goto( sequence )
TO( begin_sequence )
					nodeList->make_tree();
					if (!nodeList  ||  nodeList->type != TREE)
					{	val=Unbound;
						Goto( fetch_command )
					}
					unev = TreeP(nodeList)->tree();
					val=Unbound;
					Goto( sequence )
TO( sequence )	// Evaluate each expression in the sequence.  
					if (!running  ||  val != Unbound) 
						Goto( fetch_command )		// Stop as soon as  val!=Unbound.

					if (unev && unev->type==TREE) 
					{	if (Tree::the_generation != TreeP(unev)->generation())	// something redefined while we're running 
						{
							ProcedureP p = ProcedureP(proc);
							int linenum=0;

							for (this_line = TreeP(p->bodylist())->tree();
									this_line != unev;  this_line=LineP(this_line->tail()))
							{	// If redef isn't end of line, don't try to fix,
								// but don't blow up either. (Maybe not called from here.) 
								if (!this_line) goto nofix;
								if (this_line->is.line) linenum++;
							}
							p->untreeify();
							p->bodylist()->make_tree_from_body();
							unev = TreeP(p->bodylist())->tree();
							while (--linenum >= 0) 
							{	do next(unev);
								while (unev && unev->is.line);
							}
						}
					nofix:
						this_line = TreeP(unev)->tree();
						if (ufun  &&  ObjectP(ufun)->bit.proc_stepped) 
						{	if (tracing) 
								for (int i=1;  i < trace_level;  i++) 
									c_out << ' ';
							
//							c_out << this_line;
							theApp->edFrame->unmark(1);
							theApp->edFrame->mark(
								this_line->filename->ptr(), this_line->number, 1);
							lupdateCalls(0);
							if (updateVarsOnStep)
								lupdateVars(0);
							bool oldinstep=instep;
							instep=true;
							stepArmed=false;
							c_in.read(";->");
							instep=oldinstep;
							theApp->edFrame->unmark(1);

						}
					}

					exp=unev->first();
//					if (type_of(unev)!=LINE)
//						exp=exp;	// should never happen if everything was correct
//					else
						this_line=LineP(unev);

					if (ufun)
						if	(ObjectP(ufun)->bit.proc_stepped) 
						{	if (tracing) 
								for (int i=1;  i < trace_level;  i++) 
									c_out << ' ';
							
//							c_out << exp;//this_line->first();
							theApp->edFrame->unmark(1);
/*							if (exp  &&  exp->is.line)
								theApp->edFrame->mark(

									LineP(exp)->filename->ptr(), 
									LineP(exp)->number, 1);
*/							if (this_line  &&  this_line->is.line)
								theApp->edFrame->mark(
									this_line->filename->ptr(), 
									this_line->number, 1);
							lupdateCalls(0);
							if (updateVarsOnStep)
								lupdateVars(0);
							bool oldinstep=instep;
							instep=true;
							stepArmed=false;
							c_in.read("");//cin.read(";->");
							instep=oldinstep;
							theApp->edFrame->unmark(1);
						}
					next(unev);

					if (exp->is.list
					&& ObjectP(exp->first())->proc_or_prim()->type==TAILFORM)
					{
						PrimitiveP expProc = ObjectP(exp->first())->prim();
						Priority pri = expProc->priority;

						switch (pri)
						{
						case priority_OUTPUT: 
								errorStack.push(ObjectP(exp->first()), ufun, this_line);
								didnt_get_output = errorStack.top();
								errorStack.pop();	// pretty bad, this. Should be in some other places.
								didnt_output_name=0;

								if (!ufun && !throwing) 
								{	error(AT_TOPLEVEL, exp->first());
									val=Unbound; Goto( fetch_command )
								}else 
									switch (val_status)
									{
									case output_ok:
									case dont_care:
											val_status=value_required;
											exp = exp->second();
											Goto( tail_dispatch )

									case no_value_allowed:
									case value_required:
											val_status=value_required;
											exp = exp->second();
											unev=0;
											Goto( non_tail_eval )    // compute value then give error
									default: ;
									}
								break;

						case priority_STOP:
								if (!ufun && !throwing) 
								{	error(AT_TOPLEVEL, exp->first());
									val=Unbound; Goto( fetch_command )
								}else 
									switch (val_status)
									{
									case value_required:
									case output_ok:			didnt_output_name=fun;
									case no_value_allowed:
									case dont_care:			val=Unbound; 
																	Goto( fetch_command )
									default: ;
									}
								break;
						

						default:	// maybeoutput
								exp = exp->second();
								if (exp && !exp->first() && !exp->tail()) // new in 4.684
								{	val=Unbound;
									Goto( fetch_command )
								}
								val_status=value_maybe_ok_in_macro;
								Goto( tail_dispatch )
						}
					}
					if (!unev) 
					{	switch (val_status)
						{
						case output_ok:
						case no_value_in_macro:
								didnt_output_name=fun;
								unev = (PairP)Unbound;
								Goto( non_tail_eval )

						default: ;
						}
						Goto( tail_dispatch )
					}
					if (unev->first())
					{
						Pair &lst=*PairP(unev->first());
						if (        lst.islist()
						&&  ObjectP(lst.first())->proc_or_prim()->type==TAILFORM 
						&&  ObjectP(lst.first())->prim()->priority==priority_STOP
						&&  ufun) 
						{	switch (val_status)
							{
							case value_required:
							case output_ok:			didnt_output_name=fun;
							case no_value_allowed:
							case dont_care:			Goto( tail_dispatch )
							default: ;
							}
						}
					}
					Goto( non_tail_eval )
TO( non_tail_eval )
					varPStack << var;
					stack.push(unev, proc,  fun, ufun, this_line,  last_ufun, last_line);
					intStack << ift_iff_flag << val_status;
					var = varStack;
					tailcall = in_sequence;
					toDoStack << __sequence_continue;	Goto( dispatch )
TO( sequence_continue )
					varStack.pop_till(var);
					varPStack >> var;
					stack.pop(unev, proc,  fun, ufun, this_line,  last_ufun, last_line);

 					if (dont_fix_ift) 
					{	intStack >> val_status >> dont_fix_ift;
						dont_fix_ift=false;
					}else
						intStack >> val_status >> ift_iff_flag;

					if (stoppingFlag==MACRO_RETURN) 
					{
						if (unev==Unbound) unev=0;

						bool fromGoto=false;
						if (val)
							if (val->is.list  &&  PairP(val)->first()==NodeP(_Tag))
								fromGoto=true;
						if (fromGoto)
							unev = PairP(val)->tail();	// from goto 
						else
							unev = PairP(val)->append(unev);	// slow

						val=Unbound;
						stoppingFlag=RUN;

						if (!unev) Goto( fetch_command )
					}
					else 
						switch (val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (stoppingFlag)
								{
								case RUN:	output_node=PairP(Unbound); break;
								case STOP:	output_node=PairP(Unbound);
								case OUTPUT:stoppingFlag=RUN;
												val=output_node;
												if (val_status==value_required  &&  !throwing)
												{	if (val==Unbound) 
														didnt_output_name=_Stop;
													else
														didnt_output_name=_Output;

													error(DIDNT_OUTPUT, _Output);
												}
												Goto( fetch_command )
								default: ;
								}
						default: ;
						}

					if (val!=Unbound && !throwing) 
					{	error(!unev ?	DK_WHAT_UP
										:  DK_WHAT,  
									val);
						val=Unbound;
					}
					if (!throwing && (!unev || unev==Unbound)) 
					{
						if ((val_status != no_value_in_macro) 
						&&  (val_status != no_value_allowed) // new because of "for"
						&&  (val_status != value_maybe_ok_in_macro) // new because of "for"
						&&  (val_status != dont_care)) // new because of "for"
							error(DIDNT_OUTPUT, 0);
						Goto( fetch_command )
					}
					Goto( sequence ) 

TO( compound_apply_continue )	// Only get here if tracing 
					stack >> fun;
					--trace_level;
					if (!throwing) 
					{	for (int i=0; i < trace_level; i++) 
							writer << ' ';
						writer << fun;
						if (val==Unbound)
							writer << " stops" << endL;
						else 
							writer << " outputs " << val->maybe_quote() << endL;
					}
					Goto( fetch_command )


TO( begin_apply )							// This is for _apply.
					fun=PairP(val)->first();
					
					while (type_of(fun)==ARRAY  &&  !throwing)
						fun=ObjectP(error(APPLY_BAD_DATA, fun));

					while ((PairP(val)->tail()->isnolist() 
					&& PairP(val)->tail()->type!=ARRAY) && !throwing)
						PairP(val)->settail( 
							error(APPLY_BAD_DATA, PairP(val)->tail()) );

					if (!throwing)
					{	asp=argStack;
						if (PairP(val)->tail()->is.list)
							argStack.push_list(PairP(val)->tail());
						else if (PairP(val)->tail()->type==ARRAY)
							argStack.push(ArrayP(PairP(val)->tail()));
						val=Unbound;
					}
					if (!throwing && fun) 
					{	if (fun->is.list)		    // template 
						{	PairP lst=PairP(fun);
							if (lst->first()->is.list && lst->tail()) 
							{	if (lst->second()->is.list)		// procedure text form 
								{	proc = lst->anonymous_function();
									tracing=0;
									Goto( lambda_apply )
								}
								formals=lst->first();		// lambda form 

								varPStack << var;
								intStack << tailcall;
								tailcall = in_sequence;
								argPStack << asp;
								asp=argStack;
								argStack << formals;
								llocal((PairPP)&argStack.base[asp]);	
								// bind the formals locally 
								argPStack >> asp;
								argPStack << asp;
								intStack >> tailcall;
								
								while (formals &&  asp!=argStack  && !throwing)
								{	ObjectP(formals->first())->
										setvar( argStack.base[++asp] );
									next(formals); 
								}
								argPStack >> asp;
								argStack.pop_till(asp);
								nodeList = lst->tail();
								argPStack << qmList;

								toDoStack << __after_lambda; 
								goto lambda_questionmark;
							}else		// question-mark form 
							{	argPStack << qmList;
								qmList=asp;

								nodeList=lst;
	lambda_questionmark:
								nodeList->make_tree();
								
								if (!nodeList || nodeList->type!=TREE) 
									Goto( questionmark_failed )
								
								unev = TreeP(nodeList)->tree();

								stack << didnt_output_name;
								intStack << val_status << tailcall;

								val_status=value_maybe_ok_in_macro;

								toDoStack << __questionmark_continue;	Goto( sequence )
							}
						}else		// name of procedure to apply 
						{	int min, max, n;

							fun = fun->intern();	// 1st) try finding fun in internal routines ;-)))

							if (ObjectP(fun)->proc_or_prim()==Unbound  &&  fun!=_Null_Word)
								fun->silent_load(environment.logolib.c_str());	// 2nd) try <logolib>/<fun> 

							if (ObjectP(fun)->proc_or_prim()==Unbound  &&  fun!=_Null_Word)
								fun->silent_load(0);	// 3rd) try ./<fun>.lg (the user directory)

							proc = ObjectP(fun)->proc_or_prim();

							if (proc==Unbound && !throwing)
								val=error(DK_HOW_UNREC, fun);
							
							if (!throwing) 
							{	if (type_of(proc) == PROC_) 
								{	min= ProcedureP(proc)->min;
									max= ProcedureP(proc)->max;
								}else 
								{	Primitive &prim= *PrimitiveP(proc);

									if (prim.std < 0)		// special form
									{	error(DK_HOW_UNREC, fun);	// can't apply 
										Goto( fetch_command )
									}else 
									{	min = prim.min;
										max= prim.max;
									}
								}
								n = int(argStack.index-argStack.base)-asp;

								if (n < min) 
									error(NOT_ENOUGH, 0);
								else if (n > max && max >= 0) 
									error(TOO_MUCH, 0);
								else 
									Goto( apply_dispatch )
							}
						}
					}
					Goto( fetch_command )
TO( after_lambda )
					varStack.pop_till(var);
					varPStack >> var;
					Goto( fetch_command )


TO( questionmark_continue )
					intStack >> tailcall >> val_status;
					stack >> didnt_output_name;
					
					if (tailcall) 
					{	switch (val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (stoppingFlag)
								{
								case RUN:	output_node=PairP(Unbound); break;
								case STOP:	output_node=PairP(Unbound);
								case OUTPUT:stoppingFlag=RUN;
												val=output_node;
												if (val != Unbound)
													switch (val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
								default: ;
								}
						default: ;
						}
					}
					Goto( questionmark_failed )
TO( questionmark_failed )
					if (qmList>=0)
						argStack.pop_till(qmList);// new in 4.65
					argPStack >> qmList;
		//			argStack.pop_till(asp-1);	// bug here?
					Goto( fetch_command )

TO( macro_return )	//___________________ MACROS ______________________
					intStack >> tailcall >> val_status;

					while (!val->islist() && !throwing)
						val=error(ERR_MACRO, val);
					
					if (!throwing) 
					{	if (type_of(val)==TODO)
						{	ToDoNode &tdn = *ToDoNodeP(val);
							toDoStack << tdn.toDo;
							val = tdn.tail();
							Goto( fetch_command )
						}
						if (tailcall==in_sequence) 
						{
							PairP(val)->make_tree();
							stoppingFlag=MACRO_RETURN;

							if (val->istree()) 
								val= TreeP(val)->tree();
							else
								val=0;
							Goto( fetch_command )
						}
						nodeList= PairP(val);
						Goto( begin_sequence )
					}
					val=Unbound;
					Goto( fetch_command )

TO( runresult_continuation )
					nodeList=PairP(val);
					val_status=value_maybe_ok_in_macro;
					toDoStack <<__runresult_followup;	Goto( begin_sequence )
TO( runresult_followup )
					if (val==Unbound) val=0;
					else 
						val = new Pair(val, 0);
					Goto( fetch_command )


TO( repeat_continuation )
					intStack.push(repcount, user_repcount);
					repcount = IntP(PairP(val)->first())->int_;
					nodeList =      PairP(val)->tail ();
					val=Unbound;
					user_repcount=0;
					Goto( repeat_again )
TO( repeat_again )
					user_repcount++;
					intStack.push(val_status, tailcall);
					val_status=no_value_in_macro;
					stack << nodeList;

					toDoStack << __repeat_followup;	Goto( begin_sequence )
TO( repeat_followup )
					stack >> nodeList;
					intStack.pop(val_status, tailcall);


					if (val != Unbound && !throwing) error(DK_WHAT, val);

					if (stoppingFlag)
					{	switch (stoppingFlag)
						{
						case BREAK:	stoppingFlag=RUN;
										output_node=PairP(Unbound); 
										switch (val_status)
										{
										case value_required:
												error(DK_WHAT_UP, val);
										default: ;
										}												
										goto repeat_end;
						case CONTINUE:	stoppingFlag=RUN; 
										output_node=PairP(Unbound); 
										break;
						default: ;
						}
					}
					if (tailcall) 
					{
						switch (val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (stoppingFlag)
								{
								case RUN:	output_node=PairP(Unbound); break;
								case STOP:	output_node=PairP(Unbound);
								case OUTPUT:stoppingFlag=RUN;
												val=output_node;
												if (val!=Unbound)
													switch (val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
												goto repeat_end;
								default: ;
								}
						default:	;
						}
					}
					if (repcount >0) --repcount;	// negative means forever 
					if (repcount==0) goto repeat_end;

					File::checkBreak();
					if (running) Goto( repeat_again )

					val=Unbound;
	repeat_end:	intStack.pop(repcount, user_repcount);
					Goto( fetch_command )

TO( while_continuation )
					intStack << user_repcount;
					user_repcount = 0;
					stack << loopcond;
					loopcond = PairP(val)->first();
					 nodeList = PairP(val)->tail();
					val=Unbound;
					Goto( while_cond )
TO( while_cond )
					user_repcount++;
					intStack.push(val_status, tailcall);
					val_status=value_required;
					stack << nodeList;
					nodeList=loopcond;
					toDoStack << __while_again;	Goto( begin_sequence )
TO( while_again )
					if (val==_False)
					{	stack >> nodeList;
						intStack.pop(val_status, tailcall);
						val=Unbound;
						Goto( while_end )
					}
					val_status=no_value_in_macro;
					stack >> nodeList;	// get it again
					stack << nodeList;

					toDoStack << __while_followup;	Goto( begin_sequence )
TO( while_followup )
					stack >> nodeList;
					intStack.pop(val_status, tailcall);

					if (val != Unbound && !throwing) error(DK_WHAT, val);

					switch (stoppingFlag)
					{
					case BREAK:	stoppingFlag=RUN;
									output_node=PairP(Unbound); 
									val_status=no_value_in_macro;
									Goto( while_end );
					case CONTINUE:	stoppingFlag=RUN; 
									output_node=PairP(Unbound); 
									break;
					default: ;
					}
					if (tailcall) 
					{
						switch (val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (stoppingFlag)
								{
								case RUN:	output_node=PairP(Unbound); break;
								case STOP:	output_node=PairP(Unbound);
								case OUTPUT:stoppingFlag=RUN;
												val=output_node;
												if (val!=Unbound)
													switch (val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
												Goto( while_end );
								default: ;
								}
						default: ;
						}
					}
					File::checkBreak();
					if (running) Goto( while_cond )

					val=Unbound;
					Goto( while_end )
TO( while_end )
					stack >> loopcond;
					intStack >> user_repcount;
					Goto( fetch_command )

TO( until_continuation )
					intStack << user_repcount;
					user_repcount = 0;
					stack << loopcond;
					loopcond = PairP(val)->first();
					 nodeList = PairP(val)->tail();
					val=Unbound;
					Goto( until_cond )
TO( until_cond )
					user_repcount++;
					intStack.push(val_status, tailcall);
					val_status=value_required;
					stack << nodeList;
					nodeList=loopcond;
					toDoStack << __until_again;	Goto( begin_sequence )
TO( until_again )
					if (val==_True)
					{	stack >> nodeList;
						intStack.pop(val_status, tailcall);
						val=Unbound;
						Goto( until_end )
					}
					val_status=no_value_in_macro;
					stack >> nodeList;	// get it again
					stack << nodeList;

					toDoStack << __until_followup;	Goto( begin_sequence )
TO( until_followup )
					stack >> nodeList;
					intStack.pop(val_status, tailcall);

					if (val != Unbound && !throwing) error(DK_WHAT, val);

					switch (stoppingFlag)
					{

					case BREAK:	stoppingFlag=RUN;
									output_node=PairP(Unbound); 
									val_status=no_value_in_macro;
									Goto( until_end );
					case CONTINUE:	stoppingFlag=RUN; 
									output_node=PairP(Unbound); 
									break;
					default: ;
					}
					if (tailcall) 
					{
						switch (val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (stoppingFlag)
								{
								case RUN:	output_node=PairP(Unbound); break;
								case STOP:	output_node=PairP(Unbound);
								case OUTPUT:stoppingFlag=RUN;
												val=output_node;
												if (val!=Unbound)
													switch (val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
												Goto( until_end );
								default: ;
								}
						default: ;
						}
					}
					File::checkBreak();
					if (running) Goto( until_cond )

					val=Unbound;
					Goto( until_end )
TO( until_end )
					stack >> loopcond;
					intStack >> user_repcount;

					Goto( fetch_command )

TO( do_while_continuation )

					intStack << user_repcount;
					user_repcount = 0;
					stack << loopcond;
					loopcond = PairP(val)->first();
					 nodeList = PairP(val)->tail();
					val=Unbound;
					Goto( do_while_again )
TO( do_while_again )
					user_repcount++;
					intStack.push(val_status, tailcall);
					val_status=no_value_in_macro;
					stack << nodeList;


					toDoStack << __do_while_followup;	Goto( begin_sequence )
TO( do_while_followup )
					stack >> nodeList;
					intStack.pop(val_status, tailcall);

					if (val != Unbound && !throwing) error(DK_WHAT, val);

					switch (stoppingFlag)
					{
					case BREAK:	stoppingFlag=RUN;
									output_node=PairP(Unbound); 
									val_status=no_value_in_macro;
									Goto( do_while_end );
					case CONTINUE:	stoppingFlag=RUN; 
									output_node=PairP(Unbound); 
									break;
					default: ;
					}
					if (tailcall) 
					{
						switch (val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (stoppingFlag)
								{
								case RUN:	output_node=PairP(Unbound); break;
								case STOP:	output_node=PairP(Unbound);
								case OUTPUT:stoppingFlag=RUN;
												val=output_node;
												if (val!=Unbound)


													switch (val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
												Goto( do_while_end );
								default: ;
								}
						default: ;
						}
					}
					val_status=value_required;
					stack << nodeList;
					nodeList=loopcond;
					toDoStack << __do_while_cond;	Goto( begin_sequence )
TO( do_while_cond )
					stack >> nodeList;
					if (val==_False)
					{	val=Unbound;
						Goto( do_while_end )
					}

					File::checkBreak();
					if (running) Goto( do_while_again )

					val=Unbound;
					Goto( do_while_end )
TO( do_while_end )
					stack >> loopcond;
					intStack >> user_repcount;
					Goto( fetch_command )

TO( do_until_continuation )
					intStack << user_repcount;
					user_repcount = 0;
					stack << loopcond;
					loopcond = PairP(val)->first();
					 nodeList = PairP(val)->tail();
					val=Unbound;
					Goto( do_until_again )
TO( do_until_again )
					user_repcount++;
					intStack.push(val_status, tailcall);
					val_status=no_value_in_macro;
					stack << nodeList;

					toDoStack << __do_until_followup;	Goto( begin_sequence )
TO( do_until_followup )
					stack >> nodeList;
					intStack.pop(val_status, tailcall);

					if (val != Unbound && !throwing) error(DK_WHAT, val);

					switch (stoppingFlag)
					{
					case BREAK:	stoppingFlag=RUN;
									output_node=PairP(Unbound); 
									val_status=no_value_in_macro;
									Goto( do_until_end );
					case CONTINUE:	stoppingFlag=RUN; 
									output_node=PairP(Unbound); 
									break;
					default: ;
					}
					if (tailcall) 
					{
						switch (val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (stoppingFlag)
								{
								case RUN:	output_node=PairP(Unbound); break;
								case STOP:	output_node=PairP(Unbound);
								case OUTPUT:stoppingFlag=RUN;
												val=output_node;
												if (val!=Unbound)
													switch (val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
												Goto( do_until_end );
								default: ;

								}
						default: ;
						}
					}
					val_status=value_required;
					stack << nodeList;
					nodeList=loopcond;
					toDoStack << __do_until_cond;	Goto( begin_sequence )
TO( do_until_cond )
					stack >> nodeList;
					if (val==_True)
					{	val=Unbound;
						Goto( do_until_end )
					}
					File::checkBreak();
					if (running) Goto( do_until_again )

					val=Unbound;
					Goto( do_until_end )
TO( do_until_end )
					stack >> loopcond;
					intStack >> user_repcount;
					Goto( fetch_command )

TO( for_continuation )
					floatStack.push(fori, forend, forstep);
					stack << forname;
					PairP pval=PairP(val);
					forname= pval->first()->intern();
					fori   = FloatP(pval->second())->float_;
					forend = FloatP(pval->third ())->float_;
					forstep= FloatP(pval->fourth())->float_;
					nodeList =      pval->fifth();
					varStack << forname;
					forname->tell_shadow();
					forname->setvar( new Float(fori) );
								
					val=Unbound;
					Goto( for_again )
TO( for_again )
					intStack.push(val_status, tailcall);
					val_status=no_value_in_macro;
					stack << nodeList;

					toDoStack << __for_followup;	Goto( begin_sequence )
TO( for_followup )
					stack >> nodeList;
					intStack.pop(val_status, tailcall);

					if (val != Unbound && !throwing) error(DK_WHAT, val);

					switch (stoppingFlag)
					{
					case BREAK:	stoppingFlag=RUN;
									output_node=PairP(Unbound); 
									switch (val_status)
									{
									case value_required:
											error(DK_WHAT_UP, val);
									default: ;
									}												
									goto for_end;
					case CONTINUE:	stoppingFlag=RUN; 
									output_node=PairP(Unbound); 
									break;
					default:	;
					}
					if (tailcall) 
					{
						switch (val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (stoppingFlag)
								{
								case RUN:	output_node=PairP(Unbound); break;
								case STOP:	output_node=PairP(Unbound);
								case OUTPUT:stoppingFlag=RUN;
												val=output_node;
												if (val!=Unbound)
													switch (val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
												goto for_end;
								default: ;
								}
						default: ;
						}
					}
					if (forname->var()->type==FLOAT_)
						fori = FloatP(forname->var())->float_;
					else if (forname->var()->type==INT_)
						fori = IntP(forname->var())->int_;
					else
						error(BAD_DATA, forname);
					fori+=forstep;
					if (forstep >= 0)
					{	if (fori > forend) 
							goto for_end;
					}else
						if (fori < forend) 
							goto for_end;
					forname->setvar( new Float(fori) );

					File::checkBreak();
					if (running) Goto( for_again )

					val=Unbound;
	for_end:
					varStack.pop();
					stack >> forname;
					floatStack.pop(fori, forend, forstep);
					Goto( fetch_command )

TO( ifor_continuation )
					intStack.push(ifori, iforend, iforstep);
					stack << forname;
					PairP pval=PairP(val);
					forname= pval->first()->intern();
					ifori   = IntP(pval->second())->int_;
					iforend = IntP(pval->third ())->int_;
					iforstep= IntP(pval->fourth())->int_;
					nodeList =      pval->fifth();
					varStack << forname;
					forname->tell_shadow();
					forname->setvar( new Int(ifori) );
								
					val=Unbound;
					Goto( ifor_again )
TO( ifor_again )
					intStack.push(val_status, tailcall);
					val_status=no_value_in_macro;
					stack << nodeList;

					toDoStack << __ifor_followup;	Goto( begin_sequence )
TO( ifor_followup )
					stack >> nodeList;
					intStack.pop(val_status, tailcall);

					if (val != Unbound && !throwing) error(DK_WHAT, val);

					switch (stoppingFlag)
					{
					case BREAK:	stoppingFlag=RUN;
									output_node=PairP(Unbound); 
									switch (val_status)
									{
									case value_required:
											error(DK_WHAT_UP, val);
									default: ;
									}												
									goto ifor_end;
					case CONTINUE:	stoppingFlag=RUN; 
									output_node=PairP(Unbound); 
									break;
					default:	;
					}
					if (tailcall) 
					{
						switch (val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (stoppingFlag)
								{
								case RUN:	output_node=PairP(Unbound); break;
								case STOP:	output_node=PairP(Unbound);
								case OUTPUT:stoppingFlag=RUN;
												val=output_node;
												if (val!=Unbound)
													switch (val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;

													}
												goto ifor_end;
								default: ;
								}
						default: ;
						}
					}
					if (forname->var()->type==INT_)
						ifori = IntP(forname->var())->int_;
					else if (forname->var()->type==FLOAT_)
						ifori = int(FloatP(forname->var())->float_);
					else
						error(BAD_DATA, forname);
					ifori+=iforstep;
					if (iforstep >= 0)
					{	if (ifori > iforend) 
							goto ifor_end;
					}else
						if (ifori < iforend) 
							goto ifor_end;
					forname->setvar( new Int(ifori) );

					File::checkBreak();
					if (running) Goto( ifor_again )

					val=Unbound;
	ifor_end:	varStack.pop();
					stack >> forname;
					intStack.pop(ifori, iforend, iforstep);
					Goto( fetch_command )

TO( foreach_continuation )
					intStack << user_repcount;
					stack << loopcond;
					varPStack << var;
					var=varStack;
					varStack << _Template_Lists;
					loopcond = PairP(val)->first();
					nodeList = PairP(val)->tail ();
					user_repcount=0;
					Goto( foreach_again )
TO( foreach_again )
					if (!loopcond->first() || loopcond->first()==NodeP(_Null_Word)) 
						Goto( foreach_end )

					user_repcount++;
					_Template_Lists->setvar( loopcond );
					val=new Pair(nodeList, loopcond->firsts());
					loopcond = PairP(loopcond->butFirsts());

					intStack.push(val_status, tailcall);
					stack << nodeList;
					val_status=no_value_in_macro;

	fun=PairP(val)->first();
	
	while (type_of(fun)==ARRAY  &&  !throwing)
		fun=ObjectP(error(APPLY_BAD_DATA, fun));

	while (PairP(val)->tail()->isnolist() && !throwing)
		PairP(val)->settail( 
			error(APPLY_BAD_DATA, PairP(val)->tail()) );

	if (!throwing)
	{	asp=argStack;
		argStack.push_list(PairP(val)->tail());
		val=Unbound;
	}
	argPStack << qmList;
	qmList=asp;
					
					toDoStack << __foreach_followup;	Goto( begin_sequence )
TO( foreach_followup )
	if (qmList>=0)
		argStack.pop_till(qmList);// new in 4.65
	argPStack >> qmList;
					stack >> nodeList;

					intStack.pop(val_status, tailcall);

					if (val != Unbound && !throwing) error(DK_WHAT, val);

					switch (stoppingFlag)
					{
					case BREAK:	stoppingFlag=RUN;
									output_node=PairP(Unbound); 
									switch (val_status)
									{
									case value_required:
											error(DK_WHAT_UP, val);
									default: ;
									}												
									Goto( foreach_end );
					case CONTINUE:	stoppingFlag=RUN; 
									output_node=PairP(Unbound); 
									break;
					default:	;

					}
					if (tailcall) 
					{
						switch (val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (stoppingFlag)
								{
								case STOP:	output_node=PairP(Unbound);
								case OUTPUT:stoppingFlag=RUN;
												val=output_node;
												if (val!=Unbound)
													switch (val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;
													}
												Goto( foreach_end );
								default: ;
								}
						default: ;
						}
					}
					File::checkBreak();
					if (running) Goto( foreach_again )

					val=Unbound;
					Goto( foreach_end )
TO( foreach_end )
					varStack.pop_till(var);
					varPStack >> var;
					stack >> loopcond;
					intStack >> user_repcount;
					Goto( fetch_command )

TO( catch_continuation )
					nodeList = PairP(val)->tail ();
					catch_tag= PairP(val)->first();

					if (catch_tag->equal(_Error)) 
					{	varStack << _ErrorAction;
						_ErrorAction->setvar( Unbound );
					}
					val_status=value_maybe_ok_in_macro;

					stack << catch_tag << didnt_output_name;
					intStack << val_status << tailcall;

					toDoStack << __catch_followup;	Goto( begin_sequence )
TO( catch_followup )
					intStack >> tailcall >> val_status;
					stack >> didnt_output_name >> catch_tag;
//					argStack.pop_till(asp);

					if (tailcall) 
					{	switch (val_status)
						{
						case no_value_allowed:
						case value_required:
						case output_ok:
						case dont_care:
								switch (stoppingFlag)
								{
								case RUN:	output_node=PairP(Unbound); break;
								case STOP:	output_node=PairP(Unbound);
								case OUTPUT:stoppingFlag=RUN;
												val=output_node;
												if (val!=Unbound)
													switch (val_status)
													{
													case no_value_allowed:
													case value_required:
															error(DK_WHAT_UP, val);
													default: ;

													}
								default: ;
								}
						default: ;
						}
					}
					if ( throwing 
					&& throw_node->equal(catch_tag)) 
					{	throw_node=Unbound;
						stoppingFlag=RUN;
						val=output_node;
					}
					Goto( fetch_command )

TO( goto_continuation )
					if (!throwing) 
					{	if (!ufun) 
						{	error(AT_TOPLEVEL, _Goto);
							val=Unbound;
							Goto( fetch_command )
						}
						    proc = ObjectP(ufun)->proc_or_prim();

						nodeList = ProcedureP(proc)->bodylist();
						    unev = TreeP(nodeList)->tree();

						while (unev  &&  !checkThrowing()) 
						{
							if (unev->type==TREE)
								this_line = TreeP(unev)->tree();
							exp=unev->first();
							next(unev);

							if (exp->islist()
							&& ObjectP(exp->first())==_Tag
							&& exp->tail()
							&& exp->second()->type==QUOTE 
							&& val->equal(QuoteP(exp->second())->quote()) ) 
							{
								val = new Pair(_Tag, unev);
								stoppingFlag=MACRO_RETURN;
								Goto( fetch_command )
							}
						}
						if (!throwing)
							error(BAD_DATA_UNREC, val);
					}
					val=Unbound;
					Goto( fetch_command )

TO( end )
					tailcall=oldtailcall;
					ift_iff_flag=old_ift_iff;

					stack >> this_line >> ufun >> fun;
					varStack.pop_till(var);
					varPStack >> var;
#ifdef PROC_EVAL
					return 0;
				}
#else
				}
				return val;
}
#endif
