//___________________________________________________________________
//
//								class Event and fellows
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <stdio.h>
#include <malloc.h>

#include <wx/textctrl.h>
#include <wx/stattext.h>
#include <wx/glcanvas.h>
#include <wx/slider.h>
#include <wx/button.h>
#include <wx/tglbtn.h>
#include <wx/checkbox.h>
#include <wx/radiobut.h>
#include <wx/listctrl.h>
#include <wx/frame.h>

#include "spinctld.h"
#include "spinctli.h"

#include "Main.h"
#include "GC.h"
#include "File.h"
#include "Pair.h"
#include "Eval.h"
#include "String.h"
#include "Int.h"
#include "UInt8.h"
#include "Slider.h"
#include "IntControl.h"
#include "FloatControl.h"
#include "Button.h"
#include "ToggleButton.h"
#include "CheckBox.h"
#include "RadioButton.h"
#include "TextControl.h"
#include "ListBox.h"
#include "ChoiceBox.h"
#include "ComboBox.h"
#include "ListControl.h"
#include "Frame.h"
#include "Color.h"
#include "Point.h"
#include "Mouse.h"
#include "Console.h"
#include "Graph.h"
#include "Event.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

//___________________________________________________________________

EventFiFo::~EventFiFo() {if (allocated) free(base);}

EventFiFo &EventFiFo::operator<<(const EventP val) 
{
	if (overflow()) return *this;
	*in=val;
	if (in==max) in=base; else in++;
	if (in==out) expand();
	return *this;
}

EventFiFo &EventFiFo::operator>>(EventP &val) 
{
	val = *out;
	if (out==max) out=base; else out++;
	return *this;
}

void EventFiFo::expand()
{
	EventPP newbase = (EventP*) GC_::calloc(size+expand_size, sizeof(EventP));

	if (newbase==EventPP(0))	// no room to increase EventFiFo
	{	c_out << "; Warning: Not enough memory to aquire more events, some will be lost." << endL;
		_overflow=true;
	}
	else
	{	EventPP i=newbase;

		if (out)			// transfer old EventFiFo to new EventFiFo
		{	*this >> *i++;
			while (in!=out) 
				*this >> *i++;
		}				
		if (allocated) 
			free(base);

		allocated=true;
		size += expand_size;
		base=newbase;   max= &base[size-1];   in=i;   out= base;
	}
}

//___________________________________________________________________

EventFiFo Event::fifo(0x100);
int Event::keyboardValue=0;
int Event::sliderValue=0;
int Event::intControlValue=0;
double Event::floatControlValue=0;
wxString Event::textControlValue=wxEmptyString;
bool Event::inDispatch=false;
ListBox_ *Event::listBoxValue=0;
ChoiceBox *Event::choiceBoxValue=0;
wxString Event::comboBoxValue=wxEmptyString;
ComboBox *Event::staticComboBox=0;
ListControl *Event::listControlValue=0;
int Event::listControlRow=0, Event::listControlCol=0;
wxString Event::listControlText=wxEmptyString;

void Event::dispatch()
{
/*	if (inDispatch)
		return;	// this should prevent any stack overflows
	inDispatch=true;
*/
	EventP ev;
	while (fifo)
	{	fifo >> ev;
		ev->process();
		delete ev;
	}
	inDispatch=false;
}

void Event::eatAll()
{
	EventP ev;
	while (fifo)
	{	fifo >> ev;
		delete ev;
	}
}

//___________________________________________________________________

#define EVENT_KEY(NAME)									\
PairP Event##NAME::function=0;						\
																\
void Event##NAME::process()							\
{																\
	if (function)											\
	{	Event::keyboardValue=c;							\
		ValStatus vs=Evaluator::val_status;			\
		Evaluator::val_status = no_value_allowed;	\
		function->eval_driver();						\
		Evaluator::val_status=vs;						\
	}															\
}

EVENT_KEY(Char)
EVENT_KEY(KeyDown)
EVENT_KEY(KeyUp)
//___________________________________________________________________

#define EVENT_FRAME_KEY(NAME)							\
																\
void EventFrame##NAME::process()						\
{																\
	if ((frame!=0) && frame->on##NAME)						\
	{	Event::keyboardValue=c;							\
		ValStatus vs=Evaluator::val_status;			\
		Evaluator::val_status = no_value_allowed;	\
		frame->on##NAME->eval_driver();				\
		Evaluator::val_status=vs;						\
	}															\
}

EVENT_FRAME_KEY(Char)
EVENT_FRAME_KEY(KeyDown)
EVENT_FRAME_KEY(KeyUp)
//___________________________________________________________________

#define EVENT_GRAPH_KEY(NAME)							\
																\
void EventGraph##NAME::process()						\
{																\
	if ((graph!=0) && graph->on##NAME)						\
	{	Event::keyboardValue=c;							\
		ValStatus vs=Evaluator::val_status;			\
		Evaluator::val_status = no_value_allowed;	\
		graph->on##NAME->eval_driver();				\
		Evaluator::val_status=vs;						\
	}															\
}

EVENT_GRAPH_KEY(Char)
EVENT_GRAPH_KEY(KeyDown)
EVENT_GRAPH_KEY(KeyUp)
//___________________________________________________________________

#define EVENT_TEXT_MOUSE(NAME)								\
PairP EventTextMouse##NAME::function=0;					\
																		\
EventTextMouse##NAME::EventTextMouse##NAME(Mouse *m):	\
	Event(TextMouse##NAME),										\
	x(m->textx), y(m->texty), buttons(m->buttons) {}	\
																		\
void EventTextMouse##NAME::process()						\
{																		\
	if (function)													\
	{	console->mouse->textx = int(x);						\
		console->mouse->texty = int(y);						\
		console->mouse->buttons = buttons;					\
		ValStatus vs=Evaluator::val_status;					\
		Evaluator::val_status = no_value_allowed;			\
		function->eval_driver();								\
		Evaluator::val_status=vs;								\
	}																	\
}

EVENT_TEXT_MOUSE(LeftDown)
EVENT_TEXT_MOUSE(RightDown)
EVENT_TEXT_MOUSE(MiddleDown)

EVENT_TEXT_MOUSE(LeftUp)
EVENT_TEXT_MOUSE(RightUp)
EVENT_TEXT_MOUSE(MiddleUp)

EVENT_TEXT_MOUSE(LeftDClick)
EVENT_TEXT_MOUSE(RightDClick)
EVENT_TEXT_MOUSE(MiddleDClick)

EVENT_TEXT_MOUSE(Motion)
//___________________________________________________________________

#define EVENT_MOUSE(NAME)										\
PairP EventMouse##NAME::function=0;							\
																		\
EventMouse##NAME::EventMouse##NAME(Mouse *m):			\
	Event(Mouse##NAME),											\
	x(m->x), y(m->y), z(m->z), buttons(m->buttons) {}	\
																		\
void EventMouse##NAME::process()								\
{																		\
	if (function)													\
	{	console->mouse->x = x;									\
		console->mouse->y = y;									\
		console->mouse->z = z;									\
		console->mouse->buttons = buttons;					\
		ValStatus vs=Evaluator::val_status;					\
		Evaluator::val_status = no_value_allowed;			\
		function->eval_driver();								\
		Evaluator::val_status=vs;								\
	}																	\
}

PairP EventMouseLeftDown::function=0;

EventMouseLeftDown::EventMouseLeftDown(Mouse *m):
	Event(MouseLeftDown),
	x(m->x), y(m->y), z(m->z), buttons(m->buttons) {}

void EventMouseLeftDown::process()
{
	if (function)
	{	console->mouse->x = x;
		console->mouse->y = y;
		console->mouse->z = z;
		console->mouse->buttons = buttons;
		ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		function->eval_driver();
		Evaluator::val_status=vs;
	}
}

//EVENT_MOUSE(LeftDown)
EVENT_MOUSE(RightDown)
EVENT_MOUSE(MiddleDown)

EVENT_MOUSE(LeftUp)
EVENT_MOUSE(RightUp)
EVENT_MOUSE(MiddleUp)

EVENT_MOUSE(LeftDClick)
EVENT_MOUSE(RightDClick)
EVENT_MOUSE(MiddleDClick)

EVENT_MOUSE(Motion)
//___________________________________________________________________

#define EVENT_GRAPH_MOUSE(NAME)								\
EventGraphMouse##NAME::EventGraphMouse##NAME(Graph *g, Mouse *m):\
	EventGraph(g, GraphMouse##NAME),			\
	x(m->x), y(m->y), buttons(m->buttons) {}				\
																		\
void EventGraphMouse##NAME::process()						\
{																		\
	if ((graph!=0) && graph->onMouse##NAME)				\
	{	console->mouse->x = x;									\
		console->mouse->y = y;									\
		console->mouse->z = z;									\
		console->mouse->buttons = buttons;					\
		ValStatus vs=Evaluator::val_status;					\
		Evaluator::val_status = no_value_allowed;			\
		graph->onMouse##NAME->eval_driver();				\
		Evaluator::val_status=vs;								\
	}																	\
}

EVENT_GRAPH_MOUSE(LeftDown)
EVENT_GRAPH_MOUSE(RightDown)
EVENT_GRAPH_MOUSE(MiddleDown)

EVENT_GRAPH_MOUSE(LeftUp)
EVENT_GRAPH_MOUSE(RightUp)
EVENT_GRAPH_MOUSE(MiddleUp)

EVENT_GRAPH_MOUSE(LeftDClick)
EVENT_GRAPH_MOUSE(RightDClick)
EVENT_GRAPH_MOUSE(MiddleDClick)

EVENT_GRAPH_MOUSE(Motion)
//___________________________________________________________________

EventSlider::EventSlider(SliderP slider_):
	Event(tSlider), slider(slider_) 
{}

void EventSlider::process()
{
	if ((slider!=0) && slider->onScroll)
	{	Event::sliderValue=slider->GetValue();
		ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		slider->onScroll->eval_driver();
		Evaluator::val_status=vs;
	}
}
//___________________________________________________________________

EventIntControl::EventIntControl(IntControlP intControl_):
	Event(tIntControl), intControl(intControl_) 
{}

void EventIntControl::process()
{
	if ((intControl!=0) && intControl->onChange)
	{	Event::intControlValue=intControl->GetValue();
		ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		intControl->onChange->eval_driver();
		Evaluator::val_status=vs;
	}
}
//___________________________________________________________________

EventFloatControl::EventFloatControl(FloatControlP floatControl_):
	Event(tFloatControl), floatControl(floatControl_) 
{}

void EventFloatControl::process()
{
	if ((floatControl!=0) && floatControl->onChange)
	{	Event::floatControlValue=floatControl->GetValue();
		ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		floatControl->onChange->eval_driver();
		Evaluator::val_status=vs;
	}
}
//___________________________________________________________________

EventButton::EventButton(ButtonP button_):
	Event(tButton), button(button_) 
{}

void EventButton::process()
{
	if ((button!=0) && button->onClick)
	{	ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		button->onClick->eval_driver();
		Evaluator::val_status=vs;
	}
}
//___________________________________________________________________

EventToggleButton::EventToggleButton(ToggleButtonP toggleButton_):
	Event(tToggleButton), toggleButton(toggleButton_) 
{}

void EventToggleButton::process()
{
	if ((toggleButton!=0) && toggleButton->onClick)
	{	ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		toggleButton->onClick->eval_driver();
		Evaluator::val_status=vs;
	}
}
//___________________________________________________________________

EventCheckBox::EventCheckBox(CheckBoxP checkBox_):
	Event(tCheckBox), checkBox(checkBox_) 
{}

void EventCheckBox::process()
{
	if ((checkBox!=0) && checkBox->onClick)
	{	ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		checkBox->onClick->eval_driver();
		Evaluator::val_status=vs;
	}
}
//___________________________________________________________________

EventRadioButton::EventRadioButton(RadioButtonP radioButton_):
	Event(tRadioButton), radioButton(radioButton_) 
{}

void EventRadioButton::process()
{
	if ((radioButton!=0) && radioButton->onClick)
	{	ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		radioButton->onClick->eval_driver();
		Evaluator::val_status=vs;
	}
}
//___________________________________________________________________

#define EVENT_TEXT_KEY(NAME)							\
																\
void EventText##NAME::process()						\
{																\
	if ((textControl!=0) && textControl->on##NAME)		\
	{	Event::keyboardValue=c;							\
		ValStatus vs=Evaluator::val_status;			\
		Evaluator::val_status = no_value_allowed;	\
		textControl->on##NAME->eval_driver();		\
		Evaluator::val_status=vs;						\
	}															\
}

EVENT_TEXT_KEY(Char)
EVENT_TEXT_KEY(KeyDown)
EVENT_TEXT_KEY(KeyUp)

void EventTextChange::process()
{
	if ((textControl!=0) && textControl->onChange)
	{	Event::textControlValue=textControl->GetValue();
		ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		textControl->onChange->eval_driver();
		Evaluator::val_status=vs;
	}
}

void EventTextEnter::process()
{
	if ((textControl!=0) && textControl->onEnter)
	{	Event::textControlValue=textControl->GetValue();
		ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		textControl->onEnter->eval_driver();
		Evaluator::val_status=vs;
	}
}
//___________________________________________________________________

#define EVENT_LB_KEY(NAME)								\
																\
void EventListBox##NAME::process()					\
{																\
	if ((listBox!=0) && listBox->on##NAME)				\
	{	Event::keyboardValue=c;							\
		Event::listBoxValue=listBox;					\
		ValStatus vs=Evaluator::val_status;			\
		Evaluator::val_status = no_value_allowed;	\
		listBox->on##NAME->eval_driver();			\
		Evaluator::val_status=vs;						\
	}															\
}

EVENT_LB_KEY(Char)
EVENT_LB_KEY(KeyDown)
EVENT_LB_KEY(KeyUp)

void EventListBoxSelect::process()
{
	if ((listBox!=0) && listBox->onSelect)
	{	Event::listBoxValue=listBox;
		ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		listBox->onSelect->eval_driver();
		Evaluator::val_status=vs;
	}
}

void EventListBoxDClick::process()
{
	if ((listBox!=0) && listBox->onDClick)
	{	Event::listBoxValue=listBox;
		ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		listBox->onDClick->eval_driver();
		Evaluator::val_status=vs;
	}
}
//___________________________________________________________________

#define EVENT_CB_KEY(NAME)								\
																\
void EventChoiceBox##NAME::process()				\
{																\
	if ((choiceBox!=0) && choiceBox->on##NAME)	\
	{	Event::keyboardValue=c;							\
		Event::choiceBoxValue=choiceBox;				\
		ValStatus vs=Evaluator::val_status;			\
		Evaluator::val_status = no_value_allowed;	\
		choiceBox->on##NAME->eval_driver();			\
		Evaluator::val_status=vs;						\
	}															\
}

EVENT_CB_KEY(Char)
EVENT_CB_KEY(KeyDown)
EVENT_CB_KEY(KeyUp)

void EventChoiceBoxSelect::process()
{
	if ((choiceBox!=0) && choiceBox->onSelect)
	{	Event::choiceBoxValue=choiceBox;
		ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		choiceBox->onSelect->eval_driver();
		Evaluator::val_status=vs;
	}
}
//___________________________________________________________________

#define EVENT_COMBOBOX_KEY(NAME)						\
void EventComboBox##NAME::process()					\
{																\
	if ((comboBox!=0) && comboBox->on##NAME)		\
	{	Event::keyboardValue=c;							\
		Event::comboBoxValue=comboBox->GetValue();\
		Event::staticComboBox=comboBox;				\
		ValStatus vs=Evaluator::val_status;			\
		Evaluator::val_status = no_value_allowed;	\
		comboBox->on##NAME->eval_driver();			\
		Evaluator::val_status=vs;						\
	}															\
}

EVENT_COMBOBOX_KEY(Char)
EVENT_COMBOBOX_KEY(KeyDown)
EVENT_COMBOBOX_KEY(KeyUp)

#define EVENT_CB(NAME)									\
void EventComboBox##NAME::process()					\
{																\
	if ((comboBox!=0) && comboBox->on##NAME)		\
	{	Event::comboBoxValue=comboBox->GetValue();\
		Event::staticComboBox=comboBox;				\
		ValStatus vs=Evaluator::val_status;			\
		Evaluator::val_status = no_value_allowed;	\
		comboBox->on##NAME->eval_driver();			\
		Evaluator::val_status=vs;						\
	}															\
}

EVENT_CB(Select)
EVENT_CB(Change)
EVENT_CB(Enter)
//___________________________________________________________________

EventListControlItemSelected::EventListControlItemSelected(
	ListControl *listControl_, wxListEvent event):
EventListControl(listControl_, ListControlItemSelected)
{
	row=event.GetIndex();
	col=event.GetColumn();
	text=event.GetText();
}

EventListControlItemActivated::EventListControlItemActivated(
	ListControl *listControl_, wxListEvent event):
EventListControl(listControl_, ListControlItemActivated)
{
	row=event.GetIndex();
	col=event.GetColumn();
	text=event.GetText();
}

EventListControlColClick::EventListControlColClick(
	ListControl *listControl_, wxListEvent event):
EventListControl(listControl_, ListControlColClick)
{
	row=event.GetIndex();
	col=event.GetColumn();
	text=event.GetText();
}

#define EVENT_LC_KEY(NAME)								\
																\
void EventListControl##NAME::process()				\
{																\
	if ((listControl!=0) && listControl->on##NAME)	\
	{	Event::keyboardValue=c;							\
		Event::listControlValue=listControl;		\
		ValStatus vs=Evaluator::val_status;			\
		Evaluator::val_status = no_value_allowed;	\
		listControl->on##NAME->eval_driver();		\
		Evaluator::val_status=vs;						\
	}															\
}

EVENT_LC_KEY(Char)
EVENT_LC_KEY(KeyDown)
EVENT_LC_KEY(KeyUp)

void EventListControlItemSelected::process()
{
	if ((listControl!=0) && listControl->onItemSelected)
	{	Event::listControlValue=listControl;
		Event::listControlRow=row;
		Event::listControlCol=col;
		Event::listControlText=text;
		ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		listControl->onItemSelected->eval_driver();
		Evaluator::val_status=vs;
	}
}

void EventListControlItemActivated::process()
{
	if ((listControl!=0) && listControl->onItemActivated)
	{	Event::listControlValue=listControl;
		Event::listControlRow=row;
		Event::listControlCol=col;
		Event::listControlText=text;
		ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		listControl->onItemActivated->eval_driver();
		Evaluator::val_status=vs;
	}
}

void EventListControlColClick::process()
{
	if ((listControl!=0) && listControl->onColClick)
	{	Event::listControlValue=listControl;
		Event::listControlRow=row;
		Event::listControlCol=col;
		Event::listControlText=text;
		ValStatus vs=Evaluator::val_status;
		Evaluator::val_status = no_value_allowed;
		listControl->onColClick->eval_driver();
		Evaluator::val_status=vs;
	}
}
//___________________________________________________________________

NodeP lKeyboardValue(PairPP)
{
	return new Int(Event::keyboardValue);
}

NodeP lOnChar(PairPP arg)
{
	PairP commandlist=arg[1]->a_runnable(1);
	if (!throwing)
	{	EventChar::function=commandlist;
		if (commandlist)
			console->charMode=true;
		else
			console->charMode=false;
	}
	return Unbound;
}

#define LOGO_ON_HANDLER(NAME)						\
NodeP lOn##NAME(PairPP arg)						\
{															\
	PairP commandlist=arg[1]->a_runnable(1);	\
	if (!throwing)										\
		Event##NAME::function=commandlist;		\
	return Unbound;									\
}

LOGO_ON_HANDLER(KeyDown)
LOGO_ON_HANDLER(KeyUp)

#define LOGO_ON_TM(NAME) LOGO_ON_HANDLER(TextMouse##NAME)

LOGO_ON_TM(LeftDown)
LOGO_ON_TM(RightDown)
LOGO_ON_TM(MiddleDown)

LOGO_ON_TM(LeftUp)
LOGO_ON_TM(RightUp)
LOGO_ON_TM(MiddleUp)

LOGO_ON_TM(LeftDClick)
LOGO_ON_TM(RightDClick)
LOGO_ON_TM(MiddleDClick)

LOGO_ON_TM(Motion)

#define LOGO_ON_M(NAME) LOGO_ON_HANDLER(Mouse##NAME)

LOGO_ON_M(LeftDown)
LOGO_ON_M(RightDown)
LOGO_ON_M(MiddleDown)

LOGO_ON_M(LeftUp)
LOGO_ON_M(RightUp)
LOGO_ON_M(MiddleUp)

LOGO_ON_M(LeftDClick)
LOGO_ON_M(RightDClick)
LOGO_ON_M(MiddleDClick)

LOGO_ON_M(Motion)
