//___________________________________________________________________
//
//                        Class FloatControl
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>
#include <wx/spinctrl.h>

#include "spinctld.h"

#include "Main.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "Color.h"
#include "Float.h"
#include "Error.h"
#include "Eval.h"
#include "Stack.h"

#include "Event.h"
#include "FloatControl.h"
#include "Frame.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"
#include "Console.h"
#include "Point.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(FloatControl, wxSpinCtrlDbl)
	EVT_SPINCTRL	(-1, FloatControl::OnChange)
END_EVENT_TABLE()

FloatControl::FloatControl(FloatControlNodeP floatControlNode_, int style):
	floatControlNode(floatControlNode_)
{
	wxWindow *parent=console;
	if (floatControlNode->parentFrame)
		parent=floatControlNode->parentFrame->frame;
	else if (floatControlNode->parentGraph)
		parent=floatControlNode->parentGraph->graph;
	Create(parent, -1, 
			wxEmptyString, 
			wxDefaultPosition, 
			wxDefaultSize, 
			style, 
			1,1,1,1,-1,
			"FloatControl");
}

FloatControl::FloatControl(FloatControlNodeP floatControlNode_, 
			wxWindow* parent, wxWindowID id, 
			double value, double minValue, double maxValue, double increment,
			int digits, int format_,
			PairP onChange_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
wxSpinCtrlDbl(*parent, id, 
			wxEmptyString,
			pos, size, style, 
			minValue, maxValue, value, increment, digits,
			name),
onChange(onChange_),
floatControlNode(floatControlNode_)
{
	format=format_;
	SetDigits(digits, formatType(format));
}

FloatControl::~FloatControl()
{
	if (floatControlNode)
		floatControlNode->floatControl=0;
}

void FloatControl::OnChange(wxSpinEvent &event)
{
	Event::fifo << new EventFloatControl(this);
	event.Skip();
}

wxString FloatControl::wxstring()
{
	wxString str;
	str << "FloatControl " << GetMin() << ' ' << GetValue() << ' ' << GetMax() << ' ' 
		<< GetIncrement() << ' ' << GetDigits() << ' ' << format << ' '
		<< '[' << GetName() << "] ";
	if (onChange)
		str << onChange->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

wxDataInputStream &FloatControl::read(wxDataInputStream &istr) 
{	
	double value, min, max, increment;
	int digits;
	wxPoint pos;
	wxSize size;
	wxString name;
	
	istr >> min >> value >> max >> increment >> digits >> format
		>> name >> (int&)onChange
		>> pos.x >> pos.y >> size.x >> size.y;

	SetRange(min, max);
	SetValue(value);
	SetIncrement(increment, digits);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetDigits(digits, formatType(format));
	return istr;
}
wxDataOutputStream &FloatControl::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetMin() << GetValue() << GetMax() << GetIncrement() << GetDigits()
		<< format << GetName() 
		<< GC_::ptrToInt->find(onChange)
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y;
	return ostr;
}

PairP FloatControlNode::onChange()
{
	if (floatControl)
		return floatControl->onChange;
	return 0;
}

void FloatControlNode::setonChange(NodeP nd) 
{
	if (floatControl)
	{	floatControl->onChange = PairP(nd);  check_oldyoung(nd);
	}
}

void FloatControlNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void FloatControlNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

FloatControlNode::FloatControlNode(): 
	Node(FLOATCONTROL), floatControl(0), parentFrame(0), parentGraph(0) {}

FloatControlNode::FloatControlNode(FrameNodeP parentFrame_, 
										 GraphNodeP parentGraph_): 
	Node(FLOATCONTROL), floatControl(0), 
	parentFrame(parentFrame_),
	parentGraph(parentGraph_) 
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void FloatControlNode::free() 
{	
	if (floatControl) 
	{	floatControl->floatControlNode=0;
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		floatControl->Destroy(); 
		EventPP out=Event::fifo.out;
		EventP ev;
		while (Event::fifo)
		{	Event::fifo >> ev;
			switch (ev->type)
			{
			case Event::tFloatControl:
				if (((EventFloatControl*)ev)->floatControl==floatControl)
					((EventFloatControl*)ev)->floatControl=0;
				break;
			default:
				;
			}
		}
		Event::fifo.out=out;
		floatControl=0;
	}
}
void FloatControlNode::inc_gc() {fifo << onChange();}
int FloatControlNode::youngs() {return gmark(onChange());}
int FloatControlNode::youngsPtr() {return gmarkPtr(onChange());}
void FloatControlNode::check_oldyoungs() {check_oldyoung(onChange());}

wxDataInputStream &FloatControlNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	FrameNodeP parentFrameNode;
		GraphNodeP parentGraphNode;
		istr >> (int&)parentFrameNode >> (int&)parentGraphNode >> style;
		floatControl=new FloatControl(FloatControlNodeP(this), style);
		floatControl->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}

wxDataOutputStream &FloatControlNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (floatControl)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame) 
			<< GC_::ptrToInt->find(parentGraph) 
			<< style;
		floatControl->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void FloatControlNode::setPtrs()
{
	parentFrame=(FrameNode*)GC_::intToPtr->find(int(parentFrame));
	parentGraph=(GraphNode*)GC_::intToPtr->find(int(parentGraph));
	if (floatControl)
		setonChange((Pair*)GC_::intToPtr->find(int(floatControl->onChange)));
}

void FloatControlNode::recreate()
{
	wxWindow *p=0;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	FloatControlP f=floatControl, f2=0;
	f2=new FloatControl(FloatControlNodeP(this),
			p, -1, 
			f->GetValue(), f->GetMin(), f->GetMax(), f->GetIncrement(),
			f->GetDigits(), f->format,
			f->onChange, f->GetPosition(), f->GetSize(), style, 
			wxDefaultValidator, f->GetName());
	floatControl=f2;
	setonChange(f->onChange);
	delete f;
	floatControl=f2;
}

StringP FloatControlNode::string()
{
	if (floatControl)
		return new String(floatControl->wxstring());
	return _Null_Word->name();
}

wxString FloatControlNode::wxstring()
{
	if (floatControl)
		return floatControl->wxstring();
	return wxEmptyString;
}

NodeP lFloatControl(PairPP arg)
{
	wxWindowID id=-1;
	wxWindow *parent=graph;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=theApp->mainFrame->graphNode;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}
	wxString name=_T("FloatControl");
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	double minValue;
	{	double m=arg[3]->a_float(3);
		if throwing return Unbound;
		minValue=m;	
	}
	double value;
	{	double v=arg[4]->a_float(4);
		if throwing return Unbound;
		value=v;	
	}
	double maxValue;
	{	double m=arg[5]->a_float(5);
		if throwing return Unbound;
		maxValue=m;	
	}
	double increment;
	{	double i=arg[6]->a_float(6);
		if throwing return Unbound;
		increment=i;	
	}
	int digits=3;
	if (exist(7))
	{	int d=arg[7]->a_int(7);
		if throwing return Unbound;
		digits=d;	
	}
	PairP onChange=0;
	if (exist(8))
	{	PairP os=arg[8]->a_runnable(8);
		if throwing return Unbound;
		onChange=os;
	}
	int format=wxSpinCtrlDbl::lg_fmt;
	if (exist(9) && arg[9])
	{	int f=arg[9]->a_int(9);
		if throwing return Unbound;
		format=f;
	}
	long style = wxSP_ARROW_KEYS;
	if (exist(10) && arg[10])
	{	int s=arg[10]->a_int(10);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(11))
	{	PairP p=arg[11]->a_int_vector(11);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(12))
	{	PairP s=arg[12]->a_int_vector(12);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	FloatControlNodeP floatControlNode=new FloatControlNode(
		parentFrameNode, 
		parentGraphNode);
	FloatControlP floatControl=new FloatControl(floatControlNode,
			parent, id, 
			value, minValue, maxValue, increment, digits, format, 
			onChange, pos, size, style, wxDefaultValidator, name);
	floatControlNode->floatControl=floatControl;
	floatControlNode->style=style;
	return floatControlNode;
}

NodeP lFloatControlValue(PairPP arg)
{
	if (exist(1))
	{	FloatControlNodeP floatControl=arg[1]->a_floatControl(1);
		if throwing return Unbound;
		return new Float(floatControl->floatControl->GetValue());
	}
	else
		return new Float(Event::floatControlValue);
}

NodeP lFloatControlSetValue(PairPP arg)
{
	FloatControlNodeP floatControl=arg[1]->a_floatControl(1);

	if throwing return Unbound;
	double value=arg[2]->a_float(2);
	if throwing return Unbound;
	floatControl->floatControl->SetValue(value);
	return Unbound;
}

NodeP lFloatControlSetRange(PairPP arg)
{
	FloatControlNodeP floatControl=arg[1]->a_floatControl(1);
	if throwing return Unbound;
	double min=arg[2]->a_float(2);
	if throwing return Unbound;
	double max=arg[3]->a_float(3);
	if throwing return Unbound;
	floatControl->floatControl->SetRange(min, max);
	return Unbound;
}

NodeP lFloatControlOnChange(PairPP arg)
{
	FloatControlNodeP floatControl=arg[1]->a_floatControl(1);
	if throwing return Unbound;
	PairP function=arg[2]->a_runnable(2);
	if throwing return Unbound;
	floatControl->setonChange(function);
	return Unbound;
}

NodeP lFloatControlEnable(PairPP arg)
{
	FloatControlNodeP floatControl=arg[1]->a_floatControl(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	floatControl->floatControl->Enable(state!=0);
	return Unbound;
}

NodeP lFloatControlDestroy(PairPP arg)
{
	FloatControlNodeP floatControl=arg[1]->a_floatControl(1);
	if throwing return Unbound;
	{	FrameNodeP p=floatControl->parentFrame;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(floatControl));
			p->check_oldyoung(p->children);
		}
	}
	{	GraphNodeP p=floatControl->parentGraph;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(floatControl));
			p->check_oldyoung(p->children);
		}
	}
	floatControl->floatControl->Destroy();
	return Unbound;
}
