#ifndef FiFo_included
	#include "FiFo.h"
#endif

class PtrToInt;
class IntToPtr;
class PtrToString;
class StringToPtr;
class wxDataInputStream;
class wxDataOutputStream;

class GC_		// The Garbage Collector
{
	protected:
		NodeP
			_next,			   // node list of one generation
			oldyoung_next;	// list containing nodes that have childs

		int _marked,  // in mark marked is set to the gc run number
			gen,        // node's Generation
			age;      // How many times to GC at this generation



	public:

#ifdef _DEBUG
		const char* newFilename;
		int newLine;

		#define operator_new	void *operator new(size_t size, int blocktype, const char* newFilename_, int newLine_)

		#define alloc_dbg_info	, int blockType, const char *filename, int linenumber
		#define  malloc_GC	 _malloc_dbg(                size_t size alloc_dbg_info)
		#define  calloc_GC	 _calloc_dbg(size_t num,     size_t size alloc_dbg_info)
		#define realloc_GC	_realloc_dbg(void *memblock, size_t size alloc_dbg_info)
#else

		#define operator_new	void *operator new(size_t size)

		#define  malloc_GC	 malloc(size_t size)
		#define  calloc_GC	 calloc(size_t num, size_t size)
		#define realloc_GC	realloc(void *memblock, size_t size)

#endif

#ifdef PTR_DEBUG
		GC_(): _next(_next), oldyoung_next(oldyoung_next) {}
#else
		GC_() {}
#endif
		NodeP next_() {return _next;}
		int marked() {return _marked;}
		void setnext(NodeP nd) {_next=nd;}
		void setoldyoung_next(NodeP nd) {oldyoung_next=nd;}

		virtual void free()=0;
		virtual void inc_gc()=0;
		virtual int youngs()=0;
		virtual void check_oldyoungs()=0;

		static int run_numero() {return _run_numero;}

		void mark();	// Iterative mark procedure
		void markPtr();// mark and insert ptr into ptrToInt table

		void check_oldyoung(NodeP new_); 
		int gmark(NodeP child), 
			gmarkPtr(NodeP child);

		virtual wxDataInputStream &read(wxDataInputStream &istr)=0;
		virtual wxDataOutputStream &write(wxDataOutputStream &ostr)=0;
		virtual void setPtrs()=0;
		virtual void aftersetPtrs()=0;
		virtual void recreate()=0;
		virtual void aftersetPtrs2()=0;
		virtual void aftersetPtrs3()=0;

		static void 
			init(), run(bool full),	// collect garbage

			* malloc_GC,
			* calloc_GC,
			*realloc_GC,

			reserve_tank_fill (),
			reserve_tank_check(), 
			reserve_tank_use  (); 

		static void set_C_Stack(int stacksize, int *stack, int b);

		static void load(charP name);
		static void save(charP name);

		static bool initialized;
		static const int generations;

		static int volatile is_running, disabled;

		static PtrToInt *ptrToInt;
		static IntToPtr *intToPtr;

		static PtrToString *primitivesPtrToString;
		static StringToPtr *primitivesStringToPtr;

	protected:
		static const int 
			segment_size,
			fifo_size,
			min_free;

		static int  
			mem_free, mem_used, mem_peak, 

			mark_gen,  max_gen,
			_run_numero,
			examined;	// for GC_DEBUG

		enum Interrupt {no_interrupt, pause, stop};
		static Interrupt interrupt;

		static NodeP 
			free_node,     // first of the free nodes list
			oldyoungs,	 // list of nodes that point to younger nodes
			SpareNode;

		static PairP reserve_tank;	// list of nodes not in any generation
		static NodeP *bottom_stack;	// points to the bottom of the C program stack. See begin of main!
		static FiFo fifo;	// for resolving recursive mark() calls, which could be quite deep

		static void _cdecl mark_many(NodeP arg1, ...);
		static void _cdecl mark_manyPtr(NodeP arg1, ...);
		static void _cdecl find_many(intP manyPtrNrs, NodePP arg1, ...);
		static void mark_C_stack();	// Check C Stack for Node pointers and mark valid ones
		static void mark_C_stackPtr();	// Check C Stack for Node pointers and mark valid ones

		static NodePP mark_oldyoung(NodePP prev);	// Mark/traverse pointers to younger generations only
		static NodePP mark_oldyoungPtr(NodePP prev);	// Mark pointers for load
		static void mark_oldyoungs();	// mark pointers from old generations to young (the oldyoungs list)
		static void mark_oldyoungsPtr();	// mark oldyoungs list for load

		static void mark_phase(bool clean_all);	// mark the root (all accessable node pointers of the generations<=mark_gen)
		static void mark_phasePtr();	// mark the root (all accessable node pointers of the generations<=mark_gen)
		static int sweep_phase(int gen);	// promote marked to next gen, free the rest

		static void gc(bool full);	// Garbage Collector

		static void inc_mem_nodes();
		static NodeP newnode();

	friend class Segment;
	friend class IntToPtr;
	friend class PtrToInt;
	friend class MemoryWindow;

	friend NodeP lNodes(PairPP);

	friend void __cdecl logo_stop(int);
	friend void __cdecl logo_pause(int);
	friend int __cdecl _main(int argc, charP argv[]);
};

#define GC_included
