//___________________________________________________________________
//
//                           Class  Gauge
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/gauge.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>

#include "Main.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "Color.h"
#include "Float.h"
#include "Point.h"
#include "Error.h"
#include "Eval.h"
#include "Stack.h"

#include "Event.h"
#include "Gauge.h"
#include "Frame.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"
#include "Console.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

//___________________________________________________________________

class MyGauge : public wxGauge
{
public:
	MyGauge(wxWindow* parent, wxWindowID id, 
		int range, 
		const wxPoint& pos=wxDefaultPosition, 
		const wxSize& size=wxDefaultSize, 
		long style = 0, 
		const wxValidator& validator = wxDefaultValidator, 
		const wxString& name = "Gauge");

	// MSW sends extra kill focus event
	virtual ~MyGauge()
	{
		if (parent) 
			parent->gauge = 0;
		parent=0;
	}

	Gauge *parent;
};

MyGauge::MyGauge(wxWindow *parent_, wxWindowID id,
							  int range,
							  const wxPoint &pos, const wxSize &size,
							  long style,
							  const wxValidator& validator,
							  const wxString &name):
wxGauge(parent_, id, range, pos, size, style, validator, name)
{
	parent = (Gauge*)parent_;
}
//___________________________________________________________________

Gauge::Gauge(GaugeNodeP gaugeNode_, int style):
	label(0), gauge(0), bestSize(-999, -999),
	gaugeNode(gaugeNode_)
{
	wxWindow *parent=console;
	if (gaugeNode->parentFrame)
		parent=gaugeNode->parentFrame->frame;
	else if (gaugeNode->parentGraph)
		parent=gaugeNode->parentGraph->graph;
	create(parent, -1, 
			1, 
			wxDefaultPosition, 
			wxDefaultSize, 
			style, 
			wxDefaultValidator, 
			"Gauge");
}

Gauge::Gauge(GaugeNodeP gaugeNode_, 
			wxWindow* parent, wxWindowID id, 
			int range, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
label(0), gauge(0), bestSize(-999, -999), 
gaugeNode(gaugeNode_)
{
	bestSize = wxSize(-999, -999);
	gaugeNode->gauge=this;
	gaugeNode->style=style;
	create(parent, -1, 
			range, 
			wxDefaultPosition, 
			wxDefaultSize, 
			style, 
			wxDefaultValidator, 
			name);
}

bool Gauge::create(wxWindow* parent, wxWindowID id, 
			int range, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name)
{
	if (!wxControl::Create(parent, id, pos, size, style|wxNO_BORDER,
								wxDefaultValidator, name))
	  return false;

	wxControl::SetLabel(name);
	wxControl::SetBackgroundColour(parent->GetBackgroundColour());
	wxControl::SetForegroundColour(parent->GetForegroundColour());

	int width = size.GetWidth(), height = size.GetHeight();

	wxSize best_size(DoGetBestSize());
	if (width  == -1) width  = best_size.GetWidth();
	if (height == -1) height = best_size.GetHeight();

	if (name != wxEmptyString)
		label = new wxStaticText(this, id, name, 
			wxPoint(0, 0), wxDefaultSize, 0, "staticText");
	else
		label=0;

	if (label)
		gauge = new MyGauge(this, id, range,
			wxPoint(0,0),
			wxSize(width, height-label->GetSize().GetHeight()),
			style, validator, name);
	else
		gauge = new MyGauge(this, id,range,
			wxPoint(0,0),
			wxSize(width, height), style, validator, name);

	DoSetSize(pos.x, pos.y, width, height);
	SetBestFittingSize(wxSize(width, height));
	return true;
}

void Gauge::DoSetSize(int x, int y, int width, int height, int sizeFlags)
{
	wxSize bestSize(DoGetBestSize());
	if (width < 0)  width  = bestSize.GetWidth();
	if (height < 0) height = bestSize.GetHeight();

	wxWindow::DoSetSize(x, y, width, height, sizeFlags);

	int labelwidth  = 0;
	int labelheight = 0;
	if (label)
		label->GetSize(&labelwidth, &labelheight);

	if (gauge)   
		gauge->SetSize(0, labelheight, 
			width, height-labelheight-2);
}

wxSize Gauge::DoGetBestSize()
{
	if (bestSize.x == -999)
	{	wxGauge *t=new wxGauge((wxWindow*)this, wxID_ANY, 0,
			wxPoint(0,0), wxDefaultSize, GetWindowStyle());
		bestSize=t->GetBestSize();
		delete t;
		if (GetLabel()!=wxEmptyString)
		{	wxStaticText label((wxWindow*)this, wxID_ANY, GetLabel());
			wxSize ls=label.GetBestSize();
			bestSize.y+=ls.y;
			if (ls.x > bestSize.x)
				bestSize.x=ls.x;
		}
	}
	return bestSize;
}

Gauge::~Gauge()
{
	if (gaugeNode)
		gaugeNode->gauge=0;
	if (label)
	{	delete label;
		label=0;
	}
	if (gauge)
	{	delete gauge;
		gauge=0;
	}
}

wxString Gauge::wxstring()
{
	wxString str;
	if (gauge)
		str 
		<< "Gauge " << gauge->GetRange() << ' ' << gauge->GetValue() << ' ' 
		<< '[' << GetName() << "] "
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

wxDataInputStream &Gauge::read(wxDataInputStream &istr) 
{	
	int range, value, style;
	wxPoint pos;

	wxSize size;
	wxString name;
	unsigned char fr,fg,fb, br,bg,bb;
	
	istr >> range >> value >> name
		>> pos.x >> pos.y >> size.x >> size.y >> style
		>> fr >> fg >> fb 
		>> br >> bg >> bb;

	gauge->SetRange(range);
	gauge->SetValue(value);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetWindowStyle(style);
	SetForegroundColour(wxColour(fr,fg,fb));
	SetBackgroundColour(wxColour(br,bg,bb));
	return istr;
}
wxDataOutputStream &Gauge::write(wxDataOutputStream &ostr) 
{
	int range=0, value=0;
	if (gauge)
	{	range=gauge->GetRange();
		value=gauge->GetValue();
	}
	ostr
		<< range << value << GetName() 
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y << int(GetWindowStyle());

	{	wxColour 
			f=gauge->GetForegroundColour(), 
			b=gauge->GetBackgroundColour();
		ostr
			<< f.Red() << f.Green() << f.Blue() 
			<< b.Red() << b.Green() << b.Blue();
	}
	return ostr;
}

bool Gauge::SetBackgroundColour(const wxColour &c)
{
	if (gauge)
		return gauge->SetBackgroundColour(c);
	return false;
}

bool Gauge::SetForegroundColour(const wxColour &c)
{
	if (gauge)
		return gauge->SetForegroundColour(c);
	return false;
}

void GaugeNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void GaugeNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

GaugeNode::GaugeNode(): 
	Node(GAUGE), gauge(0), 
	parentFrame(0),
	parentGraph(0)
{}

GaugeNode::GaugeNode(FrameNodeP parentFrame_,
							GraphNodeP parentGraph_): 
	Node(GAUGE), gauge(0), 
	parentFrame(parentFrame_),
	parentGraph(parentGraph_)
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void GaugeNode::free() 
{	
	if (gauge) 
	{	gauge->gaugeNode=0;
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		gauge->Destroy(); 
		gauge=0;
	}
}

wxDataInputStream &GaugeNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	FrameNodeP parentFrameNode;
		GraphNodeP parentGraphNode;
		istr 
			>> (int&)parentFrameNode 
			>> (int&)parentGraphNode 
			>> style;
		gauge=new Gauge(GaugeNodeP(this), style);
		gauge->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}

wxDataOutputStream &GaugeNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (gauge)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame) 
			<< GC_::ptrToInt->find(parentGraph) 
			<< style;
		gauge->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void GaugeNode::setPtrs()
{
	parentFrame=(FrameNode*)GC_::intToPtr->find(int(parentFrame));
	parentGraph=(GraphNode*)GC_::intToPtr->find(int(parentGraph));
}

void GaugeNode::recreate()
{
	wxWindow *p=0;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	int range=0, value=0;
	if (gauge->gauge)
	{	range=gauge->gauge->GetRange();
		value=gauge->gauge->GetValue();
	}
	GaugeP g=gauge, 
		g2=new Gauge(GaugeNodeP(this),
			p, -1, range, 
			g->GetPosition(), g->GetSize(), style, 
			wxDefaultValidator, g->GetName());
	g2->gauge->SetRange(range);
	g2->gauge->SetValue(value);
	g2->gauge->SetForegroundColour(g->gauge->GetForegroundColour());
	g2->gauge->SetBackgroundColour(g->gauge->GetBackgroundColour());
	gauge=g2;
	delete g;
	gauge=g2;
}

StringP GaugeNode::string()
{
	if (gauge)
		return new String(gauge->wxstring());
	return _Null_Word->name();
}

wxString GaugeNode::wxstring()
{
	if (gauge)
		return gauge->wxstring();
	return wxEmptyString;
}

NodeP lGauge(PairPP arg)
{
	wxWindowID id=-1;

	wxWindow *parent=graph;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=theApp->mainFrame->graphNode;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}
	wxString name=_T("Gauge");
	if (arg[2])
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	int range;
	{	int r=arg[3]->a_int(3);
		if throwing return Unbound;
		range=r;	
	}
	int value=0;
	if (exist(4) && arg[4])
	{	int v=arg[4]->a_int(4);
		if throwing return Unbound;
		if (v > range)
		{	error(BAD_DATA, arg[4], 4);
			return Unbound;
		}
		value=v;	
	}
	long style = wxGA_HORIZONTAL;
	if (exist(5) && arg[5])
	{	int s=arg[5]->a_int(5);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(6) && arg[6])
	{	PairP p=arg[6]->a_int_vector(6);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(7))
	{	PairP s=arg[7]->a_int_vector(7);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	GaugeNodeP gaugeNode=new GaugeNode(
		parentFrameNode,
		parentGraphNode);
	GaugeP gauge=new Gauge(gaugeNode,
			parent, id, range,
			pos, size, style, wxDefaultValidator, name);
	gauge->gauge->SetValue(value);
	return gaugeNode;
}

NodeP lGaugeValue(PairPP arg)
{
	GaugeNodeP gauge=arg[1]->a_gauge(1);
	if throwing return Unbound;
	return new Int(gauge->gauge->gauge->GetValue());
}

NodeP lGaugeSetValue(PairPP arg)
{
	GaugeNodeP gauge=arg[1]->a_gauge(1);
	if throwing return Unbound;
	int value=arg[2]->a_int(2);
	if throwing return Unbound;
	gauge->gauge->gauge->SetValue(value);
	return Unbound;
}

NodeP lGaugeSetRange(PairPP arg)
{
	GaugeNodeP gauge=arg[1]->a_gauge(1);
	if throwing return Unbound;
	int range=arg[2]->a_int(2);

	if throwing return Unbound;
	gauge->gauge->gauge->SetRange(range);
	return Unbound;
}

NodeP lGaugeSetBackgroundColor(PairPP arg)
{
	GaugeNodeP gauge=arg[1]->a_gauge(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
	gauge->gauge->SetBackgroundColour(wc);
	return Unbound;
}

NodeP lGaugeSetColor(PairPP arg)
{
	GaugeNodeP gauge=arg[1]->a_gauge(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
	gauge->gauge->SetForegroundColour(wc);
	return Unbound;
}

NodeP lGaugeDestroy(PairPP arg)
{
	GaugeNodeP gauge=arg[1]->a_gauge(1);
	if throwing return Unbound;
	{	FrameNodeP p=gauge->parentFrame;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(gauge));
			p->check_oldyoung(p->children);
		}
	}
	{	GraphNodeP p=gauge->parentGraph;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(gauge));
			p->check_oldyoung(p->children);
		}
	}
	gauge->gauge->Destroy();
	return Unbound;
}
