//___________________________________________________________________
//
//                         ListBox Header
//___________________________________________________________________


extern Function 
	lListBox, lListBoxDestroy,
	lListBoxSelections, lListBoxSetSelections,
	lListBoxSetChoices, 
	lListBoxAppend, 
	lListBoxSetItem, lListBoxRemoveItem, 
	lListBoxCount,
	lListBoxSetBackgroundColor, lListBoxSetColor,
	lListBoxSetFontSize, lListBoxSetFontName,
	lListBoxSetFontStyle, lListBoxSetFontWeight,
	lListBoxOnChar, lListBoxOnKeyDown, lListBoxOnKeyUp,
	lListBoxOnSelect, lListBoxOnDClick,
	lListBoxEnable;

class wxStaticText;
class MyListBox;

class ListBox_: public wxControl
{
	public:
		wxStaticText *label;
		MyListBox *listBox;
		wxSize bestSize;
		wxArrayString choices;

		PairP onSelect, onDClick,
			onChar, onKeyUp, onKeyDown;
		ListBoxNodeP listBoxNode;

		ListBox_(ListBoxNodeP listBoxNode_, int style);
		ListBox_(ListBoxNodeP listBoxNode_,
			wxWindow* parent, wxWindowID id, 
			const wxArrayString& choices_, 
			PairP onSelect_=PairP(0),
			PairP onDClick_=PairP(0),
			const wxPoint& pos=wxDefaultPosition, 
			const wxSize& size=wxDefaultSize, 
			long style = 0, 
			const wxValidator& validator = wxDefaultValidator, 
			const wxString& name = "ListBox");

		bool create(wxWindow* parent, wxWindowID id, 
			const wxArrayString& choices_, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name);

		~ListBox_();

		virtual void DoSetSize(int x, int y, int width, int height,
						   int sizeFlags = wxSIZE_AUTO);

		virtual wxSize DoGetBestSize();

		virtual void Set(wxArrayString &choices);

		void 
			OnClose	(wxCloseEvent &event),
			OnSelect	(wxCommandEvent &event),
			OnDClick	(wxCommandEvent &event),
			OnChar	(wxKeyEvent &event),
			OnKeyDown(wxKeyEvent &event),
			OnKeyUp	(wxKeyEvent &event);

		wxString wxstring();
		wxDataInputStream &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);

		virtual bool SetBackgroundColour(const wxColour& colour);
		virtual bool SetForegroundColour(const wxColour& colour);
		virtual bool SetFont(const wxFont& font);

		DECLARE_EVENT_TABLE()
};
typedef ListBox_ *ListBoxP;

class ListBoxNode: public Node
{
	public:
		ListBoxP listBox;
		FrameNodeP parentFrame;
		GraphNodeP parentGraph;
		int style;

		static const ListBoxNode aListBoxNode;

		operator_new;
		PairP onSelect(), onDClick(), 
			onChar(), onKeyDown(), onKeyUp();
		void
			setonSelect	(NodeP function),
			setonDClick	(NodeP function),
			setonChar	(NodeP function), 
			setonKeyDown(NodeP function),
			setonKeyUp	(NodeP function);

		ListBoxNode();
		ListBoxNode(FrameNodeP parentFrame_, GraphNodeP parentGraph_);

		void free();

		void inc_gc();
		int youngs();
		int youngsPtr();
		void check_oldyoungs();
		
		wxDataInputStream &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void setPtrs();
		void recreate();

		StringP string();
		wxString wxstring();
		StringP _typeof();

		void setParentFrame(FrameNodeP frame);
		void setParentGraph(GraphNodeP graph);
};
