#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/slider.h>
#include <wx/button.h>
#include <wx/tglbtn.h>
#include <wx/checkbox.h>
#include <wx/radiobut.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>
#include <wx/frame.h>
#include <wx/glcanvas.h>
#include <wx/sizer.h>

#include <stdio.h>
#include <string.h>

#include <m_apm.h>

#include "spinctld.h"
#include "spinctli.h"

#include "Main.h"

#include "String.h"
#include "Float.h"
#include "File.h"
#include "Pair.h"
#include "List.h"
#include "WindowPair.h"
#include "Object.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "Float.h"
#include "BigFloat.h"
#include "Point.h"
#include "Array.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "FloatArray.h"
#include "Table.h"
#include "Struct.h"
#include "Bitmap.h"
#include "Texture.h"
#include "TurtleNode.h"
#include "Line.h"
#include "Runparsd.h"
#include "Tree.h"
#include "Quote.h"
#include "Colon.h"
#include "Procedur.h"
#include "GraphicNode.h"
#include "DL.h"
#include "Slider.h"
#include "Gauge.h"
#include "IntControl.h"
#include "FloatControl.h"
#include "Button.h"
#include "ToggleButton.h"
#include "CheckBox.h"
#include "RadioButton.h"
#include "StaticText.h"
#include "TextControl.h"
#include "ListBox.h"
#include "ChoiceBox.h"
#include "ComboBox.h"
#include "ListControl.h"
#include "Frame.h"
#include "Graph.h"
#include "BoxSizer.h"

#include "Error.h"
#include "Eval.h"
#include "Stack.h"
#include "Init.h"
#include "Coms.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

NodeP Node::deepCopy() 
{	NodeType t=type;
	return this;
}
NodeP Pair::deepCopy()
{	if (!this) 
		return 0;
	NodeP f=0, t=0;
	if (first())
		f=first()->deepCopy();
	if (tail())
		t=tail()->deepCopy();
	return new Pair(f, t);
}
NodeP Array::deepCopy()
{	ArrayP result=new Array(length(), origin());
	NodePP source=data(), dest=result->data();
	for (int i=length();  i;  i--)
	{	NodeP s=*source++;
		if (s)
			*dest = s->deepCopy();
		dest++;
	}
	return result;
}
// TODO: Table::deepCopy(), Struct::deepCopy()
NodeP ldeepCopy(PairPP arg) 
{
	if (!arg[1])
		return 0;
	return arg[1]->deepCopy();
}

NodeP Node::first_()
{	return string()->first();
}
NodeP String::first_() 
{	return first();
}
NodeP Pair::first_() 
{	return first();
}
NodeP Array::first_() 
{	if (length() >= 1) 
		return *(ArrayP(this)->data());
	else
		return 0;
}
NodeP FloatArray::first_() 
{	if (length() >= 1) 
		return new Float(*(FloatArrayP(this)->data()));
	else
		return 0;
}
NodeP IntArray::first_() 
{	if (length() >= 1) 
		return new Int(*(IntArrayP(this)->data()));
	else
		return 0;
}
NodeP Int16Array::first_() 
{	if (length() >= 1) 
		return new Int16(*(Int16ArrayP(this)->data()));
	else
		return 0;
}

NodeP lfirst(PairPP arg) 
{
	if (!arg[1]) 
	{	error(BAD_DATA, arg[1]); return Unbound;
	}
	return arg[1]->first_();
}

NodeP Node::butFirst()
{	return string()->butFirst();
}
NodeP String::butFirst()
{	if (length() > 1)
		return make(ptr()+1, head(), length()-1, type, copy);
	else
		return _Null_Word;
}
NodeP Pair::butFirst() 
{	return tail();
}
NodeP Array::butFirst()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	return 0;
	}
	ArrayP result=new Array(l, o);
	NodePP source=data()+1, dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}
NodeP FloatArray::butFirst()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	FloatArrayP result=new FloatArray(l, o);
	doubleP source=data()+1, dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}
NodeP IntArray::butFirst()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	IntArrayP result=new IntArray(l, o);
	intP source=data()+1, dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}
NodeP Int16Array::butFirst()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	Int16ArrayP result=new Int16Array(l, o);
	shortP source=data()+1, dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP lbutFirst(PairPP arg) 
{
	if (!arg[1]) 
	{	error(BAD_DATA, arg[1]); return Unbound;
	}
	return arg[1]->butFirst();
}

NodeP l_setFirst(PairPP arg) 
{
	PairP	  list=arg[1];
	NodeP newval=arg[2];

	if (!list || !list->is.list) 
		list = error(BAD_DATA, list);

	if (!throwing) list->setfirst(newval);
	return Unbound;
}


NodeP l_setButFirst(PairPP arg) 
{
	PairP   list=arg[1];
	NodeP newval=arg[2];

	if (!list || !list->is.list)
		list = error(BAD_DATA,list);

	if (!throwing) list->settail(newval);
	return Unbound;
}

NodeP Node::last() 
{	StringP s=string();
	if (s!=Unbound)
		return s->last();
	error(BAD_DATA, this); return Unbound;
}
NodeP String::last()	
{	return make(ptr()+length()-1, head(), 1, type, copy);
}
NodeP Pair::last()
{	PairP list=this;
	while (list->tail()) 
	{	next(list);
		if (checkThrowing()) break;
	}
	return list->first();
}
NodeP Array::last()
{	return *(data()+length()-origin());
}
NodeP FloatArray::last()
{	return new Float(*(data()+length()-origin()));
}
NodeP IntArray::last()
{	return new Int(*(data()+length()-origin()));
}
NodeP Int16Array::last()
{	return new Int16(*(data()+length()-origin()));
}

NodeP llast(PairPP arg) 
{
	NodeP nd = arg[1];
	if (!nd) {error(BAD_DATA, 0); return Unbound;}
	return nd->last();
}

NodeP Node::butLast() 
{	StringP s=string();
	if (s!=Unbound)
		return s->butLast();
	error(BAD_DATA, this); return Unbound;
}
NodeP String::butLast()
{	if (length() > 1)
		return make(ptr(), head(), length()-1, type, copy);
	else
		return _Null_Word;
}
NodeP Pair::butLast()
{	PairP list=this, result=0, tnode=0, lastnode=0;

	while (list->tail()) 
	{	tnode = new Pair(list->first(), 0);
		if (!result) 
		{	result=tnode;
			lastnode=tnode;
		}else 
		{	lastnode->settail( tnode );
			lastnode=tnode;
		}
		next(list);
		if (checkThrowing()) break;
	}
	return result;
}
NodeP Array::butLast()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	ArrayP result=new Array(l, o);
	NodePP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}
NodeP FloatArray::butLast()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	FloatArrayP result=new FloatArray(l, o);
	doubleP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}
NodeP IntArray::butLast()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	IntArrayP result=new IntArray(l, o);
	intP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}
NodeP Int16Array::butLast()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	Int16ArrayP result=new Int16Array(l, o);
	shortP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP lbutLast(PairPP arg) 
{
	if (arg[1])
		return arg[1]->butLast();
	error(BAD_DATA, arg[1]);
	return Unbound;
}

NodeP Node::firsts() {error(BAD_DATA, this); return Unbound;}
NodeP Pair::firsts()
{	PairP tnode=0, val=0, li=this;
	if (li->first())
		tnode=val= new Pair(li->first()->first_(), 0);
	else
		tnode=val= new Pair(0, 0);

	for (li=li->tail();  li;  next(li)) 
	{
		if (li->first())
			tnode->settail( new Pair(li->first()->first_(), 0) );
		else
			tnode->settail( new Pair(0, 0) );
		next(tnode);
	}
	if throwing return Unbound;
	return val;
}
NodeP Array::firsts()
{	int l=length(), o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	ArrayP result=new Array(l, o);
	NodePP source=data(), dest=result->data();
	while (l--)
	{	if (*source)
			*dest++=(*source++)->first_();
		else
			*dest++=0;
	}
	return result;
}

NodeP lfirsts(PairPP arg) 
{
	if (!arg[1]) 
	{	error(BAD_DATA, arg[1]); return Unbound;
	}
	return arg[1]->firsts();
}

NodeP Node::butFirsts() {error(BAD_DATA, this); return Unbound;}
NodeP Pair::butFirsts()
{	PairP tnode=0, val=0, li=this;
	if (li->first_())
		tnode=val = new Pair(li->first_()->butFirst(), 0);
	else
		tnode=val = new Pair(0, 0);

	for (li=li->tail();  li;  next(li)) 
	{
		if (li->first_())
			tnode->settail( new Pair(li->first_()->butFirst(), 0) );
		else
			tnode->settail( new Pair(0, 0) );
		next(tnode);
	}
	if throwing return Unbound;
	return val;
}
NodeP Array::butFirsts()
{	int l=length(), o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	ArrayP result=new Array(l, o);
	NodePP source=data(), dest=result->data();
	while (l--)
	{	if (*source)
			*dest++=(*source++)->butFirst();
		else
			*dest++=0;
	}
	return result;
}

NodeP lbutFirsts(PairPP arg) 
{
	if (!arg[1]) 
	{	error(BAD_DATA, arg[1]); return Unbound;
	}
	return arg[1]->butFirsts();
}

NodeP Node::fPut(NodeP nd) 
{	return string()->fPut(nd);
}
NodeP Object::fPut(NodeP nd)
{	return name()->fPut(nd);
}
NodeP String::fPut(NodeP nd)
{	if (!nd)
	{	error(BAD_DATA, nd); return Unbound;
	}
	StringP str=nd->string();
	return new String(str, this);
}
NodeP Pair::fPut(NodeP nd)
{	return new Pair(nd, this);
}
NodeP Array::fPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	ArrayP result=new Array(l+1, o);
	NodePP source=data(), dest=result->data();
	*dest++=nd;
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}
NodeP FloatArray::fPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	FloatArrayP result=new FloatArray(l+1, o);
	doubleP source=data(), dest=result->data();
	*dest++=nd->getnumber();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}
NodeP IntArray::fPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	IntArrayP result=new IntArray(l+1, o);
	intP source=data(), dest=result->data();
	*dest++=Int::clip(nd->getnumber());
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}
NodeP Int16Array::fPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	Int16ArrayP result=new Int16Array(l+1, o);
	shortP source=data(), dest=result->data();
	*dest++=Int16::clip(nd->getnumber());
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP lfPut(PairPP arg) 
{
	if (!arg[2]) 
	{	return new Pair(arg[1], 0);
	}
	return arg[2]->fPut(arg[1]);
}

NodeP Node::lPut(NodeP nd) 
{	return string()->lPut(nd);
}
NodeP Object::lPut(NodeP nd)
{	return name()->lPut(nd);
}
NodeP String::lPut(NodeP nd)
{	if (!nd)
	{	error(BAD_DATA, nd); return Unbound;
	}
	StringP str=nd->string();
	return new String(this, str);
}
NodeP Pair::lPut(NodeP nd)
{	PairP lst=this, val=0, lastnode=0, tnode=0;
	while (lst) 
	{	tnode = new Pair(lst->first(), 0);
		if (!val) 
			val=tnode;
		else 
			lastnode->settail( tnode );
		lastnode=tnode;
		next(lst);
		if (checkThrowing()) break;
	}
	if (!val)
		return new Pair(nd, 0);
	else
	{	lastnode->settail( new Pair(nd, 0) );
		return val;
	}
}
NodeP Array::lPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	ArrayP result=new Array(l+1, o);
	NodePP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	*dest=nd;
	return result;
}
NodeP FloatArray::lPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	FloatArrayP result=new FloatArray(l+1, o);
	doubleP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	*dest=nd->getnumber();
	return result;
}
NodeP IntArray::lPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	IntArrayP result=new IntArray(l+1, o);
	intP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	*dest=Int::clip(nd->getnumber());
	return result;
}
NodeP Int16Array::lPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	Int16ArrayP result=new Int16Array(l+1, o);
	shortP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	*dest=Int16::clip(nd->getnumber());
	return result;
}

NodeP llPut(PairPP arg) 
{
	if (!arg[2]) 
		return new Pair(arg[1], 0);
	
	return arg[2]->lPut(arg[1]);
}

WindowPairP WindowPair::wlPut(NodeP nd)
{	WindowPairP lst=this, val=0, lastnode=0, tnode=0;
	while (lst) 
	{	tnode = new WindowPair(lst->first(), 0);
		if (!val) 
			val=tnode;
		else 
			lastnode->settail( tnode );
		lastnode=tnode;
		lst=WindowPairP(lst->tail());
		if (checkThrowing()) break;
	}
	if (!val)
		return new WindowPair(nd, 0);
	else
	{	lastnode->settail( new WindowPair(nd, 0) );
		return val;
	}
}

NodeP lList(PairPP arg) {return argStack.to_list(arg);}

NodeP lisempty(PairPP arg) 
{
	return Node::boolnode( !arg[1]  ||  arg[1]==NodeP(_Null_Word) );
}

NodeP lASCII(PairPP arg) 
{
	StringP str = arg[1]->a_char(1);
	if (!throwing) 
	{	unsigned char i = backslash.decode(*str->ptr());
		return new Int(i);
	}
	return Unbound;
}

NodeP lrawASCII(PairPP arg) 
{
	StringP str = arg[1]->a_char(1);
	if (!throwing) 
	{	unsigned char i = *str->ptr();
		return new Int(i);
	}
	return Unbound;
}

NodeP lisbackslashed(PairPP arg) 
{
	StringP str = arg[1]->a_char(1);
	if (!throwing) 
	{	char i = *str->ptr();
		return Node::boolnode(backslash.is(backslash.encode(i)));
	}
	return Unbound;
}

NodeP lBackslashEncode(PairPP arg) 
{
	StringP str = arg[1]->a_char(1);
	if (!throwing) 
	{	char i = backslash.encode(*str->ptr());
		return new String(&i, 0, 1, BACKSLASHED, String::copy);
	}
	return Unbound;
}

NodeP lChar(PairPP arg) 
{
	IntP i = arg[1]->a_pos_int(1);
	if (!throwing) 
	{	char c = char(i->int_);
		return new String(&c, 0, 1, STRING, String::copy);
	}
	return Unbound;
}

NodeP lisChar(PairPP arg) 
{
	if ((type_of(arg[1])==OBJECT 
		&& ObjectP(arg[1])->name()->length()==1)
	|| (arg[1]->isstring() 
		&& StringP(arg[1])->length()==1))
		return _True;
	return _False;
}

NodeP lisAlNum(PairPP arg) 
{
	if ((type_of(arg[1])==OBJECT 
	&& ObjectP(arg[1])->name()->length()==1))
		return Node::boolnode( isalnum(*ObjectP(arg[1])->name()->ptr())!=0 );

	if ((arg[1]->isstring() 
	&& StringP(arg[1])->length()==1))
		return Node::boolnode( isalnum(*StringP(arg[1])->ptr())!=0 );

	return _False;
}

NodeP lisAlpha(PairPP arg) 
{
	if ((type_of(arg[1])==OBJECT 
	&& ObjectP(arg[1])->name()->length()==1))
		return Node::boolnode( isalpha(*ObjectP(arg[1])->name()->ptr())!=0 );

	if ((arg[1]->isstring() 
	&& StringP(arg[1])->length()==1))
		return Node::boolnode( isalpha(*StringP(arg[1])->ptr())!=0 );

	return _False;
}

NodeP lisASCII(PairPP arg) 
{
	if ((type_of(arg[1])==OBJECT 
	&& ObjectP(arg[1])->name()->length()==1))
		return Node::boolnode( isascii(*ObjectP(arg[1])->name()->ptr())!=0 );

	if ((arg[1]->isstring() 
	&& StringP(arg[1])->length()==1))
		return Node::boolnode( isascii(*StringP(arg[1])->ptr())!=0 );

	return _False;
}

NodeP lisCntrl(PairPP arg) 
{
	if ((type_of(arg[1])==OBJECT 
	&& ObjectP(arg[1])->name()->length()==1))
		return Node::boolnode( iscntrl(*ObjectP(arg[1])->name()->ptr())!=0 );

	if ((arg[1]->isstring() 
	&& StringP(arg[1])->length()==1))
		return Node::boolnode( iscntrl(*StringP(arg[1])->ptr())!=0 );

	return _False;
}

NodeP lisCSym(PairPP arg) 
{
#ifdef WIN32
	if ((type_of(arg[1])==OBJECT 
	&& ObjectP(arg[1])->name()->length()==1))
		return Node::boolnode( iscsym(*ObjectP(arg[1])->name()->ptr())!=0 );

	if ((arg[1]->isstring() 
	&& StringP(arg[1])->length()==1))
		return Node::boolnode( iscsym(*StringP(arg[1])->ptr())!=0 );

	return _False;
#else
	error(NOT_YET_SUPPORTED, new String("CSym?"));
	return Unbound;
#endif
}

NodeP lisDigit(PairPP arg) 
{
	if ((type_of(arg[1])==OBJECT 
	&& ObjectP(arg[1])->name()->length()==1))
		return Node::boolnode( isdigit(*ObjectP(arg[1])->name()->ptr())!=0 );

	if ((arg[1]->isstring() 
	&& StringP(arg[1])->length()==1))
		return Node::boolnode( isdigit(*StringP(arg[1])->ptr())!=0 );

	return _False;
}

NodeP lisGraph(PairPP arg) 
{
	if ((type_of(arg[1])==OBJECT 
	&& ObjectP(arg[1])->name()->length()==1))
		return Node::boolnode( isgraph(*ObjectP(arg[1])->name()->ptr())!=0 );

	if ((arg[1]->isstring() 
	&& StringP(arg[1])->length()==1))
		return Node::boolnode( isgraph(*StringP(arg[1])->ptr())!=0 );

	return _False;
}

NodeP lisLower(PairPP arg) 
{
	if ((type_of(arg[1])==OBJECT 
	&& ObjectP(arg[1])->name()->length()==1))
		return Node::boolnode( islower(*ObjectP(arg[1])->name()->ptr())!=0 );

	if ((arg[1]->isstring() 
	&& StringP(arg[1])->length()==1))
		return Node::boolnode( islower(*StringP(arg[1])->ptr())!=0 );

	return _False;
}

NodeP lisPrint(PairPP arg) 
{
	if ((type_of(arg[1])==OBJECT 
	&& ObjectP(arg[1])->name()->length()==1))
		return Node::boolnode( isprint(*ObjectP(arg[1])->name()->ptr())!=0 );

	if ((arg[1]->isstring() 
	&& StringP(arg[1])->length()==1))
		return Node::boolnode( isprint(*StringP(arg[1])->ptr())!=0 );

	return _False;
}

NodeP lisPunct(PairPP arg) 
{
	if ((type_of(arg[1])==OBJECT 
	&& ObjectP(arg[1])->name()->length()==1))
		return Node::boolnode( ispunct(*ObjectP(arg[1])->name()->ptr())!=0 );

	if ((arg[1]->isstring() 
	&& StringP(arg[1])->length()==1))
		return Node::boolnode( ispunct(*StringP(arg[1])->ptr())!=0 );

	return _False;
}

NodeP lisSpace(PairPP arg) 
{
	if ((type_of(arg[1])==OBJECT 
	&& ObjectP(arg[1])->name()->length()==1))
		return Node::boolnode( isspace(*ObjectP(arg[1])->name()->ptr())!=0 );

	if ((arg[1]->isstring() 
	&& StringP(arg[1])->length()==1))
		return Node::boolnode( isspace(*StringP(arg[1])->ptr())!=0 );

	return _False;
}

NodeP lisUpper(PairPP arg) 
{
	if ((type_of(arg[1])==OBJECT 
	&& ObjectP(arg[1])->name()->length()==1))
		return Node::boolnode( isupper(*ObjectP(arg[1])->name()->ptr())!=0 );

	if ((arg[1]->isstring() 
	&& StringP(arg[1])->length()==1))
		return Node::boolnode( isupper(*StringP(arg[1])->ptr())!=0 );

	return _False;
}

NodeP lisxDigit(PairPP arg) 
{
	if ((type_of(arg[1])==OBJECT 

	&& ObjectP(arg[1])->name()->length()==1))
		return Node::boolnode( isxdigit(*ObjectP(arg[1])->name()->ptr())!=0 );

	if ((arg[1]->isstring() 
	&& StringP(arg[1])->length()==1))
		return Node::boolnode( isxdigit(*StringP(arg[1])->ptr())!=0 );

	return _False;
}

NodeP Node::upper() {return error(BAD_DATA, this);}
NodeP Object::upper() {return _name->upper();}
NodeP String::upper() {return make(ptr(), 0, length(), type, copy_cap);}
NodeP Array::upper() 
{	ArrayP newar=new Array(ArrayP(this));
	NodePP d=newar->_data;
	for (int i=_length;  i;  i--)
		*d++=(*d)->upper();
	return newar;
}
NodeP Pair::upper() 
{	PairP lst=this, result=0, lastnode=0, tnode=0;
	while (lst) 
	{	NodeP fst=lst->first();
		if (!fst)
			tnode=new Pair(0,0);
		else
			tnode = new Pair(fst->lower(), 0);
		if (!result) 
			result=tnode;
		else 
			lastnode->settail( tnode );
		lastnode=tnode;
		next(lst);
		if (checkThrowing()) break;
	}
	return result;
}

NodeP lupperCase(PairPP arg) 
{
	NodeP thing=arg[1];
	if (!thing)
		return 0;
	return thing->upper();
}

NodeP Node::lower() {return error(BAD_DATA, this);}
NodeP Object::lower() {return _name->lower();}
NodeP String::lower() {return make(ptr(), 0, length(), type, copy_low);}
NodeP Array::lower() 
{	ArrayP newar=new Array(ArrayP(this));
	NodePP d=newar->_data;
	for (int i=_length;  i;  i--)
		*d++=(*d)->lower();
	return newar;
}
NodeP Pair::lower() 
{	PairP lst=this, result=0, lastnode=0, tnode=0;
	while (lst) 
	{	NodeP fst=lst->first();
		if (!fst)
			tnode=new Pair(0,0);
		else
			tnode = new Pair(fst->lower(), 0);
		if (!result) 
			result=tnode;
		else 
			lastnode->settail( tnode );
		lastnode=tnode;
		next(lst);
		if (checkThrowing()) break;
	}
	return result;
}

NodeP llowerCase(PairPP arg) 
{
	NodeP thing=arg[1];
	if (!thing)
		return 0;
	return thing->lower();
}

int   Node::count() {return string()->length();}
int String::count() {return length();}
int  Array::count() {return length();}
int IntArray::count() {return length();}
int Int16Array::count() {return length();}
int FloatArray::count() {return length();}
int   Pair::count()
{
	int c=0;
	for (PairP nd=this;  nd;  next(nd))
	{	c++;
		if (checkThrowing()) break;
	}
	return c;
}
int  Table::count()
{
	int c=0;
	for (int i=0; i<size; i++)
		for (ObjectP obj=base[i]; obj; next(obj))
		{	c++;
		}
	return c;

}
int Struct::count()
{
	int c=0;
	for (ObjectP obj=_first;  obj;  obj=ObjectP(obj->proc()))
	{	c++;
	}
	return c;
}

NodeP lcount(PairPP arg) {return new Int(arg[1]?  arg[1]->count() : 0);}

int   Node::_sizeof() {error(BAD_DATA, this); return 0;}
int  Int  ::_sizeof() {return sizeof(int);}
int  Int16::_sizeof() {return sizeof(short);}
int  Int8 ::_sizeof() {return sizeof(char);}
int UInt8 ::_sizeof() {return sizeof(unsigned char);}
int  Float::_sizeof() {return sizeof(double);}
int BigFloat::_sizeof() {return sizeof(M_APM)+1+MM_cpp_min_precision>>1;}
int Complexd::_sizeof() {return sizeof(complexd);}
int String::_sizeof() {return length();}
int Object::_sizeof() {return _name->_sizeof();}
int  Array::_sizeof() 
{	int size=0;
	NodePP d=_data;
	for (int i=_length;  i;  i--)
		size+=(*d++)->_sizeof();
	return size;
}
int IntArray::_sizeof() {return length()*sizeof(int);}
int Int16Array::_sizeof() {return length()*sizeof(short);}
int FloatArray::_sizeof() {return length()*sizeof(double);}
int   Pair::_sizeof()
{
	int size=0;
	for (PairP li=this;  li;  next(li))
	{	if (li->first())
			size+=li->first()->_sizeof();
		if (checkThrowing()) break;
	}
	return size;
}
int  Table::_sizeof()
{
	int size=0;
	for (int i=0; i<size; i++)
		for (ObjectP obj=base[i]; obj; next(obj))
			size+=obj->_sizeof();
	return size;
}
int Struct::_sizeof()
{
	int size=0;
	for (ObjectP obj=_first;  obj;  obj=ObjectP(obj->proc()))
		size+=obj->var()->_sizeof();
	return size;
}

NodeP lSizeOf(PairPP arg) {return new Int(arg[1]?  arg[1]->_sizeof() : 0);}


#define Create_typeof(x) \
	StringP  x::_typeof() {return type_##x ->name();}

Node_Type_Strings(Create_typeof, UNUSED)

StringP Object::_typeof() {return _name->_typeof();}

NodeP lTypeOf(PairPP arg) {return arg[1]?  arg[1]->_typeof() : new String("List");}


NodeP lpush(PairPP arg)
{
	NodeP stackname=arg[1]->a_name(1);

	if (!throwing)
	{	ObjectP obj=stackname->intern();
		if (obj->var()==Unbound)
			obj->setvar(new Pair(arg[2], 0));
		else
			obj->setvar(new Pair(arg[2], obj->var()));
	}
	return Unbound;
}

NodeP lpop(PairPP arg)
{
	NodeP stackname=arg[1]->a_name(1);

	if (!throwing)
	{	ObjectP obj=stackname->intern();
		NodeP var=obj->var();
		if (!var || var==Unbound || var->isnolist())
		{	error(BAD_DATA, stackname);
			return Unbound;
		}
		NodeP result=PairP(var)->first();
		obj->setvar(PairP(var)->tail());
		return result;
	}
	return Unbound;
}

NodeP lqueue(PairPP arg)
{
	NodeP queuename=arg[1]->a_name(1);

	if (!throwing)
	{	ObjectP obj=queuename->intern();
		NodeP var=obj->var();
		if (!var || var==Unbound)
			obj->setvar(new Pair(arg[2], 0));
		else if (var->isnolist())
		{	error(BAD_DATA, queuename);
			return Unbound;
		}else
		{	PairP list=PairP(var);
			while (list)
			{	if (!list->tail())
				{	list->settail(new Pair(arg[2], 0));
					return Unbound;
				}
				next(list);
			}
		}
	}
	return Unbound;
}

NodeP lWord(PairPP arg) 
{
	if (noarg)	return _Null_Word;
	int nr=1, count=0;
	StringP str=0;
	NodeType str_type=STRING;
	PairPP sp=arg;  

	while (sp!=argStack && !throwing) 
	{	if (!sp[1])
		{	sp++;
			continue;
		}
		str = (*++sp)->a_string(nr);
		if (!throwing) 
		{	if (str->is.backslashed)
				str_type=VBAR_STRING;
			count += str->length();
		}
		nr++;
	}
	if (!throwing)
		str = String::fromList(arg, count, str_type, String::copy_word);
	else
		str = StringP(Unbound);
	return str;
}

NodeP lStringBuffer(PairPP arg)
{
	int length=arg[1]->a_int(1);
	if throwing return Unbound;
	return new String(length);
}

NodeP lStringBufferToWord(PairPP arg)
{
	StringP s=arg[1]->a_string(1);
	if throwing return Unbound;
	return new String(s->ptr());
}

NodeP lcombine(PairPP arg)
{
	if (arg[2] && arg[2]->is.word)
	{	return lWord(arg);
	}
	else if (arg[2]->islist())
	{	return lfPut(arg);
	}
	else
	{	NodeType t;
		if (arg[1])
			t=arg[1]->type;
		else
			t=PAIR;
		PairPP a=arg;
		int i=1;
		while (a!=argStack)
		{	if (a[1] && a[1]->type!=t)
			{	error(BAD_DATA, arg[i], i); return Unbound;
			}
			a++;
			i++;
		}
		if (t==FLOATARRAY)
		{	PairPP sp=arg;
			int l=0;
			while (sp!=argStack)
			{	l+=FloatArrayP(sp[1])->length();
				sp++;
			}
			FloatArrayP result=new FloatArray(l);
			doubleP dr=result->data();
			sp=arg;
			while (sp!=argStack)
			{	FloatArrayP a=FloatArrayP(sp[1]);
				doubleP da=a->data();
				int la=a->length();
				while (la--)
				{	*dr++=*da++;
				}
				sp++;
			}
			return result;
		}
		else if (t==INTARRAY)
		{	PairPP sp=arg;
			int l=0;
			while (sp!=argStack)
			{	l+=IntArrayP(sp[1])->length();
				sp++;
			}
			IntArrayP result=new IntArray(l);
			intP dr=result->data();
			sp=arg;
			while (sp!=argStack)
			{	IntArrayP a=IntArrayP(sp[1]);
				intP da=a->data();
				int la=a->length();
				while (la--)
				{	*dr++=*da++;
				}
				sp++;
			}
			return result;
		}
		else if (t==INT16ARRAY)
		{	PairPP sp=arg;
			int l=0;
			while (sp!=argStack)
			{	l+=Int16ArrayP(sp[1])->length();
				sp++;
			}
			Int16ArrayP result=new Int16Array(l);
			shortP dr=result->data();
			sp=arg;
			while (sp!=argStack)
			{	Int16ArrayP a=Int16ArrayP(sp[1]);
				shortP da=a->data();
				int la=a->length();
				while (la--)
				{	*dr++=*da++;
				}
				sp++;
			}
			return result;
		}
		else if (t==ARRAY)
		{	PairPP sp=arg;
			int l=0;
			while (sp!=argStack)
			{	l+=ArrayP(sp[1])->length();
				sp++;
			}
			ArrayP result=new Array(l);
			NodePP dr=result->data();
			sp=arg;
			while (sp!=argStack)
			{	ArrayP a=ArrayP(sp[1]);
				NodePP da=a->data();
				int la=a->length();
				while (la--)
				{	*dr++=*da++;
				}
				sp++;
			}
			return result;
		}
	}
	error(BAD_DATA, arg[2], 2);
	return Unbound;
}

NodeP lSentence(PairPP arg) 
{
	NodeP val=0;
	PairP tnode=0, lastnode=0;

	while (arg!=argStack && !throwing) 
	{
		PairP th= *++arg;

		if (th->islist()) 
		{	for (PairP lst=th;  lst && !throwing;  next(lst))
			{
				tnode = new Pair(lst->first(), 0);
				if (!val) 
					val=tnode;
				else 
					lastnode->settail( tnode );
				lastnode=tnode;
			}
		}else 
		{	tnode=new Pair(th, 0);
			if (!val) 
				val=tnode;
			else 
				lastnode->settail( tnode );
			lastnode=tnode;
		}
	}
	if (throwing) return Unbound;
	return val;
}

NodeP lisList   (PairPP arg) {return Node::boolnode(arg[1]->islist());}
NodeP lisArray  (PairPP arg) {return Node::boolnode(type_of(arg[1])==ARRAY  );}
NodeP lisIntArray(PairPP arg) {return Node::boolnode(type_of(arg[1])==INTARRAY);}
NodeP lisInt16Array(PairPP arg) {return Node::boolnode(type_of(arg[1])==INT16ARRAY);}
NodeP lisFloatArray(PairPP arg) {return Node::boolnode(type_of(arg[1])==FLOATARRAY);}
NodeP lisTable  (PairPP arg) {return Node::boolnode(type_of(arg[1])==TABLE  );}
NodeP lisInt    (PairPP arg) {return Node::boolnode(type_of(arg[1])==INT_   );}
NodeP lisInt16  (PairPP arg) {return Node::boolnode(type_of(arg[1])==INT16  );}
NodeP lisInt8   (PairPP arg) {return Node::boolnode(type_of(arg[1])==INT8   );}
NodeP lisUInt8  (PairPP arg) {return Node::boolnode(type_of(arg[1])==UINT8  );}
NodeP lisFloat  (PairPP arg) {return Node::boolnode(type_of(arg[1])==FLOAT_ );}
NodeP lisBigFloat(PairPP arg) {return Node::boolnode(type_of(arg[1])==BIGFLOAT);}
NodeP lisComplex_(PairPP arg) {return Node::boolnode(type_of(arg[1])==COMPLEX);}

NodeP lisNumber(PairPP arg) 
{
	if (!arg[1])
		return _False;
	arg[1] = PairP(arg[1]->number());
	return Node::boolnode( arg[1] != Unbound );
}

NodeP lisWord(PairPP arg) 
{
	return Node::boolnode( 
		arg[1]!=Unbound  &&  arg[1]  &&  !arg[1]->isaggregate() );
}

NodeP   Node::memberP(NodeP subobj, bool notp, bool substr)
{
	if (subobj->is.aggregate) 
		return notp ?   _Null_Word  :  _False;
	else if (NodeP(this)==subobj)
		return _True;
	else
		return _False;
}
NodeP Int::memberP(NodeP subobj, bool notp, bool substr)
{	return string()->memberP(subobj, notp, substr);
}
NodeP Int16::memberP(NodeP subobj, bool notp, bool substr)
{	return string()->memberP(subobj, notp, substr);
}
NodeP Int8::memberP(NodeP subobj, bool notp, bool substr)
{	return string()->memberP(subobj, notp, substr);
}
NodeP UInt8::memberP(NodeP subobj, bool notp, bool substr)
{	return string()->memberP(subobj, notp, substr);
}
NodeP Float::memberP(NodeP subobj, bool notp, bool substr)
{	return string()->memberP(subobj, notp, substr);
}
NodeP BigFloat::memberP(NodeP subobj, bool notp, bool substr)
{	return string()->memberP(subobj, notp, substr);
}
NodeP Complexd::memberP(NodeP subobj, bool notp, bool substr)
{	return string()->memberP(subobj, notp, substr);
}
NodeP Object::memberP(NodeP subobj, bool notp, bool substr)
{	return name()->memberP(subobj, notp, substr);

}
NodeP String::memberP(NodeP subobj, bool notp, bool substr)
{
	StringP sub  = subobj->string();

	if ( sub != Unbound && this != Unbound 
	&& sub->_length <= _length 
	&& (substr || sub->_length==1) )
	{
		int len=_length - sub->_length;
		StringP tmp = String::make(_ptr, _head, sub->_length, type, copy);

		for (int i=0; i <= len; i++) 
		{	if ( sub->equalP(tmp) ) 
			{	if (notp) 
				{	tmp->_length = len+sub->_length-i;
					return tmp;
				}else 
					return _True;
			}
			tmp->_ptr++;
		}
	}
	return notp ?   _Null_Word  :  _False;
}
NodeP  Array::memberP(NodeP subobj, bool notp, bool substr)
{
	if (notp) error(BAD_DATA_UNREC, this);
	if (substr) return _False;

	NodePP d=_data;   int l=_length;

	while (--l >= 0 && !throwing) 
		if (subobj->equalP(*d++))
			return _True;
	
	return _False;
}
NodeP IntArray::memberP(NodeP subobj, bool notp, bool substr)
{
	if (notp) error(BAD_DATA_UNREC, this);
	if (substr) return _False;

	intP d=_data;   int l=_length;
	IntP i=new Int(0);

	while (--l >= 0 && !throwing) 
	{	i->int_=(*d++);
		if (subobj->equalP(i))
			return _True;
	}
	return _False;
}
NodeP Int16Array::memberP(NodeP subobj, bool notp, bool substr)
{
	if (notp) error(BAD_DATA_UNREC, this);
	if (substr) return _False;

	shortP d=_data;   int l=_length;
	Int16P i=new Int16(0);

	while (--l >= 0 && !throwing) 
	{	i->int_=(*d++);
		if (subobj->equalP(i))
			return _True;
	}
	return _False;
}
NodeP FloatArray::memberP(NodeP subobj, bool notp, bool substr)
{
	if (notp) error(BAD_DATA_UNREC, this);
	if (substr) return _False;

	doubleP d=_data;   int l=_length;
	FloatP f=new Float(0);

	while (--l >= 0 && !throwing) 
	{	f->float_=(*d++);
		if (subobj->equalP(f))
			return _True;
	}
	return _False;
}
NodeP   Pair::memberP(NodeP subobj, bool notp, bool substr)
{
	if (substr) return _False;

	PairP list=this;

	if (!subobj)
	{	while (list && !throwing)
		{	if (!list->first())
				return notp ?   list  :  NodeP(_True); 
	
			if (checkThrowing()) break;
			next(list);
		}
	}else
	{	while (list && !throwing) 
		{
			if (subobj->equalP(list->first()))
				return notp ?   list  :  NodeP(_True);
	
			if (checkThrowing()) break;
			next(list);
		}
	}
	return notp ?   NodeP(0)  :  _False;
}
NodeP Table::memberP(NodeP subobj, bool notp, bool substr)
{
	if (substr || !subobj) return _False;

	ObjectP obj=find(subobj->string());
	if (!obj)
		return _False;
	return _True;
}
NodeP lisMember   (PairPP arg)
{	return arg[2]? 
		arg[2]->memberP(arg[1], false, false) 
		: _False;
}
NodeP lisSubstring(PairPP arg)
{	return arg[2]? 
		arg[2]->memberP(arg[1], false,  true)
		: _False;
}
NodeP lMember     (PairPP arg)
{	return arg[2]? 
		arg[2]->memberP(arg[1], true, false)
		: _False;
}

NodeP lreplace(PairPP arg)
{
	StringP 
		substrp=arg[1]->a_string(1),
		newstrp=arg[2]->a_string(2),
		strp=arg[3]->a_string(3);

	if (!throwing)
	{	std::string str(strp->ptr()),
			substr(substrp->ptr()),
			newstr(newstrp->ptr());

		std::string::size_type pos=0;
		do 
		{	pos=str.find(substr, pos);
			if (pos!=std::string::npos)
			{	if (pos)
				{	str.erase(pos, substr.length());
					str.insert(pos, newstr);

					pos+=newstr.length();
				}
				else
					pos++;
			}
		} while (pos!=std::string::npos);

		StringP s=String::make(charP(str.c_str()),0,str.length(),
			STRING,String::copy);
		return s;
	}
	return Unbound;
}

bool Array::circular(NodeP _nd1, NodeP _nd2) 
{
	if (!_nd2)
	{	return false;
	}
	else if (_nd2->type == ARRAY) 
	{	ArrayP  &newArray = (ArrayP&)_nd2;
			    int i=newArray->_length;
		NodePP p=newArray->_data;
		
		if (newArray==_nd1) 
			return true;

		while (--i >= 0) 
			if (circular(_nd1, *p++)) 
				return true;

		return false;
	}
	else if (_nd2->is.list) 
	{	PairP &newList = (PairP&)_nd2;

		while (newList) 
		{	NodeP f=newList->first(),
				t=newList->tail();
			if (_nd1==f || _nd1==t)
				return true;
			if (circular(_nd1, f)) 
				return true;
			next(newList);
		}
		return false;
	}else 
		return false;
}
bool Pair::circular() 
{
	PairP list = this;

	while (list) 

	{	PairP list2=list->tail();
		while (list2)
		{	if (list==list2) 
				return true;
			next(list2);
		}
		next(list);
	}
	return false;
}

NodeP liscircular(PairPP arg)
{
	if (exist(2))
		return Node::boolnode(Array::circular(arg[1], arg[2]));
	else
		return Node::boolnode(Array::circular(arg[1], arg[1]));
}

NodeP   Node::item(int i) {error(BAD_DATA, this, 2); return Unbound;}
NodeP Object::item(int i) {return name()->item(i);}
NodeP String::item(int i)
{	if (i > 0  &&  i <= _length)

		return new String(_ptr+i-1, _head, 1, type, String::copy);
	error(BAD_DATA, new Int(i)); return Unbound;
}
NodeP   Pair::item(int i)
{	if (i >= 0) 
	{	PairP list = this;
		while (list && i--) 
			next(list);
		if (list) return list->first();
	}
	error(BAD_DATA, new Int(i)); return Unbound;
}
NodeP  Array::item(int i)
{	if (i >= 0  &&  i < _length) 
		return _data[i];
	error(BAD_DATA, new Int(i)); return Unbound;
}
NodeP IntArray::item(int i)
{	if (i >= 0  &&  i < _length) 
		return new Int(_data[i]);
	error(BAD_DATA, new Int(i)); return Unbound;
}
NodeP Int16Array::item(int i)
{	if (i >= 0  &&  i < _length) 
		return new Int16(_data[i]);
	error(BAD_DATA, new Int(i)); return Unbound;
}
NodeP FloatArray::item(int i)

{	if (i >= 0  &&  i < _length) 
		return new Float(_data[i]);
	error(BAD_DATA, new Int(i)); return Unbound;
}

#define ITEM_if_index_islist							\
	if (index->islist() && result->islist())		 \
	{	PairP i=PairP(index), r=PairP(result);		  \
		while (i && r)										   \
		{	NodeP ret=item(i->first(), r->first());    \
			if (!throwing) r->setfirst( ret );		     \
			next(i);											      \
			next(r);											       \
		}														        \
		if (!i && !r && !throwing) return result;	         \
	}

NodeP     Node::item(NodeP index, NodeP result) {return item(0);}
NodeP      Int::item(NodeP index, NodeP result)
{	return string()->item(index,result);
}
NodeP    Int16::item(NodeP index, NodeP result)
{	return string()->item(index,result);
}
NodeP    Int8::item(NodeP index, NodeP result)
{	return string()->item(index,result);
}
NodeP   UInt8::item(NodeP index, NodeP result)
{	return string()->item(index,result);
}
NodeP    Float::item(NodeP index, NodeP result)
{	return string()->item(index,result);
}
NodeP BigFloat::item(NodeP index, NodeP result)
{	return string()->item(index,result);
}
NodeP Complexd::item(NodeP index, NodeP result)
{	return string()->item(index,result);
}
NodeP   Object::item(NodeP index, NodeP result) 
{	return name()->item(index, result);
}
NodeP   String::item(NodeP index, NodeP result)
{	ITEM_if_index_islist
	else if (index->type==INT_)
		return item(IntP(index)->int_);
	error(BAD_DATA, index); return Unbound;
}
NodeP     Pair::item(NodeP index, NodeP result)
{	if (index->islist())
	{	if (result && result->islist())		 
		{	PairP i=PairP(index), r=PairP(result);		  
			while (i && r)										   
			{	NodeP ret=item(i->first(), r->first());    
				if (!throwing) r->setfirst( ret );		    
				next(i);											      
				next(r);											       
			}														        
			if (!i && !r && !throwing) return result;	         
		}
		else
		{	PairP lst=PairP(index), result=0, lastnode=0, tnode=0;
			while (lst) 
			{	NodeP fst=lst->first();
				if (!fst)
					tnode=new Pair(0,0);
				else
					tnode = new Pair(item(fst, 0), 0);
				if (!result) 
					result=tnode;
				else 
					lastnode->settail( tnode );
				lastnode=tnode;
				next(lst);
			}
			return result;
		}
	}
//	ITEM_if_index_islist
	else if (index->type==INT_)
		return item(IntP(index)->int_-1);
	else if (index->is.string)
		return item(StringP(index)->number(), 0);
	error(BAD_DATA, index); return Unbound;
}
NodeP  Array::item(NodeP index, NodeP result)
{
	if (index->type==INT_)
	{	int i=IntP(index)->int_-_origin;
		if (i >= 0  &&  i < _length) 
			return _data[i];
	}
	else if (index->type==ARRAY)
	{	ArrayP idx=ArrayP(index);
		int l=idx->length();
		ArrayP r=new Array(l);
		NodePP d=r->data(), i=idx->data();
		for (;  l;  l--)
		{	*d++=item(*i++, 0);
		}
		return r;
	}
	else if (index->islist() && result->islist())
	{	PairP i=PairP(index), r=PairP(result);
		while (i && r)
		{	NodeP ret=item(i->first(), r->first());
			if (!throwing) r->setfirst( ret );
			next(i);
			next(r);
		}
		if (!i && !r && !throwing) return result;
	}
//	ITEM_if_index_islist
	else if (index->is.string)
		return item(StringP(index)->number(), 0);
	error(BAD_DATA, index); return Unbound;
}
NodeP IntArray::item(NodeP index, NodeP result)
{	if (index->type==INT_)
	{	int i=IntP(index)->int_-_origin;
		if (i >= 0  &&  i < _length) 
			return new Int(_data[i]);
	}
	else if (index->type==INTARRAY)
	{	IntArrayP idx=IntArrayP(index);
		int l=idx->length();
		IntArrayP r=new IntArray(l);
		intP d=r->data();
		intP i=idx->data();
		for (;  l;  l--)
		{	*d++=_data[(*i++)-1];
		}
		return r;
	}
	else ITEM_if_index_islist
	error(BAD_DATA, index); return Unbound;
}
NodeP Int16Array::item(NodeP index, NodeP result)
{	ITEM_if_index_islist
	else if (index->type==INT_)
	{	int i=IntP(index)->int_-_origin;
		if (i >= 0  &&  i < _length) 
			return new Int16(_data[i]);
	}
	error(BAD_DATA, index); return Unbound;
}
NodeP FloatArray::item(NodeP index, NodeP result)
{
	if (index->type==INT_)
	{	int i=IntP(index)->int_-_origin;
		if (i >= 0  &&  i < _length) 
			return new Float(_data[i]);
	}
	else if (index->type==INTARRAY)
	{	IntArrayP idx=IntArrayP(index);
		int l=idx->length();
		FloatArrayP r=new FloatArray(l);
		doubleP d=r->data();
		intP i=idx->data();
		for (;  l;  l--)
		{	if (*i > length() || *i < origin())
			{	error(BAD_DATA, index, 2); return Unbound;
			}
			*d++=_data[(*i++)-origin()];
		}
		return r;
	}
	else if (index->type==ARRAY)
	{	ArrayP idx=ArrayP(index);
		int l=idx->length();
		FloatArrayP r=new FloatArray(l);
		doubleP d=r->data();
		NodePP i=idx->data();
		for (;  l;  l--)
		{	*d++=item(*i++, 0)->getnumber();
		}
		return r;
	}
	else ITEM_if_index_islist
	else if (index->is.string)
		return item(StringP(index)->number(), 0);
	error(BAD_DATA, index); return Unbound;
}
NodeP Table::item(NodeP index, NodeP result)
{	ObjectP obj=find(index->string());
	if (!obj)
	{	return 0;
//		error(BAD_DATA, index);
//		return Unbound;
	}
	return obj->var();
}
NodeP Struct::item(NodeP index, NodeP result)
{	ObjectP obj=find(index->string());
	if (!obj)
	{	error(BAD_DATA, index);
		return Unbound;
	}
	return obj->var();
}
NodeP lItem(PairPP arg) // _item index obj 
{
	NodeP index=arg[1], obj=arg[2], result=0;
	if (!index)
	{	error(BAD_DATA, arg[1], 1);
		return Unbound;
	}
	if (index->isstring()) index=arg[1]->a_number(1);
	if (exist(3)) result = arg[3];

	while ((!obj || obj==_Null_Word) && !throwing) 
		obj=error(BAD_DATA, obj);

	if (!throwing) return obj->item(index, result);
	return Unbound;
}


NodeP lisNameInTable(PairPP arg) // NameInTable? name table
{
	NodeP index=arg[1];
	TableP table=arg[2]->a_table(2);
	if throwing return Unbound;
	ObjectP obj=table->find(index->string());
	return Node::boolnode(obj != NodeP(0));
}

NodeP ldot(PairPP arg) // _dot( obj, index )
{
//	PairP obj=arg[1];	arg[1] = arg[2];	arg[2]=obj;
//	return _Item(arg);
	NodeP obj=arg[1], index=arg[2], result=0;
//	if (index->isstring() && index->numberP()) index=index->a_number();
	if (!index)
	{	error(BAD_DATA, arg[1], 1);
		return Unbound;
	}
	if (exist(3)) result = arg[3];

	while ((!obj || obj==_Null_Word) && !throwing) 
		obj=error(BAD_DATA, obj);

	if (!throwing) return obj->item(index, result);
	return Unbound;
}

NodeP lpick(PairPP arg)
{
	NodeP obj=arg[1];
	if (!obj) 
		return 0;

	int c=obj->count();
	if (obj->islist() || obj->isarray())
		return obj->item(Int::random(c));
	else if (obj->isstring())
		return obj->item(Int::random(c)+1);
	else if (obj->type==OBJECT)
		return ObjectP(obj)->name()->item(Int::random(c)+1);
	return Unbound;
}

NodeP Node::mdItem(NodeP index)
{	error(BAD_DATA, this); return Unbound;
}
NodeP Pair::mdItem(NodeP index)
{
	NodeP obj=this;
	if (index->islist())
	{	PairP indexlist=PairP(index);
		while (indexlist)
		{	NodeP i=indexlist->first();
			if (!i)
			{	error(BAD_DATA, index); 
				return Unbound;
			}
			if (!obj)
			{	error(BAD_DATA, this);
				return Unbound;
			}
			obj=obj->item(i, 0);
			next(indexlist);
		}
		return obj;
	}
	error(BAD_DATA, this);
	return Unbound;
}
NodeP Array::mdItem(NodeP index)
{
	NodeP obj=this;
	if (index->islist())
	{	PairP indexlist=PairP(index);
		while (indexlist)
		{	NodeP i=indexlist->first();
			if (!i)
			{	error(BAD_DATA, index); 
				return Unbound;
			}
			if (!obj)
			{	error(BAD_DATA, this);
				return Unbound;
			}
			obj=obj->item(i, 0);
			next(indexlist);
		}
		return obj;
	}
	error(BAD_DATA, this);
	return Unbound;
}

NodeP lmdItem(PairPP arg) // mdItem( indexlist, mdobj )
{
	PairP index=arg[1]->a_list(1);
	NodeP obj=arg[2];
	if throwing return Unbound;

	if (!index)
		return obj;
	if (!obj)
	{	error(BAD_DATA, obj, 2);
		return Unbound;
	}
	return obj->mdItem(index);
}

NodeP     Node::bititem(NodeP index) 
{	error(BAD_DATA, this, 2); return Unbound;
}
NodeP      Int::bititem(NodeP index)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (i<0 || i >= int(sizeof(int))<<3)
	{	error(BAD_DATA, index);
		return Unbound;

	}
	return boolnode((int_>>i) & 1);
}
NodeP    Int16::bititem(NodeP index)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (i<0 || i >= int(sizeof(short))<<3)
	{	error(BAD_DATA, index);
		return Unbound;
	}
	return boolnode((int_>>i) & 1);
}
NodeP     Int8::bititem(NodeP index)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (i<0 || i >= int(sizeof(char))<<3)
	{	error(BAD_DATA, index);
		return Unbound;
	}
	return boolnode((int_>>i) & 1);
}
NodeP    UInt8::bititem(NodeP index)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (i<0 || i >= int(sizeof(unsigned char))<<3)
	{	error(BAD_DATA, index);

		return Unbound;
	}
	return boolnode((int_>>i) & 1);
}
NodeP    Float::bititem(NodeP index)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (i<0 || i >= int(sizeof(double))<<3)
	{	error(BAD_DATA, index);
		return Unbound;
	}
	int *bits=(intP)&float_;
	return boolnode((bits[i>>5]>>(i & 31)) & 1);
}
NodeP Complexd::bititem(NodeP index)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (i<0 || i >= int(sizeof(complexd))<<3)
	{	error(BAD_DATA, index);
		return Unbound;
	}
	int *bits=(intP)&complex_;
	return boolnode((bits[i>>5]>>(i & 31)) & 1);
}
NodeP   Object::bititem(NodeP index) 
{	return name()->bititem(index);
}
NodeP   String::bititem(NodeP index)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (i<0 || i >= _length*int(sizeof(char))<<3)
	{	error(BAD_DATA, index);
		return Unbound;
	}
	return boolnode((_ptr[i>>3] >> (i & 7)) & 1);
}

NodeP lBitItem(PairPP arg) // _item( index, obj )
{
	NodeP index=arg[1], obj=arg[2];
	if (index->isstring()) index=arg[1]->a_number(1);

	while ((!obj || obj==_Null_Word) && !throwing) 
		obj=error(BAD_DATA, obj, 2);

	if (!throwing) return obj->bititem(index);
	return Unbound;
}

NodeP   Node::items(NodeP start, NodeP end) 
{	error(BAD_DATA, this, 3); return Unbound;
}
NodeP  Array::items(NodeP start, NodeP end)
{	if (start->type==INT_)
	{	if (end->type==INT_)
		{	int s=IntP(start)->int_-_origin, e=IntP(end)->int_-_origin;
			if (s >= 0  &&  s < _length)
				if (e >= 0  &&  e < _length) 
				{	ArrayP newar=new Array(e-s+1);
					for (int i=s, j=0;  i<=e;  i++, j++)
						newar->_data[j]=_data[i];
					return newar;
				}else
				{	error(BAD_DATA, end, 2); return Unbound;
				}
			else
			{	error(BAD_DATA, start, 1); return Unbound;
			}
		}
		error(BAD_DATA, end, 2); return Unbound;
	}
	error(BAD_DATA, start, 1); return Unbound;
}
NodeP IntArray::items(NodeP start, NodeP end)
{	if (start->type==INT_)
	{	if (end->type==INT_)
		{	int s=IntP(start)->int_-_origin, e=IntP(end)->int_-_origin;
			if (s >= 0  &&  s < _length)
				if (e >= 0  &&  e < _length) 
				{	IntArrayP newar=new IntArray(e-s+1);
					for (int i=s, j=0;  i<=e;  i++, j++)
						newar->_data[j]=_data[i];
					return newar;
				}else
				{	error(BAD_DATA, end, 2); return Unbound;
				}
			else
			{	error(BAD_DATA, start, 1); return Unbound;
			}
		}
		error(BAD_DATA, end, 2); return Unbound;
	}
	error(BAD_DATA, start, 1); return Unbound;
}
NodeP Int16Array::items(NodeP start, NodeP end)
{	if (start->type==INT_)
	{	if (end->type==INT_)
		{	int s=IntP(start)->int_-_origin, e=IntP(end)->int_-_origin;
			if (s >= 0  &&  s < _length)
				if (e >= 0  &&  e < _length) 
				{	Int16ArrayP newar=new Int16Array(e-s+1);
					for (int i=s, j=0;  i<=e;  i++, j++)
						newar->_data[j]=_data[i];
					return newar;
				}else
				{	error(BAD_DATA, end, 2); return Unbound;
				}
			else
			{	error(BAD_DATA, start, 1); return Unbound;
			}
		}
		error(BAD_DATA, end, 2); return Unbound;
	}
	error(BAD_DATA, start, 1); return Unbound;
}
NodeP FloatArray::items(NodeP start, NodeP end)
{	if (start->type==INT_)
	{	if (end->type==INT_)
		{	int s=IntP(start)->int_-_origin, e=IntP(end)->int_-_origin;
			if (s >= 0  &&  s < _length)
				if (e >= 0  &&  e < _length) 
				{	FloatArrayP newar=new FloatArray(e-s+1);
					for (int i=s, j=0;  i<=e;  i++, j++)
						newar->_data[j]=_data[i];
					return newar;
				}else
				{	error(BAD_DATA, end, 2); return Unbound;
				}
			else
			{	error(BAD_DATA, start, 1); return Unbound;
			}
		}
		error(BAD_DATA, end, 2); return Unbound;
	}
	error(BAD_DATA, start, 1); return Unbound;
}

NodeP lItems(PairPP arg) // _item( index, obj )
{
	NodeP start=arg[1], end=arg[2], obj=arg[3];
	if (start->isstring()) start=start->a_number(1);
	if (end  ->isstring()) end  =end  ->a_number(2);

	while ((!start || !end || !obj || obj==_Null_Word) && !throwing) 
		obj=error(BAD_DATA, obj, 3);

	if (!throwing) return obj->items(start, end);
	return Unbound;
}

NodeP Node::setitem(NodeP index, NodeP newitem, bool safe) 
{	error(BAD_DATA, this, 2); 
	return Unbound;
}
NodeP Object::setitem(NodeP index, NodeP newitem, bool safe)
{	return _name->setitem(index, newitem, safe);
}
NodeP Int::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (newitem && newitem->type==OBJECT)
	{	int bit=newitem->a_bool(3);
		if (!throwing)
		{	if (i<0 || i >= int(sizeof(int))<<3)
			{	error(BAD_DATA, index);
				return Unbound;
			}
			if (bit)
				int_|=(1<<i);
			else
				int_&=~(1<<i);
		}
	}
	else
	{	StringP num=string();
		num->setitem(index, newitem, safe);
		int_=int(num->getnumber());
	}
	return Unbound;
}
NodeP Int16::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (newitem && newitem->type==OBJECT)
	{	int bit=newitem->a_bool(3);
		if (!throwing)
		{	if (i<0 || i >= int(sizeof(short))<<3)
			{	error(BAD_DATA, index);
				return Unbound;
			}
			if (bit)
				int_|=(1<<i);
			else
				int_&=~(1<<i);
		}
	}
	else
	{	StringP num=string();
		num->setitem(index, newitem, safe);
		int_=int(num->getnumber());
	}
	return Unbound;
}
NodeP Int8::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (newitem && newitem->type==OBJECT)
	{	int bit=newitem->a_bool(3);
		if (!throwing)
		{	if (i<0 || i >= int(sizeof(char))<<3)
			{	error(BAD_DATA, index);
				return Unbound;
			}
			if (bit)
				int_|=(1<<i);
			else
				int_&=~(1<<i);
		}
	}
	else
	{	StringP num=string();
		num->setitem(index, newitem, safe);
		int_=int(num->getnumber());
	}
	return Unbound;
}
NodeP UInt8::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (newitem && newitem->type==OBJECT)
	{	int bit=newitem->a_bool(3);
		if (!throwing)
		{	if (i<0 || i >= int(sizeof(unsigned char))<<3)
			{	error(BAD_DATA, index);
				return Unbound;
			}
			if (bit)
				int_|=(1<<i);
			else
				int_&=~(1<<i);
		}
	}
	else
	{	StringP num=string();
		num->setitem(index, newitem, safe);
		int_=int(num->getnumber());
	}
	return Unbound;
}
NodeP String::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1)-1;
	StringP ch=newitem->a_char(3);
	if (!throwing)
	{	if (i<0 || i>=_length)
		{	error(BAD_DATA, index);
			return Unbound;
		}
		_ptr[i]=ch->_ptr[0];
	}
	return Unbound;
}
NodeP Array::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (safe) 
	{	while (circular(this, newitem) && !throwing) 
			newitem=error(BAD_DATA, newitem, 3);
	}
	if (!throwing) 
	{	i-=origin();

		if ((i < 0 || i >= length()) && !throwing) 
			error(BAD_DATA, index);

		if (!throwing) 
		{	_data[i] = newitem;
			check_oldyoung(newitem);
		}
	}
	return Unbound;
}
NodeP IntArray::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1),
		ni=newitem->a_int(3);
	if throwing return Unbound;
	i-=origin();

	if ((i < 0 || i >= length()) && !throwing) 
		error(BAD_DATA, index);

	if (!throwing) 
	{	_data[i] = ni;
	}
	return Unbound;
}
NodeP Int16Array::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1);
	short ni=newitem->a_int16(3);
	if throwing return Unbound;
	i-=origin();

	if ((i < 0 || i >= length()) && !throwing) 
		error(BAD_DATA, index);

	if (!throwing) 
	{	_data[i] = ni;
	}
	return Unbound;
}
NodeP FloatArray::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1);
	double ni=newitem->a_float(3);
	if throwing return Unbound;
	i-=origin();

	if ((i < 0 || i >= length()) && !throwing) 
		error(BAD_DATA, index);

	if (!throwing) 
	{	_data[i] = ni;
	}
	return Unbound;
}
NodeP Pair::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (safe) 
	{	while (Array::circular(this, newitem) && !throwing) 
			newitem=error(BAD_DATA, newitem, 3);
	}
	if (!throwing) 
	{	if ((i <= 0) && !throwing) 
			error(BAD_DATA, index);

		if (!throwing) 
		{	PairP a=this;
			while (--i && a)
			{	next(a);
			}
			if (a) a->setfirst( newitem );
			else error(BAD_DATA, index);
		}
	}
	return Unbound;
}
NodeP Table::setitem(NodeP index, NodeP newitem, bool safe)
{	StringP str=index->a_string(1);
	ObjectP obj=find(str);
	if (!obj)
		obj=insert(str);
	obj->setvar(newitem);
	return Unbound;
}
NodeP Struct::setitem(NodeP index, NodeP newitem, bool safe)
{	StringP str=index->a_string(1);
	ObjectP obj=0;
	if (String::caseIgnored)
		obj=find(StringP(str->lower()));
	else
		obj=find(str);
	if (!obj)
	{	error(BAD_DATA, index);
		return Unbound;
	}
	NodeP v=obj->var();
	if (!newitem)
	{	error(BAD_DATA, newitem, 3);
		return Unbound;
	}
	if (newitem->type==v->type)
	{	obj->setvar(newitem);
		return Unbound;
	}
	NodeP ni=0;
	if (v->type==INT_)
		ni=newitem->_int();
	else if (v->type==INT16)
		ni=newitem->toint16();
	else if (v->type==INT8)
		ni=newitem->toint8();
	else if (v->type==UINT8)
		ni=newitem->touint8();
	else if (v->type==FLOAT_)
		ni=newitem->_float();
	else if (v->type==STRING)
	{	ni=newitem->string();
		if (StringP(ni)->length() != StringP(v)->length())
		{	error(BAD_DATA, newitem, 3);
			return Unbound;
		}
	}
	else if (v->type==INTARRAY)
	{	ni=newitem->intarray();
		if throwing return Unbound;
		if (IntArrayP(ni)->length() != IntArrayP(v)->length())
		{	error(BAD_DATA, newitem, 3);
			return Unbound;
		}
	}
	else if (v->type==INT16ARRAY)
	{	ni=newitem->int16array();
		if throwing return Unbound;
		if (Int16ArrayP(ni)->length() != Int16ArrayP(v)->length())
		{	error(BAD_DATA, newitem, 3);
			return Unbound;
		}
	}
	else if (v->type==FLOATARRAY)
	{	ni=newitem->floatarray();
		if throwing return Unbound;
		if (FloatArrayP(ni)->length() != FloatArrayP(v)->length())
		{	error(BAD_DATA, newitem, 3);
			return Unbound;
		}
	}

	else
	{	error(BAD_DATA, newitem, 3);
		return Unbound;
	}
	if (ni->type!=v->type)
	{	error(BAD_DATA, newitem, 3);
		return Unbound;
	}
	obj->setvar(ni);
	return Unbound;
}
NodeP setItemHelper(PairPP arg, bool safe) 
{
	NodeP  index =arg[1];
	NodeP   obj  =arg[2];
	NodeP newitem=arg[3];

	if (!index)
		error(BAD_DATA, index);
	else
	{	if (obj)
		{	NodeP i=obj->setitem(index, newitem, safe);
			if (!throwing)
				return i;
		}else
			error(BAD_DATA, obj, 2);
	}
	return Unbound;
}

NodeP lsetItem(PairPP arg) {return setItemHelper(arg, true);}
NodeP l_setItem(PairPP arg) {return setItemHelper(arg, false);}

NodeP lmdSetItem(PairPP arg) 
{
	NodeP indexlist=arg[1];
	NodeP   obj  =arg[2];
	NodeP newitem=arg[3];

	if (!indexlist)
		error(BAD_DATA, indexlist);
	else
	{	if (obj)
		{	obj=obj->mdItem(indexlist->butLast());
			if (!obj)
			{	error(BAD_DATA, obj, 2); return Unbound;
			}
			NodeP i=obj->setitem(indexlist->last(), newitem, false);
			if (!throwing)
				return i;
		}else
			error(BAD_DATA, obj, 2);
	}
	return Unbound;
}

NodeP Node::setitems(NodeP start, NodeP newitems) 
{	error(BAD_DATA, this, 2); 
	return Unbound;
}
NodeP Array::setitems(NodeP start, NodeP newitems)
{	int i=start->a_int(1);
	if throwing return Unbound;
	if (!throwing) 
	{	i-=origin();

		if ((i < 0 || i >= length()) && !throwing) 
			error(BAD_DATA, start);

		if (!throwing) 
		{	if (type_of(newitems)==ARRAY)
			{	ArrayP n=ArrayP(newitems);
				if (i+n->_length <= _length)
				{	NodePP s=n->_data, d=&_data[i];
					for (int j=n->_length;  j;  j--) 
					{	*d = *s++;
						check_oldyoung(*d);
						d++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}
			else if (type_of(newitems)==INTARRAY)
			{	IntArrayP n=IntArrayP(newitems);
				if (i+n->length() <= _length)
				{	intP s=n->data();
					NodePP d=&_data[i];
					for (int j=n->length();  j;  j--) 
					{	*d = new Int(*s++);
						check_oldyoung(*d);
						d++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==INT16ARRAY)
			{	Int16ArrayP n=Int16ArrayP(newitems);
				if (i+n->length() <= _length)
				{	shortP s=n->data();
					NodePP d=&_data[i];
					for (int j=n->length();  j;  j--) 
					{	*d = new Int16(*s++);
						check_oldyoung(*d);
						d++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==FLOATARRAY)
			{	FloatArrayP n=FloatArrayP(newitems);
				if (i+n->length() <= _length)
				{	doubleP s=n->data();
					NodePP d=&_data[i];
					for (int j=n->length();  j;  j--) 
					{	*d = new Float(*s++);
						check_oldyoung(*d);
						d++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}else
				error(BAD_DATA, newitems, 3);
		}
	}
	return Unbound;
}
NodeP IntArray::setitems(NodeP start, NodeP newitems)
{	int i=start->a_int(1);
	if throwing return Unbound;
	if (!throwing) 
	{	i-=origin();

		if ((i < 0 || i >= length()) && !throwing) 
			error(BAD_DATA, start);

		if (!throwing) 
		{	if (type_of(newitems)==ARRAY)
			{	ArrayP n=ArrayP(newitems);
				if (i+n->length() <= _length)
				{	NodePP s=n->data();
					intP d = _data+i;
					
					for (int j=n->length();  j;  j--)
					{	NodeP h=*s;
						switch (type_of(h))
						{	case INT_:	*d++ =             IntP(h)->int_;	break;
							case FLOAT_:*d++ = Int::clip(FloatP(h)->float_);break;
							case INT16:	*d++ =           Int16P(h)->int_;	break;
							case INT8:	*d++ =            Int8P(h)->int_;	break;
							case STRING:if (h->numberP())
												*d++ = Int::clip(h->getnumber());
											else
											{	error(BAD_DATA, n, 3); 
												return IntArrayP(Unbound);
											}	
											break;
							default:
								error(BAD_DATA, n, 3);
								return IntArrayP(Unbound);
						}
						s++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}
			else if (type_of(newitems)==INTARRAY)
			{	IntArrayP n=IntArrayP(newitems);
				if (i+n->length() <= _length)
				{	intP s=n->_data, d=&_data[i];
					for (int j=n->_length;  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==INT16ARRAY)
			{	Int16ArrayP n=Int16ArrayP(newitems);
				if (i+n->length() <= _length)
				{	shortP s=n->data();
					intP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==FLOATARRAY)
			{	FloatArrayP n=FloatArrayP(newitems);
				if (i+n->length() <= _length)
				{	doubleP s=n->data();
					intP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = Int::clip(*s++);
				}else
					error(BAD_DATA, newitems, 3);
			}else
				error(BAD_DATA, newitems, 3);
		}
	}
	return Unbound;
}
NodeP Int16Array::setitems(NodeP start, NodeP newitems)
{	int i=start->a_int(1);
	if throwing return Unbound;
	if (!throwing) 

	{	i-=origin();

		if ((i < 0 || i >= length()) && !throwing) 
			error(BAD_DATA, start);

		if (!throwing) 
		{	if (type_of(newitems)==ARRAY)
			{	ArrayP n=ArrayP(newitems);
				if (i+n->length() <= _length)
				{	NodePP s=n->data();
					shortP d=&_data[i];
					
					for (int j=n->length();  j;  j--)
					{	NodeP h=*s;
						switch (type_of(h))
						{	case INT_:	*d++ =  Int16::clip(IntP(h)->int_);	break;
							case FLOAT_:*d++ =Int16::clip(FloatP(h)->float_);break;
							case INT16:	*d++ =            Int16P(h)->int_;	break;
							case INT8:	*d++ =             Int8P(h)->int_;	break;
							case STRING:if (h->numberP())
												*d++ = Int16::clip(h->getnumber());
											else
											{	error(BAD_DATA, n, 3); 
												return IntArrayP(Unbound);
											}	
											break;
							default:
								error(BAD_DATA, n, 3);	
								return IntArrayP(Unbound);
						}
						s++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}
			else if (type_of(newitems)==INTARRAY)
			{	IntArrayP n=IntArrayP(newitems);
				if (i+n->length() <= _length)
				{	intP s=n->data();
					shortP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = Int16::clip(*s++);
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==INT16ARRAY)
			{	Int16ArrayP n=Int16ArrayP(newitems);
				if (i+n->length() <= _length)
				{	shortP s=n->_data, d=&_data[i];
					for (int j=n->_length;  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==FLOATARRAY)

			{	FloatArrayP n=FloatArrayP(newitems);
				if (i+n->length() <= _length)
				{	doubleP s=n->data();
					shortP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = Int16::clip(*s++);
				}else
					error(BAD_DATA, newitems, 3);
			}else
				error(BAD_DATA, newitems, 3);
		}
	}
	return Unbound;
}
NodeP FloatArray::setitems(NodeP start, NodeP newitems)
{	int i=start->a_int(1);
	if throwing return Unbound;
	if (!throwing) 
	{	i-=origin();

		if ((i < 0 || i >= length()) && !throwing) 
			error(BAD_DATA, start);

		if (!throwing) 
		{	if (type_of(newitems)==ARRAY)
			{	ArrayP n=ArrayP(newitems);
				if (i+n->length() <= _length)
				{	NodePP s=n->data();
					doubleP d = _data+i;
					
					for (int j=n->length();  j;  j--)
					{	NodeP h=*s;
						switch (type_of(h))
						{	case INT_:	*d++ =   IntP(h)->int_;	break;
							case FLOAT_:*d++ = FloatP(h)->float_;break;
							case INT16:	*d++ = Int16P(h)->int_;	break;
							case INT8:	*d++ =  Int8P(h)->int_;	break;
							case STRING:if (h->numberP())
												*d++ = h->getnumber();
											else
											{	error(BAD_DATA, n, 3); 
												return FloatArrayP(Unbound);
											}	
											break;
							default:
								error(BAD_DATA, n, 3);
								return FloatArrayP(Unbound);
						}
						s++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}
			else if (type_of(newitems)==INTARRAY)
			{	IntArrayP n=IntArrayP(newitems);
				if (i+n->length() <= _length)
				{	intP s=n->data();
					doubleP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==INT16ARRAY)
			{	Int16ArrayP n=Int16ArrayP(newitems);
				if (i+n->length() <= _length)
				{	shortP s=n->data();
					doubleP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==FLOATARRAY)
			{	FloatArrayP n=FloatArrayP(newitems);
				if (i+n->length() <= _length)
				{	doubleP s=n->_data, d=&_data[i];
					for (int j=n->_length;  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else
				error(BAD_DATA, newitems, 3);
		}
	}
	return Unbound;
}

NodeP lsetItems(PairPP arg) 
{
	NodeP  start  =arg[1];
	NodeP   obj   =arg[2];
	NodeP newitems=arg[3];

	if (!start)
		error(BAD_DATA, start);
	else
	{	if (obj)
		{	NodeP i=obj->setitems(start, newitems);
			if (!throwing)
				return i;
		}else
			error(BAD_DATA, obj, 2);
	}
	return Unbound;
}

void Node::removeitem(NodeP index) 
{	error(BAD_DATA, this, 2); 
}
void Object::removeitem(NodeP index)
{	_name->removeitem(index);
}
void String::removeitem(NodeP index)
{	int i=index->a_int(1)-1;
	if throwing return;
	
	if (i<0 || i>=_length)
	{	error(BAD_DATA, index);
		return;
	}
	charP cp=&_ptr[i];
	while (i<_length-1)
	{	*cp=cp[1];
		cp++;
		i++;
	}
	_length--;
}
void Pair::removeitem(NodeP index)
{	int i=index->a_int(1);
	if throwing return;
	
	if (i<=0) 
		error(BAD_DATA, index);
	if throwing return;

	if (i==1)
	{	if (tail())
		{	setfirst( tail()->first() );
			settail( tail()->tail() );
		}else
		{	setfirst( 0 );
			settail( 0 );
		}
		return;
	}
	PairP a=this, last=this;
	while (--i && a)
	{	last=a;
		next(a);
	}
	if (a)
		last->settail( a->tail() );
	else 
		error(BAD_DATA, index);
}
void Table::removeitem(NodeP index)
{	StringP name=index->a_string(1);
	if throwing return;
	int i=hash(name);
	ObjectP obj=base[i], last=0;
	while (obj)
	{	if (name->equal(obj->name()))
			break;
		last=obj;
		next(obj);
	}
	if (!obj)
	{	error(BAD_DATA, index);
		return;
	}
	if (!last)
	{	base[i]=obj->tail();
	}
	else
		last->settail(obj->tail());
}

NodeP lremoveItem(PairPP arg) 
{
	NodeP index =arg[1];
	NodeP  obj  =arg[2];

	if (!index)
		error(BAD_DATA, index);
	else
	{	if (!obj)
			error(BAD_DATA, obj, 2);
		else
			obj->removeitem(index);
	}
	return Unbound;
}


NodeP Node::reverse()
{
	error(BAD_DATA, this); 
	return Unbound;
}
NodeP Object::reverse()
{
	return _name->reverse();
}
NodeP String::reverse()
{
	StringP result=make(_ptr, 0, _length, STRING, copy_reverse);
	return result;
}
NodeP Pair::reverse()
{
	PairP result=0, ins=new Pair(),
		list=this;

	while (list)
	{	ins->setfirst(list->first());
		result=ins;
		ins=new Pair(0, ins);
		next(list);
	}
	return result;
}
NodeP Array::reverse()
{
	ArrayP result=new Array(_length);

	if (!throwing)
	{	NodePP d=_data, rev=result->_data+_length-1;

		for (int i=_length; i>0; i--)
			*rev-- = *d++;

		return result;
	}
	return Unbound;
}
NodeP IntArray::reverse()
{
	IntArrayP result=new IntArray(_length);

	if (!throwing)
	{	intP d=_data, rev=result->_data+_length-1;

		for (int i=_length; i>0; i--)
			*rev-- = *d++;

		return result;
	}
	return Unbound;
}
NodeP Int16Array::reverse()
{
	Int16ArrayP result=new Int16Array(_length);

	if (!throwing)
	{	shortP d=_data, rev=result->_data+_length-1;

		for (int i=_length; i>0; i--)
			*rev-- = *d++;

		return result;
	}
	return Unbound;
}
NodeP FloatArray::reverse()
{
	FloatArrayP result=new FloatArray(_length);

	if (!throwing)
	{	doubleP d=_data, rev=result->_data+_length-1;

		for (int i=_length; i>0; i--)
			*rev-- = *d++;

		return result;
	}
	return Unbound;
}

NodeP lreverse(PairPP arg)
{
	if (!arg[1])
		return 0;

	return arg[1]->reverse();
}

NodeP Node::rotate(int places) {error(BAD_DATA, this); return Unbound;}
NodeP Array::rotate(int places)
{
	places=Int::mod(places, length());
	if (!places) return this;

	ArrayP val = new Array(length(), origin());
	NodePP
		a= data(),
		b= & val->_data[places];

	int i;
	for (i=length()-places;  i;  i--)
	{	*b = *a++;
		check_oldyoung(*b++);	
	}
	b=val->data();
	for (i=places;  i;  i--)
	{	*b = *a++;
		check_oldyoung(*b++);	
	}
	return val;
}
NodeP IntArray::rotate(int places)
{
	places=Int::mod(places, length());
	if (!places) return this;

	IntArrayP val = new IntArray(length(), origin());
	intP
		a= _data,
		b= & val->_data[places];

	int i;
	for (i=length()-places;  i;  i--)
		*b++ = *a++;

	b=val->_data;
	for (i=places;  i;  i--)
		*b++ = *a++;

	return val;
}
NodeP Int16Array::rotate(int places)
{
	places=Int::mod(places, length());
	if (!places) return this;

	Int16ArrayP val = new Int16Array(length(), origin());
	shortP

		a= _data,
		b= & val->_data[places];

	int i;
	for (i=length()-places;  i;  i--)
		*b++ = *a++;

	b=val->_data;
	for (i=places;  i;  i--)
		*b++ = *a++;

	return val;
}
NodeP FloatArray::rotate(int places)
{
	places=Int::mod(places, length());
	if (!places) return this;

	FloatArrayP val = new FloatArray(length(), origin());
	doubleP
		a= _data,
		b= & val->_data[places];

	int i;
	for (i=length()-places;  i;  i--)
		*b++ = *a++;

	b=val->_data;
	for (i=places;  i;  i--)
		*b++ = *a++;

	return val;
}

NodeP lrotate(PairPP arg) 
{
	IntP places=arg[2]->a_integer(2);
	if (throwing) return Unbound;
	if (!arg[1]) return 0;
	return arg[1]->rotate( places->int_ );
}

NodeP Node::shuffle() {error(BAD_DATA, this); return Unbound;}
NodeP Array::shuffle()
{	ArrayP result=new Array(this);
	NodePP d=result->data();
	int l=length();
	for (int i=l;  i;  i--)
	{	int
			j=Int::random(l), 
			k=Int::random(l);
		NodeP tmp=d[j];
		d[j]=d[k];
		d[k]=tmp;
	}
	return result;
}
NodeP FloatArray::shuffle()
{	FloatArrayP result=new FloatArray(this);
	doubleP d=result->data();
	int l=length();
	for (int i=l;  i;  i--)
	{	int
			j=Int::random(l), 
			k=Int::random(l);
		double tmp=d[j];
		d[j]=d[k];
		d[k]=tmp;
	}
	return result;
}
NodeP IntArray::shuffle()
{	IntArrayP result=new IntArray(this);
	intP d=result->data();
	int l=length();
	for (int i=l;  i;  i--)
	{	int
			j=Int::random(l), 
			k=Int::random(l);
		int tmp=d[j];
		d[j]=d[k];
		d[k]=tmp;
	}
	return result;
}
NodeP Int16Array::shuffle()
{	Int16ArrayP result=new Int16Array(this);
	shortP d=result->data();
	int l=length();
	for (int i=l;  i;  i--)
	{	int
			j=Int::random(l), 
			k=Int::random(l);
		short tmp=d[j];
		d[j]=d[k];
		d[k]=tmp;
	}
	return result;
}

NodeP lshuffle(PairPP arg)
{
	NodeP thing=arg[1];
	if (thing)
		return thing->shuffle();
	return 0;
}

NodeP Node::remove(NodeP) {error(BAD_DATA, this, 2); return Unbound;}
NodeP Object::remove(NodeP thing) {return name()->remove(thing);}
NodeP String::remove(NodeP thing)
{
	wxString s=ptr();
	if (!thing)
	{	error(BAD_DATA, thing);
		return Unbound;
	}
	wxString th=thing->string()->ptr();
	if (th.Length()==0) 
		return this;
	
	int pos=s.Find(th);
	unsigned int thl=th.Length();
	while (pos!=-1)
	{	s.Remove(pos, thl);
		pos=s.Find(th);
	}
	if (s.Length()==0) 
		return StringP(_Null_Word);
	return new String(s);
}
NodeP Pair::remove(NodeP thing)
{
	PairP list=this, result=0, tnode=0;
	while (list)
	{	NodeP i=list->first();
		if ((!thing && i) || (thing && !i) || (i && !i->equal(thing)))
		{	if (!tnode)
			{	result=new Pair();
				tnode=result;
			}else
			{	tnode->settail(new Pair());
				next(tnode);
			}
			tnode->setfirst(i);
		}
		next(list);
	}
	return result;
}
NodeP lremove(PairPP arg)
{
	NodeP thing=arg[1], from=arg[2];
	if (!from)
		return 0;
	return from->remove(thing);
}

NodeP Node::destructiveRemove(NodeP thing)
{
	error(BAD_DATA, this);
	return Unbound;
}
NodeP Pair::destructiveRemove(NodeP thing)
{
	PairP list=this, last=0, result=0;
	while (list)
	{	while (list && (list->first()==thing || !list->first()))
		{	if (last)
				last->settail(list->tail());
			else
				result=list->tail();
			next(list);
		}
		last=list;
		if (list)
			next(list);
	}
	return result;
}
NodeP WindowPair::destructiveRemove(NodeP thing)
{
	WindowPairP list=this, last=0, result=list;
	while (list)
	{	while (list && (list->first()==thing || !list->first()))
		{	if (last)
				last->settail(list->tail());
			else
				result=WindowPairP(list->tail());
			list=WindowPairP(list->tail());
		}
		last=list;
		if (list)
			list=WindowPairP(list->tail());
	}
	return result;
}
NodeP ldestructiveRemove(PairPP arg)
{
	NodeP thing=arg[1], from=arg[2];
	if (!from)
		return 0;
	return from->destructiveRemove(thing);
}

NodeP Node::remdup() {error(BAD_DATA, this); return Unbound;}
NodeP Object::remdup() {return name()->remdup();}
NodeP String::remdup()
{
	int i, table[256];
	for (i=255; i>=0; i--)
		table[i]=0;

	charP end=&_ptr[_length-1], s=end;
	for (i=_length; i>0; i--)
	{	table[*s--]=1;
	}
	StringP res=new String(StringP(this));
	charP s2=res->_ptr;
	s=end;
	int l=0;
	for (i=_length; i>0; i--)
	{	if (table[*s]==1)
		{	table[*s]=-1;
			*s2++=*s;
			l++;
		}
		s--;
	}
	StringP result=make(res->_ptr, 0, l, STRING, copy_reverse);
	return result;
}
NodeP Pair::remdup()
{
	int size=count();
	TableP table=new Table(size);
	PairP revlist=PairP(reverse()), list=revlist;
	while (list)
	{	NodeP f=list->first();
		if (f)
		{	StringP s=f->string();
			if (String::caseIgnored)
				s=StringP(s->lower());
			if (!table->find(s)) 
				table->insert(s);
		}
		next(list);
	}
	PairP result=0;
	list=revlist;
	while (list)
	{	NodeP f=list->first();
		if (f)
		{	StringP s=f->string();
			if (String::caseIgnored)
				s=StringP(s->lower());

			ObjectP obj=table->find(s);
			if (obj && obj->var()==Unbound) 
			{	obj->setvar(0);
				result=new Pair(f, result);
			}
		}
		next(list);
	}
	return result;
}

NodeP lremdup(PairPP arg)
{
	if (!arg[1])
		return 0;
	return arg[1]->remdup();
}

PairP Node::list() {error(BAD_DATA, this); return PairP(Unbound);}
PairP Int::list() {return string()->list();}
PairP Int16::list() {return string()->list();}
PairP Int8::list() {return string()->list();}
PairP UInt8::list() {return string()->list();}
PairP Float::list() {return string()->list();}
PairP BigFloat::list() {return string()->list();}
PairP Complexd::list() {return string()->list();}
PairP Object::list() {return name()->list();}
PairP String::list()
{
	charP cp=ptr()+length()-1;
	PairP out=0;
	for (int i=length(); i; i--, cp--)
		out=new Pair(new String(cp, 0, 1, STRING, copy), out);
	return out;
}
PairP Array::list()
{
	NodePP d = &_data[_length];
	PairP p=0;

	for (int i=_length;  i;  i--)
		p = new Pair(*--d, p);

	return p;
}
PairP IntArray::list()
{
	intP d = &_data[_length];
	PairP p=0;

	for (int i=_length;  i;  i--)
		p = new Pair(new Int(*--d), p);

	return p;
}
PairP Int16Array::list()
{
	shortP d = &_data[_length];
	PairP p=0;

	for (int i=_length;  i;  i--)
		p = new Pair(new Int16(*--d), p);

	return p;
}
PairP FloatArray::list()
{
	doubleP d = &_data[_length];
	PairP p=0;

	for (int i=_length;  i;  i--)
		p = new Pair(new Float(*--d), p);

	return p;
}
PairP Table::list() {return contentsNameVar();}
PairP Struct::list()
{
	PairP li=0, last=0, tail=0;
	for (ObjectP obj=_first;  obj;  obj=ObjectP(obj->proc()))
	{	tail=	new Pair(
				new Pair(obj->name(), 
				new Pair(obj->var(), 0)), 0);
		if (last)
			last->settail(tail);
		else
			li=tail;
		last=tail;
	}
	return li;
}

NodeP ltoList(PairPP arg) {return arg[1]->list();}


ArrayP Node::array(int origin) {error(BAD_DATA, this); return ArrayP(Unbound);}
ArrayP Object::array(int origin) {return string()->array(origin);}
ArrayP String::array(int origin) 
{
	int len=length();
	charP cp=ptr();
	ArrayP out=new Array(len, origin);
	NodePP d=out->data();
	for (int i=len; i; i--)
		*d++=new String(cp++, 0, 1, STRING, copy);
	return out;
}
ArrayP Pair::array(int origin) 
{
	int len=0;
	PairP np;
	for (np=this;  np;  next(np))	// count elements of the list
		len++;

	ArrayP array = new Array(len, origin);
	NodePP d = array->data();

	for (np=this;  np;  next(np))	// copy the pointers from the list to the new array
		*d++ = np->first();

	return array;
}
ArrayP Array::array(int origin) 
{
	return new Array(ArrayP(this));
}
ArrayP IntArray::array(int origin) 
{
	ArrayP array = new Array(_length, _origin);
	NodePP d = array->data();
	intP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = new Int(*s++);

	return array;
}
ArrayP Int16Array::array(int origin) 
{
	ArrayP array = new Array(_length, _origin);
	NodePP d = array->data();
	shortP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = new Int16(*s++);

	return array;
}
ArrayP FloatArray::array(int origin) 
{
	ArrayP array = new Array(_length, _origin);
	NodePP d = array->data();
	doubleP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = new Float(*s++);

	return array;
}

NodeP lArray(PairPP arg) 
{
	IntP length=0;
	int origin;

	if (exist(2))
		origin = arg[2]->a_int(2);
	else 
		origin=1;
	
	if (!throwing) 
	{	if (arg[1]->number() != Unbound)
		{	length = arg[1]->a_pos_int(1);
			ArrayP array = new Array(length->int_, origin);
			return array;
		}else
			 return arg[1]->array(origin);
	}
	return Unbound;
}

IntArrayP Node::intarray(int origin) 
{	error(BAD_DATA, this); return IntArrayP(Unbound);
}
IntArrayP Pair::intarray(int origin) 
{
	int len=0;
	PairP np;
	for (np=this;  np;  next(np))	// count elements of the list
		len++;

	IntArrayP intarray = new IntArray(len, origin);
	intP d = intarray->data();

	for (np=this;  np;  next(np))	// copy the pointers from the list to the new array
	{	NodeP h=np->first();
		switch (type_of(np->first()))
		{	case INT_:	*d++ =             IntP(h)->int_;		break;
			case FLOAT_:*d++ = Int::clip(FloatP(h)->float_);	break;
			case INT16: *d++ =           Int16P(h)->int_;		break;
			case INT8:	*d++ =            Int8P(h)->int_;		break;
			case STRING:if (h->numberP())
								*d++ = Int::clip(h->getnumber());
							else
							{	error(BAD_DATA, this); return IntArrayP(Unbound);
							}	
							break;
			default:
				error(BAD_DATA, this);	return IntArrayP(Unbound);
		}
	}
	return intarray;
}
IntArrayP Array::intarray(int origin) 
{
	IntArrayP intarray = new IntArray(_length, _origin);
	intP d = intarray->data();
	NodePP s=_data;

	for (int i=_length;  i;  i--)
	{	NodeP h=*s;
		switch (type_of(h))
		{	case INT_:	*d++ =             IntP(h)->int_;	break;
			case FLOAT_:*d++ = Int::clip(FloatP(h)->float_);break;
			case INT16:	*d++ =           Int16P(h)->int_;	break;
			case INT8:	*d++ =            Int8P(h)->int_;	break;
			case STRING:if (h->numberP())
								*d++ = Int::clip(h->getnumber());
							else
							{	error(BAD_DATA, this); return IntArrayP(Unbound);
							}	
							break;
			default:
				error(BAD_DATA, this);	return IntArrayP(Unbound);
		}
		s++;
	}
	return intarray;
}
IntArrayP IntArray::intarray(int origin) 
{
	return new IntArray(IntArrayP(this));
}
IntArrayP Int16Array::intarray(int origin) 
{
	IntArrayP intarray = new IntArray(_length, _origin);
	intP d = intarray->data();
	shortP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = *s++;

	return intarray;
}
IntArrayP FloatArray::intarray(int origin) 
{
	IntArrayP intarray = new IntArray(_length, _origin);
	intP d = intarray->data();
	doubleP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = Int::clip(*s++);

	return intarray;
}

NodeP lIntArray(PairPP arg) 
{
	IntP length=0;
	int origin;

	if (exist(2))
		origin = arg[2]->a_int(2);
	else 
		origin=1;
	
	if (!throwing) 
	{	if (arg[1]->number() != Unbound)
		{	length = arg[1]->a_pos_int(1);
			IntArrayP array = new IntArray(length->int_, origin);
			return array;

		}else
			return arg[1]->intarray(origin);
	}
	return Unbound;
}

Int16ArrayP Node::int16array(int origin) 
{	error(BAD_DATA, this); return Int16ArrayP(Unbound);
} 
Int16ArrayP Pair::int16array(int origin) 
{
	int len=0;
	PairP np;
	for (np=this;  np;  next(np))	// count elements of the list
		len++;

	Int16ArrayP int16array = new Int16Array(len, origin);
	shortP d = int16array->data();

	for (np=this;  np;  next(np))	// copy the pointers from the list to the new array
	{	NodeP h=np->first();
		switch (type_of(h))
		{	case INT_:	*d++ = Int16::clip(  IntP(h)->int_);	break;
			case FLOAT_:*d++ = Int16::clip(FloatP(h)->float_);	break;
			case INT16: *d++ =             Int16P(h)->int_;		break;
			case INT8:	*d++ =              Int8P(h)->int_;		break;
			case STRING:if (h->numberP())
								*d++ = Int16::clip(h->getnumber());
							else
							{	error(BAD_DATA, this); return Int16ArrayP(Unbound);
							}	
							break;
			default:
				error(BAD_DATA, this);	return Int16ArrayP(Unbound);
		}
	}
	return int16array;
}
Int16ArrayP Array::int16array(int origin) 
{
	Int16ArrayP int16array = new Int16Array(_length, _origin);
	shortP d = int16array->data();
	NodePP s=_data;

	for (int i=_length;  i;  i--)
	{	NodeP h=*s;
		switch (type_of(h))
		{	case INT_:	*d++ =   Int16::clip(IntP(h)->int_);	break;
			case FLOAT_:*d++ = Int16::clip(FloatP(h)->float_);	break;
			case INT16:	*d++ =             Int16P(h)->int_;		break;
			case INT8:	*d++ =              Int8P(h)->int_;		break;
			case STRING:if (h->numberP())
								*d++ = Int16::clip(h->getnumber());
							else
							{	error(BAD_DATA, this); return Int16ArrayP(Unbound);
							}	
							break;
			default:
				error(BAD_DATA, this);	return Int16ArrayP(Unbound);
		}
		s++;
	}
	return int16array;
}
Int16ArrayP IntArray::int16array(int origin) 
{
	Int16ArrayP int16array = new Int16Array(_length, _origin);
	shortP d = int16array->data();
	intP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = Int16::clip(*s++);

	return int16array;
}
Int16ArrayP Int16Array::int16array(int origin) 
{
	return new Int16Array(Int16ArrayP(this));
}
Int16ArrayP FloatArray::int16array(int origin) 
{
	Int16ArrayP int16array = new Int16Array(_length, _origin);
	shortP d = int16array->data();
	doubleP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = Int16::clip(*s++);

	return int16array;
}

NodeP lInt16Array(PairPP arg) 
{
	IntP length=0;
	int origin;

	if (exist(2))
		origin = arg[2]->a_int(2);
	else 
		origin=1;
	
	if (!throwing) 
	{	if (arg[1]->number() != Unbound)
		{	length = arg[1]->a_pos_int(1);
			Int16ArrayP array = new Int16Array(length->int_, origin);
			return array;
		}else
			return arg[1]->int16array(origin);
	}
	return Unbound;
}

FloatArrayP Node::floatarray(int origin) 
{	error(BAD_DATA, this); return FloatArrayP(Unbound);
}
FloatArrayP Pair::floatarray(int origin) 
{
	int len=0;
	PairP np;
	for (np=this;  np;  next(np))	// count elements of the list
		len++;

	FloatArrayP floatarray = new FloatArray(len, origin);
	doubleP d = floatarray->data();

	for (np=this;  np;  next(np))	// copy the pointers from the list to the new array
	{	NodeP h=np->first();
		switch (type_of(np->first()))
		{	case INT_:	*d++ =   IntP(h)->int_;		break;
			case FLOAT_:*d++ = FloatP(h)->float_;	break;
			case INT16: *d++ = Int16P(h)->int_;		break;
			case INT8:	*d++ =  Int8P(h)->int_;		break;
			case STRING:if (h->numberP())
								*d++ = h->getnumber();
							else
							{	error(BAD_DATA, this); return FloatArrayP(Unbound);
							}	
							break;
			default:
				error(BAD_DATA, this);	return FloatArrayP(Unbound);
		}
	}
	return floatarray;
}
FloatArrayP Array::floatarray(int origin) 
{
	FloatArrayP floatarray = new FloatArray(_length, _origin);
	doubleP d = floatarray->data();
	NodePP s=_data;

	for (int i=_length;  i;  i--)
	{	NodeP h=*s;
		switch (type_of(h))
		{	case INT_:	*d++ =   IntP(h)->int_;	break;
			case FLOAT_:*d++ = FloatP(h)->float_;break;
			case INT16:	*d++ = Int16P(h)->int_;	break;
			case INT8:	*d++ =  Int8P(h)->int_;	break;
			case STRING:if (h->numberP())
								*d++ = h->getnumber();
							else
							{	error(BAD_DATA, this); return FloatArrayP(Unbound);
							}	
							break;
			default:
				error(BAD_DATA, this);	return FloatArrayP(Unbound);
		}
		s++;
	}
	return floatarray;
}
FloatArrayP IntArray::floatarray(int origin) 
{
	FloatArrayP floatarray = new FloatArray(_length, _origin);
	doubleP d = floatarray->data();
	intP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = *s++;

	return floatarray;
}
FloatArrayP Int16Array::floatarray(int origin) 
{
	FloatArrayP floatarray = new FloatArray(_length, _origin);
	doubleP d = floatarray->data();
	shortP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = *s++;

	return floatarray;
}
FloatArrayP FloatArray::floatarray(int origin) 
{
	return new FloatArray(FloatArrayP(this));
}

NodeP lFloatArray(PairPP arg) 
{
	IntP length=0;
	int origin;

	if (exist(2))
		origin = arg[2]->a_int(2);
	else 
		origin=1;
	
	if (!throwing) 
	{	if (arg[1]->number() != Unbound)
		{	length = arg[1]->a_pos_int(1);
			FloatArrayP array = new FloatArray(length->int_, origin);
			return array;
		}else
			return arg[1]->floatarray(origin);
	}
	return Unbound;
}

TableP Node::table() {return TableP(Unbound);}
TableP Pair::table() 
{
	if (!throwing) 
	{	int len=0;
		PairP np;
		for (np=this;  np;  next(np))	// count elements of the list
		{	if (!np->first() || !np->first()->first() 
			||  !np->first()->tail())
			{	error(BAD_DATA, this);
				return TableP(Unbound);
			}
			len++;
		}
		TableP table = new Table(len);

		for (np=this;  np;  next(np))	// copy the pointers from the list to the new array
		{	StringP f=np->first()->first()->string();
			if throwing return TableP(Unbound);
			ObjectP obj=table->find(f);
			if (!obj)
				obj=table->insert(f);
			obj->setvar(np->first()->second());
		}
		return table;
	}
	return TableP(Unbound);
}

NodeP lTable(PairPP arg)
{
	if (arg[1]->number() != Unbound)
	{	int size=arg[1]->a_int(1);
		if throwing return Unbound;
		return new Table(size);
	}
	return arg[1]->table();
}

NodeP lStruct(PairPP arg)
{
	PairP data=arg[1]->a_list(1);
	if throwing return Unbound;
	return new Struct(data);
}

NodeP lmerge(PairPP arg)
{
	PairP 
		a=arg[1]->a_list(1),
		b=arg[2]->a_list(2);
	return a->merge(b);
}

NodeP lmergePairs(PairPP arg)
{
	PairP a=arg[1]->a_list(1);
	a->mergePairs();
	return a;
}

NodeP ltoListList(PairPP arg)
{
	PairP a=arg[1]->a_list(1);
	a->toListList();
	return a;
}

NodeP lmergeSort(PairPP arg)
{
	PairP a=arg[1]->a_list(1);
	return a->mergeSort();
}

NodeP lgenSym(PairPP)
{
	static char sym[255];
	Int::genSymNr++;
	sprintf(sym, "g%d", Int::genSymNr);
	return new String(sym, 0, strlen(sym), STRING, String::copy);
}

NodeP lForm(PairPP arg) 
{
	double num=arg[1]->a_float(1);
	int  width=arg[2]->a_int(2), precision=0;

	if throwing return Unbound;

	static char result[200];

	if (width < 0) 
	{	const char maxlength=20;
		static char format[maxlength];
		File::stringptr=format;
		File::stringlen=maxlength;

		StringP str=arg[3]->a_string(3);
		if (str->length() >= maxlength)
			error(BAD_DATA, arg[3], 3);

		if throwing return Unbound;

		stringbuffer << str << endL;
		*File::stringptr='\0';

//		__try {
			sprintf(result, format, num);
/*		}
		__except(1) 
		{	error(BAD_DATA, arg[1]);
		}
*/	}else
		precision = arg[3]->a_int(3);
	
	if (width >= 100) width=99;
	if (width >= 0)
		sprintf(result, "%*.*f", width, precision, num);

	return String::make(result, 0, strlen(result), STRING, String::copy);
}

NodeP lintForm(PairPP arg) 
{
	NodeP nd=arg[1]->a_number(1);
	if (nd->type==BIGFLOAT)
	{	wxString s('\0', MM_cpp_min_precision+20);
		BigFloatP(nd)->float_.toIntegerString(charP(s.c_str()));
		return new String(s);
	}
	unsigned int num=nd->a_int(1);
	int width=arg[2]->a_int(2),
		  base=arg[3]->a_int(3);

	if throwing return Unbound;

	if (width > 255 || width < 0)
	{	error(BAD_DATA, arg[2], 2);
		return Unbound;
	}
	if (base > 35 || base < 2)
	{	error(BAD_DATA, arg[3], 3);
		return Unbound;
	}


	const int maxlength=256;
	char numstr[maxlength+1];
	for (int i=0; i<maxlength; i++)
		numstr[i]=' ';
	numstr[maxlength]='\0';

	charP ciphers="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";

	charP str=&numstr[maxlength-1];
	int cipher;
	while (num)
	{	cipher = num % base;
		*str-- = ciphers[cipher];
		num/=base;
		width--;
	}
	while (width>0)
	{	str--;
		width--;
	}
	str++;
	return String::make(str, 0, strlen(str), STRING, String::copy);
}

//___________________________________________________________________
//
//                          Property Lists
//___________________________________________________________________

NodeP lPropertyList(PairPP arg) 
{
	StringP plistname = arg[1]->a_string(1);

	if throwing return 0;

	ObjectP obj = plistname->intern();
	PairP plist=obj->plist();
	if (plist)
		plist = new List(plist);

	return plist;
}

PairP Pair::getProperty(StringP name, bool before)
{
	PairP plist=this, prev=0;
	while (plist) 
	{	if (name->equal(plist->first()))
			return before ?   prev  :  plist;
		prev=plist;
		plist=plist->tail()->tail();
	}
	return 0;
}

NodeP lgetProperty(PairPP arg) 
{
	StringP plistname = arg[1]->a_string(1),
	    property_name = arg[2]->a_string(2);

	if (!throwing) 
	{	PairP	plist = plistname->intern()->plist(),  property=0;

		if (plist)
			property = plist->getProperty(property_name, false);

		if (property)
			return property->second();
	}
	return 0;
}

NodeP lputProperty(PairPP arg) 
{
	StringP plistname = arg[1]->a_string(1),
		 property_name = arg[2]->a_string(2);

	PairP newval = arg[3];

	if (!throwing) 
	{
		ObjectP obj = plistname->intern();


		if (obj->bit.plist_traced) 
		{	writer << "PutProperty "
				<< obj->maybe_quote() << ' '
				<< property_name->maybe_quote() << ' ' 
				<< newval->maybe_quote();
			if (ufun)
				writer << " in " << ufun << endL << this_line;
			writer << endL;
		}
		PairP	plist=obj->plist(),  property=0;

		if (plist)
			property = plist->getProperty(property_name, false);

		if (property)
			property->setsecond( newval );
		else
			obj->setplist( PairP(new Pair(property_name, new Pair(newval, plist))) );
	}
	return Unbound;
}

NodeP lremoveProperty(PairPP arg) 

{
	StringP plistname = arg[1]->a_string(1), 
		 property_name = arg[2]->a_string(2);

	if (!throwing) 
	{	ObjectP obj = plistname->intern();
		PairP plist=obj->plist();

		if (plist) 
		{	if (plist->first()->equal(property_name))
				obj->setplist( plist->tail()->tail() );
			else 
			{	PairP property = plist->getProperty(property_name, true);
				if (property)
					property->tail()->settail( 
						property->tail()->tail() ->tail()->tail() );
			}
		}
	}
	return Unbound;
}
