//___________________________________________________________________
//
//                        Class LogoApp
//
//                         Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/evtloop.h>
#include <wx/glcanvas.h>
#include <wx/image.h>
#include <wx/listctrl.h>
//#include <wx/html/helpctrl.h>
#include <wx/help.h>
#include <wx/splitter.h>
#include <wx/config.h>
#include <wx/datetime.h>
#include <wx/sound.h>
#include <wx/wxMidi.h>

#include <math.h>
#ifdef WIN32
	#include <mmsystem.h>
	#include <mmreg.h>
#else
	#include <sys/time.h>
#endif

#include <GL/glu.h>

#include "Main.h"
#include "Pair.h"
#include "Int.h"
#include "UInt8.h"
#include "Float.h"
#include "File.h"
#include "Object.h"
#include "String.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "Init.h"
#include "List.h"
#include "Error.h"
#include "Eval.h"
#include "Stack.h"

#include "Files.h"

#include "Color.h"
#include "Point.h"
#include "Pen.h"
#include "Font.h"
#include "Turtle.h"
#include "FTGLFontManager.h"

#include "Event.h"
#include "Console.h"
#include "Graph.h"
#include "Vars.h"
#include "Calls.h"
#include "Splitter.h"
#include "MainFrame.h"
#include "EdDefsExt.h"
#include "Edit.h"
#include "EdFrame.h"
#include "MemoryWindow.h"
#include "Midi.h"
#include "LogoApp.h"

#ifdef WIN32
	#include "io.h"
#endif

#ifdef _DEBUG
	#undef THIS_FILE
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

static char THIS_FILE[]= __FILE__;

  DECLARE_APP(LogoApp)
IMPLEMENT_APP(LogoApp)

LogoApp *theApp;
Console *console;
Graph *graph;
MemoryWindow *memoryWindow;

#define NEED_MEM_LogoApp_OnInit(x)						\
	if (!x)														\
	{	delete heapReserve;									\
		heapReserve=0;											\
		wxString msg;											\
		msg << THIS_FILE << "(" << __LINE__ << "):"	\
			#x": Insufficient Memory";						\
		wxMessageBox(msg,	"aUCBLogo Error");			\
		return false;											\
	}

bool LogoApp::OnInit()
{
	theApp=this;
	heapReserve=new char[0x100000];
	if (!heapReserve)
		return false;
	action=NOP;
	saved=false;
	console=0;
	edFrame=0;
	max.x=wxSystemSettings::GetMetric(wxSYS_SCREEN_X);
	max.y=wxSystemSettings::GetMetric(wxSYS_SCREEN_Y);
	osVersion=wxGetOsVersion(&osMajor, &osMinor);

	aUCBLogo=::aUCBLogo;
	SetAppName(aUCBLogo);

	config = new wxConfig(aUCBLogo);
	NEED_MEM_LogoApp_OnInit(config);
	config->Read("MakeBackupFiles", &backup, true);
	runDirectory = config->Read("RunDirectory", wxGetCwd());
#ifdef WIN32
	helpFileName = config->Read("HelpFileName", 
		wxGetCwd()+_T("\\aUCBLogo.chm"));
#else
	helpFileName = config->Read("HelpFileName",
		wxGetCwd()+_T("/wxhelp/aUCBLogo.hhp"));
#endif
	int posx=config->Read("PosX", (long int)(0)),
		posy=config->Read("PosY", (long int)(0)),
		sizex=config->Read("SizeX", max.x/2),
		sizey=config->Read("SizeY", int(max.y*0.96));

	int edposx=config->Read("EdPosX", max.x/2),
		edposy=config->Read("EdPosY", (long int)(0)),
		edsizex=config->Read("EdSizeX", max.x/2),
		edsizey=config->Read("EdSizeY", int(max.y*0.96));

	wxInitAllImageHandlers();
//*
	edFrame = new EdFrame("aUCBLogo", wxPoint(edposx, edposy), 
		wxSize(edsizex, edsizey), argc, argv);
	NEED_MEM_LogoApp_OnInit(edFrame);
#ifndef WIN32
	edFrame->Show();
#endif

	int fileCount=config->Read("FilesOpen", (long int)(0));
	int i;
	for (i=0;  i < fileCount;  i++)
	{	wxString name;
		name.Printf("File%i", i+1);
		edFrame->FileOpen( config->Read(name, "noname.lg") );
	}
	int selectedPage=config->Read("SelectedPage", (long int)(-1));
	if (selectedPage >= 0)
		edFrame->notebook->SetSelection(selectedPage);

#ifdef WIN32
	edFrame->Show();
#endif
	for (i=0;  i < fileCount;  i++)
	{	wxString name;
		name.Printf("Pos%i", i+1);
		((Edit*)edFrame->notebook->GetPage(i))
			->GotoPosVisible( config->Read(name, (long int)(0)) );
	}
//*/

	memposx =config->Read("MemoryPosX", max.x/2),
	memposy =config->Read("MemoryPosY", (long int)(0)),
	memsizex=config->Read("MemorySizeX", max.x/2),
	memsizey=config->Read("MemorySizeY", int(max.y*0.96));

	memoryShown=false;
	config->Read("MemoryShown", &memoryShown);
	if (memoryShown)
		memoryShow();

	wxString userName=wxGetCwd()+'/'+wxGetUserName();
	consoleFileName=userName+"-aucblogo.console";
	pictureFileName=userName+"-aucblogo.picture";
	  nodesFileName=userName+"-aucblogo.nodes";

	int caretpos=config->Read("ConsoleCaretPos", long(0));
	bool consoleAutosave, pictureAutosave, nodesAutosave;
	config->Read("ConsoleAutoSave", &consoleAutosave, true);
	config->Read("PictureAutoSave", &pictureAutosave, false);
	config->Read("NodesAutoSave", &nodesAutosave, false);
	config->Read("DoubleBuffer", &doubleBuffer, true);

	mainFrame=new MainFrame(NULL, -1, aUCBLogo, 
		wxPoint(posx, posy), wxSize(sizex, sizey),
		wxDEFAULT_FRAME_STYLE, 
		aUCBLogo, 
		consoleAutosave, caretpos, pictureAutosave, nodesAutosave);
	NEED_MEM_LogoApp_OnInit(mainFrame);
	mainFrame->Show();
	console=mainFrame->console;
	graph=mainFrame->graph;

	help=new wxHelpController();//wxHF_DEFAULT_STYLE /*| wxHF_MERGE_BOOKS*/);
	NEED_MEM_LogoApp_OnInit(help);
	//	help->AddBook(helpFileName);
	help->Initialize(helpFileName);
	ftglFontManager=new FTGLFontManager;
	NEED_MEM_LogoApp_OnInit(ftglFontManager);
	return true;
}

int LogoApp::OnExit () 
{
	config->Write("MakeBackupFiles", backup);
	config->Write("RunDirectory", runDirectory);
	config->Write("HelpFileName", helpFileName);
	delete config;
	if (help)
	{	help->Quit();
		delete help;
	}
	if (heapReserve)
		delete heapReserve;
	return 0;
}

#ifdef WIN32
	#pragma optimize("", off)
#endif
void LogoApp::doAction()
{
	switch (action)
	{
	case EXIT_LOGO:	if (graph->recorder != graph->recorderSave)
							{	action=EXIT_DELAYED;
								c_out << ";Warning: EXIT is delayed because " << endL
		<< ";the program is in the middle of GraphicStart..GraphicEnd" << endL;
								break;
							}
							if (graph->turtle->polyMode)
							{	action=EXIT_DELAYED;
								c_out << ";Warning: EXIT is delayed because " << endL
		<< ";the program is in the middle of PolyStart..PolyEnd" << endL;
								break;
							}
							if (graph->turtle->tessMode)
							{	action=EXIT_DELAYED;
								c_out << ";Warning: EXIT is delayed because " << endL
		<< ";the program is in the middle of TessStart..TessEnd" << endL;
								break;
							}
							if (graph->turtle->surfaceMode)
							{	action=EXIT_DELAYED;
								c_out << ";Warning: EXIT is delayed because " << endL
		<< ";the program is in the middle of SurfaceStart..SurfaceEnd" << endL;
								break;
							}
	case EXIT_DELAYED:if (graph->recorder != graph->recorderSave
							|| graph->turtle->polyMode
							|| graph->turtle->tessMode
							|| graph->turtle->surfaceMode)
								break;
							graph->save();
							if (mainFrame->nodesAutosave && !saved)
							{	GC_::save(charP(theApp->nodesFileName.c_str()));
								saved=true;
							}
							stoppingFlag=THROWING;
							throw_node = NodeP(_System);
							break;
	case LOAD_NODES:	
		{	action=NOP;
			StringP name=0;
			{	wxFileDialog fd(mainFrame, "Choose a Nodes file","","",
					"Nodes Files (*.zip)|*.zip|All Files (*.*)|*.*",
					wxOPEN|wxCHANGE_DIR);
				if (fd.ShowModal()==wxID_OK)
				name=new String(fd.GetPath());
			}
			if (name)
				GC_::load(name->ptr());
			break;
		}
	case SAVE_NODES:	
		{	action=NOP;
			StringP name=0;
			{	wxFileDialog fd(mainFrame, "Choose a Nodes file","","",
					"Nodes Files (*.zip)|*.zip|All Files (*.*)|*.*",
					wxSAVE|wxCHANGE_DIR);
				if (fd.ShowModal()==wxID_OK)
				name=new String(fd.GetPath());
			}
			if (name)
				GC_::save(name->ptr());
			break;
		}
	default: 
		;
	}
	Event::dispatch();
}
#ifdef WIN32
	#pragma optimize("", on)
#endif

void LogoApp::memoryShow()
{
	memoryFrame=new MemoryFrame(0, -1, "Memory", 
		wxPoint(memposx, memposy), wxSize(memsizex,memsizey),
		wxDEFAULT_FRAME_STYLE, "Memory");
	if (memoryFrame)
		memoryFrame->Show();
}	

void LogoApp::memoryHide()
{
	if (memoryFrame)
	{	delete memoryFrame;
		memoryFrame=0;
	}
}	


NodeP lLogoInstance(PairPP)
{
	return new Int(int(theApp->GetInstance()));
}

NodeP lwait(PairPP arg) 
{
	IntP i = arg[1]->a_pos_int(1);

	if throwing return Unbound;

	console->flush();

	if (i->int_ > 0) 
	{	int n=i->int_;
		while (n > 60) 
		{
#ifdef WIN32
			Sleep(1000);
#else
			sleep(1);
#endif
			n-=60;
			if (checkThrowing()) n=0;
		}
		if (n > 0)
#ifdef WIN32
			Sleep(n*1000/60);
#else
			usleep(n*1000000/60);
#endif
	}
	return Unbound;
}

NodeP lwaitMS(PairPP arg) 
{
	IntP i = arg[1]->a_pos_int(1);

	if throwing return Unbound;

	console->flush();

	if (i->int_ > 0) 
	{	int n=i->int_;
		while (n > 1000) 
		{
#ifdef WIN32

			Sleep(1000);
#else
			sleep(1);
#endif
			n-=1000;
			if (checkThrowing()) n=0;
		}
		if (n > 0)
#ifdef WIN32
			Sleep(n);
#else
			usleep(n*1000);
#endif
	}
	return Unbound;
}

void LogoApp::waituS(int n)
{
	while (n > 1000000) 
	{
#ifdef WIN32
		Sleep(1000);
#else
		sleep(1);
#endif
		n-=1000000;
		if (checkThrowing()) n=0;
	}
	if (n > 0)
#ifdef WIN32
	{	TimeUnion f,c;
		QueryPerformanceCounter(&c.i);
		QueryPerformanceFrequency(&f.i);
		if (!f.u)
			return;

		double t,t0=double(c.u)/f.u;
		do
		{	QueryPerformanceCounter(&c.i);
			t=double(c.u)/f.u;
		} while (1000000*(t-t0) < n);

	}
#else
		usleep(n);
#endif
}

NodeP lwaituS(PairPP arg) 
{
	IntP i = arg[1]->a_pos_int(1);

	if throwing return Unbound;

	if (i->int_ > 0) 
	{	int n=i->int_;
		theApp->waituS(n);
	}
	return Unbound;
}

NodeP lTimeFine(PairPP)
{
#ifdef WIN32
	TimeUnion f,c;
	QueryPerformanceFrequency(&f.i);
	QueryPerformanceCounter(&c.i);
	if (f.u)
		return new Float(double(c.u)/f.u);
	else
		return Unbound;
#else
	timeval tv;
	gettimeofday(&tv, 0);
	return new Float(tv.tv_sec+tv.tv_usec/1000000.0);
#endif
}

NodeP lTimeMilli(PairPP)
{

#ifdef WIN32
	TimeUnion f,c;
	QueryPerformanceFrequency(&f.i);
	QueryPerformanceCounter(&c.i);
	if (f.u)
		return new Float(1000*double(c.u)/f.u);
	else
		return Unbound;
#else
	timeval tv;
	gettimeofday(&tv, 0);
	return new Float(1000.0*tv.tv_sec+tv.tv_usec/1000.0);
#endif
}

NodeP lTimeu(PairPP)
{
#ifdef WIN32
	TimeUnion f,c;
	QueryPerformanceFrequency(&f.i);
	QueryPerformanceCounter(&c.i);
	if (f.u)
		return new Float(1000000*double(c.u)/f.u);
	else
		return Unbound;
#else
	timeval tv;
	gettimeofday(&tv, 0);
	return new Float(1000000.0*tv.tv_sec+tv.tv_usec);
#endif
}

NodeP lTimerFreq(PairPP)
{
#ifdef WIN32
	TimeUnion f;
	QueryPerformanceFrequency(&f.i);
	if (f.u)
		return new Float(double(f.u)/1000000);
	else
		return Unbound;
#else
	return new Int(1);
#endif
	return Unbound;
}

NodeP lTimeuRes(PairPP)
{
#ifdef WIN32
	TimeUnion f;
	QueryPerformanceFrequency(&f.i);
	if (f.u)
		return new Float(1000000./f.u);
	else
		return Unbound;
#else
	return new Int(1);
#endif
}

NodeP lMIPS(PairPP)
{
#ifdef WIN32
	TimeUnion f,c0, c;
	QueryPerformanceFrequency(&f.i);
	if (f.u)
	{	QueryPerformanceCounter(&c0.i);
		_asm
		{	mov eax, 100000000
		again:
			dec eax
			jnz again
		}
		QueryPerformanceCounter(&c.i);
		return new Float(200*double(f.u)/(c.u-c0.u));
	}else
		return Unbound;
#else
	timeval tv, tv2;
	gettimeofday(&tv, 0);
	asm("movl $100000000,%eax;  again:  decl %eax;  jnz again");
	gettimeofday(&tv2, 0);
	return new Float(200/(tv2.tv_sec-tv.tv_sec+(tv2.tv_usec-tv.tv_usec)/1000000.0));
#endif
	return Unbound;
}

NodeP lTime(PairPP)
{
	wxDateTime t=wxDateTime::UNow();
	return new List(
		new Int(t.GetHour()), new Int(t.GetMinute()), new Int(t.GetSecond()), 
		new Int(t.GetMillisecond()), 
		new Int(t.GetWeekDay()), 
		new Int(t.GetDay()), new Int(t.GetMonth()+1), new Int(t.GetYear()), _ListEnd);
	// There's a bug in GetMonth(), so we need the +1!
}

void LogoApp::Sound(double frequency, double duration)
{
	if (frequency < 19) frequency=19;

	switch (osVersion)
	{
#ifdef WIN32
	case wxWINDOWS_NT:
		Beep((int)frequency, duration); break;
/*		{	int count = (int)(1193180/frequency);
			PortOut(0x43, char(0xB6));
			PortOut(0x42,  count     && 0xff);
			PortOut(0x42, (count>>8) && 0xff);
			int bits=PortIn(0x61);
			PortOut(0x61, bits|3);
			if (duration!=INFINITE)
			{	waituS(duration*1000);
				int bits=PortIn(0x61);
				PortOut(0x61, bits & 0xF8);
			}
			break;
		}
*/	case wxWINDOWS:
	case wxWIN95:
	case wxWIN386:
		int count = (int)(1193180/frequency);
		_asm 
		{	mov al, 0xB6 // set timer 
			out 0x43, al
			mov eax, count
			out 0x42, al
			mov al, ah
			out 0x42, al
			in al, 0x61
			or al, 3 // set speaker bits
			out 0x61, al // sound on.
		}
		if (duration!=INFINITE)
		{	Sleep(duration);
			_asm 
			{	in al, 0x61
				and al, 0xf8 // reset speaker bits
				out 0x61, al // sound off.
			}
		}
#endif
	}
}

void LogoApp::Tone(double pitch, double duration)
{
	double frequency = 440*pow(2, pitch/12);
	Sound(frequency, duration);
}

NodeP lSound(PairPP arg) 
{
	double freq=arg[1]->a_float(1),
		duration=arg[2]->a_float(2);
	if throwing return Unbound;
	theApp->Sound(freq, duration);
	return Unbound;
}

NodeP lTone(PairPP arg) 
{
	double pitch=arg[1]->a_float(1),
		 duration=arg[2]->a_float(2);
	if throwing return Unbound;
	theApp->Tone(pitch, duration);
	return Unbound;
}

class SoundsThread: public wxThread
{
	int length;
	double *freq, *duration;
public:
	static volatile int finished;
	static volatile bool stop;

	SoundsThread(int length_, double *freq_, double *duration_): 
		length(length_), freq(freq_), duration(duration_) 
	{}

	ExitCode Entry()
	{	double *f=freq, *d=duration;
		for (int i=length;  i>0;  i--)
		{	theApp->Sound(*f++, *d++);
			if (stop)
				break;
		}
		finished++;
		stop=false;
		delete freq;
		delete duration;
		return 0;
	}
};

volatile int SoundsThread::finished=0;
volatile bool SoundsThread::stop=false;

NodeP lSounds(PairPP arg) 
{
	int length=0;
	double *freq=0, *duration=0;

	if (exist(2))
	{	PairP flist=arg[1]->a_list(1);
		if throwing return Unbound;
		NodeP	dnode=arg[2];
		if (!dnode)
		{	error(BAD_DATA, arg[2], 2);
			return Unbound;
		}
		length=flist->count();
		freq=new double[length]; double *f=freq;
		duration=new double[length]; double *d=duration;

		if (dnode->islist())
		{	PairP dlist=PairP(dnode);
			while (flist && flist->first() 
				 && dlist && dlist->first())
			{	*f++=flist->first()->a_float(1);
				*d++=dlist->first()->a_int(1);
				if throwing return Unbound;
				next(flist);
				next(dlist);
			}
		}
		else if (dnode->numberP())
		{	double di=dnode->getnumber();
			while (flist && flist->first())
			{	*f++=flist->first()->a_float(1);
				*d++=di;
				if throwing return Unbound;
				next(flist);
			}
		}
	}else
	{	PairP list=arg[1]->a_list(1);
		if throwing return Unbound;
		length=list->count()/2;
		freq=new double[length]; double *f=freq;
		duration=new double[length]; double *d=duration;
		while (list && list->first() && list->tail() && list->second())
		{	*f++=list->first()->a_float(1);
			*d++=list->second()->a_int(1);
			if throwing return Unbound;
			next(list);
			next(list);
		}
	}
	if throwing return Unbound;
	SoundsThread *t=new SoundsThread(length, freq, duration);
	t->Create(0x100);
	t->Run();
	return Unbound;
}

NodeP lSoundsStop(PairPP)
{
	SoundsThread::stop=true;
	return Unbound;
}

NodeP lSoundsFinished(PairPP)
{
	return new Int(SoundsThread::finished);
}

class TonesThread: public wxThread
{
	int length;
	double *pitch, *duration;
public:
	static volatile int finished;
	static volatile bool stop;

	TonesThread(int length_, double *pitch_, double *duration_): 
		length(length_), pitch(pitch_), duration(duration_) 
	{}

	ExitCode Entry()
	{	double *p=pitch, *d=duration;
		for (int i=length;  i>0;  i--)
		{	theApp->Tone(*p++, *d++);
			if (stop)
				break;
		}
		finished++;
		stop=false;
		delete pitch;
		delete duration;
		return 0;
	}
};

volatile int TonesThread::finished=0;
volatile bool TonesThread::stop=false;

NodeP lTones(PairPP arg) 
{
	int length=0;
	double *pitch=0, *duration=0;

	if (exist(2))
	{	PairP plist=arg[1]->a_list(1);
		if throwing return Unbound;
		NodeP	dnode=arg[2];
		if (!dnode)
		{	error(BAD_DATA, arg[2], 2);
			return Unbound;
		}
		length=plist->count();
		pitch=new double[length]; double *p=pitch;
		duration=new double[length]; double *d=duration;

		if (dnode->islist())
		{	PairP dlist=PairP(dnode);
			while (plist && plist->first() 
				 && dlist && dlist->first())
			{	*p++=plist->first()->a_float(1);
				*d++=dlist->first()->a_int(1);
				if throwing return Unbound;
				next(plist);
				next(dlist);
			}
		}
		else if (dnode->numberP())
		{	double di=dnode->getnumber();
			while (plist && plist->first())
			{	*p++=plist->first()->a_float(1);
				*d++=di;
				if throwing return Unbound;
				next(plist);
			}
		}
	}else
	{	PairP list=arg[1]->a_list(1);
		if throwing return Unbound;
		length=list->count()/2;
		pitch=new double[length]; double *p=pitch;
		duration=new double[length]; double *d=duration;
		while (list && list->first() && list->tail() && list->second())
		{	*p++=list->first()->a_float(1);
			*d++=list->second()->a_int(1);
			if throwing return Unbound;
			next(list);
			next(list);
		}
	}
	if throwing return Unbound;
	TonesThread *t=new TonesThread(length, pitch, duration);
	t->Create(0x100);
	t->Run();
	return Unbound;
}

NodeP lTonesStop(PairPP)
{
	TonesThread::stop=true;
	return Unbound;
}

NodeP lTonesFinished(PairPP)
{
	return new Int(TonesThread::finished);
}

NodeP lplayWave(PairPP arg)
{
#ifdef WIN32
	int flags=arg[2]->a_int(2);
	if (arg[1] && type_of(arg[1])==INT16ARRAY)
	{	PlaySound(charP(Int16ArrayP(arg[1])->data()), 
			0, (flags|SND_MEMORY) & ~SND_FILENAME);
		return Unbound;
	}
	if (arg[1] && type_of(arg[1])==INTARRAY)
	{	PlaySound(charP(IntArrayP(arg[1])->data()), 
			0, (flags|SND_MEMORY) & ~SND_FILENAME);
		return Unbound;
	}
	StringP filename=0;
	if (arg[1])
		filename=arg[1]->a_string(1);
	if (!throwing)
	{	if (arg[1])
			PlaySound(filename->ptr(), 0, (flags|SND_FILENAME) & ~SND_MEMORY);
		else
			PlaySound(0, 0, flags);
	}
#else
	StringP filename=0;
	if (arg[1])
		filename=arg[1]->a_string(1);
	if (!throwing)
		if (arg[1])
		{	wxSound s(filename->ptr());
			if (s.IsOk())
			{	wxSound::Stop();
				s.Play(wxSOUND_SYNC);//unsigned(flags));
			}
		}
		else
			wxSound::Stop();
#endif
	return Unbound;
}


#ifdef WIN32

bool inWaveOutProc=false;

struct WaveInfo
{
	HWAVEOUT hwo;
	WAVEHDR *wh;
	bool done;
	charP data;
	TimeUnion time;
};

const int WaveInfoNr=128;
WaveInfo waveInfo[WaveInfoNr]={0};

void WaveInfoDestroy()
{
	for (int i=0; i<WaveInfoNr; i++)
	{	WaveInfo &wi=waveInfo[i];
		if (wi.wh)
			delete wi.wh;
	}
}

void CALLBACK WaveOutProc(HWAVEOUT hwo, UINT uMsg, DWORD index, 
								  DWORD wh, DWORD dwParam2) 
{ 
	switch(uMsg) 
	{	case WOM_OPEN: break; 
		case WOM_DONE: 
		{	inWaveOutProc=true;
			WaveInfo &wi=waveInfo[index];
			wi.done=true;
			inWaveOutProc=false;
			break;
		}
		case WOM_CLOSE:
			break; 
	}
}

bool PlayWaveFast(charP data, int length) 
{
	WAVEHDR *wh=0;
	HWAVEOUT hwo; 
	bool samedata=false;

	while (inWaveOutProc)
		Sleep(10);

	for (int i=0; i<WaveInfoNr; i++)
	{	WaveInfo &wi=waveInfo[i];
		if (data==wi.data && wi.done)
		{	wh=wi.wh;
			hwo=wi.hwo;
			wi.done=false;
			QueryPerformanceCounter(&wi.time.i);
			break;
		}
	}
	int OFFSET_WAVEDATA=44, 
		HEADER_SIZE=OFFSET_WAVEDATA;

	WAVEFORMATEX *wf=(WAVEFORMATEX*)&data[20];
	int cbSize=128;
//	if (wf->wFormatTag != WAVE_FORMAT_PCM)
//		cbSize=wf->cbSize;

	if (!wh)
	{	
		int index=-1;
		for (int i=0; i<WaveInfoNr; i++)// try to find an empty waveInfo first
		{	WaveInfo &wi=waveInfo[i];
			if (!wi.hwo) 
			{	index=i;
				break;
			}
		}
		unsigned __int64 t=-1;
		while (index < 0)	// no empty found, so search the oldest waveInfo
		{	for (int i=0; i<WaveInfoNr; i++)
			{	WaveInfo &wi=waveInfo[i];
				if (wi.time.u < t) 
				{	index=i;
					t=wi.time.u;
				}
			}
			if (index < 0)
				Sleep(10);
		}
		if (index >= 0)
		{	WaveInfo &wi=waveInfo[index];
			if (wi.hwo && !wi.done) 
			{	if (waveOutUnprepareHeader(wi.hwo, wi.wh, sizeof(WAVEHDR)) != MMSYSERR_NOERROR) 
					return true; 
				wh=wi.wh;
				hwo=wi.hwo;
				goto weida;
			}	
		}
		wh = new WAVEHDR; 
		ZeroMemory((void*)wh, sizeof(WAVEHDR)); 
	
		wh->lpData = data+OFFSET_WAVEDATA; 
		wh->dwBufferLength = length-HEADER_SIZE-cbSize; 
		
		// Find a waveOut device and open it 
		int devs=waveOutGetNumDevs();
		for(UINT devid = 0; devid < devs; devid++) 
		{
			if(devid == devs) // Error, no free devices found 
				return true; 
			if(waveOutOpen(&hwo, WAVE_MAPPER, wf, 
					unsigned long(WaveOutProc), index, 
					CALLBACK_FUNCTION) == MMSYSERR_NOERROR) 
			{
				WaveInfo &wi=waveInfo[index];
				wi.hwo=hwo;
				wi.wh=wh;
				wi.data=data;
				wi.done=false;
				QueryPerformanceCounter(&wi.time.i);
				break; 	// Usable device found, stop searching 
			}
		} 
	}
weida:
	wh->lpData = data+OFFSET_WAVEDATA; 
	wh->dwBufferLength = length-HEADER_SIZE-cbSize; 
	wh->dwFlags=0;

	if (waveOutPrepareHeader(hwo, wh, sizeof(WAVEHDR)) != MMSYSERR_NOERROR) 
		return true; 

	if (waveOutWrite(hwo, wh, sizeof(WAVEHDR)) != MMSYSERR_NOERROR) 
		return true; 
	return false; 
}
#endif

NodeP lplayWaveFast(PairPP arg)
{
#ifdef WIN32
	if (arg[1] && type_of(arg[1])==INT16ARRAY)
	{	PlayWaveFast(charP(Int16ArrayP(arg[1])->data()), 
			Int16ArrayP(arg[1])->length()*2);
	}
	else if (arg[1] && type_of(arg[1])==INTARRAY)
	{	PlayWaveFast(charP(IntArrayP(arg[1])->data()), 
			IntArrayP(arg[1])->length()*4);
	}
#endif
	return Unbound;
}

//___________________________________________________________________
//
//                        Logo Main Run loop
//___________________________________________________________________

#define NEED_MEM_LogoApp_MainLoop(x)					\
	if (!x)														\
	{	delete heapReserve;									\
		heapReserve=0;											\
		wxString msg;											\
		msg << THIS_FILE << "(" << __LINE__ << "):"	\
			#x": Insufficient Memory";						\
		wxMessageBox(msg,	"aUCBLogo Error");			\
		return false;											\
	}

int LogoApp::MainLoop()
{
	wxEventLoop *eventLoop = new wxEventLoop,
		*oldEventLoop=wxEventLoop::GetActive();
	NEED_MEM_LogoApp_MainLoop(eventLoop);
	wxEventLoop::SetActive(eventLoop);
#ifdef WIN32
	while (Pending())
		Dispatch();
	ProcessIdle();
#else
	Yield();
#endif
	
	_main(argc, argv);  //////////  main  ///////////

	int posx, posy, sizex, sizey;
	if (mainFrame->IsIconized())
		mainFrame->Iconize(false);
	mainFrame->GetPosition(&posx, &posy);
	mainFrame->GetSize(&sizex, &sizey);
	config->Write("PosX", posx);
	config->Write("PosY", posy);
	config->Write("SizeX", sizex);
	config->Write("SizeY", sizey);

	config->Write("MemoryShown", ::memoryWindow != 0);
	if (::memoryWindow)
	{	if (memoryFrame->IsIconized())
			memoryFrame->Iconize(false);
		memoryFrame->GetPosition(&memposx , &memposy );
		memoryFrame->GetSize    (&memsizex, &memsizey);
		memoryFrame->Destroy();
	}
	config->Write("MemoryPosX", memposx);
	config->Write("MemoryPosY", memposy);
	config->Write("MemorySizeX", memsizex);
	config->Write("MemorySizeY", memsizey);

	int caretpos=mainFrame->console->GetInsertionPoint();
	config->Write("ConsoleCaretPos", caretpos);
	config->Write("ConsoleAutoSave", mainFrame->console->autosave);
	config->Write("PictureAutoSave", mainFrame->graphAutosave);
	config->Write("NodesAutoSave", mainFrame->nodesAutosave);

	if (edFrame->IsIconized())
		edFrame->Iconize(false);
	edFrame->GetPosition(&posx, &posy);
	edFrame->GetSize(&sizex, &sizey);
	config->Write("EdPosX", posx);
	config->Write("EdPosY", posy);
	config->Write("EdSizeX", sizex);
	config->Write("EdSizeY", sizey);

	config->Write("SplitterRatio" , mainFrame->splitter ->ratio);

	config->Write("SplitterRatio2", mainFrame->splitter2->ratio);
	config->Write("SplitterRatio3", mainFrame->splitter3->ratio);

	int count=edFrame->notebook->GetPageCount();
	config->Write("FilesOpen", count);

	for (int i=0; i < count;  i++)
	{	Edit *edit=(Edit*)edFrame->notebook->GetPage(i);
		wxString filename=edit->GetFilename();
		wxString name;
		name.Printf("File%i", i+1);
		config->Write(name, filename);

		int pos=edit->GetCurrentPos();
		name.Printf("Pos%i", i+1);
		config->Write(name, pos);
	}
	int selectedPage=edFrame->notebook->GetSelection();
	config->Write("SelectedPage", selectedPage);

	config->Write("DoubleBuffer", doubleBuffer);

	wxCloseEvent closeEv;
	edFrame->OnClose(closeEv);
#ifdef _WIN32
	edFrame->Destroy();
	mainFrame->Destroy();
#else
	delete edFrame;
	delete mainFrame;
#endif
	while (Pending())
		Dispatch();
	ProcessIdle();

	wxEventLoop::SetActive(oldEventLoop);
	delete eventLoop;
	theApp=0;
	delete ftglFontManager;
	delete midiSystem;
	Event::eatAll();
	return 0;
}
