#include <wx/dynlib.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/slider.h>
#include <wx/button.h>
#include <wx/radiobut.h>
#include <wx/checkbox.h>
#include <wx/tglbtn.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>
#include <wx/frame.h>
#include <wx/sizer.h>

#include <limits>

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>

#include <m_apm.h>

#ifdef WIN32
	#include <crtdbg.h>
	#include <direct.h>
#else
	#include <unistd.h>
#endif

#include "spinctld.h"
#include "spinctli.h"

#include "Main.h"

#include "Segment.h"
#include "String.h"
#include "Float.h"
#include "BigFloat.h"
#include "Complex.h"
#include "File.h"
#include "Pair.h"
#include "Procedur.h"
#include "Object.h"
#include "Table.h"
#include "Struct.h"
#include "List.h"
#include "Quote.h"
#include "Colon.h"
#include "Tree.h"
#include "Runparsd.h"
#include "WindowPair.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "Array.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "FloatArray.h"
#include "Slider.h"
#include "Gauge.h"
#include "IntControl.h"
#include "FloatControl.h"
#include "Button.h"
#include "ToggleButton.h"
#include "CheckBox.h"
#include "RadioButton.h"
#include "Frame.h"
#include "StaticText.h"
#include "TextControl.h"
#include "ListBox.h"
#include "ChoiceBox.h"
#include "ComboBox.h"
#include "ListControl.h"
#include "BoxSizer.h"

#include "Point.h"
#include "Color.h"
#include "Bitmap.h"
#include "Texture.h"
#include "TurtleNode.h"
#include "GraphicNode.h"
#include "DL.h"

#include "IntToPtr.h"
#include "PtrToInt.h"

#include "StringToPtr.h"
#include "PtrToString.h"

#include "Error.h"
#include "Eval.h"
#include "Stack.h"
#include "Init.h"

#include "Console.h"
#include "Graph.h"
#include "MainFrame.h"
#include "MemoryWindow.h"
#include "LogoApp.h"

#undef min
#undef max

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)

	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

#ifdef PTR_DEBUG
	#include "PtrDebug.cpp"
#endif

charP aUCBLogo="aUCBLogo-4.692";

ControlType stoppingFlag=RUN;

Environment environment;

Environment::Environment()
{
	  logolib=getenv("ALOGOLIB" ); if (!logolib)     logolib="lib";
	helpfiles=getenv("ALOGOHELP"); if (!helpfiles) helpfiles="help";
	   editor=getenv("AEDITOR"  ); 
	  tempdir=getenv("ATEMP"    ); if (!tempdir)     tempdir="temp";
	  comspec=getenv("COMSPEC" );  
	if (!comspec)     
#ifdef WIN32
		comspec="C:\\Windows\\System32\\command.com";
#else
		comspec="sh";
#endif

#ifdef WIN32
	fonts="C:\\Windows\\fonts\\";
	wxString substr=comspec.BeforeLast('\\');
	if (substr)
	{	substr=substr.BeforeLast('\\');
		if (substr)
			fonts=substr+"\\fonts\\";
	}
#else
	fonts="/usr/lib/X11/fonts/TTF/";
#endif

	if (editor)
		editorname = editor.AfterLast('/');
	else
		editorname=wxEmptyString;
}
Environment::~Environment()
{
}

//___________________________________________________________________
//	The following CRITICAL SECTION is quite ugly. 
//	I don`t know a better way of doing this now.  :-(

#define Node_Classes(x)												\
	x(Node)																\
	x(Int) x(Int16) x(Int8) x(UInt8)								\
	x(Float) x(BigFloat) x(Complexd)								\
	x(Array) x(IntArray) x(Int16Array) x(FloatArray)		\
	x(String)															\
	x(Pair) x(List) x(Quote) x(Runparsed) x(Tree) x(Line) \
	x(Procedure) x(Primitive) x(Object) x(Colon)				\
	x(File) x(ToDoNode) x(WindowPair)							\
	x(Table) x(Struct)												\
	x(Bitmap) x(Texture) x(TurtleNode) x(GraphicNode)		\
	x(DynamicLibrary)													\
	x(SliderNode) x(FrameNode) x(GraphNode)					\
	x(ButtonNode) x(ToggleButtonNode) x(CheckBoxNode)		\
	x(RadioButtonNode) x(GaugeNode)								\
	x(IntControlNode) x(FloatControlNode)						\
	x(StaticTextNode) x(TextControlNode) x(BoxSizerNode)	\
	x(ListBoxNode) x(ChoiceBoxNode) x(ComboBoxNode)			\
	x(ListControlNode) 

#define make_aNode(x)	const x x::a##x;

Node_Classes(make_aNode)

const int 
	Node::vmtNode      = *intP(&Node::aNode),	// get pointer to virtual method table of Node
	Node::vmtPair      = *intP(&Pair::aPair),
	Node::vmtRunparsed = *intP(&Runparsed::aRunparsed),
	Node::vmtTree      = *intP(&Tree::aTree),
	Node::vmtLine      = *intP(&Line::aLine);

#ifdef _DEBUG
	#undef new
	#define new new

	#define newOp_decl	\
		::operator new(size_t nSize, int blocktype, const char* newFilename_, int newLine_)

	#define newOp_debuginfo				\
		n->newFilename=newFilename_;	 \
		n->newLine=newLine_;
#else
	#define newOp_decl	\
		::operator new(size_t s)

	#define newOp_debuginfo
#endif

#define newOp(anything, type_, init)			  \
	newOp_decl												\
	{	union													 \
		{	int* i;											  \
			Node *n; Array *a; ToDoNode *t; File *f;	\
			Pair *p; Quote *q; Colon *c; List *li;		 \
			Runparsed *r; Tree *tr; Line *l;				  \
			Procedure *pr; Object *o;							\
			Table *tbl; Struct *strct;	Texture *tex;	    \
			SliderNode *slider; FrameNode *frame;			  \
			WindowPair *wp;											\
		};															      \
		n=GC_::newnode();										       \
		*i= *intP(&anything);								        \
		n->type=type_;											         \
		init;													             \
		newOp_debuginfo										           \
		if (memoryWindow)													   \
			memoryWindow->updateNode(n);									 \
		return n;																  \
}

#define initArray     { a->_data=0;}
#define initToDo      { t->_tail=0;}
#define initFile      { f->_tail=0; f->_name=0;}
#define initPair      { p->_tail= p->_first=0;}
#define initQuote     { q->_tail= q->_first=0;}
#define initColon     { c->_tail= c->_first=0;}
#define initList      {li->_tail=li->_first=0;}
#define initLine      { l->_tail= l->_first=0; l->number=Line::nr; l->filename=Line::fn;}
#define initRunparsed { r->_tail= r->_first=0;  r->_parsed=0;}
#define initTree      {tr->_tail=tr->_first=0; tr->_tree=0; tr->_generation=0;}
#define initProcedure {pr->_bodywords=pr->_bodylist=pr->_formals=0;}
#define initObject    {o->_name=0; o->_tail=0; o->_proc=0; o->_var=0; o->_plist=0;}
#define initTable     {tbl->base=0;}
#define initStruct    {strct->base=0; strct->_first=0;}
#define initTexture   {tex->_bitmap=0; tex->_texnr=0;}
#define initSlider    {slider->slider=0;}
#define initFrame		 {frame->frame=0;}
#define initWindowPair {wp->_tail=wp->_first=0;}

void *       Node newOp(aNode,			UNBOUND,    {})
void *        Int newOp(aInt,				INT_,       {})
void *      Int16 newOp(aInt16,			INT16,      {})
void *       Int8 newOp(aInt8,			INT8,       {})
void *      UInt8 newOp(aUInt8,			UINT8,      {})
void *      Float newOp(aFloat,			FLOAT_,     {})
void *   BigFloat newOp(aBigFloat,		BIGFLOAT,   {})
void *   Complexd newOp(aComplexd,		COMPLEX,    {})
void *  Primitive newOp(aPrimitive,		PRIM,       {})
void *     String newOp(aString,			STRING,     {})
void *     Bitmap newOp(aBitmap,			BITMAP_,    {})
void *    Texture newOp(aTexture,		TEXTURE,    initTexture)
void * TurtleNode newOp(aTurtleNode,	TURTLE_,    {})
void *GraphicNode newOp(aGraphicNode,	GRAPHIC,    {})
void *DynamicLibrary newOp(aDynamicLibrary, DYNAMIC_LIB, {})
void *      Array newOp(aArray,			ARRAY,      initArray)
void *   IntArray newOp(aIntArray,		INTARRAY,   {})
void * Int16Array newOp(aInt16Array,	INT16ARRAY, {})
void * FloatArray newOp(aFloatArray,	FLOATARRAY, {})
void *   ToDoNode newOp(aToDoNode,		TODO,       initToDo)
void *       File newOp(aFile,			FILE_,      initFile)
void *       Pair newOp(aPair,			PAIR,       initPair)
void *      Quote newOp(aQuote,			QUOTE,      initQuote)
void *      Colon newOp(aColon,			COLON,      initColon)
void *       List newOp(aList,			PAIR,			initList)	
// ...will set type to PAIR; only constructor is different from Pair
void *       Line newOp(aLine,			LINE,       initLine)
void *  Runparsed newOp(aRunparsed,		RUNPARSED,  initRunparsed)
void *       Tree newOp(aTree,			TREE,       initTree)
void * WindowPair newOp(aWindowPair,	WINDOWPAIR, initWindowPair)
void *  Procedure newOp(aProcedure,		PROC_,      initProcedure)
void *     Object newOp(aObject,			OBJECT,     initObject)
void *      Table newOp(aTable,			TABLE,      initTable)
void *     Struct newOp(aStruct,			STRUCT,     initStruct)
void * SliderNode newOp(aSliderNode,	SLIDER,		initSlider)
void *  FrameNode newOp(aFrameNode,		FRAME,		initFrame)
void *  GraphNode newOp(aGraphNode,		GRAPH,		{})
void * ButtonNode newOp(aButtonNode,	BUTTON,		{})
void *ToggleButtonNode newOp(aToggleButtonNode,	TOGGLEBUTTON,	{})
void *	 CheckBoxNode newOp(aCheckBoxNode	 ,     CHECKBOX,	{})
void * RadioButtonNode newOp(aRadioButtonNode ,  RADIOBUTTON,	{})
void *		 GaugeNode newOp(aGaugeNode		 ,			 GAUGE,	{})
void *  IntControlNode newOp(aIntControlNode	 ,	  INTCONTROL,	{})
void *FloatControlNode newOp(aFloatControlNode,	FLOATCONTROL,	{})
void *  StaticTextNode newOp(aStaticTextNode	 ,	  STATICTEXT,	{})
void * TextControlNode newOp(aTextControlNode ,	 TEXTCONTROL,	{})
void *     ListBoxNode newOp(aListBoxNode		 ,	     LISTBOX,	{})
void *   ChoiceBoxNode newOp(aChoiceBoxNode	 ,	   CHOICEBOX,	{})
void *    ComboBoxNode newOp(aComboBoxNode	 ,	    COMBOBOX,	{})
void * ListControlNode newOp(aListControlNode ,	 LISTCONTROL,	{})
void *	 BoxSizerNode newOp(aBoxSizerNode	 ,		 BOXSIZER,	{})

#ifdef _DEBUG
	#undef new
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif
//_____________________End of CRITICAL SECTION_______________________

bool checkThrowing() {return File::checkBreak() || throwing;}
void push(PairP &stack, NodeP obj) {stack = new Pair(obj, stack);}

void         Var::mark() {object->mark(); val->mark();}
void         Var::markPtr() {object->markPtr(); val->markPtr();}
void ErrorStruct::mark() 
{	fun->mark();  ufun->mark();  this_line->mark();
}
void ErrorStruct::markPtr() 
{	fun->markPtr();  ufun->markPtr();  this_line->markPtr();
}

void _cdecl logo_stop(int)
{
	if (GC_::is_running) 
		GC_::interrupt=GC_::stop;
	else
	{	File::toPending=0;
		error(STOP_ERROR, 0);
		signal(SIGINT, logo_stop);
		File::unblockInput();
	}
}

void _cdecl logo_pause(int)
{
	if (GC_::is_running) 
		GC_::interrupt=GC_::pause;
	else 
	{	File::toPending=0;
		signal(SIGABRT, logo_pause);
		lpause(argStack);

	}
}

//___________________________________________________________________

const int Node::boolerror=-99999999;

wxDataInputStream &Node::read(wxDataInputStream &istr) 
{	return istr;
}
wxDataOutputStream &Node::write(wxDataOutputStream &ostr) 
{	return ostr << type;
}

void Node::setPtrs(){}

void Node::aftersetPtrs(){}
void Node::recreate(){}

void Node::aftersetPtrs2(){}
void Node::aftersetPtrs3(){}

StringP Node::string() {return _Null_Word->name();}
NodeP  Node::   number() {return Unbound;}
double Node::getnumber() {return Float::quietNaN;}

ObjectP Node::intern() 
{
	StringP str=0;	str=string();
	if (String::caseIgnored)
		str=str->noparity_lower();
	ObjectP obj=0;	obj=root->find(str);

	if (obj)
		return obj;
	else
		return root->insert(str);
}

PairP Node::maybe_quote() 
{
	if (!this  ||  PairP(this)==Unbound  ||  is.aggregate  ||  numberP()) 
		return PairP(this);

	return new Quote(this);
}

//___________________________________________________________________
//
// Argument Type Checking Functions
// Proposal: toggle argument checking by turning a_xxx into function pointers

int Node::a_bool(int nr) 
{
	if (this)
	{	if (equal(_False)) return 0;
		if (equal(_True )) return 1;
	}
	error(BAD_DATA, this, nr);
	return -1;
}

IntP Node::a_integer(int nr) 
{
	if (!this)
	{	error(BAD_DATA, this, nr);
		return IntP(Unbound);
	}
	NodeP a=this, val;
	val = a->number();

	if (val->type==INT_) return IntP(val);
	if (val->type==FLOAT_)
	{	double f = FloatP(val)->float_;

		if (f==int(f)) return new Int(int(f));
	}
	if (val->type==INT16)
		return new Int(int(Int16P(val)->int_));
	if (val->type==INT8)
		return new Int(int(Int8P(val)->int_));
	if (val->type==UINT8)
		return new Int(int(UInt8P(val)->int_));
	
	error(BAD_DATA, this, nr); return IntP(Unbound);
}

int Node::a_int(int nr)
{
	IntP i = a_integer(nr);
	if (!throwing) 
		return i->int_;
	else
		return boolerror;
}

Int16P Node::a_integer16(int nr) 
{
	if (!this)
	{	error(BAD_DATA, this, nr);
		return Int16P(Unbound);
	}
	NodeP a=this, val;
	val = a->number();

	if (val->type==INT16) return Int16P(val);
	if (val->type==INT_)

	{	int i = IntP(val)->int_;
		if (i==short(i)) return new Int16(short(i));
	}
	if (val->type==FLOAT_)
	{	double f = FloatP(val)->float_;

		if (f==short(f)) return new Int16(short(f));
	}
	if (val->type==INT8)
		return new Int16(int(Int8P(val)->int_));
	if (val->type==UINT8)
		return new Int16(int(UInt8P(val)->int_));
	
	error(BAD_DATA, this, nr); return Int16P(Unbound);
}

short Node::a_int16(int nr)
{
	Int16P i = a_integer16(nr);
	if (!throwing) 
		return i->int_;
	else
		return short(boolerror);
}

Int8P Node::a_integer8(int nr) 
{
	if (!this)
	{	error(BAD_DATA, this, nr);
		return Int8P(Unbound);
	}
	NodeP a=this, val;
	val = a->number();

	if (val->type==INT8) return Int8P(val);
	if (val->type==INT_)
	{	int i = IntP(val)->int_;
		if (i==char(i)) return new Int8(char(i));
	}
	if (val->type==FLOAT_)
	{	double f = FloatP(val)->float_;

		if (f==char(f)) return new Int8(char(f));
	}
	if (val->type==INT16)
	{	short i = Int16P(val)->int_;
		if (i==char(i)) return new Int8(char(i));
	}
	if (val->type==UINT8)
	{	unsigned char i = UInt8P(val)->int_;
		if (i==char(i)) return new Int8(char(i));
	}
	error(BAD_DATA, this, nr); return Int8P(Unbound);
}

char Node::a_int8(int nr)
{
	Int8P i = a_integer8(nr);
	if (!throwing) 
		return i->int_;
	else
		return char(boolerror);
}

IntP Node::a_pos_int(int nr)
{


	if (!this)
	{	error(BAD_DATA, this, nr);
		return IntP(Unbound);
	}
	NodeP val = number();

	if (val->type==INT_ && IntP(val)->int_ >= 0) return IntP(val);
	if (val->type==FLOAT_)
	{	double f = FloatP(val)->float_;

		if (f==int(f)  &&  int(f) >= 0)
			return new Int(int(f));
	}
	if (val->type==INT16)
	{	short i = Int16P(val)->int_;

		if (i >= 0)
			return new Int(int(i));
	}
	if (val->type==INT8)
	{	char i = Int8P(val)->int_;

		if (i >= 0)
			return new Int(int(i));
	}
	if (val->type==UINT8)
	{	unsigned char i = UInt8P(val)->int_;
		return new Int(int(i));
	}
	error(BAD_DATA, this, nr); return IntP(Unbound);
}

double Node::a_float(int nr) 
{
	if (this)
	{	NodeP val=number();

		if (val->type==FLOAT_) return FloatP(val)->float_;
		if (val->type==INT_  ) return   IntP(val)->int_;
		if (val->type==INT16 ) return Int16P(val)->int_;
		if (val->type==INT8  ) return Int8P(val)->int_;
		if (val->type==UINT8 ) return UInt8P(val)->int_;
		if (val->type==BIGFLOAT) return BigFloatP(val)->float_.to_double();
	}
	error(BAD_DATA, this, nr);
	return Float::quietNaN;
}

NodeP Node::a_number(int nr) 
{
	if (!this)
	{	error(BAD_DATA, this, nr);
		return Unbound;
	}
	NodeP val = number();
	if (val==Unbound && !throwing) error(BAD_DATA, this, nr);
	return val;
}

ComplexdP Node::a_complex(int nr) 
{
	if (!this)
	{	error(BAD_DATA, this, nr);
		return ComplexdP(Unbound);
	}
	NodeP val = number();
	if ((val==Unbound || val->type!=COMPLEX) && !throwing) 
		error(BAD_DATA, this, nr);

	return ComplexdP(val);
}

BigFloatP Node::a_bigfloat(int nr) 
{
	if (!this)
	{	error(BAD_DATA, this, nr);
		return BigFloatP(Unbound);
	}
	NodeP val = number();
	if ((val==Unbound || val->type!=BIGFLOAT) && !throwing) 
		error(BAD_DATA, this, nr);

	return BigFloatP(val);
}

StringP Node::a_char(int nr) 
{
	if (!this)
	{	error(BAD_DATA, this, nr);
		return StringP(Unbound);
	}
	StringP str = string();
	NodeP a = this;
	while ( (str==Unbound  ||  str->length() != 1) && !throwing ) 
	{	a = error(BAD_DATA, a, nr);
		str = a->string();
	}
	return str;
}

StringP Node::a_string(int nr) 
{
	if (!this)
	{	error(BAD_DATA, this, nr);
		return StringP(Unbound);
	}
	StringP str = string();
	  NodeP a=this;

	while (str==Unbound && !throwing) 
	{	a = error(BAD_DATA, a, nr);
		str = a->string();
	}
	return str;
}

NodeP Node::a_name(int nr) 
{
	if (isaggregate())
		error(BAD_DATA, this, nr);
	return this;
}

PairP Node::a_bfable(int nr)
{
	NodeP a=this;
	while ((!a 
		|| a==Unbound 
		|| a==PairP(_Null_Word)
		|| a->type==ARRAY)  &&  !throwing) 
	{
		a = error(BAD_DATA, a, nr);
	}
	return PairP(a);
}

PairP Node::a_list(int nr) 
{
	NodeP a=this;
	while (a->isnolist() && !throwing) 
		a = error(BAD_DATA, a, nr);
	return PairP(a);
}

PairP Node::a_runnable(int nr) 
{
	if (this && is.string) 
		return StringP(this)->parser(true);

	if (islist()) return PairP(this);
	
	error(BAD_DATA, this, nr); return PairP(Unbound);
}



BitmapP Node::a_bitmap(int nr)
{
	if (this && type==BITMAP_)
		return BitmapP(this);
	error(BAD_DATA, this, nr); return BitmapP(Unbound);
}

TableP Node::a_table(int nr)
{
	if (this && type==TABLE)
		return TableP(this);
	error(BAD_DATA, this, nr); return TableP(Unbound);
}

ArrayP Node::a_array(int nr)
{
	if (this && type==ARRAY)
		return ArrayP(this);
	error(BAD_DATA, this, nr); return ArrayP(Unbound);
}

IntArrayP Node::a_intarray(int nr)
{
	if (this && type==INTARRAY)
		return IntArrayP(this);
	error(BAD_DATA, this, nr); return IntArrayP(Unbound);
}

Int16ArrayP Node::a_int16array(int nr)
{
	if (this && type==INT16ARRAY)
		return Int16ArrayP(this);
	error(BAD_DATA, this, nr); return Int16ArrayP(Unbound);
}

FloatArrayP Node::a_floatarray(int nr)
{
	if (this && type==FLOATARRAY)
		return FloatArrayP(this);
	error(BAD_DATA, this, nr); return FloatArrayP(Unbound);
}

Point Node::a_3darray(int nr)
{
	if (this && type==ARRAY)
	{	if (ArrayP(this)->length()==3)
		{	NodePP n=ArrayP(this)->data();

			Point p=Point(
				n[0]->a_float(nr), 
				n[1]->a_float(nr),
				n[2]->a_float(nr));
			if (!throwing)
				return p;
		}
	}
	error(BAD_DATA, this, nr); return Point(0,0,0);
}

PairP Node::a_vector_helper(bool floatok, bool three, bool neg, int nr) 
{
	NodeP a=this;

	while (!throwing) 
	{	if (a->islist())
		{	PairP l=PairP(a);

			if (l
			&&  l->tail()
			&&  (three?  l->tail()->tail() && !l->tail()->tail()->tail()  
						 : !l->tail()->tail()))

			{	NodeP 
					val1 = l-> first()->number(),
					val2 = l->second()->number(),
					val3 = 0;

				if (three)
					val3 = l->third()->number();

				if ( val1 != Unbound  &&  val2 != Unbound 
				&& (floatok 
					|| (   val1->type==INT_  && (neg || IntP(val1)->int_>= 0) 
						&&  val2->type==INT_  && (neg || IntP(val2)->int_>= 0) 
						&& (!three 
							 || (val3->type==INT_ && (neg || IntP(val3)->int_>= 0)))))) 
				{
					l->setfirst (val1);
					l->setsecond(val2);
					if (three) 
						l->setthird(val3);

					return l;
				}
			}
		}
		a = error(BAD_DATA, a, nr);
	}
	return PairP(Unbound);
}

PairP Node::a_vector        (int nr) 
{	return a_vector_helper( true, false, true, nr);
}

PairP Node::a_3dvector      (int nr) 
{	return a_vector_helper( true,  true, true, nr);
}

PairP Node::a_int_vector(int nr) 
{	return a_vector_helper(false, false, true, nr);
}

PairP Node::a_pos_int_vector(int nr) 
{	return a_vector_helper(false, false, false, nr);
}

PairP Node::a_rgb           (int nr) 
{	return a_vector_helper( true,  true, false, nr);
}

PairP Node::a_rgba(int nr) 
{
	NodeP a=this;
	if (a->isnolist())
	{	error(BAD_DATA, a, nr);
		return PairP(Unbound);
	}
	PairP l=PairP(a);

	for (int i=4; i && l && l->first();  i--) 
	{	NodeP val=l->first()->number();
		if (val==Unbound)
			error(BAD_DATA, a, nr);

		l->setfirst(val);
		next(l);
	}
	if (i!=0)
		error(BAD_DATA, a, nr);
	return PairP(this);
}

Color Node::a_color(int nr)
{	if (!this)
		error(BAD_DATA, this, nr);
	else if (numberP())
	{	IntP i = a_integer(nr);
		if (!throwing) 
			return i->int_;
	}
	else if (type==STRING)
	{	StringP name=StringP(StringP(this)->upper());
		wxColour wxc=(name->ptr());
		if (wxc.Ok())
			return Color(wxc.Red(), wxc.Green(), wxc.Blue());
		else
			error(BAD_DATA, this, nr);
	}
	else if (type==OBJECT)

	{	StringP name=StringP(ObjectP(this)->upper());
		wxColour wxc(name->ptr());
		if (wxc.Ok())
			return Color(wxc.Red(), wxc.Green(), wxc.Blue());
		else
			error(BAD_DATA, this, nr);
	}
	else if (is.list)
	{	int n=count();
		if (n==3)
		{	PairP rgb=a_rgb(nr);
			if throwing return 0;
			return Color(
				rgb->first ()->getnumber(), 
				rgb->second()->getnumber(), 
				rgb->third ()->getnumber());
		}
		else if (n==4)
		{	PairP rgba=a_rgba(nr);
			if throwing return 0;
			return Color(
				rgba->first ()->getnumber(), 
				rgba->second()->getnumber(), 
				rgba->third ()->getnumber(),
				rgba->fourth()->getnumber());
		}else
			error(BAD_DATA, this, nr);
	}
	else
		error(BAD_DATA, this, nr);
	return 0;
}

TurtleNodeP Node::a_turtleNode(int nr)
{
	if (this && type==TURTLE_)
		return TurtleNodeP(this);

	error(BAD_DATA, this, nr); return TurtleNodeP(Unbound);
}

GraphicNodeP Node::a_graphicNode(int nr)
{
	if (this && type==GRAPHIC)
		return GraphicNodeP(this);

	error(BAD_DATA, this, nr); return GraphicNodeP(Unbound);
}

wxDynamicLibrary *Node::a_dynamicLibrary(int nr)
{
	if (this && type==DYNAMIC_LIB)
		return DynamicLibraryP(this)->module;

	error(BAD_DATA, this, nr); return 0;
}

SliderNodeP Node::a_slider(int nr)
{
	if (this && type==SLIDER)
	{	SliderP slider=SliderNodeP(this)->slider;
		if (slider)
			return SliderNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

GaugeNodeP Node::a_gauge(int nr)
{
	if (this && type==GAUGE)
	{	GaugeP gauge=GaugeNodeP(this)->gauge;
		if (gauge)
			return GaugeNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

IntControlNodeP Node::a_intControl(int nr)
{
	if (this && type==INTCONTROL)
	{	IntControlP intControl=IntControlNodeP(this)->intControl;
		if (intControl)
			return IntControlNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

FloatControlNodeP Node::a_floatControl(int nr)
{
	if (this && type==FLOATCONTROL)
	{	FloatControlP floatControl=FloatControlNodeP(this)->floatControl;
		if (floatControl)
			return FloatControlNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

ButtonNodeP Node::a_button(int nr)
{
	if (this && type==BUTTON)
	{	ButtonP button=ButtonNodeP(this)->button;
		if (button)
			return ButtonNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

ToggleButtonNodeP Node::a_toggleButton(int nr)
{
	if (this && type==TOGGLEBUTTON)
	{	ToggleButtonP toggleButton=ToggleButtonNodeP(this)->toggleButton;
		if (toggleButton)
			return ToggleButtonNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

CheckBoxNodeP Node::a_checkBox(int nr)
{
	if (this && type==CHECKBOX)
	{	CheckBoxP checkBox=CheckBoxNodeP(this)->checkBox;
		if (checkBox)
			return CheckBoxNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

RadioButtonNodeP Node::a_radioButton(int nr)
{
	if (this && type==RADIOBUTTON)
	{	RadioButtonP radioButton=RadioButtonNodeP(this)->radioButton;
		if (radioButton)
			return RadioButtonNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

StaticTextNodeP Node::a_staticText(int nr)
{
	if (this && type==STATICTEXT)
	{	StaticTextP staticText=StaticTextNodeP(this)->staticText;
		if (staticText)
			return StaticTextNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

TextControlNodeP Node::a_textControl(int nr)
{
	if (this && type==TEXTCONTROL)
	{	TextControlP textControl=TextControlNodeP(this)->textControl;
		if (textControl)
			return TextControlNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

ListBoxNodeP Node::a_listBox(int nr)
{
	if (this && type==LISTBOX)
	{	ListBoxP listBox=ListBoxNodeP(this)->listBox;
		if (listBox)
			return ListBoxNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

ChoiceBoxNodeP Node::a_choiceBox(int nr)
{
	if (this && type==CHOICEBOX)
	{	ChoiceBoxP choiceBox=ChoiceBoxNodeP(this)->choiceBox;
		if (choiceBox)
			return ChoiceBoxNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

ComboBoxNodeP Node::a_comboBox(int nr)
{
	if (this && type==COMBOBOX)
	{	ComboBoxP comboBox=ComboBoxNodeP(this)->comboBox;
		if (comboBox)
			return ComboBoxNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

ListControlNodeP Node::a_listControl(int nr)
{
	if (this && type==LISTCONTROL)
	{	ListControlP listControl=ListControlNodeP(this)->listControl;
		if (listControl)
			return ListControlNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

FrameNodeP Node::a_frame(int nr)
{
	if (this && type==FRAME)
	{	FrameP frame=FrameNodeP(this)->frame;
		if (frame)
			return FrameNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

GraphNodeP Node::a_graph(int nr)
{
	if (this && type==GRAPH)
	{	GraphP graph=GraphNodeP(this)->graph;
		if (graph)
			return GraphNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}

BoxSizerNodeP Node::a_boxSizer(int nr)
{
	if (this && type==BOXSIZER)
	{	BoxSizerP boxSizer=BoxSizerNodeP(this)->boxSizer;
		if (boxSizer)
			return BoxSizerNodeP(this);
	}
	error(BAD_DATA, this, nr); return 0;
}
//___________________________________________________________________

const int 
	Int::max = unsigned(-1) >> 1,
	Int::min = -Int::max-1,
	Int::umax = unsigned(-1),
	Int::digits = sizeof(int)*256/100 + 3;

const __int64 Int::i64max = ((unsigned __int64)(-1)) >> 1;


int Int::genSymNr = 0;

wxDataInputStream &Int::read(wxDataInputStream &istr)
{	return istr >> int_;
}
wxDataOutputStream &Int::write(wxDataOutputStream &ostr)
{	return ostr << type << int_;
}

StringP Int::string()
{
	char s[digits];
	sprintf(s, "%i", int_);
	return new String(s, 0, strlen(s), STRING, String::copy);
}

//___________________________________________________________________

const short 
	Int16::max = (unsigned short)(-1) >> 1,
	Int16::min = -Int16::max-1,
	Int16::digits = sizeof(short)*256/100 + 3;

wxDataInputStream &Int16::read(wxDataInputStream &istr)
{	return istr >> int_;
}
wxDataOutputStream &Int16::write(wxDataOutputStream &ostr)
{	return ostr << type << int_;
}

StringP Int16::string()
{
	char s[digits];
	sprintf(s, "%i", int_);
	return new String(s, 0, strlen(s), STRING, String::copy);
}

//___________________________________________________________________

const char 
	Int8::max = (unsigned char)(-1) >> 1,
	Int8::min = -Int8::max-1,

	Int8::digits = sizeof(char)*256/100 + 3;

wxDataInputStream &Int8::read(wxDataInputStream &istr)
{	return istr >> (unsigned char&)int_;
}
wxDataOutputStream &Int8::write(wxDataOutputStream &ostr)
{	return ostr << type << (unsigned char)int_;
}

StringP Int8::string()
{
	char s[digits];
	sprintf(s, "%i", int_);
	return new String(s, 0, strlen(s), STRING, String::copy);
}

//___________________________________________________________________

const unsigned char 
	UInt8::max = (unsigned char)(-1),
	UInt8::digits = sizeof(unsigned char)*256/100 + 3;

wxDataInputStream &UInt8::read(wxDataInputStream &istr)
{	return istr >> int_;
}
wxDataOutputStream &UInt8::write(wxDataOutputStream &ostr)
{	return ostr << type << int_;
}

StringP UInt8::string()
{
	char s[digits];
	sprintf(s, "%u", int_);
	return new String(s, 0, strlen(s), STRING, String::copy);
}

//___________________________________________________________________

const int Float::digits = std::numeric_limits<double>::digits10+10;

const double 
	Float::max_ = std::numeric_limits<double>::max(),
	Float::pi = 3.141592653589793227020265931059839203954,
	Float::deg2rad = pi/180,
	Float::     epsilon = std::numeric_limits<double>::epsilon(),
	Float::         NaN = std::numeric_limits<double>::signaling_NaN(),
	Float::quietNaN = std::numeric_limits<double>::quiet_NaN();

wxDataInputStream &Float::read(wxDataInputStream &istr)
{	return istr >> float_;
}
wxDataOutputStream &Float::write(wxDataOutputStream &ostr)
{	return ostr << type << float_;
}

StringP Float::string()
{
	char format[10], s[digits];
	sprintf(format, "%%-.%ig", File::precision);
	sprintf(s, format, float_);
	return new String(s, 0, strlen(s), STRING, String::copy);
}

//___________________________________________________________________

MAPM BigFloat::deg2rad = MAPM(MM_PI)/180;

void BigFloat::free() {float_.~MAPM();}

wxDataInputStream &BigFloat::read(wxDataInputStream &istr)
{	wxString s;
	istr >> s;
	float_=s;
	return istr;
}
wxDataOutputStream &BigFloat::write(wxDataOutputStream &ostr)
{	ostr << type;
	wxString s('\0', MM_cpp_min_precision+20);
	float_.toString(charP(s.c_str()), MM_cpp_min_precision);
	return ostr << s;
}

StringP BigFloat::string()
{
	wxString s('\0', MM_cpp_min_precision+20);
	float_.toString(charP(s.c_str()), MM_cpp_min_precision);
	return new String(s);
}

double BigFloat::getnumber()
{
	double d=float_.to_double();
	if (m_apm_error)
	{	m_apm_reset_error();
		error(BAD_DATA, this);
		return Float::quietNaN;
	}
	return d;
}
//___________________________________________________________________

const int Complexd::digits = std::numeric_limits<double>::digits10+10;

wxDataInputStream &Complexd::read(wxDataInputStream &istr)
{	double r,i;
	istr >> r >> i;
	complex_=complexd(r,i);
	return istr;
}
wxDataOutputStream &Complexd::write(wxDataOutputStream &ostr)
{	return ostr << type << complex_.real() << complex_.imag();
}

StringP Complexd::string()
{
	char format[20], s[digits*2+2];
	if (complex_.imag() >= 0)
	{	sprintf(format, "%%-.%ig+%%-.%igi", 
			File::precision, File::precision);
		sprintf(s, format, complex_.real(),complex_.imag());
	}else
	{	sprintf(format, "%%-.%ig%%-.%igi", 
			File::precision, File::precision);
		sprintf(s, format, complex_.real(),complex_.imag());
	}
	return new String(s, 0, strlen(s), STRING, String::copy);
}

//___________________________________________________________________

Primitive::Primitive(PrimType &prim):
	_function(prim.function), priority(prim.priority),
	min(prim.min), std(prim.std), max(prim.max)
{
	if (prim.priority==priority_MACRO)
		type=MACRO;
	else if (prim.priority <= priority_TAIL)
		type=TAILFORM;
	else if (prim.priority == prefix)
		type=PRIM;
	else
		type=INFIX;
}

wxDataInputStream &Primitive::read(wxDataInputStream &istr) 
{	wxString name;
	istr >> name
		>> (unsigned char&)priority 
		>> (unsigned char&)min 
		>> (unsigned char&)std 
		>> (unsigned char&)max;
	_function = FunctionP(GC_::primitivesStringToPtr->find(name)); 
	if (!_function)
	{	error(FILE_ERROR, new String(
			(wxString("Function ")+name
			 +" not available in this release")));
		_function=&lerror;
	}
	return istr;
}
wxDataOutputStream &Primitive::write(wxDataOutputStream &ostr) 
{	ostr << type 
		<< GC_::primitivesPtrToString->find((void*)_function) 
		<< (unsigned char)priority 
		<< (unsigned char)min 
		<< (unsigned char)std 
		<< (unsigned char)max;
	return ostr;
}

//___________________________________________________________________

Array::Array(int length_, int origin_): Node(ARRAY)
{
	_origin=origin_;
	_length=0;
	_data = (NodeP*) GC_::calloc(length_+1, sizeof(NodeP));
	if (_data) 
		_length=length_;
}

Array::Array(ArrayP array_): Node(ARRAY)
{
	_origin = array_->_origin;
	_length=0;
	_data = (NodeP*) GC_::calloc(array_->_length+1, sizeof(NodeP));
	if (!_data)
	{	_length=0;
		return;
	}
	_length= array_->_length;
	NodePP source = array_->_data, 
		  destination = _data;

	for (int i=_length;  i;  i--)
		*destination++ = *source++;
}

void Array::free() 
{
	if (_data)
		::free(_data);
}

void Array::inc_gc()
{
	NodePP d=_data;
	for (int i=_length;  i;  i--)
		fifo << *d++;
}

int Array::youngs() 
{
	int youngs=0;
	NodePP d=_data;
	for (int i=_length;  i;  i--)
	{	NodeP child = *d++;
		youngs+=gmark(child);
	}
	return youngs;
}

int Array::youngsPtr() 
{
	int youngs=0;
	NodePP d=_data;
	for (int i=_length;  i;  i--)
	{	NodeP child = *d++;
		youngs+=gmarkPtr(child);
	}
	return youngs;
}

void Array::check_oldyoungs()
{
	NodePP d=_data;
	for (int i=_length;  i;  i--) 
		check_oldyoung(*d++);
}

void Array::setPtrs()
{
	NodePP itemptr=_data;

	for (int i=_length;  i;  i--)
	{	*itemptr=(Node*)GC_::intToPtr->find(int(*itemptr));
		check_oldyoung(*itemptr);
		itemptr++;
	}
}
wxDataInputStream &Array::read(wxDataInputStream &istr) 
{	
	int length;
	istr >> length >> _origin;

	_length=0;
	_data = (NodeP*) GC_::calloc(sizeof(NodeP), length+1);
	if (!_data)
	{	_length=0;
		return istr;
	}
	_length=length;
	NodePP itemptr=_data;
	for (int i=_length;  i;  i--)
	{	istr >> (int&)*itemptr;
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &Array::write(wxDataOutputStream &ostr) 
{
	ostr << type << _length << _origin;

	NodePP itemptr=_data;
	for (int i=_length;  i;  i--)
		ostr << GC_::ptrToInt->find(*itemptr++);
	return ostr;
}

StringP Array::string()
{
	const int maxlength=65536;
	char textbuf[maxlength];
	File::stringptr=textbuf;
	File::stringlen=maxlength-1;
	stringbuffer << NodeP(this);
	*File::stringptr='\0';
	return new String(textbuf);
}

//___________________________________________________________________

IntArray::IntArray(int length_, int origin_): Node(INTARRAY)
{
	_origin=origin_;
	_length=0;
	_data = (intP) GC_::calloc(length_+1, sizeof(int));
	if (_data) 
		_length=length_;
}

IntArray::IntArray(IntArrayP array_): Node(INTARRAY)
{
	_length= 0;
	_origin = array_->_origin;
	_data = (intP) GC_::calloc(array_->_length+1, sizeof(int));
	if (!_data) 
	{	_length=0;
		return;
	}
	_length= array_->_length;
	intP source = array_->_data, 

		  destination = _data;

	for (int i=_length;  i;  i--)
		*destination++ = *source++;
}

void IntArray::free() {::free(_data);}

wxDataInputStream &IntArray::read(wxDataInputStream &istr) 
{	
	int length;
	istr >> length >> _origin;
	_length=0;
	_data = (intP) GC_::calloc(sizeof(int), length+1);
	if (!_data) 
		return istr;
	_length=length;
	intP itemptr=_data;
	for (int i=_length;  i;  i--)
	{	istr >> (int&)*itemptr;
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &IntArray::write(wxDataOutputStream &ostr) 
{
	ostr << type << _length << _origin;
	intP itemptr=_data;
	for (int i=_length;  i;  i--)
		ostr << *itemptr++;
	return ostr;
}

StringP IntArray::string()
{
 	const int maxlength=65536;
	char textbuf[maxlength];
	File::stringptr=textbuf;
	File::stringlen=maxlength-1;
	stringbuffer << NodeP(this);
	*File::stringptr='\0';
	return new String(textbuf);
}

//___________________________________________________________________

Int16Array::Int16Array(int length_, int origin_): Node(INT16ARRAY)
{
	_origin=origin_;
	_length=0;
	_data = (shortP) GC_::calloc(length_+1, sizeof(short));
	if (!_data) 
		return;
	_length=length_;
}

Int16Array::Int16Array(Int16ArrayP array_): Node(INT16ARRAY)
{
	_length=0;
	_origin = array_->_origin;
	_data = (shortP) GC_::calloc(array_->_length+1, sizeof(short));
	if (!_data) 
		return;
	_length = array_->_length;
	shortP source = array_->_data, 
		  destination = _data;

	for (int i=_length;  i;  i--)
		*destination++ = *source++;
}

void Int16Array::free() {::free(_data);}

wxDataInputStream &Int16Array::read(wxDataInputStream &istr) 
{	
	int length;
	istr >> length >> _origin;
	_length=0;
	_data = (shortP) GC_::calloc(sizeof(short), length+1);
	if (!_data) 
		return istr;
	_length=length;
	shortP itemptr=_data;
	for (int i=_length;  i;  i--)
	{	istr >> (short&)*itemptr;
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &Int16Array::write(wxDataOutputStream &ostr) 
{
	ostr << type << _length << _origin;
	shortP itemptr=_data;
	for (int i=_length;  i;  i--)
		ostr << *itemptr++;
	return ostr;
}

StringP Int16Array::string()
{
	const int maxlength=65536;
	char textbuf[maxlength];
	File::stringptr=textbuf;
	File::stringlen=maxlength-1;
	stringbuffer << NodeP(this);
	*File::stringptr='\0';
	return new String(textbuf);
}

//___________________________________________________________________

FloatArray::FloatArray(int length_, int origin_): Node(FLOATARRAY)
{
	_origin=origin_;
	_length=0;
	_data = (doubleP) GC_::calloc(length_+1, sizeof(double));
	if (!_data) 
		return;
	_length=length_;
}

FloatArray::FloatArray(FloatArrayP array_): Node(FLOATARRAY)
{
	_origin = array_->_origin;
	_length=0;
	_data = (doubleP) GC_::calloc(array_->_length+1, sizeof(double));
	if (!_data) 
		return;
	_length= array_->_length;
	doubleP source = array_->_data, 
		  destination = _data;

	for (int i=_length;  i;  i--)
		*destination++ = *source++;
}

void FloatArray::free() {::free(_data);}

wxDataInputStream &FloatArray::read(wxDataInputStream &istr) 
{	
	int length;
	istr >> length >> _origin;
	_length=0;
	_data = (doubleP) GC_::calloc(sizeof(double), length+1);
	if (!_data)
		return istr;
	_length=length;
	doubleP itemptr=_data;
	for (int i=_length;  i;  i--)
	{	istr >> (double&)*itemptr;
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &FloatArray::write(wxDataOutputStream &ostr) 
{
	ostr << type << _length << _origin;
	doubleP itemptr=_data;
	for (int i=_length;  i;  i--)
		ostr << *itemptr++;
	return ostr;
}

StringP FloatArray::string()
{
 	const int maxlength=65536;
	char textbuf[maxlength];
	File::stringptr=textbuf;
	File::stringlen=maxlength-1;
	stringbuffer << NodeP(this);
	*File::stringptr='\0';
	return new String(textbuf);
}

//___________________________________________________________________

Backslash backslash;

Backslash::Backslash()
{
	chars="()[]{}'\b\t\n<>\r\t\n\r;|~\"\\ +-*/^=:?.";	// ' instead of \a
	size=strlen(chars);
	begin=1;	// first char used for quoteds

	int c;
	for (c=0;  c<0x100;  c++)
	{	  array[c]=char(c);
		unarray[c]=char(c);
		    _is[c]=false;
	}
	for (c=0;  c < size;  c++)
	{	  array[(unsigned char)(chars[c])] = c+begin;
		unarray[c+begin] = chars[c];
		int h=c+begin;
		if (h<7 || h>13 || (h>10 && h<13))	// exclude \a, \b, \t, \n and \r
			_is[h] = true;
	}
}


const int String::digits=200;	// max number of digits in a number
bool String::caseIgnored=true;

String::String(int length): Node(STRING) 
{
	_length=0; 
	_head= HeadP( GC_::calloc(sizeof(char), length+sizeof(int)+1) );
	if (!_head) 
	{	_ptr=0;
		return;
	}
	_length=length;
	_ptr= _head->str;
	_head->refCount=1;
}

String::String(charP str): Node(STRING) 
{
	_length=0;
	int length=strlen(str); 
	_head= HeadP( GC_::malloc(length+sizeof(int)+1) );
	if (!_head) 
	{	_ptr=0;
		return;
	}
	_length=length;
	copy(_head->str, str, _length);
	_ptr= _head->str;
	_head->refCount=1;
}

String::String(wxString str): Node(STRING) 
{
	_length=0;
	_ptr=charP(str.c_str()); 
	int length=str.Length(); 
	_head= HeadP( GC_::malloc(length+sizeof(int)+1) );
	if (!_head) 
	{	_ptr=0;
		return;
	}
	_length=length;
	copy(_head->str, _ptr, _length);
	_ptr= _head->str;
	_head->refCount=1;
}

String::String(StringP sp): Node(STRING) 
{
	int length = sp->_length;
	_length=0;
	_head= HeadP( GC_::malloc(length+sizeof(int)+1) );
	if (!_head)
	{	_ptr=0;
		return;
	}
	_length=length;
	copy(_head->str, sp->_ptr, _length);
	_ptr= _head->str;
	_head->refCount=1;
}

String::String(StringP sp, StringP sp2): Node(STRING) 
{
	int length = sp->_length + sp2->_length;
	_length=0;
	_head= HeadP( GC_::malloc(length+sizeof(int)+1) );
	if (!_head) 
	{	_ptr=0;
		return;
	}
	_length=length;
	copy(_head->str, sp->_ptr, sp->_length);
	copy(_head->str + sp->_length, sp2->_ptr, sp2->_length);
	_ptr= _head->str;
	_head->refCount=1;
}

String::String(charP ptr_, HeadP head_, int length_, NodeType typ, CopyRoutine &copy):
	Node(typ)
{
	if (!head_) 
	{	head_= HeadP( GC_::malloc(length_+sizeof(int)+1) );
		if (!head_) 
		{	_head=0;
			_length=0;
			_ptr=0;
			return;
		}
		copy(head_->str, ptr_, length_);
		ptr_= head_->str;
		head_->refCount=0;
	}
	_length = length_;
	      _ptr = ptr_;
	  _head = head_;

	_head->refCount++;
}

String::String(PairPP list, int length_, NodeType typ, PairCopyRoutine &copy):
	Node(typ)
{
	_head = HeadP( GC_::malloc(length_+sizeof(int)+1) );
	if (!_head) 
	{	_length=0;
		_ptr=0;
		return;
	}
	copy(_head->str,  list, length_);
	_ptr = _head->str;
	_length = length_;
	_head->refCount=1;
}

StringP String::make(charP ptr, HeadP head, int length, NodeType type, CopyRoutine &copy)
{
	if (length==0  &&  _Null_Word) 
		return StringP(_Null_Word);
	else
		return new String(ptr, head, length, type, copy);
}

StringP String::fromList(PairPP arg, int length, NodeType type, PairCopyRoutine &copy)
{
	if (length==0  &&  _Null_Word) 
		return StringP(_Null_Word);
	else
		return new String(arg, length, type, copy);

}

void String::free()
{
	if (_head  &&  !--_head->refCount)
		::free(_head);
}

wxDataInputStream &String::read(wxDataInputStream &istr) 
{	wxString str;
	istr >> str;

	_ptr=charP(str.c_str()); 
	int length=str.Length(); 
	_length=0;
	_head= HeadP( GC_::malloc(length+sizeof(int)+1) );
	if (!_head) 
	{	_ptr=0;
		return istr;
	}
	_length=length;
	copy(_head->str, _ptr, _length);
	_ptr= _head->str;
	_head->refCount=1;
	return istr;
}
wxDataOutputStream &String::write(wxDataOutputStream &ostr) 
{	wxString str(_ptr, _length);
	return ostr << type << str;
}

int String::numberP() 
{
	if (_length >= digits) 
		return 0;

	charP p=_ptr; 
	int l=0, left=0, right=0;	// left counts digits left of decimal point

	if (l <_length  &&  *p=='-')
		p++, l++;
	
	while (l <_length  &&  is_digit(*p))
		p++, l++, left++;
	
	if (l <_length  &&  *p=='.') 
	{	p++, l++;
		while (l <_length  &&  is_digit(*p))
			p++, l++, right++;
	}
	if (l <_length  &&  (left || right)  &&  (*p=='E' || *p=='e')) 
	{	p++, l++;
		if (l <_length  &&  (*p=='+' || *p=='-'))
			p++, l++;
		
		while (l <_length  &&  is_digit(*p))
			p++, l++, right++;
	}
	if (l <_length  &&  (left || right)  &&  (*p=='I' || *p=='i')) 
	{	p++, l++;
	}
	if ((left==0 && right==0)  ||  l != _length)
		return 0;
	else
		return right+1;
}

NodeP String::number()
{
	char s2[digits];
	charP s=s2;
	int right=numberP();

	if (_length < digits  &&  right) 
	{	if (is.backslashed)
			copy_noparity(s, _ptr, _length);
		else
			copy(s, _ptr, _length);

		if (*s=='+') ++s;
		if (s2[_length-1]=='.') 
			s2[_length-1]= 0;

		if (right-1  ||  _length > 9) 
		{	if (s2[_length-1]=='I' || s2[_length-1]=='i')
				return new Complexd(0,atof(s));
			else
				if (_length <= Float::digits)
					return new Float(atof(s));
				else
					return new BigFloat(s);
		}else 
		{	if (s2[_length-1]=='I' || s2[_length-1]=='i')
				return new Complexd(0,atol(s));
			else
				return new Int(atol(s));
		}
	}else 
		return Unbound;
}

double String::getnumber()
{
	char s2[digits];

	charP s=s2;
	int right=numberP();

	if (_length < digits  &&  right) 
	{	if (is.backslashed)
			copy_noparity(s, _ptr, _length);
		else
			copy(s, _ptr, _length);

		if (*s=='+') ++s;
		if (s2[_length-1]=='.') 
			s2[_length-1]= 0;

		if (right-1  ||  _length > 9) 
			return atof(s);
		else 
			return atol(s);
	}else 
		return Float::quietNaN;
}

StringP String::first() {return make(ptr(), head(), 1, type, copy);}

StringP String::noparity_lower() {return make(ptr(), 0, length(), STRING, copy_noparitylow);}

charP String::expand_slash() 
{
	charP result, cp, cp2;
	int i, j;
	
	for (cp=_ptr, i=0, j=_length;  --j >= 0; )
		if (backslash.is(*cp++)) 
			i++;

	result = charP( GC_::malloc(_length+i+1) );
	if (!result) return 0;

	for (cp=_ptr, cp2=result, j=_length;  --j >= 0; ) 
	{	if (backslash.is(*cp)) 
			*cp2++='\\';
		*cp2++= backslash.decode(*cp++);
	}
	*cp2='\0';
	return result;
}

charP String::copy(charP d, charP s, int n) 
{
	strncpy(d, s, n);
	d[n]='\0';
	return d;
}

charP String::copy_quote(charP d, charP s, int n) 
{
	d[0]='"';
	strncpy(d+1, s, n-1);
	d[n]='\0';
	return d;
}

charP String::copy_colon(charP d, charP s, int n) 
{
	d[0]=':';
	strncpy(d+1, s, n-1);
	d[n]='\0';
	return d;
}

charP String::copy_noparity(charP d, charP s, int n) 

{
	for (int i=0; i < n; i++)
		d[i]= backslash.decode(s[i]);

	d[n]='\0';
	return d;
}

charP String::copy_cap(charP d, charP s, int n) 
{
	charP temp=d;
	for (int i=0; i < n; i++) 
		if (islower(*s))
			*d++= toupper(*s++);
		else
			*d++=*s++;

	*d='\0';
	return temp;
}

charP String::copy_low(charP d, charP s, int n) 
{
	charP temp=d;
	for (int i=0; i < n; i++) 
		if (isupper(*s))
			*d++= tolower(*s++);
		else
			*d++=*s++;
	*d='\0';
	return temp;
}

charP String::copy_unspace(charP d, charP s, int n) 
{
	charP temp=d;
	for (int i=0; i < n; i++) 
		if (*s==' ')
		{	*d++= '-'; s++;
		}else
			*d++=*s++;
	*d='\0';
	return temp;
}

charP String::copy_noformat(charP d, charP s, int n) 
{
	charP temp=d;
	for (int i=0; i < n; i++) 
		if (*s==9 || *s==10)
		{	s++;
		}else
			*d++=*s++;
	*d='\0';
	return temp;
}

charP String::copy_noparitylow(charP d, charP s, int n) 
{
	charP temp=d;
	for (int i=0; i < n; i++) 
	{	char c = backslash.decode(*s++);
		if (isupper(c))
			*d++= tolower(c);
		else
			*d++= c;
	}
	*d='\0';
	return temp;
}

charP String::copy_mend_nosemi(charP d, charP s, int n) 
{
	bool vbar=false;
	for (int i=0; i < n; ) 
	{	while (*s=='|') 
		{	vbar = !vbar;	s++;
		}
		if (vbar) 

			if (strchr(backslash.chars, int(*s)))
				d[i++]= backslash.encode(*s++);
			else
				d[i++]=*s++;
		else 
		{	while ((*s=='~' && s[1]=='\n')) 
				while (*s=='~' && s[1]=='\n') 
					s+=2;

			if (*s !='|') 
				d[i++]=*s++;
		}
	}
	d[n]='\0';
	return d;
}

charP String::copy_mend(charP d, charP s, int n) 
{
	bool vbar=false;
	for (int i=0; i < n; ) 
	{	while (*s=='|') 
		{	vbar = !vbar;	s++;
		}
		if (vbar) 
			if (strchr(backslash.chars, (int)*s))
				d[i++]= backslash.encode(*s++);
			else
				d[i++]=*s++;
		else 
		{	while (*s==';' || (*s=='~' && s[1]=='\n')) 
			{
				while (*s=='~' && s[1]=='\n') 
					s+=2;

				if (*s==';')
					do s++; 
					while (*s !='\0' && *s !='~' && s[1] !='\n');
			}
			if (*s !='|') 
				d[i++]=*s++;
		}
	}
	d[n]='\0';
	return d;
}

charP String::copy_word(charP d, PairPP arg, int n)   // KLUDGE!
{
	while (arg!=argStack) 
	{	++arg;
		if (!*arg)
			continue;
		StringP str = (*arg)->string();
		strncpy(d, str->ptr(), str->length());
		d += str->length();
	}
	*d='\0';
	return d;
}

charP String::copy_reverse(charP d, charP s, int n) 
{
	charP temp=d;
	s+=n-1;
	for (int i=0; i < n; i++) 
		*d++=*s--;
	*d='\0';
	return temp;
}

int String::comp_noparity(charP d, charP s, int n) 
{
	for (int i=0; i < n; i++) 
	{	if (backslash.decode(*d) != backslash.decode(*s))
			return backslash.decode(*d) - backslash.decode(*s);
		d++, s++;
	}
	return 0;
}

int String::comp_case(charP d, charP s, int n) 
{
	return strncmp(d, s, n);
/*	for (int i=0; i < n; i++) 
	{	if (*d !=*s) 
			return *d - *s;
		d++, s++;
	}
	return 0;
*/
}

int String::comp_low(charP d, charP s, int n) 
{
	for (int i=0; i < n; i++) 
	{	if (*d !=*s) 
		{	if (isupper(*s)) 
			{	if (isupper(*d)) 

				{	if (tolower(*d) != tolower(*s))
						return tolower(*d) - tolower(*s);
				}else 
				{	if (*d != tolower(*s))
						return *d - tolower(*s);
				}
			}else 
				if (isupper(*d)) 
				{	if (tolower(*d) !=*s)


						return tolower(*d) - *s;
				}else 
					return *d - *s;
		}
		d++, s++;
	}
	return 0;
}

int String::comp_noparitylow(charP d, charP s, int n) 
{
	for (int i=0; i < n; i++) 
	{	char c1 = backslash.decode(*d),
			  c2 = backslash.decode(*s);
		if (c1 != c2) 
		{	if (isupper(c2)) 
			{	if (isupper(c1)) 
				{	if (tolower(c1) != tolower(c2))
						return tolower(c1) - tolower(c2);
				}else 
				{	if (c1 != tolower(c2))
						return c1 - tolower(c2);
				}
			}else 
				if (isupper(c1)) 
				{	if (tolower(c1)!=c2)
						return tolower(c1) - c2;
				}else 
					return c1 - c2;
		}
		d++, s++;
	}
	return 0;
}

//___________________________________________________________________

Pair::Pair(charP str):	
	Node(PAIR), _tail(0) {setfirst(new String(str));}

void Pair::setPtrs()
{	setfirst((Node*)GC_::intToPtr->find(int(_first)));
	settail ((Node*)GC_::intToPtr->find(int(_tail )));
}
wxDataInputStream &Pair::read(wxDataInputStream &istr) 
{	return istr >> (int&)_first >> (int&)_tail;
}
wxDataOutputStream &Pair::write(wxDataOutputStream &ostr) 
{	return ostr << type 
		<< GC_::ptrToInt->find(first()) 
		<< GC_::ptrToInt->find(tail());
}

StringP Pair::string()
{
	const int maxlength=65536;
	char textbuf[maxlength];
	File::stringptr=textbuf;
	File::stringlen=maxlength-1;
	stringbuffer << NodeP(this);
	*File::stringptr='\0';
	return new String(textbuf);
}

PairP Pair::append(NodeP tailnode)	// nondestructive append (slow!)
{
	if (!this) return PairP(tailnode);
	return new Line
		(	first(), 
			tail()->append(tailnode)
		);
}

PairP Pair::reverse_destruct()	// reverse a list destructively 
{
	PairP list=this, ret=0, temp;
	while (list)
	{	temp=list;
		next(list);
		temp->settail( ret );
		ret=temp;
	}
	return ret;
}

PairP Pair::merge(PairP b) 
{
	if (!this) return b;
	if (!b) return this;

	PairP ret=0, tnode=0, a=this;


	if (!a->first()) {error(BAD_DATA, this); return PairP(Unbound);}
	if (a->first()->compare(b->first()) < 0) 
	{	ret=a;	tnode=a;	next(a);
	}else 
	{	ret=b;	tnode=b;	next(b);
	}
	while (a  &&  b)
	{	if (!a->first()) {error(BAD_DATA, this); return PairP(Unbound);}
		if (a->first()->compare(b->first()) < 0) 
		{	tnode->settail(a);	next(a);
		}else 
		{	tnode->settail(b);	next(b);
		}
		next(tnode);
	}
	if (a)
		tnode->settail(a);
	else
		tnode->settail(b);

	return ret;
}

void Pair::mergePairs() 
{
	PairP nd=this;
	while (nd)
	{	if (!nd->tail()) break;
		nd->setfirst( nd->first()->merge(nd->second()) );
		nd->settail( nd->tail()->tail() );
		next(nd);
	}
}

void Pair::toListList() 
{

	PairP nd=this;
	while (nd) 
	{	nd->setfirst( new Pair(nd->first(), 0) );
		next(nd);
	}
}

PairP Pair::mergeSort()
{
	if (!this) return 0;
	if (!tail()) return this;
	toListList();
	while (tail())
		mergePairs();
	return first();
}

PairP Pair::one_list() 
{
	if (islist())
		return this;
	else
		return new Pair(this, 0);
}

void Pair::three_lists(PairP &proclist, PairP &varlist, PairP &plistlist) 
{
	if (this)
		if (!is.list)
			proclist = one_list();
		else if (first()->isnolist())
			proclist = this;
		else
		{	proclist = first();
			if (tail()) 
			{	varlist = second()->one_list();
				if (tail()->tail()) 
					plistlist = third()->one_list();
			}
		}

	if ( proclist->isnolist()
	||   varlist  ->isnolist()
	||   plistlist->isnolist() )
	{	error(BAD_DATA_UNREC, this);
		plistlist = varlist = proclist = NIL;
	}
}

bool Pair::all_lists()
{
	if (!this) 
		return true;

	if (first()->isnolist()) 
		return false;

	return tail()->all_lists();
}

void Pair::setflag(ObjectType flag)
{
	PairP list=this;
	while (list && !throwing) 
	{
		if (list->first()->is.aggregate) 
		{	error(BAD_DATA_UNREC, list->first());
			break;
		}
		list->first()->intern()->setflag(flag);
		next(list);
		if (checkThrowing()) break;
	}
}

void Pair::clearflag(ObjectType flag)
{
	PairP list=this;
	while (list && !throwing) 
	{
		if (list->first()->is.aggregate) 
		{	error(BAD_DATA_UNREC, list->first());
			break;
		}
		list->first()->intern()->clearflag(flag);
		next(list);
		if (checkThrowing()) break;
	}
}

NodeP Pair::bury(ObjectType flag) 
{
	PairP proclist=0, varlist=0, plistlist=0;
	three_lists(proclist, varlist, plistlist);

	int &flagi = (int&) flag;

	proclist->setflag(flag);  flagi <<= 1;
	  varlist->setflag(flag);  flagi <<= 1;

	plistlist->setflag(flag);
	return Unbound;
}

NodeP Pair::unBury(ObjectType flag) 
{
	PairP proclist=0, varlist=0, plistlist=0;
	three_lists(proclist, varlist, plistlist);

	int &flagi = (int&) flag;


	proclist->clearflag(flag);  flagi <<= 1;
	  varlist->clearflag(flag);  flagi <<= 1;
	plistlist->clearflag(flag);
	return Unbound;
}

NodeP Pair::procType(int wanted) 
{
	NodeP cell,  arg = a_name(1);
	if throwing return Unbound;
	if ((cell = arg->intern()->proc_or_prim())==Unbound)
		return _False;
	
	if (wanted==2) 
		return boolnode( arg->intern()->bit.proc_macro );

	int isprim = cell->is.primitive;
	if (!throwing) 
		return boolnode(isprim==wanted);

	return Unbound;
}

void Pair::eval_driver()		// Evaluate a line of input. 
{
	Evaluator(this, __begin_line);
}

NodeP Pair::err_eval_driver()	  // Evaluate a sequence of expressions 
{								// until we get a value to return (Called from erract)
	Evaluator::val_status = value_maybe_ok_in_macro; 
	return Evaluator(PairP(this), __begin_sequence).val;
}

void Pair::runstartup() 
{
	if (this &&  islist()) 
	{	Evaluator::val_status = no_value_allowed;
		eval_driver();
	}
	else if (this && type==OBJECT) 
	{	Evaluator::val_status = no_value_allowed;
		(new Pair(ObjectP(this), 0))->eval_driver();
	}
}
//___________________________________________________________________

#ifdef WIN32
	#pragma warning(disable: 4096)
#endif

List::List(NodeP first_, ...)
{
	if (first_==_ListEnd) return;
	setfirst(first_);

	va_list args;
	va_start(args, first_);
	NodeP arg;
	PairP lastnode=this;

	while ( (arg=va_arg(args, NodeP )) != _ListEnd ) 
	{
		PairP val = new Pair(arg, 0);
		lastnode->settail( val );
		lastnode=val;
	}
	va_end(args);
}

#ifdef WIN32
	#pragma warning(default: 4096)
#endif

List::List(PairP list)	
{
	PairP last;

	if (list)
	{	last=this;
		setfirst( list->first() );
		next(list);
	}
	while (list)
	{	PairP newpair = new Pair(list->first(), 0);
		last->settail( newpair );
		last=newpair;
		next(list);
	}
}

//___________________________________________________________________

StringP Quote::string()
{
	if (quote())
	{	StringP str = quote()->string();
		str = new String(str->ptr(), 0, str->length()+1, str->type, String::copy_quote);
		return str;
	}
	else
		return _Null_Word->name();
}

//___________________________________________________________________

void Tree::setPtrs()
{	setfirst     ((Node*)GC_::intToPtr->find(int(_first     )));
	settail      ((Node*)GC_::intToPtr->find(int(_tail      )));
	filename=   (String*)GC_::intToPtr->find(int(filename   ));
	settree      (PairP((Pair*)GC_::intToPtr->find(int(_tree      ))));
	setgeneration((Node*)GC_::intToPtr->find(int(_generation)));
}
wxDataInputStream &Tree::read(wxDataInputStream &istr) 
{	return istr >> (int&)_first >> (int&)_tail 
		>> number >> (int&)filename 
		>> (int&)_tree >> (int&)_generation;
}
wxDataOutputStream &Tree::write(wxDataOutputStream &ostr) 
{	return ostr << type
		<< GC_::ptrToInt->find(first()) 
		<< GC_::ptrToInt->find(tail())
		<< number
		<< GC_::ptrToInt->find(filename)
		<< GC_::ptrToInt->find(tree())
		<< GC_::ptrToInt->find(generation());
}

void Tree::untreeify() 
{
	if (this)
		if (is.list)
		{	((TreeP)first())->untreeify();
			((TreeP) tail())->untreeify();
			settype_LINE();
		}
}

void Tree::setline(PairP nd)  // Set the line pointer for a tree.
{
	settype_LINE();  
	settree(nd);  
	setgeneration(the_generation);
}

//___________________________________________________________________

Procedure::Procedure(PairP formals_, PairP bodylist_, PairP bodywords_, 
			   char min_, char std_, char max_):
	Node(PROC_), min(min_), std(std_), max(max_) 
{
	setformals(formals_);  setbodylist(bodylist_);  setbodywords(bodywords_);
}

void Procedure::setPtrs()
{	setformals  (PairP((Pair*)GC_::intToPtr->find(int(_formals  ))));
	setbodylist (PairP((Pair*)GC_::intToPtr->find(int(_bodylist ))));
	setbodywords(PairP((Pair*)GC_::intToPtr->find(int(_bodywords))));
}
wxDataInputStream &Procedure::read(wxDataInputStream &istr) 
{	return istr >> (int&)_formals >> (int&)_bodylist >> (int&)_bodywords
		>> (unsigned char&)min 
		>> (unsigned char&)std 
		>> (unsigned char&)max;
}
wxDataOutputStream &Procedure::write(wxDataOutputStream &ostr) 
{	return ostr << type 

		<< GC_::ptrToInt->find(formals()) 
		<< GC_::ptrToInt->find(bodylist())
		<< GC_::ptrToInt->find(bodywords())
		<< (unsigned char)min 
		<< (unsigned char)std 
		<< (unsigned char)max;
}

void Procedure::untreeify() 
{
	TreeP	body = TreeP(bodylist());

	for (TreeP body_ptr=body;  body_ptr;  body_ptr=TreeP(body_ptr->tail())) 
		TreeP(body_ptr->first())->untreeify();
	
	body->untreeify();
}

PairP Procedure::get_bodywords(NodeP name)
{
	PairP
		val = bodywords(),
		head=0, tnode=0;
	
	if (val) return val;

	ObjectP obj = name->intern();

	head = new List
		(	obj->bit.proc_macro ? __Macro_ : _To, 
			obj->name(), 
			_ListEnd
		);
	tnode=head->tail();
	val=formals();

	while (val) 
	{
		if (val->first()->is.list)
			tnode->settail
			(	new Pair
				(	new Pair
					(	new Colon(val->first()->first()), 
						val->first()->tail()
					), 
					0
				)
			);
		else 
			if ( val->first()->type == INT_)
				tnode->settail( new Pair(val->first(), 0) );
			else
				tnode->settail
				(	new Pair
					(	new Colon(val->first()),
						0
					)
				);
		next(tnode);
		next(val);
	}
	tnode = head = new Pair(head, 0);
	
	for (val=bodylist();  val;  next(val))
	{	tnode->settail( new Pair( val->first()->runparse(), 0 ) );
		next(tnode);
	}
	tnode->settail

	(	new Pair
		(	new Pair(_End, 0), 
			0
		)
	);
	return head;
}

//___________________________________________________________________

void Object::setname   (StringP    nd) {_name = nd;  check_oldyoung(nd);}
void Object::setproc_or_prim(NodeP nd) {_proc = nd;  check_oldyoung(nd);}
void Object::setproc   (ProcedureP nd) {_proc = nd;  check_oldyoung(nd);}
void Object::setplist  (PairP      nd) {_plist= nd;  check_oldyoung(nd);}

void Object::inc_gc() 
{	fifo << name() << proc_or_prim() << var() << plist() << tail();
}
int Object::youngs() 
{	return gmark(name())+gmark(proc_or_prim())+gmark(var())
			+gmark(plist())+gmark(tail());
}
int Object::youngsPtr() 
{	return gmarkPtr(name())+gmarkPtr(proc_or_prim())+gmarkPtr(var())
			+gmarkPtr(plist())+gmarkPtr(tail());
}
void Object::check_oldyoungs() 
{	check_oldyoung(name()); check_oldyoung(proc_or_prim()); 
	check_oldyoung(var()); check_oldyoung(plist()); check_oldyoung(tail());
}

void Object::setPtrs()
{	settail        (ObjectP((Object*)GC_::intToPtr->find(int(_tail ))));
	setname        (StringP((String*)GC_::intToPtr->find(int(_name ))));
	setproc_or_prim(          (Node*)GC_::intToPtr->find(int(_proc )));
	setvar         (          (Node*)GC_::intToPtr->find(int(_var  )));
	setplist       (    PairP((Pair*)GC_::intToPtr->find(int(_plist))));
}
wxDataInputStream &Object::read(wxDataInputStream &istr) 
{	return istr >> (int&)_tail >> (int&)_name 
		>> (int&)_proc >> (int&)_var >> (int&)_plist 
		>> tint;
}
wxDataOutputStream &Object::write(wxDataOutputStream &ostr) 
{	return ostr << type 
		<< GC_::ptrToInt->find(_tail) 
		<< GC_::ptrToInt->find(_name)
		<< GC_::ptrToInt->find(_proc)
		<< GC_::ptrToInt->find(_var )
		<< GC_::ptrToInt->find(_plist)

		<< tint;
}

int Object::numberP() {return name()->numberP();}
double Object::getnumber() {return name()->getnumber();}
NodeP Object::number() {return name()->number();}

void Object::print() 
{	c_out << name() << ' ';
	if (dribbler) dribbler << name() << ' ';
}

StringP Object::contents(ContentsType contents_type, ObjectType want) 
{
	switch (contents_type) 
	{
	case PROCS:
		if (proc_or_prim()==Unbound) return 0;
		break;

	case VALS:
		if (var()==Unbound) return 0;
		break;

	case PLISTS:
		if (!plist()) return 0;
		break;
	}
//	ObjectBits &bits = (ObjectBits&) want;

	if (want==NORMAL_OBJECT)
	{	if (flag==NORMAL_OBJECT)
			return name();
		else
			if (flag & BURIED ||  flag & PERMANENT)
				return 0;
			else
				return name();
	}else
		if (flag & want 
		&& (want & PERMANENT 
			|| (!(want & PERMANENT) && !(flag & PERMANENT))))
			return name();
		else
			return 0;
}

void Object::mark_all_but_objects() 
{
	name()->mark(); proc_or_prim()->mark(); 
	var()->mark(); plist()->mark();
}

void Object::tell_shadow()	// Warn the user if a local variable shadows a global one. 
{
	if (bit.val_stepped)
		error(SHADOW_WARN, this);
}

bool Object::not_local(int stackPtr)	// Check if a local variable is already in this frame 
{
	for ( ;  stackPtr > Evaluator::var;  stackPtr--) 
		if (equal(Evaluator::varStack.base[stackPtr].object)) 
			return false;
	return true;
}

int Object::find_limit() 
{
	if (!this) return -1;
	NodeP num = var()->number();
	
	if (type_of(num) == INT_) 
		return IntP(num)->int_;
	else
		return -1;
}

//___________________________________________________________________

StringP Colon::string()
{
	StringP str = colon()->string();
	str = new String(str->ptr(), 0, str->length()+1, str->type, String::copy_colon);
	return str;
}

//___________________________________________________________________

int Line::nr=1;
int Line::readnr=0;
StringP Line::fn=0;
PairP Line::filenames=0;

void Line::setPtrs()
{	setfirst(  (Node*)GC_::intToPtr->find(int(_first  )));
	settail (  (Node*)GC_::intToPtr->find(int(_tail   )));
	filename=(String*)GC_::intToPtr->find(int(filename));
}
wxDataInputStream &Line::read(wxDataInputStream &istr) 
{	return istr >> (int&)_first >> (int&)_tail 
		>> number >> (int&)filename;
}
wxDataOutputStream &Line::write(wxDataOutputStream &ostr) 
{	return ostr << type 
		<< GC_::ptrToInt->find(first()) 
		<< GC_::ptrToInt->find(tail())
		<< number
		<< GC_::ptrToInt->find(filename);
}

void Line::init()
{
	fn=new String("unnamed");
	filenames=new Pair(fn, 0);
}

StringP Line::findFile(StringP filename_)
{
	StringP f=0;
	for (PairP fnlist=Line::filenames;  fnlist;  next(fnlist))
		if (fnlist->first() && fnlist->first()->equal(filename_))
			f=StringP(fnlist->first());
	if (f)
		return f;
	else
		Line::filenames=new Pair(filename_, Line::filenames);
	return filename_;
}
//___________________________________________________________________

void Runparsed::setPtrs()
{	setfirst(  (Node*)GC_::intToPtr->find(int(_first  )));
	settail (  (Node*)GC_::intToPtr->find(int(_tail   )));
	filename=(String*)GC_::intToPtr->find(int(filename));
	setparsed( (Node*)GC_::intToPtr->find(int(_parsed)));
}
wxDataInputStream &Runparsed::read(wxDataInputStream &istr) 
{	return istr >> (int&)_first >> (int&)_tail 
		>> number >> (int&)filename >> (int&)_parsed;
}
wxDataOutputStream &Runparsed::write(wxDataOutputStream &ostr) 
{	return ostr << type 
		<< GC_::ptrToInt->find(first()) 
		<< GC_::ptrToInt->find(tail())
		<< number
		<< GC_::ptrToInt->find(filename)
		<< GC_::ptrToInt->find(parsed());
}
//___________________________________________________________________

void ToDoNode::setPtrs()
{	settail((Node*)GC_::intToPtr->find(int(_tail)));
}
wxDataInputStream &ToDoNode::read(wxDataInputStream &istr) 
{	istr >> (int&)_tail;
#ifdef PROC_EVAL
	#ifdef _DEBUG
		istr >> (int&)toDo.mnemonic;
	#endif
		istr >> (int&)toDo.instruction;
#else
	istr >> (int&)toDo.mnemonic;
#endif
	return istr;
}
wxDataOutputStream &ToDoNode::write(wxDataOutputStream &ostr) 
{	ostr << type 
		<< GC_::ptrToInt->find(tail());
#ifdef PROC_EVAL
	#ifdef _DEBUG
		ostr << toDo.mnemonic;
	#endif
		ostr << (int&)toDo.instruction;
#else
	ostr << toDo.mnemonic;
#endif
	return ostr;
}

//___________________________________________________________________
//
//                              Main
//___________________________________________________________________

#ifdef WIN32
	#pragma optimize("", off)
#endif
int _cdecl _main(int argc, charP argv[]) 
{
	PairP exec_list=0;
	GC_::bottom_stack = NodePP(&exec_list);
	// from now until exit all new vars allocated on the stack have addresses 
	// between <GC_::bottom_stack> and the recent var on the stack (<top> in GC_).

	init();
	graph->autoload();
	if (theApp->mainFrame->nodesAutosave)
	{	if (wxFileExists(theApp->nodesFileName))
			GC_::load(charP(theApp->nodesFileName.c_str()));
	}
//	signal(SIGINT , logo_stop);
//	signal(SIGABRT, logo_pause);

	argc--;
	argv++;
	if (argc==0 && console->GetLastPosition()==0) 
		c_out << "; Welcome to aUCBLogo version 4.692" << endL;

	while (argc-- > 0)
	{	NIL->silent_load(*argv);
		StringP str=0;
		{	wxString s=*argv;
			int pos=s.Find('.', true);
			s.Truncate(pos);
			pos=s.Find(
#ifdef WIN32
				'\\',
#else
				'/',
#endif
						true);
			if (pos>=0)
			{	wxString path;
				path=s.Left(pos);
				if (!wxSetWorkingDirectory(path))
					error(FILE_ERROR, new List(
						new String("couldn't change dir to '"),
						new String(path),	_ListEnd));
				s.Remove(0,pos+1);
			}
			str=new String(s);
		}
		if (!throwing)
		{	File::currentLine=str;
			exec_list = File::currentLine->parser(true);
			Evaluator::val_status = no_value_allowed;
			if (exec_list) exec_list->eval_driver();
		}
		if (theApp->action==EXIT_LOGO || theApp->action==EXIT_DELAYED)
		{	Segment::freeAll();
			return 0;
		}
		argv++;
		if (!*argv)
			break;
	}
	while (true)
	{	if throwing
		{	     if (throw_node==_System  ) break;
			else if (throw_node==_Error   ) error_print();
			else if (throw_node!=_Toplevel)
			{	error(NO_CATCH_TAG, throw_node);	error_print();
			}
			stoppingFlag=RUN;
			inpause=false;
		}
		if (stopping || outputting)
		{	c_out << "; You must be in a procedure to use OUTPUT or STOP." << endL;
			stoppingFlag=RUN;
		}
		argStack.clean();

		stepArmed=false;
		GC_::reserve_tank_check();
		File::currentLine = c_in.read("");
		// no prompt, because every line shall be executable

		if throwing continue;
		
		exec_list = File::currentLine->parser(true);
		Evaluator::val_status = no_value_allowed;
		if (exec_list) exec_list->eval_driver();
	}
	Segment::freeAll();
	return 0;
}
