//___________________________________________________________________
//
//                       Class MainFrame
//
//                          Definitons
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/textctrl.h>
#include <wx/glcanvas.h>
#include <wx/listctrl.h>
#include <wx/frame.h>
#include <wx/splitter.h>
#include <wx/menu.h>
#include <wx/dialog.h>
#include <wx/filedlg.h>
#include <wx/html/helpctrl.h>
#include <wx/msgdlg.h>
#include <wx/dc.h>

#include <math.h>

#include <GL/glu.h>

#include "Main.h"

#include "File.h"
#include "String.h"
#include "Pair.h"
#include "Object.h"

#include "Eval.h"
#include "Stack.h"
#include "Error.h"
#include "Files.h"
#include "Init.h"
#include "Workspce.h"

#include "UInt8.h"
#include "Color.h"
#include "Point.h"
#include "Pen.h"
#include "Font.h"
#include "Screen.h"
#include "Turtle.h"
#include "Console.h"
#include "Graph.h"
#include "Recorder.h"
#include "Vars.h"
#include "Calls.h"
#include "Splitter.h"
#include "MainFrame.h"
#include "AboutDlg.h"
#include "MemoryWindow.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

static char THIS_FILE[]= __FILE__;

BEGIN_EVENT_TABLE(MainFrame, wxFrame)
	EVT_CLOSE(							  MainFrame::OnClose				)
	EVT_MENU (wxID_EXIT				, MainFrame::OnExit				)
	EVT_MENU (wxID_ABOUT				, MainFrame::OnAppAbout			)
	EVT_MENU (wxID_NEW				, MainFrame::OnFileNew			)
	EVT_MENU (wxID_OPEN				, MainFrame::OnFileOpen			)
	EVT_MENU (wxID_SAVE				, MainFrame::OnFileSave			)
	EVT_MENU (ID_CONSOLE_OPEN		, MainFrame::OnConsoleOpen		)
	EVT_MENU (ID_CONSOLE_SAVE		, MainFrame::OnConsoleSave		)
	EVT_MENU (ID_CONSOLE_AUTOSAVE	, MainFrame::OnConsoleAutoSave)
	EVT_MENU (ID_PICTURE_OPEN		, MainFrame::OnPictureOpen		)
	EVT_MENU (ID_PICTURE_SAVE		, MainFrame::OnPictureSave		)
	EVT_MENU (ID_PICTURE_TEXT_OPEN, MainFrame::OnPictureTextOpen)
	EVT_MENU (ID_PICTURE_TEXT_SAVE, MainFrame::OnPictureTextSave)
	EVT_MENU (ID_PICTURE_AUTOSAVE	, MainFrame::OnPictureAutoSave)
	EVT_MENU (ID_NODES_OPEN			, MainFrame::OnNodesOpen		)
	EVT_MENU (ID_NODES_SAVE			, MainFrame::OnNodesSave		)
	EVT_MENU (ID_NODES_AUTOSAVE	, MainFrame::OnNodesAutoSave	)
	EVT_MENU (ID_EDIT_REPEAT		, MainFrame::OnEditRepeat		)
	EVT_MENU (ID_EDIT_RESET			, MainFrame::OnEditReset		)
	EVT_MENU (ID_VIEW_GRAPH			, MainFrame::OnViewGraph		)
	EVT_MENU (ID_VIEW_VARS			, MainFrame::OnViewVars			)
	EVT_MENU (ID_VIEW_CALLS			, MainFrame::OnViewCalls		)
	EVT_MENU (ID_VIEW_MEMORY		, MainFrame::OnViewMemory		)
	EVT_MENU(wxID_HELP_CONTENTS	, MainFrame::OnHelp				)
	EVT_ERASE_BACKGROUND(           MainFrame::OnErase				) 
	EVT_MOVE (                      MainFrame::OnMove				)

	EVT_MENU(wxID_PRINT				, MainFrame::OnPrint				)
	EVT_MENU(wxID_PREVIEW			, MainFrame::OnPrintPreview	)
	EVT_MENU(wxID_PRINT_SETUP		, MainFrame::OnPrintSetup		)
	EVT_MENU(ID_PAGE_SETUP			, MainFrame::OnPageSetup		)
END_EVENT_TABLE()

#define NEED_MEM(x)											\
	if (!x)														\
	{	delete theApp->heapReserve;						\
		wxString msg;											\
		msg << THIS_FILE << "(" << __LINE__ << "):"	\
			#x": Insufficient Memory";						\
		wxMessageBox(msg,	"aUCBLogo Error");			\
		return;													\
	}

MainFrame::MainFrame(wxWindow* parent, wxWindowID id, 
	const wxString& title, 
	const wxPoint& pos, 
	const wxSize& size, 
	long style, 
	const wxString& name,
	bool consoleAutosave, int caretpos, 
	bool pictureAutosave,
	bool nodesAutosave_):
wxFrame(parent, id, title, pos, size, style, name)
{
	initialized=false;
	inGraphicsMode=inSplitScreen=false;
	printData=0;
	pageSetupData=0;

	fileMenu = new wxMenu;
	NEED_MEM(fileMenu);
	fileMenu->Append(wxID_NEW   ,"&New\tCtrl-N"    ,"New File");
	fileMenu->Append(wxID_OPEN  ,"&Open...\tCtrl-O","Open File");
	fileMenu->Append(wxID_SAVE  ,"&Save...\tCtrl-S"   ,"Save File");
//	fileMenu->Append(wxID_SAVEAS,"Save &As..."     ,"Save as");
	fileMenu->AppendSeparator();
	fileMenu->Append(ID_CONSOLE_OPEN  ,"Op&en Console...","Open Console");
	fileMenu->Append(ID_CONSOLE_SAVE  ,"Sa&ve Console...","Save Console");
	fileMenu->AppendCheckItem (ID_CONSOLE_AUTOSAVE, "Autosave Console");
	fileMenu->Check(ID_CONSOLE_AUTOSAVE, consoleAutosave);
	fileMenu->AppendSeparator();
	fileMenu->Append(ID_PICTURE_OPEN  ,"Open Picture...","Open Picture");
	fileMenu->Append(ID_PICTURE_SAVE  ,"Save.Picture..","Save Picture");
	fileMenu->Append(ID_PICTURE_TEXT_OPEN  ,"Open Picture in Textmode...","Open Picture in Textmode");
	fileMenu->Append(ID_PICTURE_TEXT_SAVE  ,"Save.Picture in Textmode...","Save Picture in Textmode");
	fileMenu->AppendCheckItem (ID_PICTURE_AUTOSAVE, "Autosave Picture");
	fileMenu->Check(ID_PICTURE_AUTOSAVE, pictureAutosave);
	fileMenu->AppendSeparator();
	nodesAutosave=nodesAutosave_;
//#ifdef WIN32
	fileMenu->Append(ID_NODES_OPEN  ,"Open Nodes...","Open Nodes");
	fileMenu->Append(ID_NODES_SAVE  ,"Save Nodes...","Save Nodes");
	fileMenu->AppendCheckItem (ID_NODES_AUTOSAVE, "Autosave Nodes");
	fileMenu->Check(ID_NODES_AUTOSAVE, nodesAutosave);
	fileMenu->AppendSeparator();
//#endif
	fileMenu->Append(wxID_PRINT ,"&Print...\tCtrl-P"  ,"Print File");
	fileMenu->Append(wxID_PRINT_SETUP ,"P&rint Setup...","Print Setup");
	fileMenu->Append(ID_PAGE_SETUP, "Pa&ge Setup...", "Page setup");
	fileMenu->Append(wxID_PREVIEW ,"Pr&int Preview"  ,"Print Preview");
	fileMenu->AppendSeparator();
	fileMenu->Append(wxID_EXIT  ,"E&xit\tAlt-X"    ,"Exit");

	editMenu = new wxMenu;
	NEED_MEM(editMenu);
	editMenu->Append(wxID_UNDO  ,"&Undo\tCtrl-Z" , "Undo Edit");
	editMenu->Append(wxID_CUT   ,"Cu&t\tCtrl-X"  , "Cut Text");
	editMenu->Append(wxID_COPY  ,"&Copy\tCtrl-C" , "Copy Text");
	editMenu->Append(wxID_PASTE ,"&Paste\tCtrl-V", "Paste Text");
	editMenu->AppendSeparator();
	editMenu->Append(ID_EDIT_REPEAT ,"Repeat &last command\tF2", "Repeat last command");
	editMenu->Append(ID_EDIT_RESET  ,"&Reset\tCtrl-F2", "Call reset.lg");

	viewMenu = new wxMenu;
	NEED_MEM(viewMenu);
	viewMenu->Append(ID_VIEW_VARS ,"&Vars" ,"View Vars");
	viewMenu->Append(ID_VIEW_CALLS,"&Calls","View Calls");
	viewMenu->Append(ID_VIEW_GRAPH,"&Graph","View Graph");
	viewMenu->AppendCheckItem (ID_VIEW_MEMORY, "&Memory");
	viewMenu->Check(ID_VIEW_MEMORY, theApp->memoryShown);

	helpMenu = new wxMenu;
	NEED_MEM(helpMenu);
	helpMenu->Append(wxID_HELP_CONTENTS,"&Help\tF1","Help");
	helpMenu->AppendSeparator();
	helpMenu->Append(wxID_ABOUT,"&About","About");

	wxMenuBar *menuBar=new wxMenuBar;
	NEED_MEM(menuBar);
	menuBar->Append(fileMenu  ,"&File");
	menuBar->Append(editMenu  ,"&Edit");
	menuBar->Append(viewMenu  ,"&View");
//	menuBar->Append(windowMenu,"&Window");
	menuBar->Append(helpMenu  ,"&Help");
	SetMenuBar(menuBar);

	splitter=new Splitter(this, -1, wxDefaultPosition, 
		wxDefaultSize,	style & ~wxSP_PERMIT_UNSPLIT, "Splitter");
	NEED_MEM(splitter);
	graph=new Graph(GraphNodeP(0), splitter, -1, 
		wxDefaultPosition, wxDefaultSize, 
		style | wxFULL_REPAINT_ON_RESIZE | wxTAB_TRAVERSAL, "Graph"
#ifndef WIN32
		, 0,
#else
		, Graph::attribs,
#endif
		wxNullPalette, pictureAutosave);
	NEED_MEM(graph);
	graphAutosave=pictureAutosave;
	::graph=graph;
	graphNode=0;

	splitter2=new Splitter(splitter, -1, wxDefaultPosition, 
		wxDefaultSize,	(style | wxFULL_REPAINT_ON_RESIZE) & ~wxSP_PERMIT_UNSPLIT, "Splitter2");
	NEED_MEM(splitter2);
	splitter->SplitHorizontally(graph, splitter2, 1);
	splitter->SetMinimumPaneSize(1);
	double ratio;
	theApp->config->Read("SplitterRatio", &ratio, double(0)); 
	splitter->setRatio(ratio);


	console=new Console(splitter2, -1, "", 
		wxDefaultPosition, wxDefaultSize, 
		wxTE_MULTILINE | wxTE_PROCESS_TAB | wxTE_RICH2 | wxTE_NOHIDESEL,
		wxDefaultValidator, "Console", consoleAutosave, caretpos);
	NEED_MEM(console);
	splitter3=new Splitter(splitter2, -1, wxDefaultPosition, 
		wxDefaultSize,	style & ~wxSP_PERMIT_UNSPLIT, "Splitter3");
	NEED_MEM(splitter3);
	splitter2->SplitHorizontally(console, splitter3, int(theApp->max.y*0.9*3/4));
	splitter2->SetMinimumPaneSize(1);
	theApp->config->Read("SplitterRatio2", &ratio, double(1)); 
	splitter2->setRatio(ratio);

	vars=new Vars(splitter3, -1, wxDefaultPosition, wxDefaultSize, 
		wxLC_REPORT | wxLC_VIRTUAL);
	NEED_MEM(vars);
	calls=new Calls(splitter3, -1, wxDefaultPosition, wxDefaultSize, 
		wxLC_REPORT | wxLC_VIRTUAL);
	NEED_MEM(calls);

	splitter3->SplitHorizontally(vars, calls, 10);
	splitter3->SetMinimumPaneSize(1);
	theApp->config->Read("SplitterRatio3", &ratio, double(1)); 
	splitter3->setRatio(ratio);

	console->SetFocus();

#ifdef WIN32
	SetIcons(wxIconBundle(wxIcon(
		theApp->runDirectory+"/logo.ico", wxBITMAP_TYPE_ICO)));
#endif

	printData = new wxPrintData;
	NEED_MEM(printData);
	printData->SetOrientation(wxLANDSCAPE);
	pageSetupData = new wxPageSetupDialogData;
	NEED_MEM(pageSetupData);
	initialized=true;
}

MainFrame::~MainFrame() 
{
	if (printData)
		delete printData;
	if (pageSetupData)
		delete pageSetupData;
}

void MainFrame::setSplitter(double y)
{
	if (y==0)
		inGraphicsMode=inSplitScreen=false;
	else
		inGraphicsMode=inSplitScreen=true;
	splitter->setRatio(y);
}

void MainFrame:: TextScreen(){setSplitter(0);}
void MainFrame::splitScreen(){setSplitter(0.5);}
void MainFrame:: fullScreen(){setSplitter(1);}
void MainFrame::allFullScreen(){ShowFullScreen(true, wxFULLSCREEN_ALL);}
void MainFrame::notFullScreen(){ShowFullScreen(false);}

NodeP lTextScreen (PairPP) 
{
	theApp->mainFrame-> TextScreen(); return Unbound;
}

NodeP lsplitScreen(PairPP arg) 
{
	if (exist(1))
	{	double ratio=arg[1]->a_float(1);
		if throwing return Unbound;
		theApp->mainFrame->setSplitter(ratio); 
	}else
		theApp->mainFrame->splitScreen(); 
	return Unbound;
}

NodeP lfullScreen (PairPP) 
{
	theApp->mainFrame-> fullScreen(); return Unbound;
}

NodeP lsetVarsSplitter(PairPP arg) 
{
	double ratio=arg[1]->a_float(1);
	if throwing return Unbound;
	theApp->mainFrame->splitter2->setRatio(ratio); 
	return Unbound;
}

NodeP lsetCallsSplitter(PairPP arg) 
{
	double ratio=arg[1]->a_float(1);
	if throwing return Unbound;
	theApp->mainFrame->splitter3->setRatio(ratio); 
	return Unbound;
}

NodeP lallFullScreen(PairPP)
{
	theApp->mainFrame->allFullScreen(); return Unbound;
}

NodeP lnotFullScreen(PairPP) 
{
	theApp->mainFrame->notFullScreen(); return Unbound;
}

void MainFrame::OnMove(wxMoveEvent& event)
{
	if (initialized)
		graph->swapBuffers();
}
void MainFrame::OnErase(wxEraseEvent& event){}

void MainFrame::printPrompt(charP str) 
{
	if (inGraphicsMode && !inSplitScreen)
		lsplitScreen(argStack);
	c_out << str;
	console->flush();
}
void File::printPrompt(charP str) 
{
	theApp->mainFrame->printPrompt(str);
}

void MainFrame::OnClose(wxCloseEvent &event) 
{
	theApp->action=EXIT_LOGO;
}

void MainFrame::OnExit(wxCommandEvent &event) 
{
	wxCloseEvent cev;
	OnClose(cev);
}

void MainFrame::OnViewGraph(wxCommandEvent &event) 
{
	if (graph)
		graph->update();
}

void MainFrame::OnViewVars(wxCommandEvent &event) 
{
	if (vars)
		vars->update();
}

void MainFrame::OnViewCalls(wxCommandEvent &event) 
{
	if (calls)
		calls->update();
}

void MainFrame::OnViewMemory(wxCommandEvent &event) 
{
	if (!memoryWindow)
		theApp->memoryShow();
	else
		theApp->memoryHide();
}

void MainFrame::OnHelp(wxCommandEvent &event) 
{
	if (console)
	{	wxHelpEvent hev;
		console->OnHelp(hev);
	}
}

void MainFrame::OnAppAbout(wxCommandEvent &event)
{
	AboutDlg aboutDlg;
	aboutDlg.ShowModal();
}

void MainFrame::OnFileNew(wxCommandEvent &event)
{
	if (_True)
		leraseAll(argStack);
}

void MainFrame::OnFileOpen(wxCommandEvent &event)
{
	wxFileDialog fd(this, "Choose a Logo file","","",
		"Logo Files (*.lg)|*.lg|All Files (*.*)|*.*",
		wxOPEN|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK)
	{	wxString cstr=fd.GetPath();
		NodeP pn=new String(cstr);
		PairPP asp=argStack;
		argStack << pn;
		lload(asp);
		argStack.pop_till(asp);
	}
}

void MainFrame::OnFileSave(wxCommandEvent &event)
{
	wxFileDialog fd(this, "Choose a filename","","",
		"Logo Files (*.lg)|*.lg|All Files (*.*)|*.*",
		wxSAVE|wxOVERWRITE_PROMPT|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK)
	{	wxString cstr=fd.GetPath();
		NodeP pn=new String(cstr);
		PairPP asp=argStack;
		argStack << pn;
		lsave(asp);
		argStack.pop_till(asp);
	}
}

void MainFrame::OnConsoleOpen(wxCommandEvent &event)
{
	wxFileDialog fd(this, "Choose a Console file","","",
		"Logo Files (*.txt)|*.txt|All Files (*.*)|*.*",
		wxOPEN|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK)
	{	wxString name=fd.GetPath();
		console->LoadFile(name);
	}
}

void MainFrame::OnConsoleSave(wxCommandEvent &event)
{
	wxFileDialog fd(this, "Choose a Console file","","",
		"Logo Files (*.txt)|*.txt|All Files (*.*)|*.*",
		wxSAVE|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK)
	{	wxString name=fd.GetPath();
		console->SaveFile(name);
	}
}

void MainFrame::OnConsoleAutoSave(wxCommandEvent &event)
{
	console->autosave = !console->autosave;
}

void MainFrame::OnPictureOpen(wxCommandEvent &event)
{
	wxFileDialog fd(this, "Choose a Picture file","","",
		"Picture Files (*.zip)|*.zip|All Files (*.*)|*.*",
		wxOPEN|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK)
		graph->recorder->load(fd.GetPath().c_str());
}

void MainFrame::OnPictureSave(wxCommandEvent &event)
{
	wxFileDialog fd(this, "Choose a Picture file","","",
		"Picture Files (*.zip)|*.zip|All Files (*.*)|*.*",
		wxSAVE|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK)
		graph->recorder->save(fd.GetPath().c_str());
}

void MainFrame::OnPictureTextOpen(wxCommandEvent &event)
{
	wxFileDialog fd(this, "Choose a Picture Text file","","",
		"Picture Files (*.dat)|*.dat|All Files (*.*)|*.*",
		wxOPEN|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK)
		graph->recorder->loadText(fd.GetPath().c_str());
}

void MainFrame::OnPictureTextSave(wxCommandEvent &event)
{
	wxFileDialog fd(this, "Choose a Picture Text file","","",
		"Picture Files (*.dat)|*.dat|All Files (*.*)|*.*",
		wxSAVE|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK)
		graph->recorder->saveText(fd.GetPath().c_str());
}

void MainFrame::OnPictureAutoSave(wxCommandEvent &event)
{
	graphAutosave = !graphAutosave;
}

void MainFrame::OnNodesOpen(wxCommandEvent &event)
{
	theApp->action=LOAD_NODES;
}

void MainFrame::OnNodesSave(wxCommandEvent &event)
{
	theApp->action=SAVE_NODES;
}

void MainFrame::OnNodesAutoSave(wxCommandEvent &event)
{
	nodesAutosave = !nodesAutosave;
}

void MainFrame::OnEditRepeat(wxCommandEvent &event)
{
	console->OnEditRepeat(event);
}

void MainFrame::OnEditReset(wxCommandEvent &event)
{
	console->OnEditReset(event);
}


class MyPrintout: public wxPrintout
{
public:
	MyPrintout(wxChar *title = "My printout"):
		wxPrintout(title) {}
	bool OnPrintPage(int page);
	bool HasPage(int page);
	bool OnBeginDocument(int startPage, int endPage);
	void GetPageInfo(int *minPage, int *maxPage, int *selPageFrom, int *selPageTo);
};

bool MyPrintout::OnPrintPage(int page)
{
	wxDC *dc = GetDC();
	if (dc)
	{	if (page == 1)
			graph->print(dc);
		return true;
	}
	else
		return false;
}

bool MyPrintout::OnBeginDocument(int startPage, int endPage)
{
	if (!wxPrintout::OnBeginDocument(startPage, endPage))
		return false;

	return true;
}

void MyPrintout::GetPageInfo(int *minPage, int *maxPage, 
									  int *selPageFrom, int *selPageTo)
{
	*minPage=1;
	*maxPage=1;
	*selPageFrom=1;
	*selPageTo  =1;
}

bool MyPrintout::HasPage(int pageNum)
{
	return pageNum==1;
}


void MainFrame::OnPrint(wxCommandEvent &event)
{
	wxPrintDialogData printDialogData(*printData);

	wxPrinter printer(&printDialogData);
	MyPrintout printout("aUCBLogo");
	if (!printer.Print(this, &printout, true))
	{	if (wxPrinter::GetLastError() == wxPRINTER_ERROR)
			wxMessageBox("There was a problem printing.\n"
				"Perhaps your current printer is not set correctly?", 
				"Printing", wxOK);
	}else
	{	*printData = printer.GetPrintDialogData().GetPrintData();
	}
}

void MainFrame::OnPrintPreview(wxCommandEvent& WXUNUSED(event))
{
	// Pass two printout objects: for preview, and possible printing.
	wxPrintDialogData printDialogData(*printData);
	wxPrintPreview *preview = new wxPrintPreview(
		new MyPrintout, new MyPrintout, &printDialogData);

	if (!preview->Ok())
	{	delete preview;
		wxMessageBox("There was a problem previewing.\n"
			"Perhaps your current printer is not set correctly?",
			"Previewing", wxOK);
		return;
	}

	wxPreviewFrame *frame = new wxPreviewFrame(
		preview, this, "Print Preview", 
		wxPoint(100, 100), wxSize(800, 1024));

	frame->Centre(wxBOTH);
	frame->Initialize();
	frame->Show(TRUE);
}

void MainFrame::OnPrintSetup(wxCommandEvent& WXUNUSED(event))
{
	wxPrintDialogData printDialogData(*printData);
	wxPrintDialog printerDialog(this, &printDialogData);

	printerDialog.GetPrintDialogData().SetSetupDialog(TRUE);
	printerDialog.ShowModal();

	*printData = printerDialog.GetPrintDialogData().GetPrintData();
}

void MainFrame::OnPageSetup(wxCommandEvent& WXUNUSED(event))
{

	*pageSetupData = *printData;

	wxPageSetupDialog pageSetupDialog(this, pageSetupData);
	pageSetupDialog.ShowModal();

	*printData = pageSetupDialog.GetPageSetupData().GetPrintData();
	*pageSetupData = pageSetupDialog.GetPageSetupData();
}
