#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <malloc.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/datstrm.h>
#include <wx/wfstream.h>
#include <wx/zipstrm.h>
#include <wx/ptr_scpd.h>
#include <wx/glcanvas.h>
#include <wx/textctrl.h>
#include <wx/frame.h>

#include <GL/glu.h>

#include <m_apm.h>

#include "Main.h"

#include "Segment.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "String.h"
#include "File.h"
#include "Pair.h"
#include "List.h"
#include "Procedur.h"
#include "Object.h"
#include "Primitiv.h"
#include "Table.h"
#include "Struct.h"
#include "Tree.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "FloatArray.h"
#include "Mem.h"

#include "PtrToString.h"
#include "StringToPtr.h"

#include "Error.h"
#include "Eval.h"
#include "Stack.h"
#include "Init.h"
#include "Print.h"

#include "Event.h"
#include "Color.h"
#include "Point.h"
#include "Pen.h"
#include "Font.h"
#include "Screen.h"
#include "Turtle.h"
#include "Recorder.h"
//#include "AVIGen.h"
#include "Graph.h"
#include "Console.h"
#include "MemoryWindow.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
	#define realloc(m, len) _realloc_dbg((m), (len), _NORMAL_BLOCK, __FILE__, __LINE__)

	#define ROOT_SIZE 1021
#else
	#define ROOT_SIZE 1021	// root_size should be a prime number to achieve a fine random distribution
#endif

TableP root;	// All logo objects are stored in the root table

NodeP Unbound=0;
IntP Tree::the_generation=0;

//#define GC_DEBUG 1
//#define SEG_SIZE_ONE
//#define GC_ON_NEWNODE

const int
#ifdef SEG_SIZE_ONE
	GC_::segment_size=1,	GC_::fifo_size=100,	// Xtreme GC debugging ;-)))
#else
	#ifdef _DEBUG
		GC_::segment_size=0x4000,	GC_::fifo_size=0x100000,
	#else
//		GC_::segment_size=0x100000,	GC_::fifo_size=0x100000,	// 1 MN
		GC_::segment_size=0x10000,	GC_::fifo_size=0x100000,	// 64 k Nodes
	#endif
#endif

	// GC heuristic parameters. These parameters can be modified to fine tune
	// the performance of the GC program. The values below are a good set of
	// default parameters that should work well for most data.

	GC_::min_free = 1+GC_::segment_size*4/10,	// if fewer than min_free nodes are free after a GC run,
												 // then a new segment of nodes is allocated
	GC_::generations = 4,
												// If a node of generation gen gets older than age_threshold[gen],
	gc_age_threshold[GC_::generations] =		// then it will be promoted to an older generation.
		//{0,0,0,0};	// to deactivate generational GC
		//{1,1,1,1};
		{2,2,2,2};
		//{4,4,4,4};	// old values
		//{3, 11, 37, 113};	// optimization experiment with prime numbers

NodeP generation[GC_::generations]={0};	// list of nodes in the same generation

//___________________________________________________________________

SegmentP Segment::list=0, Segment::start=0;

Segment::Segment()
{
	base = (Object*) GC_::calloc(GC_::segment_size, sizeof(Object));
}

Segment::~Segment()
{
	free((Node*)base);
}

bool Segment::add()
{
	SegmentP newseg = new Segment;
	if ((newseg!=0) && newseg->base) 
	{	newseg->baseu=unsigned(newseg->base);
		newseg->maxu=unsigned(&newseg->base[GC_::segment_size-1]);

		newseg->_tail = list;
		if (list)
			list->prev = newseg;
		newseg->prev = 0;
		list=newseg;
		if (!start)
			start=newseg;

		for (int p=GC_::segment_size-1;  p >=0;  p--)
		{	newseg->base[p]._next = GC_::free_node;
			GC_::free_node = &newseg->base[p];
			GC_::free_node->type = NTFREE;
			*intP((Node*)GC_::free_node) = Node::vmtNode;
		}
		GC_::mem_free += GC_::segment_size;
		if (memoryWindow)
		{	wxPaintEvent pev;
			memoryWindow->OnPaint(pev);
		}
		return true;
	}else
		return false;
}

#ifdef WIN32
extern void WaveInfoDestroy();
#endif

void Segment::freeAll()
{
	theApp->mainFrame->initialized=false;
#ifdef WIN32
	WaveInfoDestroy();
#endif
	SegmentP seg=list;
	while (seg)
	{	for (int p=0;  p < GC_::segment_size;  p++)
		{	NodeP n = &seg->base[p];
			if (n)
				n->free();
		}
		next(seg);
	}
	seg=list;
	while (seg)
	{	SegmentP s=seg;
		next(seg);
		delete s;
	}
	m_apm_free_all_mem();
}

bool is_nil_or_ok(volatile Node *nd)
{
	if (!nd) return true;
	unsigned ptr = unsigned(nd);

	for (SegmentP seg=Segment::list;  seg;  next(seg)) 
	{
#ifdef SEG_SIZE_ONE
		if (ptr == seg->baseu)
			return true;
#else
		if (   ptr >= seg->baseu
		&&   ptr <= seg->maxu
		&&  (ptr - seg->baseu) % sizeof(Object) == 0)
			return true;
#endif
	}
	return false;
}

//#pragma optimize("", off)

bool is_valid_pointer(volatile Node *nd)
{
	if (!nd) return false;
	unsigned ptr = unsigned(nd);

	for (SegmentP seg=Segment::list;  seg;  next(seg)) 
	{
#ifdef SEG_SIZE_ONE
		if (ptr == seg->baseu)
			return nd->type != NTFREE;
#else
		if (   ptr >= seg->baseu
		&&   ptr <= seg->maxu
		&&  (ptr - seg->baseu) % sizeof(Object) == 0 )
			return nd->type != NTFREE;
#endif
	}
	return false;
}

//#pragma optimize("", on)

#ifdef GC_DEBUG

	#define print_gc(text, var)			  \
		File::print(cout.file(), text, var);\
		if (dribbler)								 \
			fprintf(dribbler.file(), text, var); \
		examined=0;										\

	#define print_gc_object							 \
		File::print(stdout,							  \
			"%s p=%s v=%s mark=%2i gcNo.=%2i \n",	\
			obj, obj->proc, obj->val,					 \
			obj->mark_gc, gc_run_numero);				  \

	#define inc_examined	examined++
	#define freed_sofar_init	int freed_sofar=0
	#define freed_sofar_is(n)	freed_sofar=(n)
#else
	#define print_gc(text, var)
	#define print_gc_object
	#define inc_examined
	#define freed_sofar_init
	#define freed_sofar_is(n)
#endif

//___________________________________________________________________

FiFo::~FiFo() {if (allocated) free(base);}

FiFo &FiFo::operator<<(const NodeP val) 
{
	if (overflow()) return *this;
	*in=val;
	if (in==max) in=base; else in++;
	if (in==out) expand();
	return *this;
}

FiFo &FiFo::operator>>(NodeP &val) 
{
	val = *out;
	if (out==max) out=base; else out++;
	return *this;
}





void FiFo::expand()	// transfer is untested!!!!
{
	NodePP newbase = (NodeP*) GC_::calloc(1, sizeof(NodeP)*(size+expand_size));

	if (newbase==NodePP(0))	// no room to increase FiFo
	{	c_out << "; Warning: Not enough memory to run garbage collector." << endL;
		c_out << "; GC disabled - Save important data and exit!" << endL;
		_overflow=true;
	}
	else
	{	NodePP i=newbase;

		if (out)			// transfer old FiFo to new FiFo
		{	*this >> *i++;
			while (in!=out) 
				*this >> *i++;

		}				
		if (allocated) 
			free(base);

		allocated=true;
		size += expand_size;
		base=newbase;   max= &base[size-1];   in=i;   out= base;
	}
}

//___________________________________________________________________

Table::Table(int size_): 
	Node(TABLE), allocated(false), size(0), base(0)
{	self_resize(size_);
}

Table::~Table() 
{
	if (allocated) 
	{	::free((NodeP*)base); 
		allocated=false;
	}

}

void Table::free() 
{
	if (allocated) 
	{	::free((NodeP*)base); 
		allocated=false;



	}
}

void Table::inc_gc()
{
	if (!allocated) return;

	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			fifo << obj;
}

int Table::youngs() 
{
	if (!allocated) return 0;

	int youngs=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			youngs+=gmark(obj);
	return youngs;
}

int Table::youngsPtr() 
{
	if (!allocated) return 0;

	int youngs=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			youngs+=gmarkPtr(obj);
	return youngs;
}

void Table::check_oldyoungs()
{
	if (!allocated) return;

	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			check_oldyoung(obj);
}

void Table::setPtrs()
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
	{	*itemptr=(Object*)GC_::intToPtr->find(int(*itemptr));
		check_oldyoung(*itemptr);
		itemptr++;
	}
}
wxDataInputStream &Table::read(wxDataInputStream &istr) 
{	
	int siz;
	istr >> allocated >> siz;
	size=0;
	base = (ObjectP*) GC_::calloc(sizeof(ObjectP), siz+1);
	if (!base)
		return istr;
	size=siz;
	ObjectPP itemptr=base;
	for (int i=size;  i;  i--)
	{	istr >> (int&)*itemptr;
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &Table::write(wxDataOutputStream &ostr) 
{
	ostr << type << allocated << size;

	ObjectPP itemptr=base;
	for (int i=size;  i;  i--)
		ostr << GC_::ptrToInt->find(*itemptr++);
	return ostr;


}

StringP Table::string()
{
	const int maxlength=65536;

	char textbuf[maxlength];
	File::stringptr=textbuf;
	File::stringlen=maxlength-1;
	stringbuffer << NodeP(this);
	*File::stringptr='\0';
	return new String(textbuf);
}

int Table::hash(StringP str)
{
	if (!str)
		str=_Null_Word->name();
	unsigned int *i= (unsigned int*) str->ptr(),   index=0;

	for (char len=char(str->length());  len>0;  len-=sizeof(int))
	{	switch (len)
		{	case 0:	break;
			case 1:	index^= *i & 0x000000ff;	break;
			case 2:	index^= *i & 0x0000ffff;	break;

			case 3:	index^= *i & 0x00ffffff;	break;
			case 4:	index^= *i;	break;
			default:	index^= *i++ << 5;
		}
	};
	index %= size;
	return index;
}

ObjectP Table::find(StringP str) 
{
	ObjectP obj;
	for (obj=base[hash(str)];  obj;  next(obj))
		if (str->equal(obj->name()))
			break;

	return obj;
}

ObjectP Table::operator[](StringP str) {return find(str);}


ProcedureP Table::findProc(StringP filename, int linenr)
{
	int maxlinenr=-1;


	ObjectP obj=0;
	ObjectPP itemptr=base;
	LineP line=0;
	ProcedureP result=0;

	for (int i=size;  i;  i--)
		for (obj = *itemptr++;  obj;  next(obj))
			if (obj->proc() && obj->proc()!=Unbound && !obj->bit.permanent)
			{	line=LineP(obj->proc()->bodylist());
				if (line->filename==filename)
					if (line->number <= linenr && line->number > maxlinenr)
					{	maxlinenr=line->number;
						result=obj->proc();
					}
			}

	return result;
}

ObjectP Table::insert(StringP str)
{
	ObjectP &obj=base[hash(str)];
	obj = new Object(str, ProcedureP(Unbound), Unbound, NIL, obj);
	check_oldyoung(obj);
	return obj;
}

PairP Table::objectlist()
{
	PairP li=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			push(li, obj);

	li = li->mergeSort();
	return li;
}

void Table::self_resize(int newsize)
{
	ObjectPP newbase = (ObjectP*) GC_::calloc(1, sizeof(ObjectP)*(newsize+1));

	if (newbase==ObjectPP(0))
		c_out << "; Warning: Not enough memory to resize the table." << endL;
	else
	{	if (allocated) 
		{	PairP li=0;	li=objectlist();

			::free((NodeP*)base);	base=newbase;	size=newsize;
			while (li)
			{	int i=hash(ObjectP(li->first())->name());
				base[i] = ObjectP(li->first());
				next(li);
			}
		}else
		{	allocated=true;	base=newbase;	size=newsize;
		}
	}
}

void Table::mark()
{
	_marked = run_numero();
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))

			obj->mark();

	print_gc("%2i ", examined);
}

void Table::mark_all_but_objects()
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			obj->mark_all_but_objects();

	print_gc("%2i ", examined);
}

void Table::clean()
{
	_marked = run_numero();
	mark_all_but_objects();

	print_gc("Table::clean: ", 0);
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
	{	ObjectP lastobj=0;

		for (ObjectP obj = *itemptr;  obj;  next(obj)) 
		{//	print_gc_object;

			if ( obj->marked() != GC_::run_numero()
			&&  obj->proc_or_prim()==Unbound
			&&  obj->var()==Unbound
			&& !obj->plist()
			&& !obj->bit.permanent )
			{
				inc_examined;
				if (!lastobj)
					next(*itemptr);
				else
					lastobj->settail( obj->tail() );
			}else		// has a value, don't gctwa
				lastobj=obj;
		}
		itemptr++;
	}
	print_gc("%2i collected\n", examined);
}



PairP Table::contents(ContentsType contents_type, ObjectType obj_type) 
{

	PairP contentsList=0;
	ObjectPP itemptr=base;


	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
		{
			StringP name = obj->contents(contents_type, obj_type);
			if (name)
				push(contentsList, name);
		}
	contentsList = contentsList->mergeSort();
	return contentsList;
}

PairP Table::contentsNameVar() 
{
	PairP contentsList=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
		{
			StringP name = obj->name();
			if (name)
				push(contentsList, 
					new Pair(name, 
					new Pair(obj->var(), 0)));
		}
	contentsList = contentsList->mergeSort();
	return contentsList;
}

PairP Table::contents(ObjectType obj_type)
{
	PairP ret=0;
	push(ret, contents(PLISTS, obj_type));
	push(ret, contents(  VALS, obj_type));
	push(ret, contents( PROCS, obj_type));
	return ret;
}

NodeP Table::erase_all(bool procs, bool vals, bool plists) 
{
	int redef = _RedefP->var()->equal(_True);
	
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)

		for (ObjectP obj = *itemptr++;  obj;  next(obj))
		{
			if (procs 
			&& !obj->bit.proc_buried 
			&&  obj->proc_or_prim() != Unbound 
			&& !obj->proc_or_prim()->is.primitive)
				obj->setproc( ProcedureP(Unbound) );

			if (vals  &&  !obj->bit.val_buried)
				obj->setvar( Unbound );

			if (plists  &&  !obj->bit.plist_buried)
				obj->setplist(NIL);
		}
	return Unbound;
}

//___________________________________________________________________

Struct::Struct(PairP data_): 
	Node(STRUCT), size(0), base(0), _first(0)
{
	base = (ObjectP*) GC_::calloc(1, sizeof(ObjectP)*(data_->count()+1));
	if (!base)
		return;
	size=data_->count();
	ObjectP last=0;
	PairP np;
	for (np=data_;  np;  next(np))
	{	StringP f=np->first()->first()->string();
		if (String::caseIgnored)
			f=StringP(f->lower());
		ObjectP obj=find(f);
		if (obj)
		{	error(BAD_DATA, f);
			return;
		}
		obj=insert(f);

		NodeP t=np->first()->second()->intern(), n=t;
		if (n==_Int)
			n=new Int();
		else if (n==_Int16)
			n=new Int16();

		else if (n==_Int8)
			n=new Int8();
		else if (n==_UInt8)
			n=new UInt8();
		else if (n==_Float)
			n=new Float;
		else if (n==_Complex_)
			n=new Complexd();
		else if (n==_Word_)
			n=new String();
		else if (n==_IntArray_)
			n=new IntArray();
		else if (n==_Int16Array_)
			n=new Int16Array();
		else if (n==_FloatArray_)
			n=new FloatArray();
		else 
			error(BAD_DATA, t);
		
		if throwing return;
		
		if (np->first()->tail()->tail())
		{	NodeP third=np->first()->third(), n3=third;
		
			if (t==_Int)
				n3=n3->_int();

			else if (t==_Int16)
				n3=n3->toint16();
			else if (t==_Int8)
				n3=n3->toint8();
			else if (t==_UInt8)
				n3=n3->touint8();
			else if (t==_Float)
				n3=n3->_float();
			else if (t==_Word_)
			{	if (n3->numberP())
					n3=new String(int(third->getnumber()));
				else
					n3=n3->string();
			}
			else if (t==_IntArray_)
			{	if (n3->numberP())
					n3=new IntArray(int(third->getnumber()));
				else
				{	error(BAD_DATA, third);
					return;
				}
			}
			else if (t==_Int16Array_)
			{	if (n3->numberP())
					n3=new Int16Array(int(third->getnumber()));
				else
				{	error(BAD_DATA, third);
					return;
				}
			}
			else if (t==_FloatArray_)
			{	if (n3->numberP())
					n3=new FloatArray(int(third->getnumber()));
				else
				{	error(BAD_DATA, third);
					return;
				}
			}
			else
			{	error(BAD_DATA, third);
				return;
			}
			if (n->type!=n3->type && !(n->is.string && n3->is.string))
			{	error(BAD_DATA, third);
				return;
			}
			n=n3;
		}
		obj->setvar(n);
		if (last)
			last->setproc(ProcedureP(obj));
		else
			setfirst(obj);
		last=obj;
	}
	last->setproc(ProcedureP(0));
}

void Struct::free() 
{
	::free((NodeP*)base); 
}

void Struct::inc_gc()
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			fifo << obj;
}

int Struct::youngs() 
{
	int youngs=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			youngs+=gmark(obj);
	return youngs;
}

int Struct::youngsPtr() 
{
	int youngs=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			youngs+=gmarkPtr(obj);
	return youngs;
}

void Struct::check_oldyoungs()
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			check_oldyoung(obj);


}

void Struct::setPtrs()
{
	setfirst(ObjectP((Object*)GC_::intToPtr->find(int(_first))));
	ObjectPP itemptr=base;


	for (int i=size;  i;  i--)
	{	*itemptr=(Object*)GC_::intToPtr->find(int(*itemptr));
		check_oldyoung(*itemptr);
		itemptr++;
	}
}
wxDataInputStream &Struct::read(wxDataInputStream &istr) 
{	
	int siz;
	istr >> siz >> (int&)_first;

	base = (ObjectP*) GC_::calloc(sizeof(ObjectP), siz+1);
	if (!base)
		return istr;
	size=siz;
	ObjectPP itemptr=base;
	for (int i=size;  i;  i--)
	{	istr >> (int&)*itemptr;
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &Struct::write(wxDataOutputStream &ostr) 
{
	ostr << type << size 
		<< GC_::ptrToInt->find(first());

	ObjectPP itemptr=base;
	for (int i=size;  i;  i--)
		ostr << GC_::ptrToInt->find(*itemptr++);
	return ostr;
}

StringP Struct::string()
{
	const int maxlength=65536;
	char textbuf[maxlength];
	File::stringptr=textbuf;
	File::stringlen=maxlength;
	stringbuffer << NodeP(this);
	*File::stringptr='\0';
	return new String(textbuf);
}


int Struct::hash(StringP str)
{
	unsigned int *i= (unsigned int*) str->ptr(),   index=0;

	for (char len=char(str->length());  len>0;  len-=sizeof(int))
	{	switch (len)
		{	case 0:	break;
			case 1:	index^= *i & 0x000000ff;	break;
			case 2:	index^= *i & 0x0000ffff;	break;
			case 3:	index^= *i & 0x00ffffff;	break;
			case 4:	index^= *i;	break;
			default:	index^= *i++ << 5;
		}
	};
	index %= size;
	return index;

}

ObjectP Struct::find(StringP str) 
{
	ObjectP obj;
	for (obj=base[hash(str)];  obj;  next(obj))
		if (str->equal(obj->name()))
			break;

	return obj;
}

ObjectP Struct::operator[](StringP str) {return find(str);}

ObjectP Struct::insert(StringP str)
{
	ObjectP &obj=base[hash(str)];
	obj = new Object(str, ProcedureP(Unbound), Unbound, NIL, obj);
	check_oldyoung(obj);
	return obj;
}

void Struct::mark()
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			obj->mark();

	print_gc("%2i ", examined);
}

//___________________________________________________________________

bool GC_::initialized=false;
int volatile 
	GC_::is_running=false, 
	GC_::disabled=false;

int 
	GC_::mem_free=0,	// how many free nodes are available in the allocated segments
	GC_::mem_peak=0,	// peak memory (when most nodes were newed while least freed)
	GC_::mem_used=0,	// newed-freed nodes

	GC_::mark_gen=0,  GC_::max_gen=0,
	GC_::_run_numero=0,
	GC_::examined=0;

GC_::Interrupt GC_::interrupt = GC_::no_interrupt;

NodeP
	GC_::free_node=0,	// list of free nodes
	GC_::oldyoungs=0,	// list of nodes that point to younger nodes
	GC_::SpareNode=0;	// here goes newnode if no free_node is available

PairP GC_::reserve_tank=0;	// list of nodes not in any generation
NodeP *GC_::bottom_stack=0;	// points to the bottom of the C program stack. See begin of main!

FiFo GC_::fifo(GC_::fifo_size);

PtrToInt *GC_::ptrToInt;
IntToPtr *GC_::intToPtr;

PtrToString *GC_::primitivesPtrToString;
StringToPtr *GC_::primitivesStringToPtr;

void GC_::init()
{
	memset(generation, 0, sizeof(generation));
	Segment::add();
	reserve_tank_fill();
	oldyoungs = Unbound = new Node();
	SpareNode=new Node(SPARE_NODE);
	root=new Table(ROOT_SIZE);
}

void GC_::reserve_tank_fill() 
{
	PairP p=0;
	for (int i=50;  i;  i--)		// make pairs not in any generation
	{	if (!free_node) break;

		PairP newnd = PairP(free_node);   
		free_node = newnd->next_();	// remove newnd from the free_node list

		inc_mem_nodes();
		newnd->setnext(0);
		newnd->setoldyoung_next(0);

		newnd->settype_PAIR();
		newnd->_tail=p;	p=newnd;	// insert newnd into reserve tank list
		newnd->setfirst(0);
	}
	reserve_tank=p;
}

void GC_::reserve_tank_check() 
{
	if (!reserve_tank) 
		reserve_tank_fill();
}

void GC_::reserve_tank_use() 
{
	PairP nd=reserve_tank;	reserve_tank=0;
	while(nd) 
	{	nd->setnext(free_node);	free_node=nd;	// insert nd into the free_node list
		nd->settype_NTFREE();
		mem_free++;  mem_used--;
		next(nd);
	}
}


void GC_::mark()	// Iterative mark procedure
{
	if (!this) return;
	if (gen > mark_gen) return;   // too old
	if (marked()==run_numero()) return;  // already marked
	
	fifo << NodeP((Node*)this);

	while (fifo)
	{	NodeP nd;	fifo >> nd;
		if (nd)
			if (nd->gen <= mark_gen)
				if (nd->marked() != run_numero())
				{	nd->_marked = run_numero();		inc_examined;
					nd->inc_gc();
				}
	}
}

void GC_::markPtr()	// Iterative mark procedure
{
	if (!this) return;
	if (gen > mark_gen) return;   // too old
	if (marked()==run_numero()) return;  // already marked
	
	fifo << NodeP((Node*)this);



	while (fifo)
	{	NodeP nd;	fifo >> nd;

		if (nd)
			if (nd->marked() != run_numero())
			{	nd->_marked = run_numero();		inc_examined;
				ptrToInt->insertNode(nd);
				nd->inc_gc();
			}
	}
}

int GC_::gmark(NodeP child)
{
	if (child)
		if (child->gen < gen)
		{	child->mark();
			return 1;
		}
	return 0;
}

int GC_::gmarkPtr(NodeP child)
{
	if (child)
		if (child->gen < gen)

		{	child->markPtr();
			return 1;
		}
	return 0;
}

void _cdecl GC_::mark_many(NodeP arg1, ...)
{
	va_list args;
	va_start(args, arg1);

	NodeP arg=arg1;
	while (arg!=_ListEnd)
	{	arg->mark();
		arg = va_arg(args, NodeP);
	}
	va_end(args);
	_ListEnd->mark();
	print_gc(";Globals %2i + ", examined);
}

void _cdecl GC_::mark_manyPtr(NodeP arg1, ...)
{
	va_list args;
	va_start(args, arg1);

	NodeP arg=arg1;
	while (arg!=_ListEnd)
	{	arg->markPtr();
		arg = va_arg(args, NodeP);
	}
	va_end(args);
	_ListEnd->markPtr();
	print_gc(";Globals %2i + ", examined);

}

void GC_::mark_C_stack()	// Check C Stack for Node pointers and mark valid ones
{
	NodeP  top=0,				// top is just now allocated on the Stack, 
			*top_stack = &top;  //top_stack points to it

	if ((unsigned int)top_stack < (unsigned int)bottom_stack)	// check direction stack grows
	{	for (NodeP *sp=top_stack;  (unsigned int)sp <= (unsigned int)bottom_stack;  sp++)
			if (is_valid_pointer(*sp)) 
				(*sp)->mark();
	}else
	{	for (NodeP *sp=top_stack;  (unsigned int)sp >= (unsigned int)bottom_stack;  sp--)

			if (is_valid_pointer(*sp)) 
				(*sp)->mark();
	}
	print_gc("C stack %2i + ", examined);
}

void GC_::mark_C_stackPtr()	// Check C Stack for Node pointers and mark valid ones
{
	NodeP  top=0,				// top is just now allocated on the Stack, 
			*top_stack = &top;  //top_stack points to it

	if ((unsigned int)top_stack < (unsigned int)bottom_stack)	// check direction stack grows
	{	for (NodeP *sp=top_stack;  (unsigned int)sp <= (unsigned int)bottom_stack;  sp++)
			if (is_valid_pointer(*sp)) 
				(*sp)->markPtr();
	}else
	{	for (NodeP *sp=top_stack;  (unsigned int)sp >= (unsigned int)bottom_stack;  sp--)
			if (is_valid_pointer(*sp)) 
				(*sp)->markPtr();
	}
	print_gc("C stack %2i + ", examined);
}

NodePP GC_::mark_oldyoung(NodePP oldyoung)	// Mark/traverse pointers to younger generations only
{
	Node &nd = **oldyoung;

	if (nd.gen <= mark_gen)		// too young
		return &(nd.oldyoung_next);

	if (!nd.youngs())		// nd no longer points to a node of a younger generation
	{
		*oldyoung = nd.oldyoung_next;  nd.oldyoung_next=0;	// remove nd from the list of oldyoungs
		return oldyoung;
	}
	return &(nd.oldyoung_next);
}

NodePP GC_::mark_oldyoungPtr(NodePP oldyoung)	// Mark/traverse pointers to younger generations only
{
	Node &nd = **oldyoung;

	if (!nd.youngsPtr())	// nd no longer points to a node of a younger generation
	{
		*oldyoung = nd.oldyoung_next;  nd.oldyoung_next=0;	// remove nd from the list of oldyoungs
		return oldyoung;
	}
	return &(nd.oldyoung_next);
}


void GC_::mark_oldyoungs()	// mark pointers from old generations to young
{
	NodePP oldyoung = &oldyoungs;  

	while (*oldyoung != Unbound)
		oldyoung = mark_oldyoung(oldyoung);

	print_gc("oldyoungs %2i marked ", examined);
}

void GC_::mark_oldyoungsPtr()	// mark pointers from old generations to young
{
	NodePP oldyoung = &oldyoungs;  

	while (*oldyoung != Unbound)
		oldyoung = mark_oldyoungPtr(oldyoung);

	print_gc("oldyoungs %2i marked ", examined);
}

void _cdecl GC_::find_many(intP manyPtrNrs, NodePP arg1, ...)
{
	va_list args;
	va_start(args, arg1);

	NodePP arg=arg1;
	int i=0;
	while (*arg!=_ListEnd)
	{	*arg = (Node*)intToPtr->find(manyPtrNrs[i]);
		 arg = va_arg(args, NodePP);
		i++;
	}
	va_end(args);
	_ListEnd = (Object*)intToPtr->find(manyPtrNrs[i]);
	print_gc(";Globals %2i + ", examined);
}

//===================================================================
#define MANY_NODES(x)		\
	x(GC_::SpareNode)			\
	x(_Not_Enough_node)		\
	x(Tree::the_generation) \
	x(File::line) x(File::currentLine) \
	x(throw_node)  x(err_mesg)		\
	x(fun) x(ufun) x(this_line)	\
	x(Evaluator::last_ufun)	\
	x(Evaluator::last_call)	\
	x(Evaluator::last_line)	\
	x(Evaluator::loopcond)	\
	x(Evaluator::forname)	\
	x(didnt_output_name) x(output_node) \
	x(File::Null  )	x(File::Stdin   ) x(File::Stdout) \
	x(File::CIn   )	x(File::COut    ) \
	x(File::Reader)	x(File::Writer  ) \
	x(File::Loader)	x(File::Dribbler) \
	x(File::Stringbuffer) x(File::list) \
	x(File::TextCtrl) x(File::textControl) \
	x(Line::filenames)	\
	x(EventChar   ::function)	\
	x(EventKeyDown::function)	\
	x(EventKeyUp  ::function)	\
	x(EventTextMouseLeftDown	::function)	\
	x(EventTextMouseRightDown	::function)	\
	x(EventTextMouseMiddleDown	::function)	\
	x(EventTextMouseLeftUp		::function)	\
	x(EventTextMouseRightUp		::function)	\
	x(EventTextMouseMiddleUp	::function)	\
	x(EventTextMouseLeftDClick	::function)	\
	x(EventTextMouseRightDClick::function)	\
	x(EventTextMouseMiddleDClick::function)\
	x(EventTextMouseMotion		::function)	\
	x(EventMouseLeftDown		::function)	\
	x(EventMouseRightDown	::function)	\
	x(EventMouseMiddleDown	::function)	\
	x(EventMouseLeftUp		::function)	\
	x(EventMouseRightUp		::function)	\
	x(EventMouseMiddleUp		::function)	\
	x(EventMouseLeftDClick	::function)	\
	x(EventMouseRightDClick	::function)	\
	x(EventMouseMiddleDClick::function)	\
	x(EventMouseMotion		::function)	\
	x(theApp->mainFrame->graphNode)	\
	x(_ListEnd)
// ^^^^^^^^^^^======== COUNT them!!!! Important for GC_::load !!!
#define HOW_MANY_NODES 56

#define LIST_MANY(x) x,
#define LIST_MANY_ADDRESS(x) (NodePP)&x,

void GC_::mark_phase(bool clean_root)	// mark all accessable node pointers of generations<=mark_gen
{
mark_phase:
	_run_numero++;

	mark_many(Unbound, R_Texture::textures, 
		MANY_NODES(LIST_MANY) 0);
	Evaluator::     stack.mark();
	Evaluator::  varStack.mark();
	Evaluator::errorStack.mark();
	             argStack.mark();
	print_gc("NodePStacks %2i + ", examined);

	mark_C_stack();

	mark_oldyoungs();		// mark pointers from old generations to young

	if (fifo.overflow()) return;

	if (root)
		if (clean_root) 
		{	root->clean();	clean_root=false;
			goto mark_phase;
		}
		else
			root->mark();
}

void GC_::mark_phasePtr()	// mark all accessable node pointers of generations<=mark_gen
{
	_run_numero++;
	mark_gen=max_gen;

	mark_manyPtr(Unbound, MANY_NODES(LIST_MANY) 0);
	Evaluator::     stack.markPtr();
	Evaluator::  varStack.markPtr();
	Evaluator::errorStack.markPtr();
	             argStack.markPtr();
	print_gc("NodePStacks %2i + ", examined);

	mark_C_stackPtr();
	mark_oldyoungsPtr();	// mark pointers from old generations to young
	if (fifo.overflow()) 
		return;

	if (root)
		root->markPtr();
}

void GC_::check_oldyoung(NodeP new_child) 
{
	if (new_child)
		if (new_child->gen < gen  &&  !oldyoung_next)
		{
			oldyoung_next = oldyoungs;	
			oldyoungs=(Node*)this;	// insert this into the list of oldyoung pointers
		}
}


int GC_::sweep_phase(int gen)	// promote marked to next gen, free the rest
{
	freed_sofar_init;
	int num_freed=0;

	for (int i=gen;  i >=0;  i--)
	{
		NodePP lastnd = &generation[i];

		for (NodeP nd= *lastnd;  nd;  nd = *lastnd) 
		{
			if (nd->marked() != run_numero()) 			// free the nd
			{	num_freed++,  mem_used--;   mem_free++;

				*lastnd = nd->_next;					   // remove nd from its generation
				nd->_next=free_node;   free_node=nd;	// insert nd into the free_node list

				if (nd->oldyoung_next) 
				{	NodePP prev = &oldyoungs;
					while (*prev != nd)					// search nd in the list of oldyoungs (expensive!)
						prev = &((*prev)->oldyoung_next);

					*prev = nd->oldyoung_next;   nd->oldyoung_next=0;	// remove nd from the list of oldyoungs
				}
				nd->free();
				nd->settype_NTFREE();
				if (memoryWindow)
					memoryWindow->updateNode(nd);
			}else
			{	nd->age--;
				if (nd->age==0  &&  i < generations-1)		// promote to next gen
				{
					*lastnd = nd->_next;									// remove nd form its gen


					nd->_next= generation[i+1];    generation[i+1] = nd;	// insert nd into next gen

					nd->gen = i+1;
					nd->age = gc_age_threshold[i+1];

					nd->check_oldyoungs();

					if (max_gen==i) 
						max_gen++;
				}else								// keep in this gen
					lastnd = &(nd->_next);
			}
		}
		print_gc("%i+", num_freed - freed_sofar);
		freed_sofar_is(num_freed);
	}
	print_gc("=%2i freed\n", num_freed);
	return num_freed;
}

	
void GC_::gc(bool full)	// Garbage Collector
{
	if (!initialized)
		return;

	static int next_gen=0;

	if (disabled)
	{	if (!free_node)
			if (!Segment::add())
				error(OUT_OF_MEM_UNREC, 0);
		return;
	}
	if (fifo.overflow()) 
	{	if (!Segment::add() && !free_node)

			error(OUT_OF_MEM_UNREC, 0);
		return;
	}
	checkThrowing();

	int gen =		// deepest generation to garbage collect 
		mark_gen = full ?  max_gen
							 :  next_gen;

	bool clean_all =		// garbage collect root objects
		(gen==max_gen  &&  max_gen > 1) || full;

	mark_phase(clean_all);

#ifdef _DEBUG
	int num_freed =
#endif
		sweep_phase(gen);

	if (mem_free > min_free)
		next_gen=0;			// check only nursery gen next time
	else if (gen < max_gen)
		next_gen=gen+1;	// need to check more gens next time
	else	// all gens have been collected; allocate more memory
	{	if (!Segment::add()  &&  mem_free < 50  &&  !free_node)
			error(OUT_OF_MEM_UNREC, 0);
		next_gen=0;	
	}
	if (theApp && !theApp->heapReserve)
	{	theApp->heapReserve=new char[0x110000];
		if (theApp->heapReserve)	// well, at least *try* to fix it
		{	delete theApp->heapReserve;
			theApp->heapReserve=new char[0x100000];
		}
	}
}

void GC_::inc_mem_nodes()
{
	mem_used++;
	mem_free--;
	if (mem_used > mem_peak) 
		mem_peak=mem_used;
}

void GC_::run(bool full) 
{
	interrupt=no_interrupt;
	is_running++;
	gc(full);
	is_running=0;

	if (interrupt) 
	{	if (interrupt==pause) 
			logo_pause(0);
		else 
			logo_stop(0);
	}
}

void *GC_::malloc_GC
{
	void *mem=::malloc(size);
	if (!mem)
	{	run(true);				 // collect garbage 
		mem=::malloc(size);	// and try again
		if (!mem) 
			error(OUT_OF_MEM, 0);
	}
	return mem;
}

void *GC_::calloc_GC
{
	void *mem=::calloc(num, size);
	if (!mem)
	{	run(true);						 // collect garbage 
		mem=::calloc(num, size);	// and try again
		if (!mem) 
			error(OUT_OF_MEM, 0);
	}
	return mem;
}

void *GC_::realloc_GC
{
	void *mem=::realloc(memblock, size);
	if (!mem)
	{	run(true);								 // collect garbage 
		mem=::realloc(memblock, size);	// and try again
		if (!mem) 
			error(OUT_OF_MEM, 0);
	}
	return mem;
}

NodeP GC_::newnode() 
{
#ifndef GC_ON_NEWNODE
	while (!free_node && !throwing) 
#endif
		run(false);

	if (!free_node)
	{	if (SpareNode->oldyoung_next) 
		{	NodePP prev = &oldyoungs;
			while (*prev != SpareNode)	// search nd in the list of oldyoungs
				prev = &((*prev)->oldyoung_next);

			*prev = SpareNode->oldyoung_next;   
			SpareNode->oldyoung_next=0;	// remove nd from the list of oldyoungs
		}
		SpareNode->free();
		return SpareNode;
	}
	NodeP newnd=free_node;	free_node=newnd->_next;	// remove newnd from the free_node list

	newnd->_next= generation[0];	generation[0]=newnd;	// insert newnd into the nursery generation

	newnd->oldyoung_next=0;
	newnd->_marked=0; 
	newnd->gen=0;
	newnd->age = gc_age_threshold[0];

	inc_mem_nodes();
	return newnd;
}
//___________________________________________________________________

wxDataInputStream &operator>>(wxDataInputStream &istr, Node &i)
{	return i.read(istr);
}
wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Node &i)
{	return i.write(ostr);
}
//___________________________________________________________________

#ifdef WIN32
	#pragma optimize("", off)
#endif
void GC_::set_C_Stack(int stacksize, int *stack, int bi)
{
	int top=0,					 // top is just now allocated on the Stack, 
		*top_stack = &top;  //top_stack points to it
	int b;

	int *sp=(int*)bottom_stack;
	if ((unsigned int)bottom_stack < (unsigned int)top_stack)	// check direction stack grows
	{	if (top_stack-(int*)bottom_stack <= stacksize+64)
			set_C_Stack(stacksize, stack, bi);
		for (int i=0; i < stacksize; i++)
			*sp++=stack[i];
		delete stack;
	}
	else
	{	int real_stacksize=(int*)bottom_stack-top_stack;
		if (real_stacksize <= stacksize+64)
			set_C_Stack(stacksize, stack, bi);
		b=int(bottom_stack-bi);
		sp=(int*)bottom_stack;
		for (int i=0; i < stacksize; i++)
			*sp--=stack[i];
		delete stack;
	}
#ifdef WIN32
	_asm mov esp,b
	_asm pop ebp
	_asm pop eax
	_asm mov esp,ebp
	_asm pop ebp
	_asm ret;
#else
	asm("mov %0,%%esp": :"g" (b));
	asm("pop %ebp");
	asm("pop %eax");
	asm("mov %ebp,%esp");
	asm("pop %ebp");
	asm("ret");
#endif
}
#ifdef WIN32
	#pragma optimize("", on)
#endif


wxDECLARE_SCOPED_PTR(wxZipEntry, wxZipEntryPtr);

void GC_::load(charP name)
{
	root->erase_all(true, true, true);
	run(true);

	wxFFileInputStream *input=new wxFFileInputStream(name);
	if (!input->Ok())
		return;
	
	wxZipInputStream *zip=new wxZipInputStream(*input);
	wxZipEntryPtr *entry=new wxZipEntryPtr;

	if (entry->reset(zip->GetNextEntry()), entry->get() != 0)
	{
		wxDataInputStream *istr=new wxDataInputStream(*zip);

		int i, tableSize, rootPtrNr, manyPtrNrs[HOW_MANY_NODES];
		*istr >> tableSize >> rootPtrNr;
		for (i=0; i < HOW_MANY_NODES; i++)
			*istr >> manyPtrNrs[i];

		argStack.read(*istr);
		Evaluator::stack.read(*istr);
		Evaluator::intStack.read(*istr);
		Evaluator::floatStack.read(*istr);
		Evaluator::toDoStack.read(*istr);
		Evaluator::varStack.read(*istr);
		Evaluator::errorStack.read(*istr);
		Evaluator::argPStack.read(*istr);
		Evaluator::varPStack.read(*istr);
	
		intToPtr=new IntToPtr(tableSize);
		RECORDER.intToPtr=new IntToPtr(Recorder::tableSize);
		primitivesStringToPtr=new StringToPtr(
			ROOT_SIZE, Primitive::primitives);

		int stacksize;
		*istr >> stacksize;
		int *intstack=new int[stacksize], 
			*argpstack=new int[stacksize], 
//			*varpstack=new int[stacksize], 
				 *stack=new int[stacksize],
		*consolestack=new int[stacksize],
		  *stackstack=new int[stacksize];
		for (i=0; i<stacksize; i++)
			*istr >> intstack[i];
		for (i=0; i<stacksize; i++)
			*istr >> argpstack[i];
//		for (i=0; i<stacksize; i++)
//			*istr >> varpstack[i];
		for (i=0; i<stacksize; i++)
			*istr >> consolestack[i];
		int bi;
		*istr >> bi;
		for (i=0; i<stacksize; i++)
			*istr >> stack[i];
		for (i=0; i<stacksize; i++)
			*istr >> stackstack[i];
		int v;
		*istr >> v;
		Evaluator::var = v;
		*istr
			>> Evaluator::dont_fix_ift
			>> Evaluator::ift_iff_flag
			>> Evaluator::repcount
			>> Evaluator::user_repcount	// counters for repeat & while
			>> Evaluator::ifori
			>> Evaluator::iforend
			>> Evaluator::iforstep	// ...and for for
			>> Evaluator::trace_level	// indentation level when tracing
			>> Evaluator::fori
			>> Evaluator::forend
			>> Evaluator::forstep	// counters for for
			>> (int&)Evaluator::tailcall
			>> (int&)Evaluator::val_status;
		bool insertMode;
		*istr	>> insertMode >> console->charMode;
		if (!insertMode)
			console->toggleInsertMode();
		*istr >> graph->doubleBuffer
			>> Int::genSymNr;

/*		bool avi;	// shall append to Video Stream, but crashes!
		*istr >> avi;
		if (avi)		// if there was a AVI Video Stream open it again
		{	wxString *filename=new wxString;
			int rate, frame;
			*istr >> *filename >> rate >> frame;
			if (graph->aviGen)
			{	graph->aviGen->ReleaseEngine();
				delete graph->aviGen;
				graph->aviGen=0;
			}
			graph->aviGen=new AVIGen(charP(filename->c_str()), 
											 graph, rate);
			int err=graph->aviGen->InitEngine();//true, frame);
			//delete filename;
		}

*/
		intToPtr->readNodes(*istr);

		root=(Table*)intToPtr->find(rootPtrNr);
		find_many(manyPtrNrs, MANY_NODES(LIST_MANY_ADDRESS) 0);
		argStack.setPtrs();
		Evaluator::stack.setPtrs();
		Evaluator::varStack.setPtrs();
		Evaluator::errorStack.setPtrs();

#define FIND_AND_SET_PERMANENT(x) \
		if (_##x) _##x=root->find(_##x->name());
#define FIND_AND_SET_PERMANENT_TYPE(x) \
		if (type_##x) type_##x=root->find(type_##x->name());

		Permanent_Objects(FIND_AND_SET_PERMANENT)
		Permanent_Objects2(FIND_AND_SET_PERMANENT)
		Node_Type_Strings(FIND_AND_SET_PERMANENT_TYPE, UNUSED)

		for (i=0; i<stacksize; i++)
		{	if (intstack[i])
				stack[i]=int(intToPtr->find(intstack[i]));
			if (argpstack[i])
				stack[i]=int(argpstack[i]+argStack.base);
//			if (varpstack[i])
//				stack[i]=int(varpstack[i]+Evaluator::varStack.base);
			if (consolestack[i])
				stack[i]=int(console);
			if (stackstack[i])
				stack[i]=int(bottom_stack-stackstack[i]);
		}
		delete intstack;
		delete argpstack;
//		delete varpstack;
		delete consolestack;
		delete stackstack;

		delete primitivesStringToPtr;
		delete RECORDER.intToPtr;
		delete intToPtr;
		delete istr;
		delete entry;
		delete zip;
		delete input;
		Event::eatAll();
		set_C_Stack(stacksize, stack, bi);
	}
}

void GC_::save(charP name)
{
	NodeP  top=0,				// top is just now allocated on the Stack, 
			*top_stack = &top;  //top_stack points to it

	run(true);
	ptrToInt=new PtrToInt(mem_used);
	primitivesPtrToString=new PtrToString(
		ROOT_SIZE, Primitive::primitives);

	mark_phasePtr();

	wxFFileOutputStream output(name);
	wxZipOutputStream zip(output);
	zip.PutNextEntry(_T("nodes.dat"));
	wxDataOutputStream ostr(zip);

	Recorder::ptrToInt=new PtrToInt(Recorder::tableSize);

#define FIND_AND_WRITE_PTR(x) << ptrToInt->find(x)

	ostr << mem_used 
		<< ptrToInt->find(root)
		MANY_NODES(FIND_AND_WRITE_PTR);

	argStack.write(ostr);
	Evaluator::stack.write(ostr);
	Evaluator::intStack.write(ostr);
	Evaluator::floatStack.write(ostr);
	Evaluator::toDoStack.write(ostr);
	Evaluator::varStack.write(ostr);

	Evaluator::errorStack.write(ostr);
	Evaluator::argPStack.write(ostr);
	Evaluator::varPStack.write(ostr);

	int stacksize;
	if ((unsigned int)bottom_stack < (unsigned int)top_stack)	// check direction stack grows
	{	stacksize=top_stack-bottom_stack+1;
		ostr << stacksize;
		int *intstack=new int[stacksize];
		int i=0;
		NodeP *sp;
		for (sp=bottom_stack;  (unsigned int)sp <= (unsigned int)top_stack;  sp++, i++)
			if (is_valid_pointer(*sp)) 
				intstack[i]=ptrToInt->find(*sp);
			else
				intstack[i]=0;
		for (i=0; i<stacksize; i++)
			ostr << intstack[i];
		delete intstack;
		for (sp=bottom_stack;  (unsigned int)sp <= (unsigned int)top_stack;  sp++)
			ostr << int(*sp);
	}else
	{	stacksize=bottom_stack-top_stack+1;
		ostr << stacksize;
		int *intstack=new int[stacksize],
			*argpstack=new int[stacksize],
//			*varpstack=new int[stacksize],
			*consolestack=new int[stacksize],
			*stackstack=new int[stacksize];
		int i=0;
		NodeP *sp;
		for (sp=bottom_stack;  (unsigned int)sp >= (unsigned int)top_stack;  sp--, i++)
		{	if (is_valid_pointer(*sp)) 
				intstack[i]=ptrToInt->find(*sp);
			else
				intstack[i]=0;

			unsigned ptr = unsigned(*sp);
			if ( ptr >= unsigned(argStack.base )
			&&   ptr <= unsigned(argStack.index)
			&&  (ptr -  unsigned(argStack.base )) % sizeof(PairPP) == 0 )
				argpstack[i]=NodePP((NodeP*)&*(*sp))-argStack.base;
			else
				argpstack[i]=0;

/*			if ( ptr >= unsigned(Evaluator::varStack.base )
			&&   ptr <= unsigned(Evaluator::varStack.index)
			&&  (ptr -  unsigned(Evaluator::varStack.base )) % sizeof(PairPP) == 0 )
				varpstack[i]=VarP((Var*)&*(*sp))-Evaluator::varStack.base;
			else
				varpstack[i]=0;
*/
			if (int(*sp)==int(console))
				consolestack[i]=1;
			else
				consolestack[i]=0;
		}

		for (i=0; i<stacksize; i++)
			ostr << intstack[i];
		for (i=0; i<stacksize; i++)
			ostr << argpstack[i];
/*		for (i=0; i<stacksize; i++)
			ostr << varpstack[i];
*/		for (i=0; i<stacksize; i++)
			ostr << consolestack[i];
		delete argpstack;
//		delete varpstack;
		delete consolestack;
		i=0;
		for (sp=bottom_stack;  (unsigned int)sp >= (unsigned int)top_stack;  sp--, i++)
			intstack[i]=int(*sp);

		int b;
#ifdef WIN32
		_asm mov b,ebp
#else
		asm("mov %%ebp,%0":"=g" (b));
#endif
		int bi=intP(bottom_stack)-intP(b);
		ostr << bi;
		int *p=intP(intstack[bi]);
		for (i=0; i<stacksize; i++)
			ostr << intstack[i];

		for (i=0; i<stacksize; i++)
		{	p=intP(intstack[i]);
			unsigned n=intP(bottom_stack)-p;
			if (n<unsigned(stacksize) || n>unsigned(-stacksize))
				stackstack[i]=n;
			else
				stackstack[i]=0;
		}
		for (i=0; i<stacksize; i++)
			ostr << stackstack[i];

		delete stackstack;
		delete intstack;
	}
	int v=Evaluator::var;
	ostr << v 
		<< Evaluator::dont_fix_ift
		<< Evaluator::ift_iff_flag
		<< Evaluator::repcount
		<< Evaluator::user_repcount	// counters for repeat & while
		<< Evaluator::ifori
		<< Evaluator::iforend
		<< Evaluator::iforstep	// ...and for for
		<< Evaluator::trace_level	// indentation level when tracing
		<< Evaluator::fori
		<< Evaluator::forend
		<< Evaluator::forstep	// counters for for
		<< Evaluator::tailcall
		<< Evaluator::val_status
		<< console->insertMode
		<< console->charMode
		<< graph->doubleBuffer
		<< Int::genSymNr;

/*	if (graph->aviGen)
	{	ostr.Write8(true); 
		ostr
			<< graph->aviGen->filename 
			<< graph->aviGen->rate 
			<< graph->aviGen->frame;
		graph->aviGen->ReleaseEngine();
		delete graph->aviGen;
		graph->aviGen=0;
	}
	else
		ostr << false;
*/	
	ptrToInt->writeNodes(ostr);

	delete RECORDER.ptrToInt;
	delete primitivesPtrToString;
	delete ptrToInt;
}
//___________________________________________________________________

NodeP lGC(PairPP arg) 
{
	GC_::run(exist(1));
	return Unbound;
}

NodeP lNodes(PairPP) 
{
	GC_::run(true);	// get real in-use figures

	int peak=GC_::mem_peak,
		used=GC_::mem_used;
	GC_::mem_peak=GC_::mem_used;

	return new Pair
	(	new Int(used),
		new Pair
		(	new Int(peak), 
			0
		)
	);
}

NodeP lloadNodes(PairPP arg)
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	GC_::load(name->ptr());
	return Unbound;
}

NodeP lsaveNodes(PairPP arg)
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	GC_::save(name->ptr());
	return Unbound;
}

NodeP lshrinkStacks(PairPP)
{
	Evaluator::intStack.shrink();
	Evaluator::varPStack.shrink();
	Evaluator::argPStack.shrink();
	Evaluator::floatStack.shrink();
	Evaluator::stack.shrink();
	argStack.shrink();
	Evaluator::varStack.shrink();
	Evaluator::toDoStack.shrink();
	Evaluator::errorStack.shrink();
	return Unbound;
}

NodeP lsetStackNoisy(PairPP arg)
{
	bool noisy=arg[1]->a_bool(1)!=0;
	if throwing return Unbound;
	IntStack::noisy=noisy;
	return Unbound;
}
