//___________________________________________________________________
//
//                       Class MemoryWindow
//
//                           Definitons
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/frame.h>
#include <wx/dcclient.h>
#include <wx/dcmemory.h>
#include <wx/textctrl.h>
#include <wx/tooltip.h>
#include <wx/tipwin.h>
#include <wx/scrolwin.h>
#include <wx/app.h>
#include <wx/menu.h>

#include "Main.h"
#include "Node.h"
#include "Object.h"
#include "UInt8.h"
#include "String.h"
#include "Pair.h"
#include "File.h"
#include "Color.h"
#include "Segment.h"
#include "Console.h"
#include "MainFrame.h"
#include "MemoryWindow.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(MemoryFrame, wxFrame)
//	EVT_CLOSE	(MemoryFrame::OnClose		)
END_EVENT_TABLE()

MemoryFrame::MemoryFrame(wxWindow* parent_, wxWindowID id, 
						const wxString& title, 
						const wxPoint& pos, 
						const wxSize& size, 
						long style, 
						const wxString& name):
wxFrame(parent_, id, title, pos, size, style, name)
{
	memoryWindow=new MemoryWindow(this, -1, "Memory", 
		wxPoint(0, 0), wxSize(size.x-20,size.y),
		wxVSCROLL | wxALWAYS_SHOW_SB, "Memory");
	::memoryWindow=memoryWindow;
}

bool MemoryFrame::Destroy()
{
	GetPosition(&theApp->memposx , &theApp->memposy );
	GetSize    (&theApp->memsizex, &theApp->memsizey);
	if (memoryWindow->tip)
		memoryWindow->tip->Destroy();
	return wxFrame::Destroy();
}

BEGIN_EVENT_TABLE(MemoryWindow, wxScrolledWindow)
	EVT_ERASE_BACKGROUND(MemoryWindow::OnErase) 
	EVT_SIZE		(MemoryWindow::OnSize		)
	EVT_PAINT	(MemoryWindow::OnPaint		)
	EVT_MOTION	(MemoryWindow::OnMouseMove	)
END_EVENT_TABLE()

MemoryWindow::MemoryWindow(wxWindow* parent_, wxWindowID id, 
	const wxString& title, 
	const wxPoint& pos, 
	const wxSize& size, 
	long style, 
	const wxString& name):
//wxFrame(parent, id, title, pos, size, style, name)
wxScrolledWindow(parent_, id, pos, size, style, name),
parent(parent_)
{
	width=size.x;
	height=size.y*4;
	SetVirtualSize(width, height);
	blockx=5;
	blocky=5;
	blocksx=width/blockx;
	SetScrollRate(0, blocky);
	EnableScrolling(false, true);

	inMouseMove=false;

	dc=new wxClientDC(this);
	pen=new wxPen();
	brush=new wxBrush();
#ifdef WIN32
	wxToolTip::SetDelay(0);
#endif
	tip=0;

	palsize=0x1000;
	palette   =new Color[palsize];
	paletteBGR=new Color[palsize];
	int i;
	for (i=0; i< palsize; i++)
	{	Color &c=palette[i];
		switch (i)
		{
		case nt_ENUM:		c=0;			break;
		case UNBOUND:		c=0xffff00;	break;
		case STRING:
		case BACKSLASHED:
		case VBAR_STRING:	c=0xff;		break;
		case INT_:			c=0xff00;	break;
		case FLOAT_:		c=0xff00;	break;
		case COMPLEX:		c=0xff00;	break;
		case INT16:			c=0xff00;	break;
		case INT8:			c=0xff00;	break;
		case UINT8:			c=0xff00;	break;
		case QUOTE:			c=0xff0000;	break;
		case COLON:			c=0xff0000;	break;
		case OBJECT:		c=0xff7f7f;	break;
		case PAIR:			c=0xff0000;	break;
		case LINE:			c=0xff0000;	break;
		case RUNPARSED:	c=0xff0000;	break;
		case TREE:			c=0xff0000;	break;
		case TODO:			c=0xff0000;	break;
		case FILE_:			c=0xff0000;	break;
		case PROC_:			c=0xff0000;	break;
		case ARRAY:			c=0xffff;	break;
		case BITMAP_:		c=0xffff;	break;
		case TEXTURE:		c=0xffff;	break;
		case TURTLE_:		c=0xffff;	break;
		case TABLE:			c=0xffff;	break;
		case STRUCT:		c=0xffff;	break;
		case GRAPHIC:		c=0xffff;	break;
		case DYNAMIC_LIB:	c=0xffff;	break;
		case INTARRAY:		c=0xffff;	break;
		case INT16ARRAY:	c=0xffff;	break;
		case FLOATARRAY:	c=0xffff;	break;
		case PRIM:
		case MACRO:
		case TAILFORM:
		case INFIX:			c=0xff00ff;	break;
		default:				c=0xffffff;
		}
	}
	for (i=0; i< palsize; i++)
		paletteBGR[i]=	((palette[i] & 0xff) << 16)
						|   (palette[i] & 0xff00)
						|	((palette[i] & 0xff0000) >> 16);
}

MemoryWindow::~MemoryWindow() 
{
	dc->SetPen(wxNullPen);
	delete pen;
	dc->SetBrush(wxNullBrush);
	delete brush;
	delete dc;
	delete palette;
	delete paletteBGR;
	if (theApp)
		theApp->mainFrame->viewMenu->Check(ID_VIEW_MEMORY, false);
	memoryWindow=0;
}

void MemoryWindow::OnErase(wxEraseEvent &event)
{
//	event.Skip();
}

void MemoryWindow::OnSize(wxSizeEvent &event)
{
	wxSize size=event.GetSize();
	width=size.x-20;
	SetVirtualSize(width, height);
	blocksx=width/blockx;
	SetScrollRate(0, blocky);
	event.Skip();
}

void MemoryWindow::OnPaint(wxPaintEvent &event)
{
	if (!dc)
		return;
	wxPoint pt = dc->GetDeviceOrigin();
	dc->SetDeviceOrigin(pt.x,
                       pt.y-m_yScrollPosition*m_yScrollPixelsPerLine);
	wxMemoryDC mdc;
	int winwidth, winheight;
	GetClientSize(&winwidth, &winheight);
	int y=m_yScrollPosition*m_yScrollPixelsPerLine;
	int size=winwidth*(winheight/blocky+3)*blocky;
	ColorP pixels=new Color[size], 
		pix=pixels, 
		end=&pixels[size];
	int x=0, restx=winwidth % blockx, maxx=winwidth-restx-blockx, 
		n=0, nodeNr=y*(winwidth/blockx);
	SegmentP seg;
	for (seg=Segment::start;  seg;  seg=seg->prev) 
	{	if (nodeNr-n >= 0  &&  nodeNr-n < GC_::segment_size)
		{	for (ObjectP nd=&seg->base[nodeNr-n]; 
					unsigned(nd) <= seg->maxu && pix < end; nd++)
			{	int c;
				if (nd->type < palsize)
					c=paletteBGR[nd->type];
				else
					c=0xffffff;
				ColorP start=pix, first=pix;
				for (int j=blocky-1; j; j--)
				{	for (int i=blockx-1; i; i--)
						*pix++=c;	// don't mess with SetPixel, this way is much faster
					*pix++=0;
					first+=winwidth;
					pix=first;
				}
				pix=start+blockx;
				x+=blockx;
				if (x > maxx)
				{	start=first=pix;
					for (int j=blocky-1; j; j--)
					{	for (int i=restx; i; i--)
							*pix++=0;
						first+=winwidth;
						pix=first;
					}
					pix=start+restx;
					x=0;
					pix+=(blocky-2)*winwidth;
					for (int i=winwidth; i; i--)
						*pix++=0;
				}
			}
			if (seg)
				seg=seg->prev;
			break;
		}
		n+=GC_::segment_size;
	}
	for (;  seg;  seg=seg->prev) 
	{	for (ObjectP nd=seg->base; unsigned(nd) <= seg->maxu && pix < end; nd++)
		{	int c;
			if (nd->type < palsize)
				c=paletteBGR[nd->type];
			else
				c=0xffffff;
			ColorP start=pix, first=pix;
			for (int j=blocky-1; j; j--)
			{	for (int i=blockx-1; i; i--)
					*pix++=c;	// don't mess with SetPixel, this way is much faster
				*pix++=0;
				first+=winwidth;
				pix=first;
			}
			pix=start+blockx;
			x+=blockx;
			if (x > maxx)
			{	start=first=pix;
				for (int j=blocky-1; j; j--)
				{	for (int i=restx; i; i--)
						*pix++=0;
					first+=winwidth;
					pix=first;
				}
				pix=start+restx;
				x=0;
				pix+=(blocky-2)*winwidth;
				for (int i=winwidth; i; i--)
					*pix++=0;
			}
		}
	}
	if (pix < end)
	{	ColorP first=pix;
		for (int j=blocky-1; j>0; j--)
		{	for (int i=winwidth-x; i>0; i--)
				*pix++=0x808080;
			first+=winwidth;
			pix=first;
		}
		pix-=x;
		for (int i=x; i>0; i--)
			*pix++=0;
		while (pix < end)
			*pix++=0x808080;
	}
	wxBitmap bitmap((const char*)pixels, winwidth, winheight, 32);
	mdc.SelectObject(bitmap);

	dc->Blit(0,y, winwidth,winheight, &mdc, 0,0);
	delete pixels;
	dc->SetDeviceOrigin(pt.x, pt.y);
	event.Skip();
}

void MemoryWindow::updateNode(NodeP nd)
{
//	console->dispatchMessages();
	if (!dc || !pen || !brush)
		return;

	int nodeNr, i=0;
	for (SegmentP seg=Segment::start;  seg;  seg=seg->prev) 
	{	if (unsigned(nd) >= unsigned(seg->base)  &&  unsigned(nd) <= seg->maxu)
		{	nodeNr=(Object*)&*nd-(Object*)seg->base;
			nodeNr+=GC_::segment_size*i;
			break;
		}
		i++;
	}
	int x,y,c;
	x=blockx*(nodeNr % blocksx);
	y=blocky*(nodeNr / blocksx);
	y-=m_yScrollPosition*m_yScrollPixelsPerLine*blocky;
	if (nd->type < palsize)
		c=palette[nd->type];
	else
		c=0xffffff;
	pen->SetColour(c);
	brush->SetColour(c);
	dc->SetPen(*pen);
	dc->SetBrush(*brush);
//	wxPoint pt = dc->GetDeviceOrigin();
//	dc->SetDeviceOrigin(pt.x,
//		pt.y-m_yScrollPosition*m_yScrollPixelsPerLine*blocky);
	dc->SetUserScale(1,1);
	dc->SetLogicalScale(1,1);
#ifdef WIN32
	dc->SetSystemScale(1,1);
#endif
	dc->DrawRectangle(x,y, blockx-1, blocky-1);
//	dc->SetDeviceOrigin(pt.x, pt.y);
	dc->SetPen(wxNullPen);
	dc->SetBrush(wxNullBrush);
}

void MemoryWindow::OnMouseMove(wxMouseEvent &event)
{
	if (!GC_::initialized || !dc)
		return;
	if (inMouseMove)
		return;
	inMouseMove=true;
	wxPoint pt = dc->GetDeviceOrigin();
	dc->SetDeviceOrigin(pt.x,
		pt.y-m_yScrollPosition*m_yScrollPixelsPerLine*blocky);
	wxPoint p=event.GetLogicalPosition(*dc);
	dc->SetDeviceOrigin(pt.x, pt.y);
	int x,y, nodeNr;
	x=p.x/blockx;
	y=p.y/blocky;
	nodeNr=x+y*blocksx;
	NodeP nd=0;
	int n=0;
	for (SegmentP seg=Segment::start;  seg;  seg=seg->prev)
	{	if (nodeNr-n >= 0  &&  nodeNr-n < GC_::segment_size)
		{	nd=&seg->base[nodeNr-n];
			break;
		}
		n+=GC_::segment_size;
	}
	wxString text, ndtext;
	text << nodeNr << "                            \n";
	if (nd)
	{	StringP s=0;
		s=nd->_typeof();
		if (s)
		{	if (s->is.string)
				text << s->ptr();
			else
				s=s;
			if (nd)
			{	if (nd->is.string)
				{	
					if (StringP(nd)->ptr()[StringP(nd)->length()] != '\0')
						s=String::make(StringP(nd)->ptr(), 0, 
											StringP(nd)->length(), 
											nd->type, String::copy);
					else
						s=StringP(nd);
				}
				else
					s=nd->string();
				if (s && s->is.string)
				{	text << '\n';
					ndtext=s->ptr();
					ndtext.Replace("\t", " ");
					text << ndtext;
				}
			}
		}
	}
	else
		text << "List\n[]";
	if (tip)
		delete tip;
	tip = new wxTipWindow(this, text, 1024, &tip);
	int	x0=(event.GetX()/blockx)*blockx, 
			y0=(event.GetY()/blocky)*blocky;
	ClientToScreen(&x0, &y0);
	wxRect rect(x0, y0, blockx, blocky);
	if (tip)
		tip->SetBoundingRect(rect);
	inMouseMove=false;
}
