#ifndef GC_included
	#include "GC.h"
#endif

#define OPERATORS_DECL															\
	virtual NodeP																	\
		BitNot(), random(),														\
		_int(), toint16(), toint8(), touint8(),							\
		truncate(), round(), _float(), _bigfloat(),						\
		Sin(), Cos(), Tan(),														\
		ArcSin(), ArcCos(),														\
		radSin(), radCos(), radTan(),											\
		radArcSin(), radArcCos(),												\
		abs(), Sqr(), Sqrt(), exp(), LN(), Log10(),						\
		real(), imag(), conjugate(), Signum(),								\
		add(NodeP arg), sub(NodeP arg),										\
		mul(NodeP arg), div(NodeP arg),										\
		mod(NodeP arg), Remainder(NodeP arg),								\
		aShift(NodeP arg), lShift(NodeP arg),								\
		BitOr(NodeP arg), BitAnd(NodeP arg), BitXOr(NodeP arg),		\
		Power(NodeP arg), ArcTan(NodeP arg), radArcTan(NodeP arg);	\
																						\
	virtual void																	\
		xCopy(NodeP arg), xAdd(NodeP arg), xSub(NodeP arg),			\
		 xMul(NodeP arg), xDiv(NodeP arg), xMod(NodeP arg);			\

#define GRAPHIC_DECL											\
	virtual void												\
		setPixel(NodeP color),								\
		setPixelXY(NodeP yin, NodeP color),				\
		setPixelXYZ(NodeP yin, NodeP zin, NodeP color),	\
		line(NodeP color),									\
		setXYZ(NodeP ynode, NodeP znode);				\

extern NodeP Unbound;

enum NodeType
{
	nt_LIST     =1<<0,
	nt_WORD     =1<<1,
	nt_AGGREGATE=1<<2,
	nt_PRIM     =1<<3,
	nt_STRING   =1<<4,
	nt_NUMBER   =1<<5,
	nt_PUNCT    =1<<6,
	nt_BACKSL   =1<<7,
	nt_LINE     =1<<8,
	nt_ENUM     =1<<9,

	    UNBOUND = 0,
	     NTFREE = nt_ENUM,
	 SPARE_NODE = nt_ENUM*2,

	     STRING = nt_WORD | nt_STRING,
	BACKSLASHED = nt_WORD | nt_STRING | nt_BACKSL,
	VBAR_STRING = nt_WORD | nt_STRING | nt_BACKSL | nt_ENUM,

	       INT_ = nt_WORD | nt_NUMBER,
	     FLOAT_ = nt_WORD | nt_NUMBER | nt_ENUM,
	    COMPLEX = nt_WORD | nt_NUMBER | nt_ENUM*2,
	      INT16 = nt_WORD | nt_NUMBER | nt_ENUM*3,
	       INT8 = nt_WORD | nt_NUMBER | nt_ENUM*4,
	      UINT8 = nt_WORD | nt_NUMBER | nt_ENUM*5,
	   BIGFLOAT = nt_WORD | nt_NUMBER | nt_ENUM*6,

	      QUOTE = nt_WORD | nt_PUNCT | nt_LINE,
	      COLON = nt_WORD | nt_PUNCT | nt_LINE | nt_ENUM,

	     OBJECT = nt_WORD,

	       PAIR = nt_AGGREGATE | nt_LIST,
	       LINE = nt_AGGREGATE | nt_LIST | nt_LINE,
	  RUNPARSED = nt_AGGREGATE | nt_LIST | nt_LINE | nt_ENUM,
	       TREE = nt_AGGREGATE | nt_LIST | nt_LINE | nt_ENUM*2,
	       TODO = nt_AGGREGATE | nt_LIST | nt_ENUM*4,
			FILE_ = nt_AGGREGATE | nt_LIST | nt_ENUM*5,
	       PNIL = nt_AGGREGATE | nt_LIST | nt_ENUM*6,
	 WINDOWPAIR = nt_AGGREGATE | nt_LIST | nt_ENUM*7,

			PROC_ = nt_AGGREGATE | nt_ENUM, 
	      ARRAY = nt_AGGREGATE | nt_ENUM*2,
	    BITMAP_ = nt_AGGREGATE | nt_ENUM*3,
	    TEXTURE = nt_AGGREGATE | nt_ENUM*4,
	    TURTLE_ = nt_AGGREGATE | nt_ENUM*5,
	      TABLE = nt_AGGREGATE | nt_ENUM*6,
	     STRUCT = nt_AGGREGATE | nt_ENUM*7,
		 GRAPHIC = nt_AGGREGATE | nt_ENUM*8,
	DYNAMIC_LIB = nt_AGGREGATE | nt_ENUM*9,
		INTARRAY = nt_AGGREGATE | nt_ENUM*10,
	 INT16ARRAY = nt_AGGREGATE | nt_ENUM*11,
	 FLOATARRAY = nt_AGGREGATE | nt_ENUM*12,
	     SLIDER = nt_AGGREGATE | nt_ENUM*13,
	      GAUGE = nt_AGGREGATE | nt_ENUM*14,
	 INTCONTROL = nt_AGGREGATE | nt_ENUM*15,
	FLOATCONTROL= nt_AGGREGATE | nt_ENUM*16,
	      FRAME = nt_AGGREGATE | nt_ENUM*17,
	      GRAPH = nt_AGGREGATE | nt_ENUM*18,
	     BUTTON = nt_AGGREGATE | nt_ENUM*19,
	TOGGLEBUTTON= nt_AGGREGATE | nt_ENUM*20,
	   CHECKBOX = nt_AGGREGATE | nt_ENUM*21,
	RADIOBUTTON = nt_AGGREGATE | nt_ENUM*22,
	 STATICTEXT = nt_AGGREGATE | nt_ENUM*23,
	TEXTCONTROL = nt_AGGREGATE | nt_ENUM*24,
		 LISTBOX = nt_AGGREGATE | nt_ENUM*25,
	  CHOICEBOX = nt_AGGREGATE | nt_ENUM*26,
	   COMBOBOX = nt_AGGREGATE | nt_ENUM*27,
	LISTCONTROL = nt_AGGREGATE | nt_ENUM*28,
		BOXSIZER = nt_AGGREGATE | nt_ENUM*29,

			 PRIM = nt_PRIM,
	      MACRO = nt_PRIM | nt_ENUM,
	   TAILFORM = nt_PRIM | nt_ENUM*2,
	      INFIX = nt_PRIM | nt_ENUM*3,
};

enum Priority
{	priority_STOP  =0,
	priority_OUTPUT=1,
	priority_MAYBE =2,  priority_TAIL=2,
	priority_MACRO =3,
	prefix = 10,
	priority_assignment  = prefix-3,
	priority_logical     = prefix-2,
	priority_bit         = prefix-1,
	priority_prefix      = prefix,
	priority_comparsion  = prefix+1,
	priority_addition    = prefix+2,
	priority_product     = prefix+3,
	priority_power       = prefix+4,
	priority_faculty     = prefix+5,
	priority_tight_minus = prefix+6,
	priority_dot         = prefix+7,
};

enum ObjectType
{	// Object flags.  Ones settable by users via bury_helper must come in threes
	// for proc, val, plist even if meaningless for some of those.
	NORMAL_OBJECT= 0,
	PROC_BURIED =1<<0,  VAL_BURIED =1<<1,  PLIST_BURIED =1<<2,
	PROC_TRACED =1<<3,  VAL_TRACED =1<<4,  PLIST_TRACED =1<<5,
	PROC_STEPPED=1<<6,  VAL_STEPPED=1<<7,  PLIST_STEPPED=1<<8,
	PROC_MACRO  =1<<9,    PERMANENT=1<<10,

	 BURIED = PROC_BURIED  | VAL_BURIED  | PLIST_BURIED,
	 TRACED = PROC_TRACED  | VAL_TRACED  | PLIST_TRACED,
	STEPPED = PROC_STEPPED | VAL_STEPPED | PLIST_STEPPED,
};

class Point;
class wxDynamicLibrary;
class Slider;
class Frame;
class wxDataInputStream;
class wxDataOutputStream;

class Node: public GC_	// is the basic data structure of Logo.
{
	public:
		struct Bits
		{	bool
				list       :1,
				word       :1,
				aggregate  :1,
				primitive  :1,
				string     :1,
				number     :1,
				punct      :1,
				backslashed:1,
				line       :1;
		};
		union {NodeType type; Bits is;};

		static const Node aNode;
		static const int vmtNode, vmtPair, vmtRunparsed, vmtLine, vmtTree;
		static const int boolerror;

		operator_new;
		Node(): type(UNBOUND) {}
		Node(NodeType type_): type(type_) {}

		virtual void free() {}
		virtual void inc_gc() {}
		virtual int youngs() {return 0;}
		virtual int youngsPtr() {return 0;}
		virtual void check_oldyoungs() {}

		virtual wxDataInputStream &read(wxDataInputStream &istr);
		virtual wxDataOutputStream &write(wxDataOutputStream &ostr);
		virtual void setPtrs();
		virtual void aftersetPtrs();
		virtual void recreate();
		virtual void aftersetPtrs2();
		virtual void aftersetPtrs3();

		virtual StringP   string();
		virtual    int   numberP() {return 0;}
		virtual  NodeP    number();
		virtual double getnumber(); 

		OPERATORS_DECL
		virtual int compare(NodeP nd);
		virtual ObjectP intern();
		virtual int count();
		virtual int _sizeof();
		virtual StringP _typeof();
		virtual bool equalP(NodeP b);
		virtual NodeP deepCopy();
		virtual NodeP first_();
		virtual NodeP butFirst();
		virtual NodeP firsts();
		virtual NodeP butFirsts();
		virtual NodeP last();
		virtual NodeP butLast();
		virtual NodeP fPut(NodeP nd);
		virtual NodeP lPut(NodeP nd);
		virtual double _min(double d);
		virtual double _max(double d);
		virtual NodeP Norm();
		virtual NodeP maxNorm();
		virtual NodeP resize(int newsize);
		virtual NodeP lowPassFilter(double weight);
		virtual NodeP saturateAbove(double limit);
		virtual NodeP saturateBelow(double limit);
		virtual NodeP    item(int index);
		virtual NodeP    item(NodeP index, NodeP result);
		virtual NodeP   items(NodeP start, NodeP end);
		virtual NodeP  mdItem(NodeP index);
		virtual NodeP bititem(NodeP index);
		virtual NodeP setitem(NodeP index, NodeP newitem, bool safe);
		virtual NodeP setitems(NodeP start, NodeP newitems);
		virtual NodeP memberP(NodeP subobj, bool notp, bool substr);
		virtual void removeitem(NodeP index);
		virtual NodeP reverse();
		virtual NodeP rotate(int places);
		virtual NodeP shuffle();
		virtual NodeP remove(NodeP thing);
		virtual NodeP destructiveRemove(NodeP thing);
		virtual NodeP remdup();
		virtual NodeP upper();
		virtual NodeP lower();

		virtual PairP list();
		virtual ArrayP array(int origin=1);
		virtual IntArrayP intarray(int origin=1);
		virtual Int16ArrayP int16array(int origin=1);
		virtual FloatArrayP floatarray(int origin=1);
		virtual TableP table();

		virtual FileP findInFilelist(bool remove);
		virtual Priority priority();
		virtual PairP gather_args(PairPP args, bool inparen, ObjectPP ifnode);
		virtual NodeP readbin(FileP file);
		virtual void setParentFrame(FrameNodeP frame) {}
		virtual void setParentGraph(GraphNodeP graph) {}
		GRAPHIC_DECL

		void settype_NTFREE()   {type=NTFREE;    *intP(this) = vmtNode;     }
		void settype_PAIR()     {type=PAIR;      *intP(this) = vmtPair;     }
		void settype_RUNPARSED(){type=RUNPARSED; *intP(this) = vmtRunparsed;}
		void settype_TREE()     {type=TREE;      *intP(this) = vmtTree;     }
		void settype_LINE()     {type=LINE;      *intP(this) = vmtLine;     }

		bool islist()     {return this? is.list      : true ;}
		bool isnolist()   {return this? !is.list     : false;}
		bool isaggregate(){return this? is.aggregate : true ;}
		bool isstring()   {return this? is.string    : false;}
		bool isarray()    {return this? type==ARRAY  : false;}
		bool istree()     {return this? type==TREE   : false;}

		bool equal(NodeP b) {return this? !compare(b) : b? false : true;}

		PairP maybe_quote();
		void silent_load(const char *prefix);

		static NodeP boolnode(bool tf);
		static void checkboolerror(int tf, NodeP a, NodeP b);

		      int a_bool      (int nr);	// these are for argument type checking
		     IntP a_integer   (int nr);
		      int a_int       (int nr);
		   Int16P a_integer16 (int nr);
		    short a_int16     (int nr);
		    Int8P a_integer8  (int nr);
		     char a_int8      (int nr);
		     IntP a_pos_int   (int nr);
		   double a_float     (int nr);
		    NodeP a_number    (int nr);
	   ComplexdP a_complex   (int nr);
		BigFloatP a_bigfloat  (int nr);
		  StringP a_char      (int nr);
		  StringP a_string    (int nr);
		    NodeP a_name      (int nr);
		    PairP a_bfable    (int nr);
		    PairP a_list      (int nr);
		    PairP a_runnable  (int nr);
		  BitmapP a_bitmap    (int nr);
	      TableP a_table     (int nr);
	      ArrayP a_array     (int nr);
	   IntArrayP a_intarray  (int nr);
	 Int16ArrayP a_int16array(int nr);
	 FloatArrayP a_floatarray(int nr);
		    Point a_3darray   (int nr);
		    PairP a_vector_helper(bool floatok, bool three, bool neg, int nr);
		    PairP a_vector    (int nr);
		    PairP a_3dvector  (int nr);
		    PairP a_rgb       (int nr);
		    PairP a_rgba      (int nr);
		    PairP a_int_vector(int nr);
		    PairP a_pos_int_vector(int nr);
		    Color a_color			(int nr);
	 TurtleNodeP a_turtleNode	(int nr);
   GraphicNodeP a_graphicNode	(int nr);
   wxDynamicLibrary *a_dynamicLibrary(int nr);
	 SliderNodeP a_slider		(int nr);
	  GaugeNodeP a_gauge			(int nr);
	  IntControlNodeP a_intControl	(int nr);
	FloatControlNodeP a_floatControl	(int nr);
			ButtonNodeP a_button			(int nr);
	ToggleButtonNodeP a_toggleButton	(int nr);
		CheckBoxNodeP a_checkBox	(int nr);
	RadioButtonNodeP a_radioButton(int nr);
			FrameNodeP a_frame		(int nr);
			GraphNodeP a_graph		(int nr);
	 StaticTextNodeP a_staticText	(int nr);
	TextControlNodeP a_textControl(int nr);
		 ListBoxNodeP a_listBox		(int nr);
	  ChoiceBoxNodeP a_choiceBox	(int nr);
	   ComboBoxNodeP a_comboBox	(int nr);
	ListControlNodeP a_listControl(int nr);
		BoxSizerNodeP a_boxSizer	(int nr);
};

inline NodeType type_of(NodeP nd) {return nd? nd->type : PNIL;}

extern wxDataInputStream &operator>>(wxDataInputStream &istr, bool &b);
extern wxDataOutputStream &operator<<(wxDataOutputStream &ostr, bool &b);

#define Node_included
