#ifndef Node_included
	#include "Node.h"
#endif
#ifndef Primitive_included
	#include "Primitiv.h"
#endif

enum ContentsType {PROCS, VALS, PLISTS};

struct ObjectBits
{	bool
		proc_buried : 1,  val_buried : 1,  plist_buried : 1,
		proc_traced : 1,  val_traced : 1,  plist_traced : 1,
		proc_stepped: 1,  val_stepped: 1,  plist_stepped: 1,
		proc_macro  : 1,    permanent: 1;
};

class Object: public Node
{
	ObjectP _tail;
	StringP _name;
	NodeP _proc;
	NodeP _var;
	PairP _plist;

	public:
		union {ObjectType flag; ObjectBits bit; int tint;};
		static const Object aObject;

		operator_new;

		   ObjectP tail() {return _tail;}
		   StringP name() {return _name;}
		NodeP proc_or_prim() {return _proc;}
		ProcedureP proc() {return ProcedureP(_proc);}
		PrimitiveP prim() {return PrimitiveP(_proc);}
		     NodeP var()  {return _var;}
		     PairP plist(){return _plist;}

		void settail (ObjectP    nd) {_tail = nd;  check_oldyoung(nd);}
		void setname (StringP    nd);
		void setproc_or_prim(NodeP nd);
		void setproc (ProcedureP nd);
		void setprim (PrimitiveP nd) {_proc = nd;  check_oldyoung(nd);}
		void setvar  (NodeP      nd) {_var  = nd;  check_oldyoung(nd);}
		void setplist(PairP      nd);

		Object(): 
			Node(OBJECT), _tail(0), _name(0), _proc(0), _var(0), _plist(0)
		{flag=ObjectType(0);}

		Object(StringP name_, ProcedureP proc_, NodeP var_, PairP plist_, ObjectP tail_): 
			Node(OBJECT), flag(NORMAL_OBJECT) 
		{setname(name_), setproc(proc_), setvar(var_), setplist(plist_), settail(tail_);}

		Object(StringP name_, PrimitiveP prim_, NodeP var_, PairP plist_, ObjectP tail_): 
			Node(OBJECT), flag(NORMAL_OBJECT) 
		{setname(name_), setprim(prim_), setvar(var_), setplist(plist_), settail(tail_);}

		void inc_gc();
		int youngs();
		int youngsPtr();
		void check_oldyoungs();

		void setPtrs();
		wxDataInputStream &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);

		ObjectP intern() {return this;}
		StringP string() {return name();}
		   int   numberP(); 
		double getnumber(); 
		NodeP     number();
		OPERATORS_DECL
		int compare(NodeP nd);
		int _sizeof();
		StringP _typeof();
		NodeP fPut(NodeP nd);
		NodeP lPut(NodeP nd);
		NodeP item(int i);
		NodeP item(NodeP index, NodeP result);
		NodeP bititem(NodeP index);
		NodeP memberP(NodeP subobj, bool notp, bool substr);
		NodeP setitem(NodeP index, NodeP newitem, bool safe);
		void removeitem(NodeP index);
		NodeP reverse();
		NodeP remove(NodeP thing);
		NodeP remdup();
		NodeP upper();
		NodeP lower();
		PairP list();
		ArrayP array(int origin);

		void   setflag(int f) {tint  |=   f;}
		void clearflag(int f) {tint &= ~f;}

		void mark_all_but_objects();
		Priority priority();
		void tell_shadow();
		bool not_local(int stackPtr);

		int find_limit();
		void print();
		StringP contents(ContentsType contents_type, ObjectType want_buried);

	friend class Evaluator;
};

#define Object_included
