//___________________________________________________________________
//
//                         Class RadioButton
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/radiobut.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/textctrl.h>

#include "Main.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "Color.h"
#include "Float.h"
#include "Point.h"
#include "Eval.h"
#include "Stack.h"

#include "Event.h"
#include "RadioButton.h"
#include "Frame.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"
#include "Error.h"
#include "Graph.h"
#include "Console.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(RadioButton, wxRadioButton)
	EVT_RADIOBUTTON	(-1, RadioButton::OnClick	)
END_EVENT_TABLE()

RadioButton::RadioButton(RadioButtonNodeP radioButtonNode_):
	radioButtonNode(radioButtonNode_)
{
	wxWindow *parent=console;
	if (radioButtonNode->parentFrame)
		parent=radioButtonNode->parentFrame->frame;
	if (radioButtonNode->parentGraph)
		parent=radioButtonNode->parentGraph->graph;
	Create(parent, -1, 
			wxEmptyString, 
			wxDefaultPosition, 
			wxDefaultSize, 
			0, 
			wxDefaultValidator, 
			"RadioButton");
}

RadioButton::RadioButton(RadioButtonNodeP radioButtonNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxString &label,
			PairP onClick_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
wxRadioButton(parent, id, label, 
			pos, size, style, 
			validator, name),
onClick(onClick_),
radioButtonNode(radioButtonNode_)
{
	radioButtonNode->radioButton=this;
}

RadioButton::~RadioButton()
{
	if (radioButtonNode)
		radioButtonNode->radioButton=0;
}

void RadioButton::OnClick(wxCommandEvent &event)
{
	Event::fifo << new EventRadioButton(this);
}

wxString RadioButton::wxstring()
{
	wxString str;
	str << "RadioButton [" << GetLabel() << "] " 
		<< '[' << GetName() << "] ";
	if (onClick)
		str << onClick->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " 
		<< GetValue();
	return str;
}

wxDataInputStream &RadioButton::read(wxDataInputStream &istr) 
{	
	wxString label, name;
	wxPoint pos;
	wxSize size;
	int style, state;
	
	istr >> label >> name >> (int&)onClick
		>> pos.x >> pos.y >> size.x >> size.y >> style >> state;

	SetLabel(label);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetWindowStyle(style);
	SetValue(state!=0);
	return istr;
}
wxDataOutputStream &RadioButton::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetLabel() << GetName() 
		<< GC_::ptrToInt->find(onClick)
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y
		<< int(GetWindowStyle()) << int(GetValue());
	return ostr;
}

PairP RadioButtonNode::onClick()
{
	if (radioButton)
		return radioButton->onClick;
	return 0;
}

void RadioButtonNode::setonClick(NodeP nd) 
{
	if (radioButton)
	{	radioButton->onClick = PairP(nd);  check_oldyoung(nd);
	}
}

void RadioButtonNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void RadioButtonNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

RadioButtonNode::RadioButtonNode(): 
	Node(RADIOBUTTON), radioButton(0), 
	parentFrame(0),
	parentGraph(0)
{}

RadioButtonNode::RadioButtonNode(FrameNodeP parentFrame_,
											GraphNodeP parentGraph_): 
	Node(RADIOBUTTON), radioButton(0), 
	parentFrame(parentFrame_),
	parentGraph(parentGraph_)
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void RadioButtonNode::free() 
{	
	if (radioButton) 
	{	radioButton->radioButtonNode=0;
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		radioButton->Destroy(); 
		EventPP out=Event::fifo.out;
		EventP ev;
		while (Event::fifo)
		{	Event::fifo >> ev;
			switch (ev->type)
			{
			case Event::tRadioButton:
				if (((EventRadioButton*)ev)->radioButton==radioButton)
					((EventRadioButton*)ev)->radioButton=0;
				break;
			default:
				;
			}
		}
		Event::fifo.out=out;
		radioButton=0;
	}
}
void RadioButtonNode::inc_gc() {fifo << onClick();}
int RadioButtonNode::youngs() {return gmark(onClick());}
int RadioButtonNode::youngsPtr() {return gmarkPtr(onClick());}
void RadioButtonNode::check_oldyoungs() {check_oldyoung(onClick());}

wxDataInputStream &RadioButtonNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	FrameNodeP parentFrameNode;
		GraphNodeP parentGraphNode;
		istr 
			>> (int&)parentFrameNode 
			>> (int&)parentGraphNode;
		radioButton=new RadioButton(RadioButtonNodeP(this));
		radioButton->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}

wxDataOutputStream &RadioButtonNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (radioButton)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame)
			<< GC_::ptrToInt->find(parentGraph);
		radioButton->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void RadioButtonNode::setPtrs()
{
	parentFrame=(FrameNode*)GC_::intToPtr->find(int(parentFrame));
	parentGraph=(GraphNode*)GC_::intToPtr->find(int(parentGraph));
	if (radioButton)
	{	setonClick((Pair*)GC_::intToPtr->find(int(onClick())));
	}
}

void RadioButtonNode::recreate()
{
	wxWindow *p=0;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	RadioButtonP r=radioButton, 
		r2=new RadioButton(RadioButtonNodeP(this),
				p, -1, r->GetLabel(),
				r->onClick, r->GetPosition(), r->GetSize(), r->GetWindowStyle(), 
				wxDefaultValidator, r->GetName());
	r2->SetValue(r->GetValue());
	radioButton=r2;
	setonClick(r->onClick);
	delete r;
	radioButton=r2;
}

StringP RadioButtonNode::string()
{
	if (radioButton)
		return new String(radioButton->wxstring());
	return _Null_Word->name();
}

wxString RadioButtonNode::wxstring()
{
	if (radioButton)
		return radioButton->wxstring();
	return wxEmptyString;
}

NodeP lRadioButton(PairPP arg)
{
	wxWindow *parent=graph;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=theApp->mainFrame->graphNode;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}
	wxString label=wxEmptyString;
	{	StringP l=arg[2]->a_string(2);
		if throwing return Unbound;
		label=l->ptr();
	}
	PairP onClick=0;
	if (exist(3))
	{	PairP oc=arg[3]->a_runnable(3);
		if throwing return Unbound;
		onClick=oc;
	}
	long style = 0;
	if (exist(4))
	{	int s=arg[4]->a_int(4);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(5))
	{	PairP p=arg[5]->a_int_vector(5);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(6))
	{	PairP s=arg[6]->a_int_vector(6);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	wxWindowID id=-1;
	if (exist(7))
	{	int i=arg[7]->a_int(7);
		if throwing return Unbound;
		id=i;	
	}
	wxString name=label;
	if (exist(8))
	{	StringP n=arg[8]->a_string(8);
		if throwing return Unbound;
		name=n->ptr();
	}
	RadioButtonNodeP radioButtonNode=new RadioButtonNode(
		parentFrameNode,
		parentGraphNode);
	new RadioButton(radioButtonNode,
			parent, id, label, 
			onClick, pos, size, style, wxDefaultValidator, name);
	return radioButtonNode;
}

NodeP lRadioButtonOnClick(PairPP arg)
{
	RadioButtonNodeP radioButton=arg[1]->a_radioButton(1);
	if throwing return Unbound;
	PairP function=arg[2]->a_runnable(2);
	if throwing return Unbound;
	radioButton->setonClick(function);
	return Unbound;
}

NodeP lRadioButtonDestroy(PairPP arg)
{
	RadioButtonNodeP radioButton=arg[1]->a_radioButton(1);
	if throwing return Unbound;
	{	FrameNodeP p=radioButton->parentFrame;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(radioButton));
			p->check_oldyoung(p->children);
		}
	}
	{	GraphNodeP p=radioButton->parentGraph;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(radioButton));
			p->check_oldyoung(p->children);
		}
	}
	radioButton->radioButton->Destroy();
	return Unbound;
}

NodeP lRadioButtonSet(PairPP arg)
{
	RadioButtonNodeP radioButton=arg[1]->a_radioButton(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	radioButton->radioButton->SetValue(state!=0);
	return Unbound;
}

NodeP lRadioButtonValue(PairPP arg)
{
	RadioButtonNodeP radioButton=arg[1]->a_radioButton(1);
	if throwing return Unbound;
	bool state=radioButton->radioButton->GetValue();
	return Node::boolnode(state);
}

NodeP lRadioButtonEnable(PairPP arg)
{
	RadioButtonNodeP radioButton=arg[1]->a_radioButton(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	radioButton->radioButton->Enable(state!=0);
	return Unbound;
}

