//___________________________________________________________________
//
//                         Class Screen_
//
//                          Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/glcanvas.h>
#include <math.h>
#include <GL/glu.h>

#include "Main.h"
#include "Error.h"
#include "Init.h"
#include "File.h"
#include "List.h"
#include "UInt8.h"

#include "Point.h"
#include "Color.h"
#include "Screen.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

// Be careful: managing screen coordinates can really be mind-boggeling ;-))

Screen_::Screen_()
{
	perspective=textured=false;
	Min=Point(-400,-300,-1000000); 
	Max=Point( 400, 300, 1000000);
	size = (Max-Min).abs();
	base=Matrix(Point(1,0,0), Point(0,1,0), Point(0,0,1));
	scale=center0=center=center3d=0;
	base_scaled=0;
	distance=1000;
	color=0xffffffff;
	winrangex=winrangey=0;
	parent=this; // The "parent=this"-instruction rigorously 
				   // prevents accidental setting of bogus pointers.
}

Screen_::~Screen_() {}

Screen_ &Screen_::operator=(const Screen_ &b) // only use if you know what you do!
{
	x=b.x; y=b.y; z=b.z;
	distance=b.distance;
	Min=b.Min; Max=b.Max; 
	size=b.size; scale=b.scale; 
	base=b.base;
	base_scaled=b.base_scaled;
	center=b.center; center0=b.center0;
	return *this;
}

void Screen_::create(Screen_ *_parent)
{
	parent=_parent;
	create(parent->winsizex, parent->winsizey);
}

void Screen_::create(int wx, int wy)
{
	winsizex=wx;
	winsizey=wy;
}

void Screen_::destroy()
{
}

void Screen_::clear() 
{
	glClearColor(
		float(color.r)/255, 
		float(color.g)/255, 
		float(color.b)/255, 
		float(color.a)/255);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
}

void Screen_::setColor(Color c) 
{
	color=c;
	clear();
}
bool Screen_::setCoors(Rect &rect)
{
	Point winsize=Point(
		rect.right-rect.left, 
		rect.bottom-rect.top,
		2000000);
	center0=center=(winsize/2).round()-Point(.5,.5,.5);
									// so ]-0.5...0.5] will be displayed at 0
	scale=winsize/size;
	scale=scale.x<scale.y? scale.x: scale.y;

	base_scaled.x=base.x*scale.x;	
	base_scaled.y=base.y*scale.y;	
	base_scaled.z=base.z*scale.x;
	return true;
}

Point Screen_::rotate(Point &point)
{
	return base_scaled*point;
}

Point Screen_::rotateNotScaled(Point &point)
{
	return base*point;
}

void Screen_::map(Point &point)
{
	Point pc=point-center3d;
	Point p = rotate(pc);
	double d=distance-p.z;
	if (d==0) d=1e-6;
	double sc=distance/d;
	sx=(int)( p.x*sc+center.x);
	sy=(int)(-p.y*sc+center.y);
}

void Screen_::setEye(Point eye_, Point center3d_, Point up_)
{
	eye=eye_; center3d=center3d_; up=up_;
	distance=(eye-center3d).Norm();
	reset3dCoors();
}

void Screen_::reset3dCoors()
{
	Point p=eye-center3d;
	double pn=p.Norm();
	if (pn==0)
	{	error(BAD_DATA,0);
		return;
	}
	p=p/p.Norm();

	base.x=up.X(p);
	base.y=p.X(base.x);
	base.z=p;

	base=base.invert();

	base_scaled.x=base.x*scale.x;	
	base_scaled.y=base.y*scale.y;	
	base_scaled.z=base.z*scale.x;	
}

NodeP Screen_::matrix()
{
	Point center_ = (center-center0)/scale;
	return new List
	(	new List
		(	new Float(base.x.x),
			new Float(base.x.y),
			new Float(base.x.z), _ListEnd
		),
		new List
		(	new Float(base.y.x),
			new Float(base.y.y), 
			new Float(base.y.z), _ListEnd
		), 
		new List
		(	new Float(base.z.x),
			new Float(base.z.y), 
			new Float(base.z.z), _ListEnd
		), 
		new List
		(	new Float( center_.x),
			new Float(-center_.y), 
			new Float( center_.z), _ListEnd
		), 
		_ListEnd
	);
}

void Screen_::setEye()
{
	glLoadIdentity();
	gluLookAt(
		   eye.x,   eye.y,   eye.z, 
		center.x,center.y,center.z, 
		    up.x,    up.y,    up.z);
}

void Screen_::getMatrices()
{
	glGetDoublev(GL_MODELVIEW_MATRIX, modelview_matrix);
	glGetDoublev(GL_PROJECTION_MATRIX, projection_matrix);
	glGetIntegerv(GL_VIEWPORT, viewport);
}
