//___________________________________________________________________
//
//                              Stack
//
//                           Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/app.h>
#include <wx/datstrm.h>

#include <stdio.h>
#include <memory.h>

#include "Main.h"

#include "File.h"
#include "Pair.h"
#include "Array.h"
#include "Object.h"

#include "Eval.h"
#include "Error.h"
#include "Stack.h"

#include "IntToPtr.h"
#include "PtrToInt.h"

#include "LogoApp.h"

#define DEBUG_STACK
//#define DEBUG_STACK2
//#define _DEBUG

#ifndef max
	#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
	#define realloc(m, len) _realloc_dbg((m), (len), _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

const int PAGE_SIZE=0x100;

bool IntStack::noisy=false;

IntStack::IntStack() 
{
	size=0;
	base=intP(GC_::calloc(PAGE_SIZE, sizeof(int)));
	if (!base)
		return;
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_Stack
	item=0;
#endif
}

IntStack::~IntStack() 
{
	free(base);
#ifdef DEBUG_STACK2
	if (index != base)
		c_out << "Warning: IntStack has still " 
			<< int(index-base) << " elements on destruction" << endL;
#endif
}

bool IntStack::grow()
{
	intP ob=base;
	base=intP(GC_::realloc(base, sizeof(int)*size*2));
	if (!base)
	{	base=ob;
		return false;	
	}
	size*=2;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (noisy)
		c_out << "IntStack size is now grown to " << size << endL;
#endif
	return true;
}

bool IntStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(s)/log(2));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	intP ob=base;
	base=intP(GC_::realloc(base, sizeof(int)*s));
	if (!base)
	{	base=ob;
		return false;	
	}
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (noisy)
		c_out << "IntStack size is now shrinked to " << size << endL;
#endif
	return true;
}

#ifdef _DEBUG
	void IntStack::push(int val) 
	{
		if (index+1 > max)
			if (!grow())
				return;
		*++index=val;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void IntStack::pop(int &val) 
	{
		if (index>=base) 
		{	val = *index;  *index--=0;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << "Error: IntStack underflow on pop(" << val << ')' << endL;
	}
	IntStack &IntStack::operator<<(const int val)
	{	push(val);
		return *this;
	}

	IntStack &IntStack::operator>>(int &val)
	{	pop(val);
		return *this;
	}
	IntStack &IntStack::operator>>(bool &val)
	{	pop((int&)val);
		return *this;
	}
	IntStack &IntStack::operator>>(TailCall &val)
	{	pop((int&)val);
		return *this;
	}
	IntStack &IntStack::operator>>(ValStatus &val)
	{	pop((int&)val);
		return *this;
	}
	void IntStack::push(int i1, int i2)
	{	*this << i1 << i2;
	}
	void IntStack::push(int i1, int i2, int i3)
	{	*this << i1 << i2 << i3;
	}
	void IntStack::pop(int &i1, int &i2)
	{	*this >> i2 >> i1;
	}
	void IntStack::pop(int &i1, int &i2, int &i3)
	{	*this >> i3 >> i2 >> i1;
	}
	void IntStack::pop(ValStatus &i1, TailCall &i2)
	{	*this >> i2 >> i1;
	}
	void IntStack::pop(ValStatus &i1, TailCall &i2, int &i3)
	{	*this >> i3 >> i2 >> i1;
	}
#endif

wxDataInputStream &IntStack::read(wxDataInputStream &istr) 
{
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	int val;
		istr >> val;
		*this << val;
	}
	return istr;
}
wxDataOutputStream &IntStack::write(wxDataOutputStream &ostr) 
{
	int siz=index-base;
	ostr << siz;

	intP itemptr=base;
	for (int i=siz;  i>=1;  i--)
		ostr << *++itemptr;
	return ostr;
}


FloatStack::FloatStack() 
{
	size=0;
	base=doubleP(GC_::calloc(PAGE_SIZE, sizeof(double)));
	if (!base)
		return;
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=0;
#endif
}

FloatStack::~FloatStack() 
{
	free(base);
#ifdef DEBUG_STACK2
	if (index != base)
		c_out << "Warning: FloatStack has still " 
			<< int(index-base) << " elements on destruction" << endL;
#endif
}

bool FloatStack::grow()
{
	doubleP ob=base;
	base=doubleP(GC_::realloc(base, sizeof(double)*size*2));
	if (!base)
	{	base=ob;
		return false;	
	}
	size*=2;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << "FloatStack size is now grown to " << size << endL;
#endif
	return true;
}

bool FloatStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(s)/log(2));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	doubleP ob=base;
	base=doubleP(GC_::realloc(base, sizeof(double)*s));
	if (!base)
	{	base=ob;
		return false;	
	}
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << "FloatStack size is now shrinked to " << size << endL;
#endif
	return true;
}

#ifdef _DEBUG
	void FloatStack::push(double val) 
	{	if (index+2 > max)
			if (!grow())
				return;
		*++index=val;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void FloatStack::pop(double &val) 
	{	if (index>=base) 
		{	val = *index;  *index--=0;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << "Error: FloatStack underflow on pop(" << val << ')' << endL;
	}
	FloatStack &FloatStack::operator<<(const double val)
	{	push(val);
		return *this;
	}

	FloatStack &FloatStack::operator>>(double &val)
	{	pop(val);
		return *this;
	}
	void FloatStack::push(double d1, double d2, double d3)
	{	*this << d1 << d2 << d3;
	}
	void FloatStack::pop(double &d1, double &d2, double &d3)
	{	*this >> d3 >> d2 >> d1;
	}
#endif

wxDataInputStream &FloatStack::read(wxDataInputStream &istr) 
{
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	double val;
		istr >> val;
		*this << val;
	}
	return istr;
}
wxDataOutputStream &FloatStack::write(wxDataOutputStream &ostr) 
{
	int siz=index-base;
	ostr << siz;

	doubleP itemptr=base;
	for (int i=siz;  i>=1;  i--)
		ostr << *++itemptr;
	return ostr;
}


NodePStack::NodePStack()
{
	size=0;
	base=(NodeP*)GC_::calloc(PAGE_SIZE, sizeof(NodeP));
	if (!base)
	{	size=0;
		return;
	}
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=0;
#endif
}

NodePStack::~NodePStack() 
{
	free(base);
	clean();
}

bool NodePStack::grow()
{
	NodeP *ob=base;
	base=(NodeP*)GC_::realloc(base, sizeof(NodeP)*size*2);
	if (!base)
	{	base=ob;
		return false;	
	}
	size*=2;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << "NodePStack size is now grown to " << size << endL;
#endif
	return true;
}

bool NodePStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(s)/log(2));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	NodeP *ob=base;
	base=(NodeP*)GC_::realloc(base, sizeof(NodeP)*s);
	if (!base)
	{	base=ob;
		return false;	
	}
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << "NodePStack size is now shrinked to " << size << endL;
#endif
	return true;
}

#ifdef _DEBUG

	void NodePStack::clean()
	{
		if (index!=base && theApp 
		&& theApp->action!=EXIT_LOGO 
		&& theApp->action!=EXIT_DELAYED)
		{	c_out << "; Warning: NodePStack is still containing " 
				<< int(index-base) <<" elements!" << endL;

			while (index!=base)
			{	NodeP nd; pop(nd); 
//				File::print(stdout, nd);
			}
			c_out << endL;
		}		
	}

	void NodePStack::push(NodeP val) 
	{	if (index+1 > max)
			if (!grow())
				return;
		*++index=val;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void NodePStack::pop(NodeP &val) 
	{
		if (index>base) 
		{	val = *index;  *index--=0;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << "Error: NodePStack underflow on pop(" << val << ')' << endL;
		if (type_of(val)==NTFREE)
			c_out << "Error in NodePStack::pop(" << val
				<< "): NodePStack was damaged during garbge collection" << endL;
	}
	void NodePStack::pop() 
	{
		if (index>base) 
		{	*index--=0;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << "Error: NodePStack underflow on pop()" << endL;
	}
	void NodePStack::push_list(PairP head) 
	{
		while (head)
		{	push(NodeP(head->first()));
			next(head);
		}
	}
	void NodePStack::push(ArrayP a) 
	{
		int l=a->length();
		NodePP d=a->data();
		while (l--)
			push(*d++);
	}
	void NodePStack::pop_till(PairPP sp)
	{
		while (index > (NodeP*)sp)
			pop();
	}
	NodePStack &NodePStack::operator<<(const NodeP val)
	{	push(val);
		return *this;
	}
	NodePStack &NodePStack::operator>>(NodeP &val)
	{	pop(val);
		return *this;
	}
	NodePStack &NodePStack::operator>>(PairP &val)
	{	pop((NodeP&)val);
		return *this;
	}
	NodePStack &NodePStack::operator>>(LineP &val)
	{	pop((NodeP&)val);
		return *this;
	}
	NodePStack &NodePStack::operator>>(ObjectP &val)
	{	pop((NodeP&)val);
		return *this;
	}
	NodePStack &NodePStack::operator>>(ProcedureP &val)
	{	pop((NodeP&)val);
		return *this;
	}
	void NodePStack::push(
		const NodeP n1, const NodeP n2, const NodeP n3, const NodeP n4, 
		const NodeP n5, const NodeP n6, const NodeP n7) 
	{	
		*this << n1 << n2 << n3 << n4 << n5 << n6 << n7;
	}
	void NodePStack::push(
		const NodeP n1, const NodeP n2, const NodeP n3, const NodeP n4, 
		const NodeP n5, const NodeP n6, const NodeP n7, const NodeP n8) 
	{	
		*this << n1 << n2 << n3 << n4 << n5 << n6 << n7 << n8;
	}
	void NodePStack::pop(
		NodeP &n1, NodeP &n2, 
		NodeP &n3, ObjectP &n4, LineP &n5, 
		ObjectP &n6, LineP &n7)
	{
		*this >> n7 >> n6 >> n5 >> n4 >> n3 >> n2 >> n1;
	}
	void NodePStack::pop(
		PairP &n1, NodeP &n2, 
		NodeP &n3, ObjectP &n4, LineP &n5, 
		ObjectP &n6, LineP &n7)
	{
		*this >> n7 >> n6 >> n5 >> n4 >> n3 >> n2 >> n1;
	}
#else

void NodePStack::push(ArrayP a) 
{
	int l=a->length();
	NodePP d=a->data();
	while (l--)
	{	if (index+1 > max)
			if (!grow())
				return;
		*++index=*d++;
	}
}

#endif

void NodePStack::mark()
{
	for (NodeP *i=base; i<=index; i++)
		(*i)->mark();
}

void NodePStack::markPtr()
{
	for (NodeP *i=base; i<=index; i++)
		(*i)->markPtr();
}

wxDataInputStream &NodePStack::read(wxDataInputStream &istr) 
{
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	NodeP val;
		istr >> (int&)val;
		*this << val;
	}
	return istr;
}
wxDataOutputStream &NodePStack::write(wxDataOutputStream &ostr) 
{
	int siz=index-base;
	ostr << siz;

	NodePP itemptr=base;
	for (int i=siz;  i>=1;  i--)
		ostr << GC_::ptrToInt->find(*++itemptr);
	return ostr;
}

void NodePStack::setPtrs()
{
	int siz=index-base;
	NodePP itemptr=base;
	for (int i=siz;  i>=1;  i--)
	{	itemptr++;
		*itemptr = (Node*)GC_::intToPtr->find(int(*itemptr));
	}
}

PairP NodePStack::to_list(PairPP sp)
{
	PairP list=0;
	NodePP i=index;
	while (i != NodePP(sp))
		::push(list, *i--);
	return list;
}

PairP NodePStack::to_list(int i)
{
	return to_list((PairPP)&base[i]);
}

PairP NodePStack::to_list_reversed(PairPP sp)
{
	PairP list=0;
	while (index != NodePP(sp))
		::push(list, *++sp);
	return list;
}


PairPPStack::PairPPStack()
{
	size=0;
	base=(PairPP*)GC_::calloc(PAGE_SIZE, sizeof(PairPP));
	if (!base)
	{	size=0;
		return;
	}
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=0;
#endif
}

PairPPStack::~PairPPStack() 
{
	free(base);
#ifdef DEBUG_STACK2
	if (index!=base)
		c_out << "Warning: PairPPStack is still containing "
			<< int(index-base) << " elements on destruction!" << endL;
#endif
}

bool PairPPStack::grow()
{
	PairPP *ob=base;
	base=(PairPP*)GC_::realloc(base, sizeof(PairPP)*size*2);
	if (!base)
	{	base=ob;
		return false;	
	}
	size*=2;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << "NodePStack size is now " << size << endL;
#endif
	return true;
}

bool PairPPStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(s)/log(2));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	PairPP *ob=base;
	base=(PairPP*)GC_::realloc(base, sizeof(PairPP)*s);
	if (!base)
	{	base=ob;
		return false;	
	}
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << "IntStack size is now " << size << endL;
#endif
	return true;
}

#ifdef _DEBUG
	void PairPPStack::push(PairPP val) 
	{	if (index+1 > max)
			if (!grow())
				return;
		*++index=val;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void PairPPStack::pop(PairPP &val) 
	{	if (index>=base) 
		{	val = *index;  *index--=0;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << "Error: PairPPStack underflow on pop(" << val << ')' << endL;
	}
	PairPPStack &PairPPStack::operator<<(const PairPP val)
	{	push(val);
		return *this;
	}
	PairPPStack &PairPPStack::operator>>(PairPP &val)
	{	pop(val);
		return *this;
	}
#endif

wxDataInputStream &PairPPStack::read(wxDataInputStream &istr) 
{
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	int nr;
		istr >> nr;
		*this << nr+(PairP*)argStack.base;
	}
	return istr;
}
wxDataOutputStream &PairPPStack::write(wxDataOutputStream &ostr) 
{
	int siz=index-base;
	ostr << siz;

	PairPP *itemptr=base;
	for (int i=siz;  i>=1;  i--)
	{	int nr=*++itemptr-(PairP*)argStack.base;
		ostr << nr;
	}
	return ostr;
}


VarStack::VarStack()
{
	size=0;
	base=(VarP)GC_::calloc(PAGE_SIZE, sizeof(Var));
	if (!base)
	{	size=0;
		return;
	}
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=0;
#endif
}

VarStack::~VarStack() 
{
	free(base);
#ifdef DEBUG_STACK2
	if (index!=base)
		c_out << "Warning: VarStack is still containing "
			<< int(index-base) << " elements on destruction!" << endL;
#endif
}

bool VarStack::grow()
{
	VarP ob=base;
	base=(VarP)GC_::realloc(base, sizeof(Var)*size*2);
	if (!base)
	{	base=ob;
		return false;	
	}
	size*=2;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << "NodePStack size is now grown to " << size << endL;
#endif
	return true;
}

bool VarStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(s)/log(2));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	VarP ob=base;
	base=VarP(GC_::realloc(base, sizeof(Var)*s));
	if (!base)
	{	base=ob;
		return false;	
	}
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << "VarStack size is now shrinked to " << size << endL;
#endif
	return true;
}

#ifdef _DEBUG
	void VarStack::push(ObjectP object) 
	{	if (index+2 > max)
			if (!grow())
				return;
		++index;
		index->object=object;
		index->val=object->var();
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void VarStack::pop() 
	{
		if (index>base) 
		{	index->object->setvar( index->val );
			index->object=0; index->val=0;
			index--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << "Error: VarStack underflow on pop(" << index->object << ')' << endL;
		if (type_of(index->object)==NTFREE)
			c_out << "Error in VarStack::pop(" << index->object
				<< "): VarStack was damaged during garbge collection" << endL;
		if (type_of(index->val)==NTFREE)
			c_out << "Error in VarStack::pop(" << index->val
				<< "): VarStack was damaged during garbge collection" << endL;
	}
	VarStack &VarStack::operator<<(const ObjectP object)
	{	push(object);
		return *this;
	}

	void VarStack::pop_till(VarP oldSP)	// pop saved vars till old stack pointer
	{	while (index > oldSP)
			pop();
	}
#endif

void VarStack::mark()
{
	for (Var *i=base; i<=index; i++)
		i->mark();
}

void VarStack::markPtr()
{
	for (Var *i=base; i<=index; i++)
		i->markPtr();
}
void VarStack::setPtrs()
{
	int size=index-base;
	VarP itemptr=base;
	for (int i=size;  i>=1;  i--)
	{	itemptr++;
		itemptr->object = (Object*)GC_::intToPtr->find(int(itemptr->object));
		itemptr->val    =   (Node*)GC_::intToPtr->find(int(itemptr->val   ));
	}
}
wxDataInputStream &VarStack::read(wxDataInputStream &istr) 
{
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	ObjectP object;
		NodeP val;
		istr 
			>> (int&)object
			>> (int&)val;
		if (index+2 > max)
			if (!grow())
				return istr;
		++index;
		index->object=object;
		index->val=val;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	return istr;
}
wxDataOutputStream &VarStack::write(wxDataOutputStream &ostr) 
{
	int siz=index-base;
	ostr << siz;

	VarP itemptr=base;
	for (int i=siz;  i>=1;  i--)
	{	itemptr++;
		ostr 
			<< GC_::ptrToInt->find(itemptr->object)
			<< GC_::ptrToInt->find(itemptr->val);
	}
	return ostr;
}


VarPStack::VarPStack()
{
	size=0;
	base=(VarP*)GC_::calloc(PAGE_SIZE, sizeof(VarP));
	if (!base)
	{	size=0;
		return;
	}
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=-1;
#endif
}

VarPStack::~VarPStack() 
{
	free(base);
#ifdef DEBUG_STACK2
	if (index!=base && theApp)
		c_out << "Warning: VarPStack is still containing " 
			<< int(index-base) << " elements on destruction!" << endL;
#endif
}

bool VarPStack::grow()
{
	VarP *ob=base;
	base=(VarP*)GC_::realloc(base, sizeof(VarP)*size*2);
	if (!base)
	{	base=ob;
		return false;	
	}
	size*=2;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << "NodePStack size is now " << size << endL;
#endif
	return true;
}

bool VarPStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(s)/log(2));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	VarP *ob=base;
	base=(VarP*)GC_::realloc(base, sizeof(VarP)*s);
	if (!base)
	{	base=ob;
		return false;	
	}
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << "IntStack size is now " << size << endL;
#endif
	return true;
}

#ifdef _DEBUG
	void VarPStack::push(VarP val) 
	{	if (index+1 > max)
			if (!grow())
				return;
		*++index=val;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void VarPStack::pop(VarP &val) 
	{
		if (index>=base) 
		{	val = *index;	*index--=0;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << "Error: VarPStack underflow on pop(" << val << ')' << endL;
	}
	VarPStack &VarPStack::operator<<(const VarP val)
	{	push(val);
		return *this;
	}
	VarPStack &VarPStack::operator>>(VarP &val)
	{	pop(val);
		return *this;
	}
#endif

wxDataInputStream &VarPStack::read(wxDataInputStream &istr) 
{
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	int nr;
		istr >> nr;
		*this << VarP(nr+VarP(Evaluator::varStack.base));
	}
	return istr;
}
wxDataOutputStream &VarPStack::write(wxDataOutputStream &ostr) 
{
	int siz=index-base;
	ostr << siz;

	VarP *itemptr=base;
	for (int i=siz;  i>=1;  i--)
	{	int nr=*++itemptr-VarP(Evaluator::varStack.base); // special case here for argPStack!!!
		ostr << nr;
	}
	return ostr;
}


ToDoStack::ToDoStack() 
{
	size=0;
	base=(ToDo*)GC_::calloc(PAGE_SIZE, sizeof(ToDo));
	if (!base)
	{	size=0;
		return;
	}
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=0;
#endif
}

ToDoStack::~ToDoStack() 
{
	free(base);
#ifdef DEBUG_STACK2
	if (index != base)
		c_out << "Warning: ToDoStack has still "
			<< int(index-base) << " elements on destruction" << endL;
#endif
}

bool ToDoStack::grow()
{
	ToDo *ob=base;
	base=(ToDo*)GC_::realloc(base, sizeof(ToDo)*size*2);
	if (!base)
	{	base=ob;
		return false;	
	}
	size*=2;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << "ToDoStack size is now grown to " << size << endL;
#endif
	return true;
}

bool ToDoStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(s)/log(2));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	ToDo *ob=base;
	base=(ToDo*)GC_::realloc(base, sizeof(ToDo)*s);
	if (!base)
	{	base=ob;
		return false;	
	}
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << "ToDoStack size is now shrinked to " << size << endL;
#endif
	return true;
}

#ifdef _DEBUG
	void ToDoStack::push(ToDo val) 
	{	if (index+2 > max)
			if (!grow())
				return;
		*++index = val;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void ToDoStack::pop(ToDo &val) 
	{	if (index>=base) 
		{	val= *index;	*index-- = eval_dummy_label;	
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << "Error: ToDoStack underflow on pop(" 
				<< val.mnemonic << ')' << endL;
	}
	ToDoStack &ToDoStack::operator<<(const ToDo val)
	{	push(val);
		return *this;
	}
	ToDoStack &ToDoStack::operator>>(ToDo &val)
	{	pop(val);
		return *this;
	}
#endif

wxDataInputStream &ToDoStack::read(wxDataInputStream &istr) 
{
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	ToDo td;
#ifdef PROC_EVAL
	#ifdef _DEBUG
		istr >> (int&)td.mnemonic;
	#endif
		istr >> td.i;
#else
		istr >> (int&)td.mnemonic;
#endif
		*this << td;
	}
	return istr;
}
wxDataOutputStream &ToDoStack::write(wxDataOutputStream &ostr) 
{
	int siz=index-base;
	ostr << siz;

	ToDo *itemptr=base;
	for (int i=siz;  i>=1;  i--)
	{	itemptr++;
#ifdef PROC_EVAL
	#ifdef _DEBUG
		ostr << (int&)itemptr->mnemonic;
	#endif
		ostr << itemptr->i;
#else
		ostr << (int&)itemptr->mnemonic;
#endif
	}
	return ostr;
}

//#define _DEBUG
ErrorStack::ErrorStack() 
{
	size=0;
	base=(ErrorStruct*)GC_::calloc(PAGE_SIZE, sizeof(ErrorStruct));
	if (!base)
	{	size=0;
		return;
	}
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=0;
#endif
}

ErrorStack::~ErrorStack() 
{
	free(base);
#ifdef DEBUG_STACK2
	if (index != base)
		c_out << "Warning: ErrorStack has still "
			<< int(index-base) << " elements on destruction" << endL;
#endif
}

bool ErrorStack::grow()
{
	ErrorStruct *ob=base;
	base=(ErrorStruct*)GC_::realloc(base, sizeof(ErrorStruct)*size*2);
	if (!base)
	{	base=ob;
		return false;	
	}
	size*=2;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << "ErrorStack size is now grown to " << size << endL;
#endif
	return true;
}

bool ErrorStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(s)/log(2));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	ErrorStruct *ob=base;
	base=(ErrorStruct*)GC_::realloc(base, sizeof(ErrorStruct)*s);
	if (!base)
	{	base=ob;
		return false;	
	}
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << "ErrorStack size is now shrinked to " << size << endL;
#endif
	return true;
}

#ifdef _DEBUG
	void ErrorStack::push(ObjectP fun_, ObjectP ufun_, LineP this_line_) 
	{	if (index+2 > max)
			if (!grow())
				return;
		++index;
		index->fun = fun_;
		index->ufun = ufun_;	
		index->this_line = this_line_;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void ErrorStack::pop() 
	{
		if (index>=base) 
		{	index->fun=index->ufun=0; index->this_line=0;
			index--;  
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << "Error: ErrorStack underflow on pop("
				<< index->fun << ", " << index->ufun << ", " 
				<< index->this_line << ')' << endL;
	}
#endif

void ErrorStack::mark()
{
	for (ErrorStruct *i=base; i<=index; i++)
		i->mark();
}
void ErrorStack::markPtr()
{
	for (ErrorStruct *i=base; i<=index; i++)
		i->markPtr();
}
wxDataInputStream &ErrorStack::read(wxDataInputStream &istr) 
{
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	ErrorStruct e;
		istr 
			>> (int&)e.fun
			>> (int&)e.ufun
			>> (int&)e.this_line;
		push(e.fun, e.ufun, e.this_line);
	}
	return istr;
}
wxDataOutputStream &ErrorStack::write(wxDataOutputStream &ostr) 
{
	int siz=index-base;
	ostr << siz;

	ErrorStructP itemptr=base;
	for (int i=siz;  i>=1;  i--)
	{	itemptr++;
		ostr 
			<< GC_::ptrToInt->find(itemptr->fun      )
			<< GC_::ptrToInt->find(itemptr->ufun     )
			<< GC_::ptrToInt->find(itemptr->this_line);
	}
	return ostr;
}
void ErrorStack::setPtrs()
{
	int siz=index-base;
	ErrorStructP itemptr=base;
	for (int i=siz;  i>=1;  i--)
	{	itemptr++;
		itemptr->fun   = (Object*)GC_::intToPtr->find(int(itemptr->fun ));
		itemptr->ufun  = (Object*)GC_::intToPtr->find(int(itemptr->ufun));
		itemptr->this_line=(Line*)GC_::intToPtr->find(int(itemptr->this_line));
	}
}
