//___________________________________________________________________
//
//                         Class StaticText
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>

#include "Main.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "Eval.h"
#include "Stack.h"
#include "UInt8.h"
#include "Color.h"
#include "Float.h"
#include "Point.h"

#include "Event.h"
#include "StaticText.h"
#include "Frame.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"
#include "Error.h"
#include "Console.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

StaticText::StaticText(StaticTextNodeP staticTextNode_):
	staticTextNode(staticTextNode_)
{
	wxWindow *parent=console;
	if (staticTextNode->parentFrame)
		parent=staticTextNode->parentFrame->frame;
	else if (staticTextNode->parentGraph)
		parent=staticTextNode->parentGraph->graph;
	Create(parent, -1, 
			wxEmptyString, 
			wxDefaultPosition, 
			wxDefaultSize, 
			0, 
			"StaticText");
}

StaticText::StaticText(StaticTextNodeP staticTextNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxString &label, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxString& name):
wxStaticText(parent, id, label, pos, size, style, name),
staticTextNode(staticTextNode_)
{
	staticTextNode->staticText=this;
	SetLabel(label);
}

StaticText::~StaticText()
{
	if (staticTextNode)
		staticTextNode->staticText=0;
}

wxString StaticText::wxstring()
{
	wxString str;
	str << "StaticText " << GetLabel() << ' ' 
		<< '[' << GetName() << "] "
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

wxDataInputStream &StaticText::read(wxDataInputStream &istr) 
{	
	wxString label, name, fontName;
	wxPoint pos;
	wxSize size;
	int style, fontSize, fontStyle, fontWeight;
	unsigned char fr,fg,fb, br,bg,bb;
	
	istr >> label >> name
		>> pos.x >> pos.y >> size.x >> size.y >> style 
		>> fr >> fg >> fb 
		>> br >> bg >> bb 
		>> fontSize >> fontName >> fontStyle >> fontWeight;

	SetLabel(label);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetWindowStyle(style);
	SetForegroundColour(wxColour(fr,fg,fb));
	SetBackgroundColour(wxColour(br,bg,bb));
	wxFont f=GetFont();
	f.SetPointSize(fontSize);
	f.SetFaceName(fontName);
	f.SetStyle(fontStyle);
	f.SetWeight(fontWeight);
	SetFont(f);
	return istr;
}
wxDataOutputStream &StaticText::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetLabel() << GetName() 
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y << int(GetWindowStyle());
	{	wxColour f=GetForegroundColour(), b=GetBackgroundColour();
		ostr
			<< f.Red() << f.Green() << f.Blue() 
			<< b.Red() << b.Green() << b.Blue();
	}
	wxFont f=GetFont();
	ostr << f.GetPointSize() << f.GetFaceName() 
		<< f.GetStyle() << f.GetWeight();
	return ostr;
}

void StaticTextNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void StaticTextNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

StaticTextNode::StaticTextNode(): 
	Node(STATICTEXT), staticText(0), parentFrame(0), parentGraph(0) {}

StaticTextNode::StaticTextNode(FrameNodeP parentFrame_, 
										 GraphNodeP parentGraph_): 
	Node(STATICTEXT), staticText(0), 
	parentFrame(parentFrame_), 
	parentGraph(parentGraph_) 
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void StaticTextNode::free() 
{	
	if (staticText) 
	{	staticText->staticTextNode=0;
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		staticText->Destroy(); 
		staticText=0;
	}
}

wxDataInputStream &StaticTextNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	FrameNodeP parentFrameNode;
		GraphNodeP parentGraphNode;
		istr >> (int&)parentFrameNode >> (int&)parentGraphNode;
		staticText=new StaticText(StaticTextNodeP(this));
		staticText->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}

wxDataOutputStream &StaticTextNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (staticText)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame)
			<< GC_::ptrToInt->find(parentGraph);
		staticText->write(ostr);
	}

	else
		ostr << 0;
	return ostr;
}

void StaticTextNode::setPtrs()
{
	parentFrame=(FrameNode*)GC_::intToPtr->find(int(parentFrame));
	parentGraph=(GraphNode*)GC_::intToPtr->find(int(parentGraph));
}

void StaticTextNode::recreate()
{
	wxWindow *p=0;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	StaticTextP s=staticText, s2=0;
	s2=new StaticText(StaticTextNodeP(this),
			p, -1, s->GetLabel(), 
			s->GetPosition(), s->GetSize(), 
			s->GetWindowStyle(), s->GetName());
	staticText=s2;
	s2->SetForegroundColour(s->GetForegroundColour());
	s2->SetBackgroundColour(s->GetBackgroundColour());
	wxFont f=s->GetFont();
	s2->SetFont(f);
	delete s;
	staticText=s2;
}

StringP StaticTextNode::string()
{
	if (staticText)
		return new String(staticText->wxstring());
	return _Null_Word->name();
}

wxString StaticTextNode::wxstring()
{
	if (staticText)
		return staticText->wxstring();
	return wxEmptyString;
}

NodeP lStaticText(PairPP arg)
{
	wxWindowID id=-1;
	wxWindow *parent=graph;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=theApp->mainFrame->graphNode;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}
	wxString label=wxEmptyString;
	{	StringP l=arg[2]->a_string(2);
		if throwing return Unbound;
		label=l->ptr();
	}
	long style = 0;
	if (exist(3))
	{	int s=arg[3]->a_int(3);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(4))
	{	PairP p=arg[4]->a_int_vector(4);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(5))
	{	PairP s=arg[5]->a_int_vector(5);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	wxString name=_T("StaticText");
	if (exist(6))
	{	StringP n=arg[6]->a_string(6);
		if throwing return Unbound;
		name=n->ptr();
	}
	StaticTextNodeP staticTextNode=new StaticTextNode(
		parentFrameNode,
		parentGraphNode);
	new StaticText(staticTextNode,
			parent, id, label, pos, size, style, name);
	return staticTextNode;
}

NodeP lStaticTextLabel(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	return new String(staticText->staticText->GetLabel());
}

NodeP lStaticTextSetLabel(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	StringP label=arg[2]->a_string(2);
	if throwing return Unbound;
	wxString wxlabel=label->ptr();
	staticText->staticText->SetLabel(wxlabel);
	return Unbound;
}

NodeP lStaticTextSetBackgroundColor(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
	staticText->staticText->SetBackgroundColour(wc);
	return Unbound;
}

NodeP lStaticTextSetColor(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
	staticText->staticText->SetForegroundColour(wc);
	return Unbound;
}

NodeP lStaticTextSetFontSize(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=staticText->staticText->GetFont();
	f.SetPointSize(s);
	staticText->staticText->SetFont(f);
	return Unbound;
}

NodeP lStaticTextSetFontName(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	StringP s=arg[2]->a_string(2);
	if throwing return Unbound;
	wxFont f=staticText->staticText->GetFont();
	f.SetFaceName(s->ptr());
	staticText->staticText->SetFont(f);
	return Unbound;
}

NodeP lStaticTextSetFontStyle(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=staticText->staticText->GetFont();
	f.SetStyle(s);
	staticText->staticText->SetFont(f);
	return Unbound;
}

NodeP lStaticTextSetFontWeight(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=staticText->staticText->GetFont();
	f.SetWeight(s);
	staticText->staticText->SetFont(f);
	return Unbound;
}

NodeP lStaticTextDestroy(PairPP arg)
{
	StaticTextNodeP staticText=arg[1]->a_staticText(1);
	if throwing return Unbound;
	{	FrameNodeP p=staticText->parentFrame;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(staticText));
			p->check_oldyoung(p->children);
		}
	}
	{	GraphNodeP p=staticText->parentGraph;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(staticText));
			p->check_oldyoung(p->children);
		}
	}
	staticText->staticText->Destroy();
	return Unbound;
}
