//___________________________________________________________________
//
//                          Class TextControl
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/textctrl.h>
#include <wx/stattext.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>

#include <math.h>

#include "Main.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "Float.h"
#include "Point.h"
#include "Eval.h"
#include "Stack.h"

#include "Event.h"
#include "TextControl.h"
#include "Frame.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"
#include "Error.h"
#include "Color.h"
#include "Graph.h"
#include "Console.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

//___________________________________________________________________

class MyTextCtrl : public wxTextCtrl
{
public:
	MyTextCtrl(wxWindow *parent, wxWindowID id,
				  const wxString &value = wxEmptyString,
              const wxPoint &pos = wxDefaultPosition,
              const wxSize &size = wxDefaultSize,
              long style = 0,
              const wxValidator& validator = wxDefaultValidator,
              const wxString &name = wxTextCtrlNameStr);



	// MSW sends extra kill focus event
	virtual ~MyTextCtrl()
	{
		if (parent) 
			parent->text = 0;
		parent=0;
	}

	TextControl *parent;

	void OnChar(wxKeyEvent &event);         // pass chars to TextControl
	void OnKeyDown(wxKeyEvent &event);
	void OnKeyUp(wxKeyEvent &event);

private:
	DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(MyTextCtrl,wxTextCtrl)
	EVT_CHAR		(MyTextCtrl::OnChar		)
	EVT_KEY_DOWN(MyTextCtrl::OnKeyDown	)
	EVT_KEY_UP	(MyTextCtrl::OnKeyUp		)
END_EVENT_TABLE()

MyTextCtrl::MyTextCtrl(wxWindow *parent_, wxWindowID id,
							  const wxString &value,
							  const wxPoint &pos, const wxSize &size,
							  long style,
							  const wxValidator& validator,
							  const wxString &name):
wxTextCtrl(parent_, id, value, pos, size, style, validator, name)
{
	parent = (TextControl*)parent_;
}

void MyTextCtrl::OnChar(wxKeyEvent &event)
{
	if (parent) 
		parent->OnChar(event);
	event.Skip();
}

void MyTextCtrl::OnKeyDown(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyDown(event);
	event.Skip();
}

void MyTextCtrl::OnKeyUp(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyUp(event);
	event.Skip();
}
//___________________________________________________________________

BEGIN_EVENT_TABLE(TextControl, wxControl)
	EVT_CHAR			(		TextControl::OnChar		)
	EVT_KEY_DOWN	(		TextControl::OnKeyDown	)
	EVT_KEY_UP		(		TextControl::OnKeyUp		)
	EVT_TEXT			(-1,	TextControl::OnChange	)
	EVT_TEXT_ENTER	(-1,	TextControl::OnEnter		)
END_EVENT_TABLE()

TextControl::TextControl(TextControlNodeP textControlNode_, int style):
label(0), text(0), bestSize(-999, -999),
onChange(0), onEnter(0),
onChar(0), onKeyUp(0), onKeyDown(0),
textControlNode(textControlNode_),
insertMode(true), writeBufferSize(0x1000)
{
	wxWindow *parent=console;
	if (textControlNode->parentFrame)
		parent=textControlNode->parentFrame->frame;
	else if (textControlNode->parentGraph)
		parent=textControlNode->parentGraph->graph;
	create(parent, -1, wxEmptyString, 
			wxDefaultPosition, wxDefaultSize, style,
			wxDefaultValidator, "TextControl");
}

TextControl::TextControl(TextControlNodeP textControlNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxString &value_, 
			PairP onChange_,
			PairP onEnter_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
label(0), text(0), bestSize(-999, -999), value(value_),
onChange(onChange_), onEnter(onEnter_),
onChar(0), onKeyUp(0), onKeyDown(0), 

textControlNode(textControlNode_),
insertMode(true), writeBufferSize(0x1000)
{
	bestSize = wxSize(-999, -999);
	textControlNode->textControl=this;
	textControlNode->style=style;
	create(parent, id, value, pos, size, style, validator, name);
}

bool TextControl::create(wxWindow* parent, wxWindowID id, 
			const wxString &value_, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name)
{
	if (!wxControl::Create(parent, id, pos, size, style|wxNO_BORDER,
								wxDefaultValidator, name))
	  return false;

	wxControl::SetLabel(name);
	wxControl::SetBackgroundColour(parent->GetBackgroundColour());
	wxControl::SetForegroundColour(parent->GetForegroundColour());

	int width = size.GetWidth(), height = size.GetHeight();

	wxSize best_size(DoGetBestSize());
	if (width  == -1) width  = best_size.GetWidth();
	if (height == -1) height = best_size.GetHeight();

	if (name != wxEmptyString)
		label = new wxStaticText(this, id, name, 
			wxPoint(0, 0), wxDefaultSize, 0, "staticText");
	else
		label=0;

	if (label)
		text = new MyTextCtrl(this, id, value_,
			wxPoint(0,0),
			wxSize(width, height-label->GetSize().GetHeight()),
			style, validator, name);
	else
		text = new MyTextCtrl(this, id, value_,
			wxPoint(0,0),
			wxSize(width, height), style, validator, name);

	value=wxEmptyString;

	DoSetSize(pos.x, pos.y, width, height);
	SetBestFittingSize(wxSize(width, height));
	return true;
}

void TextControl::DoSetSize(int x, int y, int width, int height, int sizeFlags)
{
	wxSize bestSize(DoGetBestSize());
	if (width < 0)  width  = bestSize.GetWidth();
	if (height < 0) height = bestSize.GetHeight();

	wxWindow::DoSetSize(x, y, width, height, sizeFlags);

	int labelwidth  = 0;
	int labelheight = 0;
	if (label)
		label->GetSize(&labelwidth, &labelheight);

	if (text)   
		text->SetSize(0, labelheight, 
			width, height-labelheight-2);
}

wxSize TextControl::DoGetBestSize()
{
	if (bestSize.x == -999)
	{	wxTextCtrl *t=new wxTextCtrl((wxWindow*)this, wxID_ANY, GetValue(),
			wxPoint(0,0), wxDefaultSize, GetWindowStyle());
		bestSize=t->GetBestSize();
		delete t;
		if (GetLabel()!=wxEmptyString)
		{	wxStaticText label((wxWindow*)this, wxID_ANY, GetLabel());
			wxSize ls=label.GetBestSize();
			bestSize.y+=ls.y;
			if (ls.x > bestSize.x)
				bestSize.x=ls.x;
		}
	}
	return bestSize;
}

TextControl::~TextControl()
{
	if (textControlNode)
		textControlNode->textControl=0;
	if (label)

	{	delete label;
		label=0;
	}
	if (text)
	{	delete text;
		text=0;
	}

	EventPP out=Event::fifo.out;
	EventP ev;
	while (Event::fifo)
	{	Event::fifo >> ev;
		switch (ev->type)
		{
		case Event::TextControlChange:
		case Event::TextControlEnter:
		case Event::TextControlChar:
		case Event::TextControlKeyDown:
		case Event::TextControlKeyUp:
			if (((EventText*)ev)->textControl==this)
				((EventText*)ev)->textControl=0;
			break;
		default:
			;
		}
	}
	Event::fifo.out=out;
}

void TextControl::SetValue(wxString &value)
{
	text->SetValue(value);
}

wxString TextControl::GetValue() const
{
	if (text)
		return text->GetValue();
	else
		return value;
}

#define TC_ON(NAME)																\
void TextControl::On##NAME(wxKeyEvent &event)							\
{																						\
	Event::fifo << new EventText##NAME(this, event.GetKeyCode());	\
	event.Skip();																	\
}

TC_ON(Char)
TC_ON(KeyDown)
TC_ON(KeyUp)

void TextControl::OnChange(wxCommandEvent &event)
{

	Event::fifo << new EventTextChange(this);
	event.Skip();
}

void TextControl::OnEnter(wxCommandEvent &event)
{
	Event::fifo << new EventTextEnter(this);
	event.Skip();
}

wxString TextControl::wxstring()
{
	wxString str;
	str << "TextControl " << GetValue() << ' ' 
		<< '[' << GetName() << "] ";
	if (onChange)
		str << onChange->string()->ptr() << ' ';
	else
		str << "[] ";
	if (onEnter)
		str << onEnter->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

wxDataInputStream &TextControl::read(wxDataInputStream &istr) 
{	
	wxPoint pos;
	wxSize size;
	wxString value, name, fontName;
	int fontSize, fontStyle, fontWeight, insertionPoint;
	unsigned char fr,fg,fb, br,bg,bb;
	
	istr >> value >> name 
		>> (int&)onChange 
		>> (int&)onEnter
		>> (int&)onChar
		>> (int&)onKeyDown
		>> (int&)onKeyUp
		>> pos.x >> pos.y >> size.x >> size.y
		>> (int&)insertMode >> insertionPoint
		>> fr >> fg >> fb 
		>> br >> bg >> bb 
		>> fontSize >> fontName >> fontStyle >> fontWeight;

	SetValue(value);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	if (!insertMode)
		toggleInsertMode();
	if (text)
		text->SetInsertionPoint(insertionPoint);
	SetForegroundColour(wxColour(fr,fg,fb));
	SetBackgroundColour(wxColour(br,bg,bb));
	wxFont f=GetFont();
	f.SetPointSize(fontSize);
	f.SetFaceName(fontName);
	f.SetStyle(fontStyle);
	f.SetWeight(fontWeight);
	SetFont(f);
	return istr;
}
wxDataOutputStream &TextControl::write(wxDataOutputStream &ostr) 
{
//	flush();
	ostr
		<< GetValue() << GetName() 
		<< GC_::ptrToInt->find(onChange)
		<< GC_::ptrToInt->find(onEnter )
		<< GC_::ptrToInt->find(onChar)
		<< GC_::ptrToInt->find(onKeyDown)
		<< GC_::ptrToInt->find(onKeyUp)
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y
		<< int(insertMode);
	if (text)
		ostr << int(text->GetInsertionPoint());
	else
		ostr << 0;

	{	wxColour 
			f=text->GetForegroundColour(), 
			b=text->GetBackgroundColour();
		ostr
			<< f.Red() << f.Green() << f.Blue() 
			<< b.Red() << b.Green() << b.Blue();
	}
	wxFont f=text->GetFont();
	ostr 
		<< f.GetPointSize() << f.GetFaceName() 
		<< f.GetStyle() << f.GetWeight();
	return ostr;
}

void TextControl::flush()
{
	long from, to;
	if (text)
		text->GetSelection(&from, &to);
#ifdef WIN32
	if (!insertMode)
		to=from+writeBuffer.length();
	if (text)
		text->Replace(from, to, writeBuffer);
#else
	long ip=text->GetInsertionPoint();
	if (from < 0)
		from=ip;
	if (to < 0)
		to=ip;
	if (!insertMode)
	{	to=from+writeBuffer.length();
		if (to > text->GetLastPosition())
			to=text->GetLastPosition();
//		Remove(from, to);
	}
	text->Replace(from, to, writeBuffer);
//	WriteText(writeBuffer);
//	long ipnew=ip+writeBuffer.length();
//	SetInsertionPoint(ipnew);
#endif
	writeBuffer="";
}

void TextControl::toggleInsertMode()
{
	flush();
	insertMode=!insertMode;
#ifdef _WIN32
	keybd_event( VK_INSERT, 0x0, KEYEVENTF_EXTENDEDKEY , 0 );
	keybd_event( VK_INSERT, 0x0, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
#else
	text->EmulateKeyPress(wxKeyEvent(WXK_INSERT));
//	error(NOT_YET_SUPPORTED, new String("Console::toggleInsertMode"));
#endif
	console->dispatchMessages();
}

void TextControl::setInsertMode()
{
	if (!insertMode)
		toggleInsertMode();
}

void TextControl::setOverwriteMode()
{
	if (insertMode)
		toggleInsertMode();
}

NodeP TextControl::Cursor() 
{
	if (!text)
		return new Pair(new Int(0), new Pair(new Int(0),0));

	long pos=text->GetInsertionPoint();
	long x,y;
	text->PositionToXY(pos, &x, &y);
	return new Pair
	(	new Int(x),
		new Pair(new Int(y),0) );
}

NodeP TextControl::setCursor(PairP pos) 
{
	if (!throwing && text) 
	{	flush();
		wxPoint coord;
		coord.x=IntP(pos-> first())->int_;
		coord.y=IntP(pos->second())->int_;

		if (coord.y >= text->GetNumberOfLines())
		{	int end=text->GetLastPosition();
			text->SetInsertionPoint(end);
			long x,y;
			text->PositionToXY(end, &x, &y);
			int missingLines=coord.y-y;
			wxString CRLFs;
			while (missingLines)
			{
#ifdef WIN32
				CRLFs+="\r\n";
#else
				CRLFs+="\n";
#endif
				missingLines--;
			}
			*text << CRLFs;
		}
		if (coord.x > text->GetLineLength(coord.y))
		{	text->SetInsertionPoint(
				text->XYToPosition( text->GetLineLength(coord.y), coord.y ));
			text->WriteText(wxString(
				coord.x+1-text->GetLineLength(coord.y), ' ').c_str());
		}
		text->SetInsertionPoint(text->XYToPosition(coord.x, coord.y));
	}
	return Unbound;
}

bool TextControl::SetBackgroundColour(const wxColour &c)
{
	if (text)
		return text->SetBackgroundColour(c);
	return false;
}

bool TextControl::SetForegroundColour(const wxColour &c)
{
	if (text)
		return text->SetForegroundColour(c);
	return false;
}

bool TextControl::SetFont(const wxFont &f)
{
	if (text)
		return text->SetFont(f);
	return false;
}

//___________________________________________________________________

#define NODE_ON(x)								\
PairP TextControlNode::on##x()				\
{														\
	if (textControl)								\
		return textControl->on##x;				\
	return 0;										\
}														\
														\
void TextControlNode::seton##x(NodeP nd)	\
{														\
	if (textControl)								\
	{	textControl->on##x = PairP(nd);		\
		check_oldyoung(nd);						\
	}													\
}

NODE_ON(Char)
NODE_ON(KeyDown)
NODE_ON(KeyUp)
NODE_ON(Change)

NODE_ON(Enter)

void TextControlNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void TextControlNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

TextControlNode::TextControlNode(): 
	Node(TEXTCONTROL), textControl(0), 
	parentFrame(0), 
	parentGraph(0) 
{}

TextControlNode::TextControlNode(FrameNodeP parentFrame_, 
											GraphNodeP parentGraph_): 
	Node(TEXTCONTROL), textControl(0), 
	parentFrame(parentFrame_),
	parentGraph(parentGraph_) 
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void TextControlNode::free() 
{	
	if (textControl) 
	{	textControl->textControlNode=0;
		textControl->Destroy(); 
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		EventPP out=Event::fifo.out;
		EventP ev;
		while (Event::fifo)
		{	Event::fifo >> ev;
			switch (ev->type)
			{
			case Event::TextControlChange:
			case Event::TextControlEnter:
			case Event::TextControlChar:
			case Event::TextControlKeyDown:
			case Event::TextControlKeyUp:
				if (((EventText*)ev)->textControl==textControl)
					((EventText*)ev)->textControl=0;
				break;
			default:
				;
			}
		}
		Event::fifo.out=out;
		textControl=0;
	}
}
void TextControlNode::inc_gc() 
{	fifo << onChange() << onEnter()
		 << onChar() << onKeyDown() << onKeyUp();
}
int TextControlNode::youngs() 
{	return gmark(onChange())+gmark(onEnter())
		+gmark(onChar())+gmark(onKeyDown())+gmark(onKeyUp());
}
int TextControlNode::youngsPtr() 
{	return gmarkPtr(onChange())+gmarkPtr(onEnter())
		+gmarkPtr(onChar())+gmarkPtr(onKeyDown())+gmarkPtr(onKeyUp());
}
void TextControlNode::check_oldyoungs() 
{	check_oldyoung(onChange());
	check_oldyoung(onEnter());
	check_oldyoung(onChar());
	check_oldyoung(onKeyDown());
	check_oldyoung(onKeyUp());
}

wxDataInputStream &TextControlNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	FrameNodeP parentFrameNode;
		GraphNodeP parentGraphNode;
		istr >> (int&)parentFrameNode >> (int&)parentGraphNode >> style;
		textControl=new TextControl(TextControlNodeP(this), style);
		textControl->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}
wxDataOutputStream &TextControlNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (textControl)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame) 
			<< GC_::ptrToInt->find(parentGraph)
			<< style;
		textControl->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void TextControlNode::setPtrs()
{
	parentFrame=(FrameNode*)GC_::intToPtr->find(int(parentFrame));
	parentGraph=(GraphNode*)GC_::intToPtr->find(int(parentGraph));
	if (textControl)
	{	TextControlP t=textControl;
		setonChange ((Pair*)GC_::intToPtr->find(int(t->onChange )));
		setonEnter  ((Pair*)GC_::intToPtr->find(int(t->onEnter  )));
		setonChar   ((Pair*)GC_::intToPtr->find(int(t->onChar   )));
		setonKeyDown((Pair*)GC_::intToPtr->find(int(t->onKeyDown)));
		setonKeyUp  ((Pair*)GC_::intToPtr->find(int(t->onKeyUp  )));
	}
}

void TextControlNode::recreate()
{
	wxWindow *p=0;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	TextControlP t=textControl, t2=0;
	wxPoint pos=t->GetPosition();
	t2=new TextControl(TextControlNodeP(this),
			p, -1, t->GetValue(), 
			t->onChange, t->onEnter, pos, t->GetSize(), 
			style, wxDefaultValidator, t->GetName());
	textControl=t2;
	setonChange (t->onChange );
	setonEnter  (t->onEnter	 );
	setonChar   (t->onChar   );
	setonKeyDown(t->onKeyDown);
	setonKeyUp  (t->onKeyUp  );
	if (!t->insertMode)
		t2->toggleInsertMode();
	t2->text->SetInsertionPoint(t->text->GetInsertionPoint());
	t2->text->SetForegroundColour(t->text->GetForegroundColour());
	t2->text->SetBackgroundColour(t->text->GetBackgroundColour());
	wxFont f=t->text->GetFont();
	t2->text->SetFont(f);
	delete t;
	textControl=t2;
}

StringP TextControlNode::string()
{
	if (textControl)
		return new String(textControl->wxstring());
	return _Null_Word->name();
}

wxString TextControlNode::wxstring()
{
	if (textControl)
		return textControl->wxstring();
	return wxEmptyString;
}

NodeP lTextControl(PairPP arg)
{
	wxWindowID id=-1;
	wxWindow *parent=graph;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=theApp->mainFrame->graphNode;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}
	wxString name=_T("TextControl");
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	wxString value;
	if (exist(3))
	{	StringP v=arg[3]->a_string(3);
		if throwing return Unbound;
		value=v->ptr();	
	}
	PairP onChange=0;
	if (exist(4))
	{	PairP oc=arg[4]->a_runnable(4);
		if throwing return Unbound;
		onChange=oc;
	}
	PairP onEnter=0;
	if (exist(5))
	{	PairP oe=arg[5]->a_runnable(5);
		if throwing return Unbound;
		onEnter=oe;
	}
	long style = 0;
	if (exist(6))
	{	int s=arg[6]->a_int(6);
		if throwing return Unbound;
		style=s;	

	}
	wxPoint pos=wxDefaultPosition;
	if (exist(7))
	{	PairP p=arg[7]->a_int_vector(7);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(8))
	{	PairP s=arg[8]->a_int_vector(8);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	TextControlNodeP textControlNode=new TextControlNode(
		parentFrameNode, 
		parentGraphNode);
	new TextControl(textControlNode,
			parent, id, value, 
			onChange, onEnter, 
			pos, size, style, wxDefaultValidator, name);
	return textControlNode;
}

NodeP lTextControlValue(PairPP arg)
{
	if (exist(1))
	{	TextControlNodeP textControl=arg[1]->a_textControl(1);
		if throwing return Unbound;
		return new String(textControl->textControl->GetValue());
	}
	else
		return new String(Event::textControlValue);
}

NodeP lTextControlSetValue(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	StringP value=arg[2]->a_string(2);
	if throwing return Unbound;
	textControl->textControl->text->SetValue(value->ptr());
	return Unbound;
}

NodeP lTextControlWrite(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	StringP value=arg[2]->a_string(2);
	if throwing return Unbound;
	textControl->textControl->text->WriteText(value->ptr());
	return Unbound;
}

NodeP lTextControlAppend(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	StringP value=arg[2]->a_string(2);
	if throwing return Unbound;
	textControl->textControl->text->AppendText(value->ptr());
	return Unbound;
}

#define _TC_ON(NAME)									\
NodeP lTextControlOn##NAME(PairPP arg)			\
{															\
	TextControlNodeP textControl=arg[1]->a_textControl(1);\
	if throwing return Unbound;					\
	PairP function=arg[2]->a_runnable(2);		\
	if throwing return Unbound;					\
	textControl->seton##NAME(function);			\
	return Unbound;									\
}

_TC_ON(Change)
_TC_ON(Enter)
_TC_ON(Char)
_TC_ON(KeyDown)
_TC_ON(KeyUp)

NodeP lTextControlSetInsertionPointEnd(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	textControl->textControl->text->SetInsertionPointEnd();
	return Unbound;
}

NodeP lTextControlInsertMode(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	textControl->textControl->setInsertMode();
	return Unbound;
}

NodeP lTextControlOverwriteMode(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	textControl->textControl->setOverwriteMode();
	return Unbound;
}

NodeP lTextControlCursor(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	return textControl->textControl->Cursor();
}

NodeP lTextControlSetCursor(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	PairP pos=arg[2]->a_pos_int_vector(2);
	if throwing return Unbound;
	return textControl->textControl->setCursor(pos);
}

NodeP lTextControlSetBackgroundColor(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
	textControl->textControl->SetBackgroundColour(wc);
	return Unbound;
}


NodeP lTextControlSetColor(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
	textControl->textControl->SetForegroundColour(wc);
	return Unbound;
}

NodeP lTextControlSetFontSize(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=textControl->textControl->GetFont();
	f.SetPointSize(s);
	textControl->textControl->SetFont(f);
	return Unbound;
}

NodeP lTextControlSetFontName(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	StringP s=arg[2]->a_string(2);
	if throwing return Unbound;
	wxFont f=textControl->textControl->GetFont();
	f.SetFaceName(s->ptr());
	textControl->textControl->SetFont(f);
	return Unbound;
}

NodeP lTextControlSetFontStyle(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=textControl->textControl->GetFont();
	f.SetStyle(s);
	textControl->textControl->SetFont(f);
	return Unbound;
}

NodeP lTextControlSetFontWeight(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=textControl->textControl->GetFont();
	f.SetWeight(s);
	textControl->textControl->SetFont(f);
	return Unbound;

}

NodeP lTextControlEnable(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	textControl->textControl->Enable(state!=0);
	return Unbound;
}

NodeP lTextControlDestroy(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	{	FrameNodeP p=textControl->parentFrame;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(textControl));
			p->check_oldyoung(p->children);
		}
	}
	{	GraphNodeP p=textControl->parentGraph;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(textControl));
			p->check_oldyoung(p->children);
		}
	}
	textControl->textControl->Destroy();
	return Unbound;
}
