//___________________________________________________________________
//
//                        Class  ToggleButton
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#ifdef WIN32
	#include <wx/tglbtn.h>
#else
	#include <wx/checkbox.h>
#endif
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>

#include "Main.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "Float.h"
#include "Eval.h"
#include "Stack.h"
#include "Color.h"
#include "Point.h"

#include "Event.h"
#include "ToggleButton.h"
#include "Frame.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"
#include "Error.h"
#include "Graph.h"
#include "Console.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

#ifdef WIN32
BEGIN_EVENT_TABLE(ToggleButton, wxToggleButton)
	EVT_TOGGLEBUTTON	(-1, ToggleButton::OnClick	)
END_EVENT_TABLE()
#else
BEGIN_EVENT_TABLE(ToggleButton, wxCheckBox)
	EVT_CHECKBOX	(-1, ToggleButton::OnClick	)
END_EVENT_TABLE()
#endif

ToggleButton::ToggleButton(ToggleButtonNodeP toggleButtonNode_):
	toggleButtonNode(toggleButtonNode_)
{
	wxWindow *parent=console;
	if (toggleButtonNode->parentFrame)
		parent=toggleButtonNode->parentFrame->frame;
	else if (toggleButtonNode->parentGraph)
		parent=toggleButtonNode->parentGraph->graph;
	Create(parent, -1, 
			wxEmptyString, 
			wxDefaultPosition, 
			wxDefaultSize, 
			0, 
			wxDefaultValidator, 
			"ToggleButton");
}

ToggleButton::ToggleButton(ToggleButtonNodeP toggleButtonNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxString &label,
			PairP onClick_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
#ifdef WIN32
wxToggleButton(parent, id, label,
			pos, size, style,
			validator, name),
#else
wxCheckBox(parent, id, label,
			pos, size, style,
			validator, name),
#endif
onClick(onClick_),
toggleButtonNode(toggleButtonNode_)
{}

ToggleButton::~ToggleButton()
{
	if (toggleButtonNode)
		toggleButtonNode->toggleButton=0;
}

void ToggleButton::OnClick(wxCommandEvent &event)
{
	Event::fifo << new EventToggleButton(this);
}

wxString ToggleButton::wxstring()
{
	wxString str;
	str << "ToggleButton [" << GetLabel() << "] " 
		<< '[' << GetName() << "] ";
	if (onClick)
		str << onClick->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " 
		<< GetValue();
	return str;
}

wxDataInputStream &ToggleButton::read(wxDataInputStream &istr) 
{	
	wxString label, name;
	wxPoint pos;
	wxSize size;
	int style, state;
	
	istr >> label >> name >> (int&)onClick
		>> pos.x >> pos.y >> size.x >> size.y >> style >> state;

	SetLabel(label);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetWindowStyle(style);
	SetValue(state!=0);
	return istr;
}
wxDataOutputStream &ToggleButton::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetLabel() << GetName() 
		<< GC_::ptrToInt->find(onClick)
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y
		<< int(GetWindowStyle()) << int(GetValue());
	return ostr;
}

PairP ToggleButtonNode::onClick()
{
	if (toggleButton)
		return toggleButton->onClick;
	return 0;
}

void ToggleButtonNode::setonClick(NodeP nd) 
{
	if (toggleButton)
	{	toggleButton->onClick = PairP(nd);  check_oldyoung(nd);
	}
}

void ToggleButtonNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void ToggleButtonNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

ToggleButtonNode::ToggleButtonNode(): 
	Node(TOGGLEBUTTON), toggleButton(0), 
	parentFrame(0), 
	parentGraph(0) 
{}

ToggleButtonNode::ToggleButtonNode(FrameNodeP parentFrame_, 
											  GraphNodeP parentGraph_): 
	Node(TOGGLEBUTTON), toggleButton(0), 
	parentFrame(parentFrame_),
	parentGraph(parentGraph_)
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void ToggleButtonNode::free() 
{	
	if (toggleButton) 
	{	toggleButton->toggleButtonNode=0;
		toggleButton->Destroy(); 
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		EventPP out=Event::fifo.out;
		EventP ev;
		while (Event::fifo)
		{	Event::fifo >> ev;
			switch (ev->type)
			{
			case Event::tToggleButton:
				if (((EventToggleButton*)ev)->toggleButton==toggleButton)
					((EventToggleButton*)ev)->toggleButton=0;
				break;
			default:
				;
			}
		}
		Event::fifo.out=out;
		toggleButton=0;
	}
}
void ToggleButtonNode::inc_gc() {fifo << onClick();}
int ToggleButtonNode::youngs() {return gmark(onClick());}
int ToggleButtonNode::youngsPtr() {return gmarkPtr(onClick());}
void ToggleButtonNode::check_oldyoungs() {check_oldyoung(onClick());}

wxDataInputStream &ToggleButtonNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	FrameNodeP parentFrameNode;
		GraphNodeP parentGraphNode;
		istr >> (int&)parentFrameNode >> (int&)parentGraphNode;
		toggleButton=new ToggleButton(ToggleButtonNodeP(this));
		toggleButton->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}

wxDataOutputStream &ToggleButtonNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (toggleButton)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame) 
			<< GC_::ptrToInt->find(parentGraph);
		toggleButton->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void ToggleButtonNode::setPtrs()
{
	parentFrame=(FrameNode*)GC_::intToPtr->find(int(parentFrame));
	parentGraph=(GraphNode*)GC_::intToPtr->find(int(parentGraph));
	if (toggleButton)
	{	setonClick((Pair*)GC_::intToPtr->find(int(toggleButton->onClick)));
	}
}

void ToggleButtonNode::recreate()
{
	wxWindow *p=0;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	ToggleButtonP t=toggleButton, t2=0;
	wxPoint pos=t->GetPosition();
	t2=new ToggleButton(ToggleButtonNodeP(this),
			p, -1, t->GetLabel(), 
			t->onClick, pos, t->GetSize(), t->GetWindowStyle(), 
			wxDefaultValidator, t->GetName());
	t2->SetValue(t->GetValue());
	toggleButton=t2;
	setonClick(t->onClick);
	delete t;
	toggleButton=t2;
}

StringP ToggleButtonNode::string()
{
	if (toggleButton)
		return new String(toggleButton->wxstring());
	return _Null_Word->name();
}

wxString ToggleButtonNode::wxstring()
{
	if (toggleButton)
		return toggleButton->wxstring();
	return wxEmptyString;
}

NodeP lToggleButton(PairPP arg)
{
	wxWindowID id=-1;
	wxWindow *parent=graph;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=theApp->mainFrame->graphNode;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}
	wxString label=wxEmptyString;
	{	StringP l=arg[2]->a_string(2);
		if throwing return Unbound;
		label=l->ptr();
	}
	PairP onClick=0;
	{	PairP oc=arg[3]->a_runnable(3);
		if throwing return Unbound;
		onClick=oc;
	}
	long style = 0;
	if (exist(4))
	{	int s=arg[4]->a_int(4);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(5))
	{	PairP p=arg[5]->a_int_vector(5);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(6))
	{	PairP s=arg[6]->a_int_vector(6);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	wxString name=label;
	if (exist(7))
	{	StringP n=arg[7]->a_string(7);
		if throwing return Unbound;
		name=n->ptr();
	}
	ToggleButtonNodeP toggleButtonNode=new ToggleButtonNode(
		parentFrameNode,
		parentGraphNode);
	ToggleButtonP toggleButton=new ToggleButton(toggleButtonNode,
			parent, id, label, 
			onClick, pos, size, style, wxDefaultValidator, name);
	toggleButtonNode->toggleButton=toggleButton;
	return toggleButtonNode;
}

NodeP lToggleButtonOnClick(PairPP arg)
{
	ToggleButtonNodeP toggleButton=arg[1]->a_toggleButton(1);
	if throwing return Unbound;
	PairP function=arg[2]->a_runnable(2);
	if throwing return Unbound;
	toggleButton->setonClick(function);
	return Unbound;
}

NodeP lToggleButtonEnable(PairPP arg)
{
	ToggleButtonNodeP toggleButton=arg[1]->a_toggleButton(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	toggleButton->toggleButton->Enable(state!=0);
	return Unbound;
}

NodeP lToggleButtonDestroy(PairPP arg)
{
	ToggleButtonNodeP toggleButton=arg[1]->a_toggleButton(1);
	if throwing return Unbound;
	toggleButton->toggleButton->Destroy();
	return Unbound;
}

NodeP lToggleButtonSetValue(PairPP arg)
{
	ToggleButtonNodeP toggleButton=arg[1]->a_toggleButton(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	toggleButton->toggleButton->SetValue(state!=0);
	return Unbound;
}

NodeP lToggleButtonValue(PairPP arg)
{
	ToggleButtonNodeP toggleButton=arg[1]->a_toggleButton(1);
	if throwing return Unbound;
	bool state=toggleButton->toggleButton->GetValue();
	return Node::boolnode(state);
}

