//___________________________________________________________________
//
//                          Class Turtle
//
//                           Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/app.h>
#include <wx/textctrl.h>
#include <wx/listctrl.h>
#include <wx/glcanvas.h>
#include <wx/frame.h>
#include <wx/splitter.h>
#include <wx/html/helpctrl.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include <math.h>

#ifndef WIN32
	#include <setjmp.h>
	#define FT_FREETYPE_H </usr/include/freetype2/freetype/freetype.h>
	#define FT_ERRORS_H </usr/include/freetype2/freetype/fterrors.h>
	#define FT_CONFIG_CONFIG_H </usr/include/freetype2/freetype/config/ftconfig.h>
	#define FT_CONFIG_OPTIONS_H </usr/include/freetype2/freetype/config/ftoption.h>
	#define FT_CONFIG_STANDARD_LIBRARY_H </usr/include/freetype2/freetype/config/ftstdlib.h>
	#define FT_MODULE_ERRORS_H </usr/include/freetype2/freetype/ftmoderr.h>
	#define FT_BEGIN_HEADER
	#define FT_END_HEADER
	#define FT_ERROR_DEFINITIONS_H </usr/include/freetype2/freetype/fterrdef.h>
	#define FT_TYPES_H </usr/include/freetype2/freetype/fttypes.h>
	#define FT_SYSTEM_H </usr/include/freetype2/freetype/ftsystem.h>
	#define FT_IMAGE_H </usr/include/freetype2/freetype/ftimage.h>
	#define FT_GLYPH_H </usr/include/freetype2/freetype/ftglyph.h>
#endif
#include <FTGL/FTGLPolygonFont.h>

#include "Main.h"

#include "String.h"

#include "Int.h"
#include "UInt8.h"
#include "Float.h"
#include "List.h"

#include "Error.h"
#include "Init.h"
#include "File.h"


#include "Point.h"
#include "Color.h"
#include "Pen.h"
#include "Screen.h"
#include "FTGLFontManager.h"
#include "StringToPtr.h"
#include "Font.h"
#include "Turtle.h"
#include "Mouse.h"

#include "Bitmap.h"
#include "Recorder.h"

#include "Console.h"
#include "Graph.h"
#include "Vars.h"
#include "Calls.h"
#include "Splitter.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Shadow.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
	#define realloc(m, len) _realloc_dbg((m), (len), _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

//___________________________________________________________________
//
//                          Portable Part
//___________________________________________________________________

// This part should be portable and reentrant, 
//	because it is the core of the turtle graphics.
// It only uses standard functions, which appear in the Machine-specific Part.
// If one machine does not have that function, write a dummy. 
// You could also leave the machine-specific code in a separate file, 
// but I think it is better if it is here altogether, because it is ONE class.
//
// You may ask, why it is necessary to have such a big class. 
// First, the task drawing wrapping Turtle Graphics simply is not so easy. 
// Second, I didn`t have so much time as to split it up once more, 
//	but simplifying is still wanted.
// Third, I didn`t want to leave it global, 
//	because I really really wanted it to be reentrant.
// And in global functions reentrancy also certainly is possible 
//	but not as easy as in a Class.
//___________________________________________________________________

PointFiFo::~PointFiFo() {if (allocated) free(base);}

PointFiFo &PointFiFo::operator<<(const Point val) 
{
	if (overflow()) return *this;
	*in=val;
	if (in==max) in=base; else in++;
	if (in==out) expand();
	return *this;
}

PointFiFo &PointFiFo::operator>>(Point &val) 
{
	val = *out;
	if (out==max) out=base; else out++;
	return *this;
}

PointFiFo &PointFiFo::operator=(PointFiFo &b)
{
	PointP bout=b.out;
	in=out=base;
	Point p;
	while (b)
	{	b >> p;
		*this << p;
	}
	b.out=bout;
	return *this;
}

void PointFiFo::expand()	// transfer is untested!!!!
{
	PointP newbase = (Point*) GC_::calloc(1, sizeof(Point)*(size+expand_size));

	if (newbase==PointP(0))	// no room to increase FiFo
	{	c_out << "; Warning: Not enough memory to expand PointFiFo!" << endL;
		_overflow=true;
	}
	else
	{	PointP i=newbase;

		if (out)			// transfer old FiFo to new FiFo
		{	*this >> *i++;
			while (in!=out) 
				*this >> *i++;
		}				
		if (allocated) 
			free(base);

		allocated=true;
		size += expand_size;
		base=newbase;   max= &base[size-1];   in=i;   out= base;
	}
}

//___________________________________________________________________

ColorFiFo::~ColorFiFo() {if (allocated) free(base);}

ColorFiFo &ColorFiFo::operator<<(const Color val) 
{
	if (overflow()) return *this;
	*in=val;
	if (in==max) in=base; else in++;
	if (in==out) expand();
	return *this;
}

ColorFiFo &ColorFiFo::operator>>(Color &val) 
{
	val = *out;
	if (out==max) out=base; else out++;
	return *this;
}

ColorFiFo &ColorFiFo::operator=(ColorFiFo &b)
{
	ColorP bout=b.out;
	in=out=base;
	Color p;
	while (b)
	{	b >> p;
		*this << p;
	}
	b.out=bout;
	return *this;
}

void ColorFiFo::expand()	// transfer is untested!!!!
{
	ColorP newbase = (Color*) GC_::calloc(1, sizeof(Color)*(size+expand_size));

	if (newbase==ColorP(0))	// no room to increase FiFo
	{	c_out << "; Warning: Not enough memory to expand ColorFiFo!" << endL;
		_overflow=true;
	}
	else
	{	ColorP i=newbase;

		if (out)			// transfer old FiFo to new FiFo
		{	*this >> *i++;
			while (in!=out) 
				*this >> *i++;
		}				
		if (allocated) 
			free(base);

		allocated=true;
		size += expand_size;
		base=newbase;   max= &base[size-1];   in=i;   out= base;
	}
}
//___________________________________________________________________

FTGLFontManager *ftglFontManager=0;

FTGLFontManager::FTGLFontManager()
{
	fonts=new StringToPtr(31);
}
FTGLFontManager::~FTGLFontManager()
{
	clear();
	delete fonts;
}

void FTGLFontManager::clear()
{
	StringToPtr::PtrStringPairP obj;
	for (int i=0; i < fonts->size; i++)
	{	for (obj=fonts->data[i];  obj;  obj=obj->tail)
			delete FTFontP(obj->p);
	}
	delete fonts;
	fonts=new StringToPtr(31);
}

FTFontP FTGLFontManager::getFont(const char *filename, int size)
{
	wxString uniqueName=filename;
	uniqueName << size;
	FTFontP result = FTFontP(fonts->find(uniqueName));

	if (result)
		return result;

	wxString fullname = 
		wxString(environment.fonts)+filename+".ttf";

	if (!File::exists(fullname))
	{	error(FILE_ERROR, new String(
			wxString("Font ")+fullname+" doesn't exist"));
		return 0;
	}

	FTFontP font = new FTGLPolygonFont(fullname.c_str());

	if (!font)
	{	error(FILE_ERROR, new String(
			wxString("Font ")+fullname+" failed to open"));

		delete font;
		return 0;
	}
	if (!font->FaceSize( size))
	{	error(FILE_ERROR, new Pair(
			new String(
				wxString("Font ")+filename+" failed to set size "), 
			new Int(size)));
		delete font;
		return 0;
	}
	fonts->insert(uniqueName, font);
	return font;
}
//___________________________________________________________________

FTFont *Font_::ftfont=0;

Font_::Font_()
{	
	sizex=20; sizey=20; weight=400; alignx=aligny=0; 
#ifdef WIN32
	name="Times";
#else
	name="luxirr";
#endif
	ftfont=0;
}

Font_::~Font_()
{
}

void Font_::load()
{
	ftfont=ftglFontManager->getFont(name.c_str(), int(sizex));
}
//___________________________________________________________________

ShadowObject::~ShadowObject()
{
	if (planes) free(planes);
	if (points) free(points);
	if (colors) free(colors);
	if (texs) free(texs);
}

void ShadowObject::checkPoints(int i)
{
	if (ptsize <= i)
	{	ptsize+=dptsize;
		points=(Point*)realloc(points, ptsize*sizeof(Point));
		colors=(Color*)realloc(colors, ptsize*sizeof(Color));
		texs=(Point*)realloc(texs, ptsize*sizeof(Point));
		dptsize*=2;
	}
}
void ShadowObject::checkPlanes(int i)
{
	if (plsize <= i)
	{	int oplsize=plsize;
		plsize+=dplsize;
		planes=(sPlane*)GC_::realloc(planes, plsize*sizeof(sPlane));
		for (int i=oplsize; i<plsize;  i++)
		{	planes[i].neigh[0]=planes[i].neigh[1]=planes[i].neigh[2]=0;
		}
		dplsize*=2;
	}
}

void ShadowObject::clear()
{
	if (planes) free(planes);
	if (points) free(points);
	if (colors) free(colors);
	if (texs) free(texs);
	planes=0;
	points=0;
	colors=0;
	texs=0;
	ptsize=plsize=nPoints=nPlanes=0;
	dptsize=dplsize=256;
}

int ShadowObject::read(char *st)
{
	FILE *file;
	int i,j;

	file = fopen(st, "r");
	if (!file) return FALSE;
	//points
	fscanf(file, "%d", &(nPoints));
	for (i=1;  i<=nPoints;  i++)
	{	checkPoints(i);
		fscanf(file, "%f", &(points[i].x));
		fscanf(file, "%f", &(points[i].y));
		fscanf(file, "%f", &(points[i].z));
	}
	//planes
	fscanf(file, "%d", &(nPlanes));
	for (i=0;  i<nPlanes;  i++)
	{	checkPlanes(i);
		fscanf(file, "%d", &(planes[i].p[0]));
		fscanf(file, "%d", &(planes[i].p[1]));
		fscanf(file, "%d", &(planes[i].p[2]));

		for (j=0;  j<3;  j++)
		{	fscanf(file, "%f", &(planes[i].normals[j].x));
			fscanf(file, "%f", &(planes[i].normals[j].y));
			fscanf(file, "%f", &(planes[i].normals[j].z));
		}
	}
	return TRUE;
}

void ShadowObject::setConnectivity()
{
	// connectivity procedure - based on Gamasutra's article

	int p1i, p2i, p1j, p2j;
	int P1i, P2i, P1j, P2j;
	int i,j,ki,kj;

	for (i=0;  i<nPlanes-1;  i++)
		for (j=i+1;  j<nPlanes;  j++)
			for (ki=0;  ki<3;  ki++)
				if (!planes[i].neigh[ki])
				{	for (kj=0;  kj<3;  kj++)
					{	p1i=ki;
						p1j=kj;
						p2i=(ki+1)%3;
						p2j=(kj+1)%3;

						p1i=planes[i].p[p1i];
						p2i=planes[i].p[p2i];
						p1j=planes[j].p[p1j];
						p2j=planes[j].p[p2j];

						P1i=((p1i+p2i)-abs(p1i-p2i))/2;
						P2i=((p1i+p2i)+abs(p1i-p2i))/2;
						P1j=((p1j+p2j)-abs(p1j-p2j))/2;
						P2j=((p1j+p2j)+abs(p1j-p2j))/2;

						if ((P1i==P1j) && (P2i==P2j))	//they are neighbours
						{	planes[i].neigh[ki] = j+1;	  
							planes[j].neigh[kj] = i+1;	  
						}
					}
				}
}

void ShadowObject::calcPlane(sPlane *plane)
{
	// function for computing a plane equation given 3 points
	Point v[4];
	int i;

	for (i=0;  i<3;  i++)
	{	v[i+1].x = points[plane->p[i]].x;
		v[i+1].y = points[plane->p[i]].y;
		v[i+1].z = points[plane->p[i]].z;
	}
	plane->PlaneEq.a = 
		v[1].y*(v[2].z-v[3].z) + 
		v[2].y*(v[3].z-v[1].z) + 
		v[3].y*(v[1].z-v[2].z);
	plane->PlaneEq.b = 
		v[1].z*(v[2].x-v[3].x) + 
		v[2].z*(v[3].x-v[1].x) + 
		v[3].z*(v[1].x-v[2].x);
	plane->PlaneEq.c = 
		v[1].x*(v[2].y-v[3].y) + 
		v[2].x*(v[3].y-v[1].y) + 
		v[3].x*(v[1].y-v[2].y);
	plane->PlaneEq.d =-( 
		v[1].x*(v[2].y*v[3].z - v[3].y*v[2].z) +
		v[2].x*(v[3].y*v[1].z - v[1].y*v[3].z) +
		v[3].x*(v[1].y*v[2].z - v[2].y*v[1].z) );
}

void ShadowObject::draw()
{
	int i, j;

	glBegin(GL_TRIANGLES);
	for (i=0;  i<nPlanes;  i++)
	{	for (j=0;  j<3;  j++)
		{
			glColor4ubv((unsigned char*)&colors[planes[i].p[j]]);
			glNormal3dv((double*)&planes[i].normals[j]);
			glVertex3dv((double*)&points[planes[i].p[j]]);
		}
	}
	glEnd();
}

void ShadowObject::castShadows(Point lp, double lp4)
{
	const double f=100;
	int i, j, k, jj;
	int p1, p2;
	Point v1, v2;
	double	side;

	//set visual parameter
	for (i=0;  i<nPlanes;  i++)
	{	// check to see if light is in front or behind the plane (face plane)
		side= planes[i].PlaneEq.a*lp.x+
				planes[i].PlaneEq.b*lp.y+
				planes[i].PlaneEq.c*lp.z+
				planes[i].PlaneEq.d*lp4;
		if (side >0) 
			planes[i].visible = TRUE;
		else 
			planes[i].visible = FALSE;
	}
	glDisable(GL_LIGHTING);
	glDepthMask(GL_FALSE);
	glDepthFunc(GL_LEQUAL);

	glEnable(GL_STENCIL_TEST);
	glColorMask(0, 0, 0, 0);
	glStencilFunc(GL_ALWAYS, 1, 0xffffffff);

	for (int l=1;  l<=2;  l++)	// loop to avoid strange artefacts
	{
		// first pass, stencil operation decreases stencil value
		glFrontFace(GL_CCW);
		glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
		for (i=0;  i<nPlanes;  i++)
		{	if (planes[i].visible)
				for (j=0;  j<3;  j++)
				{	k = planes[i].neigh[j];
					if ((!k) || (!planes[k-1].visible))
					{	// here we have an edge, we must draw a polygon
						p1 = planes[i].p[j];
						jj = (j+1)%3;
						p2 = planes[i].p[jj];

						//calculate the length of the vector
						v1.x = (points[p1].x - lp.x)*f;
						v1.y = (points[p1].y - lp.y)*f;
						v1.z = (points[p1].z - lp.z)*f;

						v2.x = (points[p2].x - lp.x)*f;
						v2.y = (points[p2].y - lp.y)*f;
						v2.z = (points[p2].z - lp.z)*f;
						
						//draw the polygon
						glBegin(GL_TRIANGLE_STRIP);
							glVertex3d(	points[p1].x,
											points[p1].y,
											points[p1].z);
							glVertex3d(	points[p1].x + v1.x,
											points[p1].y + v1.y,
											points[p1].z + v1.z);

							glVertex3d(	points[p2].x,
											points[p2].y,
											points[p2].z);
							glVertex3d(	points[p2].x + v2.x,
											points[p2].y + v2.y,
											points[p2].z + v2.z);
						glEnd();
					}
				}
		}

		// second pass, stencil operation increases stencil value
		glFrontFace(GL_CW);
		glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
		for (i=0;  i<nPlanes;  i++)
		{	if (planes[i].visible)
				for (j=0;  j<3;  j++)
				{	k = planes[i].neigh[j];
					if ((!k) || (!planes[k-1].visible))
					{	// here we have an edge, we must draw a polygon
						p1 = planes[i].p[j];
						jj = (j+1)%3;
						p2 = planes[i].p[jj];

						//calculate the length of the vector
						v1.x = (points[p1].x - lp.x)*f;
						v1.y = (points[p1].y - lp.y)*f;
						v1.z = (points[p1].z - lp.z)*f;

						v2.x = (points[p2].x - lp.x)*f;
						v2.y = (points[p2].y - lp.y)*f;
						v2.z = (points[p2].z - lp.z)*f;
						
						//draw the polygon
						glBegin(GL_TRIANGLE_STRIP);
							glVertex3d(	points[p1].x,
											points[p1].y,
											points[p1].z);
							glVertex3d(	points[p1].x + v1.x,
											points[p1].y + v1.y,
											points[p1].z + v1.z);

							glVertex3d(	points[p2].x,
											points[p2].y,
											points[p2].z);
							glVertex3d(	points[p2].x + v2.x,
											points[p2].y + v2.y,
											points[p2].z + v2.z);
						glEnd();
					}
				}
		}
		lp*=1.05;
	}

	glFrontFace(GL_CCW);
	glColorMask(1, 1, 1, 1);

	//draw a shadowing rectangle covering the entire screen
	glColor4ubv((unsigned char*)&TURTLE->shadowcolor);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glStencilFunc(GL_NOTEQUAL, 0, 0xffffffff);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
	glPushMatrix();
	glLoadIdentity();
	glBegin(GL_TRIANGLE_STRIP);
		glVertex3f(-2000, 2000,1);
		glVertex3f(-2000,-2000,1);
		glVertex3f( 2000, 2000,1);
		glVertex3f( 2000,-2000,1);
	glEnd();
	glPopMatrix();
	glDisable(GL_BLEND);

	glDepthFunc(GL_LEQUAL);
	glDepthMask(GL_TRUE);
	glEnable(GL_LIGHTING);
	glDisable(GL_STENCIL_TEST);
	glShadeModel(GL_SMOOTH);
}
//___________________________________________________________________

Turtle::Turtle(bool makeOld)
{
	size=30; half_bottom=size/3; 
	shown=direct=true;
	cylinderLines=outOfBounds=inEraseMode=false;
	mode=wrapmode;
	heading=wrapDelta=0;

	polyMode=polyShown=tessMode=surfaceMode=false;
	polyPointNr=tessPointNr=0;
	texpos=0;
	tessWindingRule=0;
	surfaceShown=tessShown=false;
	surfaceColumnNr=0;
	shadows=false;
	shadowcolor=Color(0,0,0,0.6);
//	refCount=0;
	sinCos();
	reset3d();
}

Turtle::~Turtle()
{
}

Turtle &Turtle::operator=(const Turtle &b) 
{
	x=b.x; y=b.y; z=b.z;
	pen=b.pen; font=b.font;
	shown=b.shown; outOfBounds=outOfBounds; direct=b.direct;
	polyMode=b.polyMode;
	polyShown=b.polyShown;
	tessMode=b.tessMode;
	surfaceMode=b.surfaceMode;
	cylinderLines=b.cylinderLines;
	mode=b.mode; heading=b.heading; 
	shoulder=b.shoulder; head=b.head;
	size=b.size;	half_bottom=b.half_bottom; 
	_sin=b._sin; _cos=b._cos; 
	u=b.u;
	hbase=b.hbase;
	surfaceShown=b.surfaceShown;
	tessShown=b.tessShown;
	surfaceColumnNr=b.surfaceColumnNr;
	return *this;
}

Turtle &Turtle::operator=(const Point &b)
{
	x=b.x; y=b.y; z=b.z; 
	return *this;
}

void Turtle::create(Recorder *recorder_, Screen_ *screen_)
{
	recorder=recorder_;
	screen=screen_;
	pen.size_rotated=pen.size;//*screen->scale;
	pen.create(pen.color);
}

void Turtle::destroy()
{
	pen.destroy();
}

void Turtle::reset3d()
{
	u=Matrix(Point(1,0,0),Point(0,1,0),Point(0,0,1));
	hbase=u;
}

void Turtle::prepareToDraw(void)
{
	if (theApp->mainFrame)
	{	if (theApp->mainFrame->inGraphicsMode) 
			return;
		else if (console)
		{	console->dispatchMessages();
			theApp->mainFrame->splitScreen();
		}
	}
}


void Turtle::clearScreen() 
{
	screen->clear();
	if (direct)
		mySwapBuffers();
}


void Turtle::setScreenColor(Color c)
{
	c=Pen::realcolor(c);
	screen->setColor(c);
}

void Turtle::setColor(Color c)
{
	draw();
	if (!inEraseMode) 
	{	pen.setColor(c);
	}
	draw();
}

void Turtle::plainXOrPen()
{
	static unsigned char white[]={0xff,0xff,0xff,0xff};
	glColor4ubv(white);
	glEnable(GL_COLOR_LOGIC_OP);
	glLogicOp(GL_INVERT);

}

void Turtle::setPenSize(double sizeX, double sizeY) 
{
	pen.size.x=sizeX;
	pen.size.y=sizeY;
	pen.size.z=0;
	pen.size_rotated=pen.size;//*screen->scale;

//	double w[2];
//	glGetDoublev(GL_LINE_WIDTH_RANGE, w);
//	pen.wide=(pen.size.x > w[1]);
	pen.wide=(pen.size.x > 0);
		
	pen.setColor(inEraseMode ?   screen->color  :  pen.color);
}

void Turtle::setPenMode(Pen::WinMode newmode) 
{
	int rop2_mode, newpc;
	pen.mode=newmode;
	inEraseMode=newmode==Pen::WIN_ERASE;
	
	if (newmode==Pen::WIN_REVERSE)
		rop2_mode=Pen::XOR_PEN;
	else
		rop2_mode=Pen::COPY_PEN;

	if (rop2_mode==Pen::XOR_PEN)
	{	glEnable(GL_COLOR_LOGIC_OP);

		glLogicOp(GL_INVERT);
	}else
		glDisable(GL_COLOR_LOGIC_OP);
	
	if (newmode==Pen::WIN_ERASE)
		newpc=screen->color;
	else
		newpc=pen.color;
	
	pen.setColor(newpc);
}

void Turtle::penErase()   {setPenMode(Pen::WIN_ERASE  );}
void Turtle::penDown()    {setPenMode(Pen::WIN_DOWN   );}
void Turtle::penReverse() {setPenMode(Pen::WIN_REVERSE);}

void Turtle::restorePen() 

{
	if (in(screen->Min, screen->Max))
		moveTo();
	setPenMode(pen.mode);
	pen.setColor(inEraseMode ?   screen->color  :  pen.color);
}

NodeP Turtle::getNodePenMode()
{
	if (inEraseMode)
		return new String("erase");
	GLboolean logic=glIsEnabled(GL_LOGIC_OP);
	if (logic==GL_TRUE)
	{	GLint mode;
		glGetIntegerv(GL_LOGIC_OP_MODE, &mode);
		if (mode==GL_XOR)
			return new String("reverse");
	}
	return new String("paint");
}

void Turtle::moveTo() {moveTo(*this);}

void Turtle::moveTo(Point target) 
{
	*this=target;
	if (recorder->recording)
		*recorder << new R_MoveTo(*this,*recorder,target);
}
void Turtle::lineTo(Point target)
{
	if (pen.down) 
	{	if (polyMode)
		{	if (polyPointNr==0)
			{	polypt1=target;
				polyc1=pen.color;
				polyPointNr++;
				if (screen->textured) 
					texpt1=texpos;
			}else if (polyPointNr==1)
			{	polypt2=target;
				polyv1=polypt2-polypt1;
				polyc2=pen.color;
				polyPointNr++;
				if (screen->textured) 
					texpt2=texpos;
			}else if (polyPointNr==2)
			{	
				Point 
					polypt3=target,
					&v1=polyv1,
					v2=polypt3-polypt2;
				normal=v1.X(v2);
				double norm=normal.Norm();

				if (fabs(norm) > Float::epsilon)
				{	normal=normal/norm;				
				}
				glBegin(GL_POLYGON);
				if (fabs(norm) > Float::epsilon)
				{	glNormal3dv((double*)&normal);
				}else
				{/*	error(BAD_DATA, 
						new String("polygon normal nullvector"));
					polymode=false;
					return;
				*/}
				if (screen->textured) glTexCoord2dv(doubleP(&texpt1));
				glColor4ubv((unsigned char*)&polyc1);
				glNormal3dv((double*)&normal);
				glVertex3dv(doubleP(&polypt1));
				if (screen->textured) glTexCoord2dv(doubleP(&texpt2));
				glColor4ubv((unsigned char*)&polyc2);
				glNormal3dv((double*)&normal);
				glVertex3dv(doubleP(&polypt2));
				if (screen->textured) glTexCoord2dv(doubleP(&texpos));
				glColor4ubv((unsigned char*)&pen.color);
				glNormal3dv((double*)&normal);
				glVertex3dv(doubleP(&polypt3));
				
				if (recorder->recording)
					*recorder << new R_PolyStart(*this, normal, polypt3, texpos);

				if (shadows)
				{	ShadowObject *so=graph->shadowObject;
					int &i=so->nPoints, &j=so->nPlanes;
					so->checkPoints(i);
					so->points[i]=polypt1;
					i++;
					so->checkPoints(i);
					so->points[i]=polypt2;
					i++;
					so->checkPoints(i);
					so->points[i]=polypt3;
					i++;
					so->checkPlanes(j);
					so->planes[j].p[0]=i-3;
					so->planes[j].p[1]=i-2;
					so->planes[j].p[2]=i-1;
					for (int k=0;  k<3;  k++)
						so->planes[j].normals[k]=normal;
					j++;
					i++;
					polypt2=polypt3;
				}
				polyPointNr++;
			}else
			{	if (screen->textured) 
					glTexCoord2dv(doubleP(&texpos));
				glNormal3dv((double*)&normal);
				glVertex3dv(doubleP(&target));
				if (shadows)
				{	ShadowObject *so=graph->shadowObject;
					int &i=so->nPoints, &j=so->nPlanes;
					so->checkPoints(i);
					so->points[i]=polypt1;
					i++;
					so->checkPoints(i);
					so->points[i]=polypt2;
					i++;
					so->checkPoints(i);
					so->points[i]=target;
					i++;
					so->checkPlanes(j);
					so->planes[j].p[0]=i-3;
					so->planes[j].p[1]=i-2;
					so->planes[j].p[2]=i-1;
					for (int k=0;  k<3;  k++)
						so->planes[j].normals[k]=normal;
					j++;
					i++;
					polypt1=polypt2;
					polypt2=target;
				}
				if (recorder->recording)
					*recorder << new R_PolyPoint(target, texpos);
				*this=target;
			}
		}else if (tessMode)
		{	(*tessFiFo) << target;
			(*tessFiFo) << texpos;
			(*tessFiFo) << Point(*(double*)&pen.color,0,0);
			tessPointNr++;
		}else if (surfaceMode)
		{	(*surfaceFiFo) << target;
			(*surfaceFiFo) << texpos;
			(*colorFiFo) << pen.color;
		}else
		{	if (screen->perspective)
			{	if (cylinderLines)
				{	glPushMatrix();
					glTranslated(x,y,z);

					Point v=target-(*this);
				//*
					double xz, theta, phi;

					xz=::sqrt(Float::sqr(v.x)+Float::sqr(v.z));
					if (fabs(xz) < Float::epsilon
					 && fabs(v.y) < Float::epsilon)
						theta=0;
					else

						theta=atan2(xz, v.y)/Float::deg2rad;

					if (fabs(v.x) < Float::epsilon
					 && fabs(v.z) < Float::epsilon)
						phi=0;
					else
						phi=atan2(v.x, v.z)/Float::deg2rad;

					glRotated(phi,0,1,0);
					glRotated(theta-90,1,0,0);
				/*/
					glRotated(roll(),0,1,0);
					glRotated(pitch()-90,1,0,0);
					glRotated(getHeading(),0,1,0);
				//*/
					double height=v.Norm();
						GLUquadricObj *q=gluNewQuadric();
						gluCylinder(q,
							pen.size.x/2,pen.size.x/2, 
							height, int(pen.size.x/4+10),1);
						gluDeleteQuadric(q);
					glPopMatrix();
				}else
				{	if (pen.wide)
					{	glPushMatrix();
						glTranslated(x,y,z);

						Point v=target-(*this),
							ortho=u.x*pen.size.x/2;
						Point
							p1=-ortho,
							p2=v-ortho,
							p3=v+ortho,
							p4=+ortho;

						glBegin(GL_QUADS);
						Point n=-u.z;
						glNormal3dv((double*)&n);
						glVertex3dv((double*)&p1);
						glVertex3dv((double*)&p2);
						glVertex3dv((double*)&p3);
						glVertex3dv((double*)&p4);
						glEnd();

//						glLineWidth(1);
//						glBegin(GL_LINES);
//						glVertex3dv((double*)&p1);
//						glVertex3dv((double*)&p2);
//						glVertex3dv((double*)&p3);
//						glVertex3dv((double*)&p4);
//						glEnd();
						glPopMatrix();
					}
					else
					{	glBegin(GL_LINES);
						glVertex3dv((double*)this);
						glVertex3dv((double*)&target);
						glEnd();
					}
				}
				if (pen.size.x > 0 && pen.roundLineEnds)
				{	glPushMatrix();
					glTranslated(x,y,z);
					if (cylinderLines)
					{	GLUquadricObj *q=gluNewQuadric();
						gluSphere(q, pen.size.x/2, 
							int(pen.size.x/8+8), int(pen.size.x/8+8));
//						gluDeleteQuadric(q);
						glPopMatrix();

						glPushMatrix();
						glTranslated(target.x,target.y,target.z);
	
//						q=gluNewQuadric();
						gluSphere(q, pen.size.x/2, 
							int(pen.size.x/8+8), int(pen.size.x/8+8));
						gluDeleteQuadric(q);
					}
					else
					{	double r=roll(), p=pitch(), h=-getHeading();

						glRotated(r,0,1,0);
						glRotated(p,1,0,0);
						glRotated(h,0,0,1);
					
						GLUquadricObj *q=gluNewQuadric();
						if (screen->textured)
							gluQuadricTexture(q, GL_TRUE);
						gluPartialDisk(q, 0, pen.size.x/2, 
							int(3*pen.size.x), 1, 90, 180);
						glPopMatrix();

						glPushMatrix();
						glTranslated(target.x,target.y,target.z);
						glRotated(r,0,1,0);
						glRotated(p,1,0,0);
						glRotated(h,0,0,1);
					
						gluPartialDisk(q, 0, pen.size.x/2, 
							int(3*pen.size.x), 1, -90, 180);

						gluDeleteQuadric(q);
					}
					glPopMatrix();
				}
				if (cylinderLines)
				{	if (recorder->recording)
					{	*recorder << new R_LineTo3d(*this, *recorder, target);
			//			mySwapBuffers();
					}
				}else
					if (recorder->recording)
					{	*recorder << new R_LineTo(*this, *recorder, target);
			//			mySwapBuffers();
					}
			}
			else // !screen->perspective
			{	if (pen.wide)
				{	Point ortho=Point(target.y-this->y, this->x-target.x, 0);
					ortho=(ortho/ortho.Norm()/2)*pen.size.x;
					Point
						p1=Point(*this)+ortho,
						p2=target+ortho,
						p3=target-ortho,
						p4=Point(*this)-ortho;

					glBegin(GL_QUADS);
					glVertex3dv((double*)&p1);
					glVertex3dv((double*)&p2);
					glVertex3dv((double*)&p3);
					glVertex3dv((double*)&p4);
					glEnd();
					glLineWidth(1);
					glBegin(GL_LINES);
					glVertex3dv((double*)&p1);
					glVertex3dv((double*)&p2);
					glVertex3dv((double*)&p3);
					glVertex3dv((double*)&p4);
					glEnd();
				}else
				{	glBegin(GL_LINES);
					glVertex3dv((double*)this);
					glVertex3dv((double*)&target);
					glEnd();
				}
				if (pen.size.x > 0 && pen.roundLineEnds)
				{	glPushMatrix();
					glTranslated(x,y,z);
					double startangle=
						-atan2(target.y-y, target.x-x)/Float::deg2rad;

						GLUquadricObj *q=gluNewQuadric();
						gluPartialDisk(q, 0, 
							pen.size.x/2, int(3*pen.size.x/2), 1, startangle+180, 180);
						gluDeleteQuadric(q);
					glPopMatrix();

					glPushMatrix();
					glTranslated(target.x,target.y,target.z);
						q=gluNewQuadric();
						gluPartialDisk(q, 0, 
							pen.size.x/2, int(3*pen.size.x/2), 1, startangle, 180);
						gluDeleteQuadric(q);
					glPopMatrix();
				}
				if (recorder->recording)
				{	*recorder << new R_LineTo(*this, *recorder, target);
		//			mySwapBuffers();
				}
			}
		}
	}
	else
	{	if (recorder->recording)
			*recorder << new R_MoveTo(*this, *recorder, target);
	}
	Float::reset_FPU();
}

void Turtle::moveTo(double x, double y) {moveTo(Point(x,y,z));}
void Turtle::lineTo(double x, double y) {lineTo(Point(x,y,z));}

void Turtle::polyStart()
{
	prepareToDraw();
	draw();
	polyShown=shown;
	shown=false;
	polyMode=true;
	polyPointNr=0;
}

void Turtle::polyEnd()
{
	polyMode=false;
	glEnd();
	shown=polyShown;
	draw();
	if (recorder->recording)
		*recorder << new R_PolyEnd();
}

int Turtle::tess_nPoints, Turtle::tess_prim;
Point Turtle::tessNormal;

void CALLBACK Turtle::tessglBegin(GLenum prim)
{	glBegin(prim);
	glNormal3dv((double*)&tessNormal);
	if (TURTLE->shadows)
	{	tess_nPoints=graph->shadowObject->nPoints;
		tess_prim=prim;
	}
	if (RECORDER.recording)
		RECORDER << new R_glBegin(prim, tessNormal);
}
void CALLBACK Turtle::tessglVertex(void *data)
{	PointP p=PointP(data), t=p+1;
	ColorP c=(ColorP)(p+2);
	glColor4ubv((GLubyte*)c);
	glTexCoord2dv(doubleP(t));
	glVertex3dv(doubleP(p));
/*	if (c->r==0 && c->b==0)
	{	c_out 
		<< p->x << ' '
		<< p->y << ' '
		<< p->z << ' '
		<< (int)c->r << ' ' 
		<< (int)c->g << ' ' 
		<< (int)c->b << ' ' 
		<< (int)c->a << endL;
	}
*/
	if (TURTLE->shadows)
	{	ShadowObject *so=graph->shadowObject;
		int &i=so->nPoints, &j=so->nPlanes;
		so->checkPoints(i);
		so->points[i]=*p;
		i++;
	}
	if (RECORDER.recording)
		RECORDER << new R_tessPoint(*p, *t, *c);
}
void CALLBACK Turtle::tessglEnd()
{	glEnd();
	if (TURTLE->shadows)
	{	ShadowObject *so=graph->shadowObject;
		int i=tess_nPoints, &j=so->nPlanes;
		switch (tess_prim)
		{
		case GL_TRIANGLE_STRIP:
			while (i < so->nPoints-2)
			{	int k;
				so->checkPlanes(j);
				so->planes[j].p[0]=i;
				so->planes[j].p[1]=i+1;
				so->planes[j].p[2]=i+2;
				for (k=0;  k<3;  k++)
					so->planes[j].normals[k]=tessNormal;
				j++;
				i++;
				if (i >= so->nPoints-2)
					break;
				so->checkPlanes(j);
				so->planes[j].p[0]=i+1;
				so->planes[j].p[1]=i;
				so->planes[j].p[2]=i+2;
				for (k=0;  k<3;  k++)
					so->planes[j].normals[k]=tessNormal;
				j++;
				i++;
			}
			break;
		case GL_TRIANGLE_FAN:
			{	int i0=tess_nPoints;
				while (i < so->nPoints-2)
				{	so->checkPlanes(j);
					so->planes[j].p[0]=i0;
					so->planes[j].p[1]=i+1;
					so->planes[j].p[2]=i+2;
					for (int k=0;  k<3;  k++)
						so->planes[j].normals[k]=tessNormal;
					j++;
					i++;
				}
			}
			break;
		case GL_TRIANGLES:
			{	while (i < so->nPoints-2)
				{	so->checkPlanes(j);
					so->planes[j].p[0]=i;
					so->planes[j].p[1]=i+1;
					so->planes[j].p[2]=i+2;
					for (int k=0;  k<3;  k++)
						so->planes[j].normals[k]=tessNormal;
					j++;
					i+=3;
				}
			}
			break;
		default:
			i=i;
		}
	}
	if (RECORDER.recording)
		RECORDER << new R_PolyEnd();
}

void CALLBACK Turtle::tessglCombine(PointP p, 
												PointPointColorP d[4], 
												GLfloat w[4], void **out)
{
//c_out << " x0=" << d[0]->p.x << "  y0=" << d[0]->p.y << "  z0=" << d[0]->p.z << " w0=" << w[0] << endL;
//c_out << " x1=" << d[1]->p.x << "  y1=" << d[1]->p.y << "  z1=" << d[1]->p.z << " w1=" << w[1] << endL;
//c_out << " x2=" << d[2]->p.x << "  y2=" << d[2]->p.y << "  z2=" << d[2]->p.z << " w2=" << w[2] << endL;
//c_out << " x3=" << d[3]->p.x << "  y3=" << d[3]->p.y << "  z3=" << d[3]->p.z << " w3=" << w[3] << endL;
//c_out << endL;

//c_out << "tx0=" << d[0]->t.x << " ty0=" << d[0]->t.y << " w0=" << w[0] << endL;
//c_out << "tx1=" << d[1]->t.x << " ty1=" << d[1]->t.y << " w1=" << w[1] << endL;
//c_out << "tx2=" << d[2]->t.x << " ty2=" << d[2]->t.y << " w2=" << w[2] << endL;
//c_out << "tx3=" << d[3]->t.x << " ty3=" << d[3]->t.y << " w3=" << w[3] << endL;
//c_out << endL;

//c_out << " r=" << int(d[0]->c.r) << "  g=" << int(d[0]->c.g) << "  b=" << int(d[0]->c.b) << " a=" << int(d[0]->c.a) << endL;

	(*TURTLE->tessFiFo) << (*p);
	Point tex;
	Color c;
	if (d[3])
	{	c.r=(unsigned char)(w[0]*d[0]->c.r + w[1]*d[1]->c.r + w[2]*d[2]->c.r + w[3]*d[3]->c.r);
		c.g=(unsigned char)(w[0]*d[0]->c.g + w[1]*d[1]->c.g + w[2]*d[2]->c.g + w[3]*d[3]->c.g);
		c.b=(unsigned char)(w[0]*d[0]->c.b + w[1]*d[1]->c.b + w[2]*d[2]->c.b + w[3]*d[3]->c.b);
		c.a=(unsigned char)(w[0]*d[0]->c.a + w[1]*d[1]->c.a + w[2]*d[2]->c.a + w[3]*d[3]->c.a);
	}else if (d[2])
	{	c.r=(unsigned char)(w[0]*d[0]->c.r + w[1]*d[1]->c.r + w[2]*d[2]->c.r);
		c.g=(unsigned char)(w[0]*d[0]->c.g + w[1]*d[1]->c.g + w[2]*d[2]->c.g);
		c.b=(unsigned char)(w[0]*d[0]->c.b + w[1]*d[1]->c.b + w[2]*d[2]->c.b);
		c.a=(unsigned char)(w[0]*d[0]->c.a + w[1]*d[1]->c.a + w[2]*d[2]->c.a);
	}else if (d[1])
	{	c.r=(unsigned char)(w[0]*d[0]->c.r + w[1]*d[1]->c.r);
		c.g=(unsigned char)(w[0]*d[0]->c.g + w[1]*d[1]->c.g);

		c.b=(unsigned char)(w[0]*d[0]->c.b + w[1]*d[1]->c.b);
		c.a=(unsigned char)(w[0]*d[0]->c.a + w[1]*d[1]->c.a);
	}else if (d[0])
	{	c.r=(unsigned char)(w[0]*d[0]->c.r);
		c.g=(unsigned char)(w[0]*d[0]->c.g);
		c.b=(unsigned char)(w[0]*d[0]->c.b);
		c.a=(unsigned char)(w[0]*d[0]->c.a);
	}
	if (d[3])
	{	tex.x=(w[0]*d[0]->t.x + w[1]*d[1]->t.x + w[2]*d[2]->t.x + w[3]*d[3]->t.x);
		tex.y=(w[0]*d[0]->t.y + w[1]*d[1]->t.y + w[2]*d[2]->t.y + w[3]*d[3]->t.y);
	}else if (d[2])
	{	tex.x=(w[0]*d[0]->t.x + w[1]*d[1]->t.x + w[2]*d[2]->t.x);
		tex.y=(w[0]*d[0]->t.y + w[1]*d[1]->t.y + w[2]*d[2]->t.y);
	}else if (d[1])
	{	tex.x=(w[0]*d[0]->t.x + w[1]*d[1]->t.x);
		tex.y=(w[0]*d[0]->t.y + w[1]*d[1]->t.y);
	}else if (d[0])
	{	tex.x=(w[0]*d[0]->t.x);
		tex.y=(w[0]*d[0]->t.y);
	}
	tex.z=0;
//c_out << "tx=" << tex.x << " ty=" << tex.y << endL;
//c_out << endL;
	(*TURTLE->tessFiFo) << tex;
	(*TURTLE->tessFiFo) << Point(*(double*)&c,0,0);
	if (TURTLE->tessFiFo->in-3 >= TURTLE->tessFiFo->base)
		*out = TURTLE->tessFiFo->in-3; 
	else
		*out = TURTLE->tessFiFo->max-3; 
}

void CALLBACK Turtle::tessglError(GLenum errorCode)
{
   const char *estring;

   estring = (char*)gluErrorString(errorCode);
   c_out << "Tessellation Error: " << estring << endL;
}

void Turtle::tessStart()
{
	prepareToDraw();
	draw();
	tessShown=shown;
	shown=false;
	tessMode=true;
	tessFiFo=new PointFiFo(0x1000);
/*	(*tessFiFo) << *this;
	(*tessFiFo) << Point(*(double*)&pen.color,0,0);
*/	tessPointNr=0;

	tess=gluNewTess();
	gluTessCallback(tess, GLU_TESS_BEGIN, (void (CALLBACK *) ())tessglBegin);  
	gluTessCallback(tess, GLU_TESS_COMBINE, (void (CALLBACK *) ())tessglCombine);
	gluTessCallback(tess, GLU_TESS_VERTEX, (void (CALLBACK *) ())tessglVertex);  
	gluTessCallback(tess, GLU_TESS_END, tessglEnd);  
	gluTessCallback(tess, GLU_TESS_ERROR, (void (CALLBACK *) ())tessglError);

	gluTessProperty(tess, GLU_TESS_WINDING_RULE, tessWindingRule);
	gluTessBeginPolygon(tess, 0);
}

void Turtle::tessDraw()
{
	if (tessPointNr < 3)
	{	error(BAD_DATA, new String(

			"Need at least three points for tesselation!"));
		return;
	}
	PointP p1,p2,p3;
	p1=tessFiFo->out++;
	tessFiFo->out++;
	tessFiFo->out++;
	p2=tessFiFo->out++;
	tessFiFo->out++;
	tessFiFo->out++;
	p3=tessFiFo->out++;
	tessFiFo->out++;
	tessFiFo->out++;

	Point v1=*p2-*p1, v2=*p3-*p2;
	tessNormal=v1.X(v2);
	double norm=tessNormal.Norm();

	if (norm > Float::epsilon)
		tessNormal=tessNormal/norm;				
	
	gluTessBeginContour(tess);
	if (norm > Float::epsilon)
		gluTessNormal(tess, tessNormal.x,tessNormal.y,tessNormal.z);
	
	gluTessVertex(tess, doubleP(p1), doubleP(p1));
	gluTessVertex(tess, doubleP(p2), doubleP(p2));
	gluTessVertex(tess, doubleP(p3), doubleP(p3));

	while (tessFiFo->out < tessFiFo->in)
	{	PointP p=tessFiFo->out++;	tessFiFo->out++;	tessFiFo->out++;
		gluTessVertex(tess, doubleP(p), doubleP(p));
	}
}

void Turtle::tessContour()
{
	if (!tessMode)
		return;

	tessDraw();
	gluTessEndContour(tess); 
}

void Turtle::tessEnd()
{
	if (!tessMode)
		return;

	tessDraw();
	tessMode=false;
	gluTessEndContour(tess); 
	gluTessEndPolygon(tess); 
	gluDeleteTess(tess);
	delete tessFiFo;
	shown=tessShown;
	draw();
}

void Turtle::surfaceStart()
{
	prepareToDraw();
	draw();
	surfaceShown=shown;
	shown=false;
	surfaceMode=true;
	surfaceFiFo=new PointFiFo(0x1000);
	surfaceFiFo2=new PointFiFo(0x1000);
	surfaceFiFo3=new PointFiFo(0x1000);
	surfaceFiFo00=new PointFiFo(0x1000);
	surfaceFiFo01=new PointFiFo(0x1000);
	surfaceFiFo02=new PointFiFo(0x1000);
	colorFiFo=new ColorFiFo(0x1000);
	colorFiFo2=new ColorFiFo(0x1000);
	colorFiFo00=new ColorFiFo(0x1000);
	colorFiFo01=new ColorFiFo(0x1000);
	surfaceColumnNr=0;
}

void Turtle::surfaceColumn()
{
	if (!surfaceMode)
		return;

	if (shadows)
	{	surfaceColumnSh();
		return;
	}

	PointFiFo *h=surfaceFiFo;
	surfaceFiFo=surfaceFiFo2;
	surfaceFiFo2=surfaceFiFo3;
	surfaceFiFo3=h;

	ColorFiFo *hc=colorFiFo;
	colorFiFo=colorFiFo2;
	colorFiFo2=hc;

	if (surfaceColumnNr==0)
	{	*surfaceFiFo00=*surfaceFiFo3;
		*colorFiFo00=*colorFiFo2;
	}else if (surfaceColumnNr==1)
	{	*surfaceFiFo01=*surfaceFiFo3;
		*colorFiFo01=*colorFiFo2;
	}else if (surfaceColumnNr==2)
	{	*surfaceFiFo02=*surfaceFiFo3;
	}
	if (surfaceColumnNr >= 2)
	{	glBegin(GL_QUAD_STRIP);

		Point p1,t1, p2,t2, p3,t3, p4,t4, o1,o3, dummy;
		(*surfaceFiFo3) >> p1 >> t1 >> p3 >> t3;
		(*surfaceFiFo2) >> p2 >> t2 >> p4 >> t4;
		(*surfaceFiFo ) >> o1 >> dummy >> o3 >> dummy;
		Color c1,c2,c3,c4;
		(*colorFiFo2) >> c1 >> c3;
		(*colorFiFo ) >> c2 >> c4;

		Point normal=(p1-p2).X(p2-p4);
		double norm=normal.Norm();
		if (norm > Float::epsilon)
		{	normal/=norm;
			glNormal3dv((double*)&normal);
		}
		glColor4ubv((unsigned char*)&c1);
		glTexCoord2dv(doubleP(&t1));
		glVertex3dv(doubleP(&p1));

		Point onormal=(p2-o1).X(p2-p4);
		double onorm=onormal.Norm();
		if (onorm > Float::epsilon)
		{	onormal/=onorm;
			glNormal3dv((double*)&onormal);
		}
		glColor4ubv((unsigned char*)&c2);
		glTexCoord2dv(doubleP(&t2));
		glVertex3dv(doubleP(&p2));

		normal=(p3-p1).X(p3-p4);
		norm=normal.Norm();
		if (norm > Float::epsilon)
		{	normal/=norm;
			glNormal3dv((double*)&normal);
		}
		glColor4ubv((unsigned char*)&c3);
		glTexCoord2dv(doubleP(&t3));
		glVertex3dv(doubleP(&p3));

		onormal=(p4-p2).X(p4-o3);
		onorm=onormal.Norm();
		if (onorm > Float::epsilon)
		{	onormal/=onorm;
			glNormal3dv((double*)&onormal);
		}
		glColor4ubv((unsigned char*)&c4);
		glTexCoord2dv(doubleP(&t4));
		glVertex3dv(doubleP(&p4));
		if (recorder->recording)
			*recorder << new R_QuadStripStart(normal,onormal, 
				c1,p1,t1, c2,p2,t2, c3,p3,t3, c4,p4,t4);

		c1=c3;	p1=p3;	t1=t3;
		c2=c4;	p2=p4;	t2=t4;
		o1=o3;

		while (*surfaceFiFo && *surfaceFiFo2)
		{	(*surfaceFiFo3) >> p3 >> t3;
			(*surfaceFiFo2) >> p4 >> t4;
			(*surfaceFiFo ) >> o3 >> dummy;
			(*colorFiFo2) >> c3;
			(*colorFiFo ) >> c4;

			normal=(p3-p1).X(p3-p4);
			norm=normal.Norm();
			if (norm > Float::epsilon)
			{	normal/=norm;
				glNormal3dv((double*)&normal);
			}
			glColor4ubv((unsigned char*)&c3);
			glTexCoord2dv(doubleP(&t3));
			glVertex3dv(doubleP(&p3));

			Point onormal;
			onormal=(p4-p2).X(p4-o3);
			double onorm=onormal.Norm();
			if (onorm > Float::epsilon)
			{	onormal/=onorm;
				glNormal3dv((double*)&onormal);
			}
			glColor4ubv((unsigned char*)&c4);
			glTexCoord2dv(doubleP(&t4));
			glVertex3dv(doubleP(&p4));

			if (recorder->recording)
				*recorder << new R_QuadStrip(
					normal,onormal, c3,p3,t3, c4,p4,t4);
			c1=c3;	p1=p3;	t1=t3;
			c2=c4;	p2=p4;	t2=t4;
			o1=o3;
		}
		glEnd();
		if (recorder->recording)
			*recorder << new R_QuadStripEnd();
		surfaceFiFo3->reset_out();
		surfaceFiFo2->reset_out();
		surfaceFiFo->reset();
		colorFiFo2->reset_out();
		colorFiFo->reset();
//mySwapBuffers();
	}
	surfaceColumnNr++;
}

void Turtle::surfaceColumnSh()
{
	if (!surfaceMode)
		return;

	PointFiFo *h=surfaceFiFo;
	surfaceFiFo=surfaceFiFo2;
	surfaceFiFo2=surfaceFiFo3;
	surfaceFiFo3=h;

	ColorFiFo *hc=colorFiFo;
	colorFiFo=colorFiFo2;
	colorFiFo2=hc;

	if (surfaceColumnNr==0)
	{	*surfaceFiFo00=*surfaceFiFo3;
		*colorFiFo00=*colorFiFo2;
	}else if (surfaceColumnNr==1)
	{	*surfaceFiFo01=*surfaceFiFo3;
		*colorFiFo01=*colorFiFo2;
	}else if (surfaceColumnNr==2)
	{	*surfaceFiFo02=*surfaceFiFo3;
	}
	if (surfaceColumnNr >= 2)
	{	glBegin(GL_QUAD_STRIP);

		Point p1,t1, p2,t2, p3,t3, p4,t4, o1,o3, dummy;
		(*surfaceFiFo3) >> p1 >> t1 >> p3 >> t3;
		(*surfaceFiFo2) >> p2 >> t2 >> p4 >> t4;
		(*surfaceFiFo ) >> o1 >> dummy >> o3 >> dummy;
		Color c1,c2,c3,c4;
		(*colorFiFo2) >> c1 >> c3;
		(*colorFiFo ) >> c2 >> c4;

		Point normal2=(p1-p2).X(p2-p4);
		double norm=normal2.Norm();
		if (norm > Float::epsilon)
		{	normal2/=norm;
			glNormal3dv((double*)&normal2);
		}
		glColor4ubv((unsigned char*)&c1);
		glTexCoord2dv(doubleP(&t1));
		glVertex3dv(doubleP(&p1));
		ShadowObject *so=graph->shadowObject;
		int i=so->nPoints,
			j=so->nPlanes;
		so->checkPoints(i);
		so->points[i]=p1;
		so->colors[i]=c1;
		so->texs[i]=t1;
		i++;

		Point onormal2=(p2-o1).X(p2-p4);
		double onorm=onormal2.Norm();
		if (onorm > Float::epsilon)
		{	onormal2/=onorm;
			glNormal3dv((double*)&onormal2);
		}
		glColor4ubv((unsigned char*)&c2);
		glTexCoord2dv(doubleP(&t2));
		glVertex3dv(doubleP(&p2));
		so->checkPoints(i);
		so->points[i]=p2;
		so->colors[i]=c2;
		so->texs[i]=t2;
		i++;

		Point normal=(p3-p1).X(p3-p4);
		norm=normal.Norm();
		if (norm > Float::epsilon)
		{	normal/=norm;
			glNormal3dv((double*)&normal);
		}
		glColor4ubv((unsigned char*)&c3);
		glTexCoord2dv(doubleP(&t3));
		glVertex3dv(doubleP(&p3));
		so->checkPoints(i);
		so->points[i]=p3;
		so->colors[i]=c3;
		so->texs[i]=t3;
		i++;

		Point onormal=(p4-p2).X(p4-o3);
		onorm=onormal.Norm();
		if (onorm > Float::epsilon)
		{	onormal/=onorm;
			glNormal3dv((double*)&onormal);
		}
		glColor4ubv((unsigned char*)&c4);
		glTexCoord2dv(doubleP(&t4));
		glVertex3dv(doubleP(&p4));
		so->checkPoints(i);
		so->points[i]=p4;
		so->colors[i]=c4;
		so->texs[i]=t4;
		i++;

		so->checkPlanes(j);
		so->planes[j].p[2]=i-1;
		so->planes[j].p[1]=i-2;
		so->planes[j].p[0]=i-3;
		so->planes[j].normals[2]=onormal;
		so->planes[j].normals[1]=normal;
		so->planes[j].normals[0]=onormal2;
		so->planes[j].neigh[0]=j+2;
		so->planes[j].neigh[1]=j+4;
		j++;

		so->checkPlanes(j);
		so->planes[j].p[2]=i-4;
		so->planes[j].p[1]=i-3;
		so->planes[j].p[0]=i-2;
		so->planes[j].normals[2]=normal2;
		so->planes[j].normals[1]=onormal2;
		so->planes[j].normals[0]=normal;
		so->planes[j].neigh[0]=j;
		j++;

		if (recorder->recording)
			*recorder << new R_QuadStripStart(normal,onormal, 
				c1,p1,t1, c2,p2,t2, c3,p3,t3, c4,p4,t4);

		c1=c3;	p1=p3;	t1=t3;
		c2=c4;	p2=p4;	t2=t4;
		o1=o3;
		normal2=normal;
		onormal2=onormal;

		while (*surfaceFiFo && *surfaceFiFo2)
		{	(*surfaceFiFo3) >> p3 >> t3;
			(*surfaceFiFo2) >> p4 >> t4;
			(*surfaceFiFo ) >> o3 >> dummy;
			(*colorFiFo2) >> c3;
			(*colorFiFo ) >> c4;

			normal=(p3-p1).X(p3-p4);
			norm=normal.Norm();
			if (norm > Float::epsilon)
			{	normal/=norm;
				glNormal3dv((double*)&normal);
			}
			glColor4ubv((unsigned char*)&c3);
			glTexCoord2dv(doubleP(&t3));
			glVertex3dv(doubleP(&p3));
			so->checkPoints(i);
			so->points[i]=p3;
			so->colors[i]=c3;
			so->texs[i]=t3;
			i++;

			Point onormal;
			onormal=(p4-p2).X(p4-o3);
			double onorm=onormal.Norm();
			if (onorm > Float::epsilon)
			{	onormal/=onorm;
				glNormal3dv((double*)&onormal);
			}
			glColor4ubv((unsigned char*)&c4);
			glTexCoord2dv(doubleP(&t4));
			glVertex3dv(doubleP(&p4));
			so->checkPoints(i);
			so->points[i]=p4;
			so->colors[i]=c4;
			so->texs[i]=t4;
			i++;

			so->checkPlanes(j);
			so->planes[j].p[2]=i-1;
			so->planes[j].p[1]=i-2;
			so->planes[j].p[0]=i-3;
			so->planes[j].normals[2]=onormal;
			so->planes[j].normals[1]=normal;
			so->planes[j].normals[0]=onormal2;
			so->planes[j].neigh[0]=j+2;
			if (*surfaceFiFo)
				so->planes[j].neigh[1]=j+4;
			j++;

			so->checkPlanes(j);
			so->planes[j].p[2]=i-4;
			so->planes[j].p[1]=i-3;
			so->planes[j].p[0]=i-2;
			so->planes[j].normals[2]=normal2;
			so->planes[j].normals[1]=onormal2;
			so->planes[j].normals[0]=normal;
			so->planes[j].neigh[0]=j;
			so->planes[j].neigh[1]=j-2;
			j++;

			if (recorder->recording)
				*recorder << new R_QuadStrip(
					normal,onormal, c3,p3,t3, c4,p4,t4);
			c1=c3;	p1=p3;	t1=t3;
			c2=c4;	p2=p4;	t2=t4;
			o1=o3;
			normal2=normal;
			onormal2=onormal;
		}
		glEnd();
		if (recorder->recording)
			*recorder << new R_QuadStripEnd();
		surfaceFiFo3->reset_out();
		surfaceFiFo2->reset_out();
		surfaceFiFo->reset();
		colorFiFo2->reset_out();
		colorFiFo->reset();

		so->nPoints=i;
		so->nPlanes=j;
//mySwapBuffers();
	}
	surfaceColumnNr++;
}

void Turtle::surfaceEnd()
{
	if (!surfaceMode)
		return;

	delete surfaceFiFo;
	delete surfaceFiFo2;
	delete surfaceFiFo3;
	delete colorFiFo;
	delete colorFiFo2;

	surfaceFiFo2=surfaceFiFo02;
	surfaceFiFo3=surfaceFiFo01;
	surfaceFiFo=surfaceFiFo00;

	colorFiFo2=colorFiFo01;
	colorFiFo=colorFiFo00;

	if (shadows)
		surfaceColumnSh();
	else
		surfaceColumn();

	surfaceMode=false;
	delete surfaceFiFo00;
	delete surfaceFiFo01;
	delete surfaceFiFo02;
	delete colorFiFo00;
	delete colorFiFo01;
	shown=surfaceShown;
	draw();
}

void Turtle::setPenSize(Point size) {setPenSize(size.x, size.y);}

inline void Turtle::sinCos()
{
	double real_heading=(-heading + 90)*Float::deg2rad;
	_cos= Float::cutError(cos(real_heading));
	_sin= Float::cutError(sin(real_heading));
	shoulder=Point(_sin,-_cos, 0)*half_bottom;
		 head=Point(_cos, _sin, 0)*size;
}

void Turtle::drawRaw() 
{
//	prepareToDraw();

	static Turtle saved(false);
	saved = *this;
	pen.color=0xffffff;
	plainXOrPen();
	pen.down=true;
	setPenSize(0,0);
   Point
		 left = *this-shoulder,
		right = *this+shoulder,
		  top = *this+head;

	bool r=recorder->recording;
	recorder->recording=false;

	moveTo(right);	*this=right;
	lineTo(left);	*this=left;
	lineTo(top);	*this=top;
	lineTo(right);	*this=right;
	moveTo(saved);	*this = saved;
	restorePen();
	setPenSize(saved.pen.size.x, saved.pen.size.y);

	recorder->recording=r;
}

void Turtle::drawp() 
{
	if (shown)
	{	static Turtle saved(false);
		saved = *this;
		pen.color=0xffffffff;
		plainXOrPen();
		pen.down=true;
		setPenSize(0,0);

		glDisable(GL_DEPTH_TEST);
		Point
			 left = *this-u.x*half_bottom,
			right = *this+u.x*half_bottom,
			  top = *this+u.y*size;

		bool r=recorder->recording;
		recorder->recording=false;

		moveTo(right);	*this=right;
		lineTo(left);	*this=left;
		lineTo(top);	*this=top;
		lineTo(right);	*this=right;
		moveTo(saved);	*this = saved;
		
		glEnable(GL_DEPTH_TEST);
		restorePen();

		recorder->recording=r;
	}
}

void Turtle::checkMinX() {if (x-size < screen->Min.x) {x+=screen->size.x; drawRaw(); x-=screen->size.x;}}
void Turtle::checkMaxX() {if (x+size >=screen->Max.x) {x-=screen->size.x; drawRaw(); x+=screen->size.x;}}
void Turtle::checkMinY() {if (y-size < screen->Min.y) {y+=screen->size.y; drawRaw(); checkMaxX(); checkMinX(); y-=screen->size.y;}}
void Turtle::checkMaxY() {if (y+size >=screen->Max.y) {y-=screen->size.y; drawRaw(); checkMaxX(); checkMinX(); y+=screen->size.y;}}

void Turtle::getInWindow() 
{	if (screen->size.x!=0.0 && screen->size.y!=0.0 && screen->size.z!=0.0)
		*this=(*this-screen->Min) % screen->size + screen->Min;	// get *this in window
}

void Turtle::draw() 
{
	if (screen->perspective)
		drawp();
	else if (shown)
	{	if (mode==wrapmode)
		{	if (!in(screen->Min, screen->Max))
			{	Point saved=*this;
				getInWindow();
				checkMinX(); checkMaxX(); checkMinY(); checkMaxY(); drawRaw();
				*this=saved;
			}else
			{	checkMinX(); checkMaxX(); checkMinY(); checkMaxY(); drawRaw();
			}
		}else 
			drawRaw();
	}
}

void Turtle::setHeading(double angle) 
{
	if (screen->perspective)
	{
/*		double c=cos(angle*Float::deg2rad), s=sin(angle*Float::deg2rad);
		Point
		 uy=hbase.y*c+hbase.x*s;
		drawp();

		u.x=hbase.x*c-hbase.y*s;
		u.y=uy;
		u.z=hbase.z;
		drawp();
		heading = Float::mod(angle, 360);
*/		right(angle-getHeading());
	}
	else
	{	//if (angle != heading)
		{	draw();
			heading = Float::mod(angle, 360);
			sinCos();
			draw();
		}
	}
}

void Turtle::right(double angle) 
{	
	if (screen->perspective)
	{	double c=cos(angle*Float::deg2rad), s=sin(angle*Float::deg2rad);
		Point 
		 uy=u.y*c+u.x*s;
		drawp();
		u.x=u.x*c-u.y*s;
		u.y=uy;
		drawp();
		heading+=Float::mod(angle, 360);

	}
	else
		setHeading(heading+angle);
}

void Turtle::upPitch(double angle) 
{	
	double c=cos(angle*Float::deg2rad), s=sin(angle*Float::deg2rad);
	Point 
	 uz=u.z*c-u.y*s;
	drawp();
	u.y=u.y*c+u.z*s;
	u.z=uz;
	drawp();
	hbase=u;
	heading=0;
}

void Turtle::rightRoll(double angle) 
{	
	double c=cos(angle*Float::deg2rad), s=sin(angle*Float::deg2rad);
	Point 
	 uz=u.z*c+u.x*s;
	drawp();
	u.x=u.x*c-u.z*s;
	u.z=uz;
	drawp();
	hbase=u;
	heading=0;
}

void Turtle::spinZ(double angle) 
{	
	double c=cos(angle*Float::deg2rad), s=sin(angle*Float::deg2rad);
	Point ux, uy, uz;
	drawp();
	ux.x = u.x.x*c-u.x.y*s;
	ux.y = u.x.y*c+u.x.x*s;
	ux.z = u.x.z;
	u.x=ux;

	uy.x = u.y.x*c-u.y.y*s;
	uy.y = u.y.y*c+u.y.x*s;
	uy.z = u.y.z;
	u.y=uy;

	uz.x = u.z.x*c-u.z.y*s;
	uz.y = u.z.y*c+u.z.x*s;
	uz.z = u.z.z;
	u.z=uz;

	drawp();
	hbase=u;
	heading=0;
}


void Turtle::spinY(double angle) 
{	
	double c=cos(angle*Float::deg2rad), s=sin(angle*Float::deg2rad);
	Point ux, uy, uz;
	drawp();
	ux.z = u.x.z*c-u.x.x*s;
	ux.x = u.x.x*c+u.x.z*s;
	ux.y = u.x.y;
	u.x=ux;

	uy.z = u.y.z*c-u.y.x*s;
	uy.x = u.y.x*c+u.y.z*s;
	uy.y = u.y.y;
	u.y=uy;

	uz.z = u.z.z*c-u.z.x*s;
	uz.x = u.z.x*c+u.z.z*s;
	uz.y = u.z.y;
	u.z=uz;

	drawp();
	hbase=u;
	heading=0;
}

void Turtle::spinX(double angle) 
{	
	double c=cos(angle*Float::deg2rad), s=sin(angle*Float::deg2rad);
	Point ux, uy, uz;
	drawp();
	ux.y = u.x.y*c-u.x.z*s;
	ux.z = u.x.z*c+u.x.y*s;
	ux.x = u.x.x;
	u.x=ux;

	uy.y = u.y.y*c-u.y.z*s;
	uy.z = u.y.z*c+u.y.y*s;
	uy.x = u.y.x;
	u.y=uy;

	uz.y = u.z.y*c-u.z.z*s;
	uz.z = u.z.z*c+u.z.y*s;
	uz.x = u.z.x;
	u.z=uz;

	drawp();
	hbase=u;
	heading=0;
}

double Turtle::getHeading()
{
	if (screen->perspective)
	{	double result;


		if ((1.0-fabs(u.z.y)) < Float::epsilon)
		{	result=atan2(u.y.x, u.x.x)/Float::deg2rad;
		}else
		{	result=atan2(-u.x.y, u.y.y)/Float::deg2rad;
		}

		if (result < 0.0) return result + 360.0; else return result;
	}else
		return heading;
}

double Turtle::pitch()
{
//*
	double ry=atan2(-u.z.x, u.z.z), a;

	if (fabs(fabs(ry)-90.0*Float::deg2rad) < Float::epsilon)
		a=u.z.x/sin(-ry);
	else
		a=u.z.z/cos(ry);

	return atan2(-u.z.y, a)/Float::deg2rad;
/*/
	double xx=::sqrt(Float::sqr(u.z.x)+Float::sqr(u.z.z));
	if (u.z.z < 0)
		xx=-xx;
	if (fabs( xx  ) < Float::epsilon 
	 && fabs(u.z.y) < Float::epsilon)
		return 0;
	return atan2(u.z.y, // why is this not correct?
		xx)/Float::deg2rad;
//*/
}

double Turtle::roll() 
{
	if (fabs(u.z.z) < Float::epsilon
	 && fabs(u.z.x) < Float::epsilon)
		return 0;
	else
		return atan2(u.z.x, u.z.z)/Float::deg2rad;
}

void Turtle::setRoll(double angle)
{
	angle = Float::mod(angle, 360);
	rightRoll(angle-roll());
}

void Turtle::setPitch(double angle)
{
	angle = Float::mod(angle, 360);
	upPitch(angle-pitch());
}

void Turtle::setOrientation(Point o)
{
	draw();
	reset3d();
/*	right    (-getHeading());
	upPitch  (-pitch());
	rightRoll(-roll());
*/
	setRoll   (o.x);
	setPitch  (o.y);
	setHeading(o.z);
	draw();
}

Point Turtle::orientation()
{
	Point ori;
	ori.x=roll();
	ori.y=pitch();
	ori.z=getHeading();
	return ori;
}

void Turtle::show()
{
	prepareToDraw();
	if (!shown) 
	{	shown=true;
		draw();
	}
}

void Turtle::hide()
{
	prepareToDraw();
	if (shown) 
	{	draw();
		shown=false;
	}
}

bool Turtle::wrapLeft(double d, Point end)
{
	double yi, newd;
	if (end.x < screen->Min.x) 
	{
		yi = y-(x-screen->Min.x)*(end.y-y)/(end.x-x);

		if (screen->Min.y <= yi  &&  yi <= screen->Max.y) 
		{	lineTo(screen->Min.x, yi);
			if (mode==wrapmode) 
			{
				newd = d*(end.x-screen->Min.x)/(end.x-x);
				*this = Point(screen->Max.x, yi, 0);
				if (newd*d > 0) 
				{	wrapDelta=newd;
					return true;
				}
			}else 
				error(TURTLE_OUT_OF_BOUNDS, 0);	    
		}
	}
	return false;
}

bool Turtle::wrapRight(double d, Point end)
{
	double yi, newd;
	if (end.x > screen->Max.x) 
	{
		yi = y+(screen->Max.x-x)*(end.y-y)/(end.x-x);

		if (screen->Min.y <= yi  &&  yi <= screen->Max.y) 
		{	lineTo(screen->Max.x, yi);
			if (mode==wrapmode) 
			{
				newd = d*(end.x-screen->Max.x)/(end.x-x);
				*this = Point(screen->Min.x, yi, 0);
				if (newd*d > 0) 
				{	wrapDelta=newd;
					return true;
				}

			}else 
				error(TURTLE_OUT_OF_BOUNDS, 0);
		}
	}
	return false;
}

bool Turtle::wrapUp(double d, Point end)
{
	double xi, newd;
	if (end.y > screen->Max.y) 
	{
		xi = x-(y-screen->Max.y)*(end.x-x)/(end.y-y);

		if (screen->Min.x <= xi  &&  xi < screen->Max.x) 
		{	lineTo(xi, screen->Max.y);
			if (mode==wrapmode) 
			{
				newd = d*(end.y-screen->Max.y)/(end.y-y);
				*this = Point(xi, screen->Min.y, 0);
				if (newd*d > 0) 
				{	wrapDelta=newd;
					return true;
				}
			}else 
				error(TURTLE_OUT_OF_BOUNDS, 0);
		}else 
			if (xi <= screen->Min.x  &&  screen->Max.x<=xi)
			{
				xi= screen->Max.x<=xi ?   screen->Max.x  :  screen->Min.x;

				lineTo(xi, screen->Max.y);
				if (mode==wrapmode) 
				{
					newd = d*(end.y-screen->Max.y)/(end.y-y);
					*this = Point(xi, screen->Min.y, 0);
					if (newd*d > 0)
					{	wrapDelta=newd;
						return true;
					}
				}else 
					error(TURTLE_OUT_OF_BOUNDS, 0);
			}
	}
	return false;
}

bool Turtle::wrapDown(double d, Point end)
{
	double xi, newd;
	if (end.y < screen->Min.y) 

	{
		xi = x+(screen->Min.y-y)*(end.x-x)/(end.y-y);

		if (screen->Min.x <= xi  &&  xi <= screen->Max.x) 
		{	lineTo(xi, screen->Min.y);
			if (mode==wrapmode) 
			{
				newd = d*(end.y-screen->Min.y)/(end.y-y);
				*this = Point(xi, screen->Max.y, 0);
				if (newd*d > 0) 
				{	wrapDelta=newd;
					return true;
				}
			}else 
				error(TURTLE_OUT_OF_BOUNDS, 0);
		}else 
			if (xi <= screen->Min.x  &&  screen->Max.x <= xi) 
			{
				xi= screen->Max.x <= xi ?   screen->Max.x  :  screen->Min.x;

				lineTo(xi, screen->Min.y);
				if (mode==wrapmode) 
				{
					newd = d*(end.y-screen->Min.y)/(end.y-y);
					*this = Point(xi, screen->Max.y, 0);
					if (newd*d > 0) 
					{	wrapDelta=newd;
						return true;
					}
				}else 
					error(TURTLE_OUT_OF_BOUNDS, 0);
			}
	}
	return false;
}

void Turtle::forward(double d) 
{
	Point delta, target, endPoint;
	bool wraping=false;

	wrapDelta=0;
	draw();
	if (screen->perspective)
	{	target = *this+u.y*d;
		lineTo(target);
		*this=target;
	}
	else
	{	do
		{	if (wrapDelta!=0) 
				d=wrapDelta;
			delta = Point(_cos, _sin, 0)*d;

			if (wrapDelta==0.) 
			{	endPoint = *this+delta;
				if (mode==wrapmode && !in(screen->Min, screen->Max))
					getInWindow();
			}
			target = *this+delta;
			if (checkThrowing()) return;
//			if (pen.internalMode==Pen::Reverse && pen.down && d > 0) 
//				lineTo(target);

			if (mode==windowmode || target.in(screen->Min, screen->Max)) 
			{	lineTo(target);
				break;
			}
			wraping = wrapUp(d, target) || wrapDown(d, target) || wrapLeft(d, target) || wrapRight(d, target); 
			if (wraping)
				moveTo(*this);
		}
		while (wraping);

//		if (pen.internalMode==Pen::Reverse && pen.down && d < 0) 
//			lineTo(*this);

		*this=endPoint;
	}
	draw();
	outOfBounds=false;
}


double Turtle::towards_helper(Point end) 
{
	if (end != *this) 
	{	double a;
		a=-(atan2(end.y-y, end.x-x) / Float::deg2rad - 90);
		return a < 0 ?   360+a  :  a;
	}
	return 0;
}

void Turtle::setPos(Point target) 
{
	if (!throwing) 
	{	prepareToDraw();
		bool targetInScreen=target.in(screen->Min, screen->Max),
			wrapping= 
			( !targetInScreen
			|| !this->in(screen->Min, screen->Max));

		if (mode==fencemode && !targetInScreen)
			error(TURTLE_OUT_OF_BOUNDS, 0);
		else 
			if (mode==wrapmode && (wrapping || outOfBounds)) 
			{	draw();
				bool savedShown=shown;
				shown=false;
				double save_heading=heading;
				setHeading(towards_helper(target));

				forward((target - *this).Norm());

				setHeading(save_heading);
				shown=savedShown;
				draw();
				outOfBounds=wrapping;
			}else 
			{	outOfBounds=false;
				draw();

				lineTo(target); *this=target;
				draw();
			}
	}
}

double Turtle::distance(Point target)
{
	return (target-*this).Norm();
}

void Turtle::setPos(double x, double y, double z) {setPos(Point(x,y,z));}

void Turtle::setPos(NodeP nodeX, NodeP nodeY, NodeP nodeZ) 
{
	Point target;
  	if (!throwing) 
	{	if (nodeX) target.x=nodeX->getnumber(); else target.x=x;
		if (nodeY) target.y=nodeY->getnumber(); else target.y=y;
		if (nodeZ) target.z=nodeZ->getnumber(); else target.z=z;
		setPos(target);
	}
}

void Turtle::_setPos(Point target) 
{
	if (!throwing) 
	{	prepareToDraw();
		draw();
		lineTo(target);	*this=target;
		draw();
	}
}

void Turtle::_setPos(double x, double y, double z) {_setPos(Point(x,y,z));}


void Turtle::_setPos(NodeP nodeX, NodeP nodeY, NodeP nodeZ) 
{
	Point target;
  	if (!throwing) 
	{	if (nodeX) target.x=nodeX->getnumber(); else target.x=x;
		if (nodeY) target.y=nodeY->getnumber(); else target.y=y;
		if (nodeZ) target.z=nodeZ->getnumber(); else target.z=z;
		_setPos(target);
	}
}

void Turtle::setSpherePos(double r, double phi, double theta)
{
	setPos(
		r*cos(phi*Float::deg2rad)*sin(theta*Float::deg2rad), 
		r*cos(theta*Float::deg2rad), 
		r*sin(phi*Float::deg2rad)*sin(theta*Float::deg2rad));
}

void Turtle::setCylinderPos(double r, double phi, double y)
{
	setPos(r*cos(phi*Float::deg2rad), y, r*sin(phi*Float::deg2rad));
}

void Turtle::home()
{
	outOfBounds=false;
	setPos(0., 0., 0.);
	setHeading(0);
	reset3d();
}

NodeP Turtle::matrix()
{
/*	Point saved=*this;
	if (mode==wrapmode && !in(screen->Min, screen->Max))
		getInWindow();

	Point basex_,basey_,basez_, trans_,
		&centr_ = screen->center - screen->center0;

//	&centr_ = screen->center - screen->center0,

	if (screen->perspective)
	{	basex_ = screen->rotateNotScaled(u.x);
		basey_ = screen->rotateNotScaled(u.y);
		basez_ = screen->rotateNotScaled(u.z);
	}else
	{	basex_ = screen->rotateNotScaled(Point(_sin,-_cos, 0));
		basey_ = screen->rotateNotScaled(Point(_cos, _sin, 0));
		basez_ = screen->rotateNotScaled(Point(   0,    0, 1));
	}
	trans_ = screen->rotateNotScaled(Point(x,y,z));

	*this=saved;
	return new List
	(	new List

		(	new Float(basex_.x),
			new Float(basex_.y), 
			new Float(basex_.z), _ListEnd),
		new List
		(	new Float(basey_.x),
			new Float(basey_.y), 
			new Float(basey_.z), _ListEnd), 
		new List
		(	new Float(basez_.x),
			new Float(basez_.y), 
			new Float(basez_.z), _ListEnd), 
		new List
		(	new Float(trans_.x+centr_.x),
			new Float(trans_.y-centr_.y),	
			new Float(trans_.z+centr_.z),	_ListEnd), _ListEnd);
*/
	glPushMatrix();
		glTranslated(x,y,z);
		if (screen->perspective)
		{	glRotated(roll(),0,1,0);
			glRotated(pitch(),1,0,0);
			glRotated(-getHeading(),0,0,1);
		}else
			glRotated(-heading,0,0,1);


		double m[16];
		glGetDoublev(GL_MODELVIEW_MATRIX, m);
	glPopMatrix();

	return new List
	(	new List
		(	new Float(m[0]),
			new Float(m[1]), 
			new Float(m[2]), _ListEnd),
		new List
		(	new Float(m[4]),
			new Float(m[5]), 
			new Float(m[6]), _ListEnd), 
		new List
		(	new Float(m[8]),
			new Float(m[9]), 
			new Float(m[10]), _ListEnd), 
		new List
		(	new Float(m[12]),
			new Float(m[13]),	
			new Float(m[14]),	_ListEnd), _ListEnd);
}

void Turtle::setMatrix(Point basex,Point basey,Point basez,Point trans)
{
	draw();
/*	screen->base.x = basex;
	screen->base.y = basey;
	screen->base.z = basez;

	screen->base_scaled.x=screen->base.x*screen->scale.x;
	screen->base_scaled.y=screen->base.y*screen->scale.y;
	screen->base_scaled.z=screen->base.z*screen->scale.z;

	screen->center = screen->center0+trans*screen->scale;
*/
	double m[16]=
	{	basex.x,basex.y,basex.z,0,
		basey.x,basey.y,basey.z,0,
		basez.x,basez.y,basez.z,0,
		trans.x,trans.y,trans.z,1};

	for (int i=0; i<16; i++)
		screen->modelview_matrix[i]=m[i];

	glLoadMatrixd(screen->modelview_matrix);

	if (recorder->recording)
		*recorder << new R_SetMatrix(basex,basey,basez,trans);
	draw();
}

void Turtle::setMatrix(PairP basex,PairP basey,PairP basez,PairP trans)
{
	Point 
		basex_ = Point(
			basex-> first()->getnumber(),
			basex->second()->getnumber(), 
			basex-> third()->getnumber()),
		basey_ = Point(
			basey-> first()->getnumber(),
			basey->second()->getnumber(),
			basey-> third()->getnumber()),
		basez_ = Point(
			basez-> first()->getnumber(),

			basez->second()->getnumber(),
			basez-> third()->getnumber()),
		trans_ = Point(
			trans-> first()->getnumber(),
			trans->second()->getnumber(),
			trans-> third()->getnumber());

	setMatrix(basex_,basey_,basez_,trans_);
}

void Turtle::setTurtleMatrix()
{
/*	Point saved=*this;
	if (mode==wrapmode && !in(screen->Min, screen->Max))
		getInWindow();

	Point basex,basey,basez, trans;
	
	if (screen->perspective)
	{	basex=screen->rotateNotScaled(u.x);
		basey=screen->rotateNotScaled(u.y);
		basez=screen->rotateNotScaled(u.z);
	}else
	{	basex = screen->rotateNotScaled(Point(_sin,-_cos, 0)),
		basey = screen->rotateNotScaled(Point(_cos, _sin, 0)),
		basez = screen->rotateNotScaled(Point(  0 ,   0 , 1));
	}
	trans = screen->rotateNotScaled(Point(x,-y,z));

	*this=saved;
	screen->base.x=basex;
	screen->base.y=basey;
	screen->base.z=basez;

	screen->base_scaled.x=basex*screen->scale.x;
	screen->base_scaled.y=basey*screen->scale.y;
	screen->base_scaled.z=basez*screen->scale.z;


	screen->center+= trans*screen->scale;

	if (recorder->recording)
		*recorder << new R_SetMatrix(basex,basey,basez,trans);
*/


	glTranslated(x,y,z);
	if (screen->perspective)
	{	glRotated(roll(),0,1,0);
		glRotated(pitch(),1,0,0);
		glRotated(-getHeading(),0,0,1);
	}else
		glRotated(-heading,0,0,1);
	glGetDoublev(GL_MODELVIEW_MATRIX, screen->modelview_matrix);
	if (recorder->recording)
		*recorder << new R_SetMatrix4x4(screen->modelview_matrix);
}

void Turtle::wrap()
{
	draw();
	mode=wrapmode;
	getInWindow();
	moveTo(*this);
	draw();
}

void Turtle::arc(double angle, double radius)
{
	if (!pen.down) 
		return;
	static Turtle saved(false);
	saved=*this;

	prepareToDraw();
	draw();
	shown=false;

	if (screen->perspective)
	{	Point point = u.y*radius;

		pen.down=false;
		setPos(saved+point); // jump to begin of first line segment without drawing
		pen.down=true;

		int count = int(4*ceil(fabs(angle*radius/360))); // calculate resolution parameters
		if (count)
		{	double delta=Float::deg2rad*angle/count,
				radAngle=0;

			for (int i=0; i<count; i++) // draw each line segment of arc (will do wrap)
			{	radAngle+=delta;
				point=(u.y*cos(radAngle)+u.x*sin(radAngle))*radius;
				setPos(saved+point);
			}
		}
	}
	else
	{	double headingAngle = 90-heading, 
			radAngle = headingAngle*Float::deg2rad;

		Point point = Point(cos(radAngle), sin(radAngle), 0)*radius;

		pen.down=false;
		setPos(saved+point); // jump to begin of first line segment without drawing
		pen.down=true;

		int count = int(4*ceil(fabs(angle*radius/360))); // calculate resolution parameters
		if (count)
		{	double delta=-Float::deg2rad*angle/count; // heed the minus! It`s because Logo angles go strangely clockwise.
			int i;
			for (i=0; i<count; i++) // draw each line segment of arc (will do wrap)
			{	radAngle+=delta;
				point=Point(cos(radAngle), sin(radAngle), 0)*radius;
				setPos(saved+point);
			}
			--count;	// This section is neccessary because of an OpenGL bug (gaps!)
			radAngle-=delta/2;
			for (i=0; i<count; i++) // draw each line segment of arc (will do wrap)
			{	radAngle-=delta;
				point=Point(cos(radAngle), sin(radAngle), 0)*radius;
				setPos(saved+point);
			}
		}
	}
	*this=saved;
	moveTo();
	draw();
}

void Turtle::EllipseArc(double angle, double xRadius, 
								double yRadius, double startAngle)
{
	if (!pen.down) 
		return;
	static Turtle saved(false);
	saved=*this;

	prepareToDraw();
	draw();
	shown=false;

	if (screen->perspective)
	{	double radStart = startAngle*Float::deg2rad;

		Point 
			point =-u.x*xRadius*sin(radStart)
					 -u.y*yRadius*cos(radStart);

		pen.down=false;
		setPos(saved+point); // jump to begin of first line segment without drawing
		pen.down=true;

		// calculate resolution parameters

		int count = int(4*ceil(fabs(angle*(yRadius+xRadius)/2/360)));
		if (count)
		{	double delta=Float::deg2rad*angle/count;
		
			for (int i=0; i<count; i++) // draw each line segment of arc (will do wrap)
			{	radStart+=delta;
				point =-u.x*xRadius*sin(radStart)
						 -u.y*yRadius*cos(radStart);
				setPos(saved+point);
			}
		}
	}
	else
	{	double headingAngle = 90-heading, 
			radHead  = headingAngle*Float::deg2rad,
			radStart = (270-startAngle)*Float::deg2rad;

		Point 
			ex= Point( cos(radHead),sin(radHead),0),
			ey= Point(-sin(radHead),cos(radHead),0),

			point = ex*yRadius*sin(radStart)
					 -ey*xRadius*cos(radStart);

		pen.down=false;
		setPos(saved+point); // jump to begin of first line segment without drawing
		pen.down=true;

		// calculate resolution parameters
		int count = int(4*ceil(fabs(angle*(yRadius+xRadius)/2/360)));
		if (count)
		{	double delta=-Float::deg2rad*angle/count; // heed the minus! It`s because Logo angles go strangely clockwise.
			int i;

			for (i=0; i<count; i++) // draw each line segment of arc (will do wrap)
			{	radStart+=delta;
				point = ex*yRadius*sin(radStart)
						 -ey*xRadius*cos(radStart);
				setPos(saved+point);
			}
			--count;	// This section is neccessary because of an OpenGL bug (gaps!)
			radStart-=delta/2;
			for (i=0; i<count; i++) // draw each line segment of arc (will do wrap)
			{	radStart-=delta;
				point = ex*yRadius*sin(radStart)
						 -ey*xRadius*cos(radStart);
				setPos(saved+point);
			}
		}

	}
	*this=saved;
	moveTo();
	draw();
}

void Turtle::fillRect(Point a, Point b)
{
	draw();
	Point old=*this;
	if (!screen->perspective 
	&& mode==wrapmode 
	&& !in(screen->Min, screen->Max))
		getInWindow();

	glPushMatrix();
		glColor4ubv((unsigned char*)&pen.floodcolor);
		glTranslated(x,y,z);
		if (screen->perspective)
		{	glRotated(roll(),0,1,0);
			glRotated(pitch(),1,0,0);
			glRotated(-getHeading(),0,0,1);
		}else
			glRotated(-heading,0,0,1);
		glBegin(GL_QUADS); 
			if (screen->perspective)
			{	Point 
					v1=Point(a.x,a.y,0)-Point(a.x,b.y,0),
					v2=Point(b.x,b.y,0)-Point(a.x,b.y,0),
					normal=v1.X(v2);
				double n=normal.Norm();

				if (n!=0)
				{	normal/=n;
					glNormal3dv((double*)&normal);
				}
			}
			if (screen->textured)
			{	glTexCoord2d(0,0);
				glVertex2d(a.x, a.y); 
				glTexCoord2d(1,0);
				glVertex2d(b.x, a.y); 
				glTexCoord2d(1,1);
				glVertex2d(b.x, b.y); 
				glTexCoord2d(0,1);
				glVertex2d(a.x, b.y); 
			}else
			{	glVertex2d(a.x, a.y); 
				glVertex2d(b.x, a.y); 
				glVertex2d(b.x, b.y); 
				glVertex2d(a.x, b.y); 
			}
		glEnd(); 
		glColor4ubv((unsigned char*)&pen.color);
	glPopMatrix();
	*this=old;
	draw();
}

void Turtle::fillEllipse(Point radius)
{
	if (radius.x > 0)
	{	draw();
		Point old=*this;

		if (!screen->perspective 
		&& mode==wrapmode 
		&& !in(screen->Min, screen->Max))
			getInWindow();

		glPushMatrix();
			glColor4ubv((byte*)&pen.floodcolor);
			glTranslated(x,y,z);
			if (screen->perspective)
			{	glRotated(roll(),0,1,0);
				glRotated(pitch(),1,0,0);
				glRotated(-getHeading(),0,0,1);
			}else
				glRotated(-heading,0,0,1);

			glScaled(1,radius.y/radius.x,1);
				GLUquadricObj *q=gluNewQuadric();
				if (screen->textured)
					gluQuadricTexture(q, GL_TRUE);
				gluDisk(q, 0, radius.x, int(3*radius.x), 1);
				gluDeleteQuadric(q);
			glColor4ubv((byte*)&pen.color);
		glPopMatrix();
		*this=old;
		draw();
	}
}

void Turtle::sphere(double radius, int slices, int stacks)
{
	if (radius > 0)
	{	draw();
		Point old=*this;
		if (!screen->perspective 
		&& mode==wrapmode 
		&& !in(screen->Min, screen->Max))
			getInWindow();


		glPushMatrix();
			glTranslated(x,y,z);
			glRotated(roll(),0,1,0);
			glRotated(pitch()-90,1,0,0);
			glRotated(getHeading(),0,1,0);
			GLUquadricObj *q=gluNewQuadric();
			if (screen->textured)
				gluQuadricTexture(q, GL_TRUE);
			if (slices==0)
				slices=int(radius/2+10);
			if (stacks==0)
				stacks=int(radius/2+10);
			gluSphere(q, radius, slices, stacks);
			gluDeleteQuadric(q);
		glPopMatrix();
		*this=old;
		draw();
	}
}

void Turtle::ellipsoid(Point radius, int slices, int stacks)
{
	double r=radius.Norm();
	if (r > 0)
	{	draw();
		Point old=*this;
		if (!screen->perspective 
		&& mode==wrapmode 
		&& !in(screen->Min, screen->Max))
			getInWindow();

		glPushMatrix();
			glTranslated(x,y,z);
			glRotated(roll(),0,1,0);
			glRotated(pitch()-90,1,0,0);
			glRotated(getHeading(),0,1,0);
			glScaled(radius.x/100,radius.z/100,radius.y/100);
			glEnable(GL_AUTO_NORMAL);
			GLUquadricObj *q=gluNewQuadric();
			if (screen->textured)
				gluQuadricTexture(q, GL_TRUE);
			if (slices==0)
				slices=int(r/4+10);
			if (stacks==0)
				stacks=int(r/4+10);
			gluSphere(q, 100, slices, stacks);
			gluDeleteQuadric(q);
			glDisable(GL_AUTO_NORMAL);
		glPopMatrix();
		*this=old;
		draw();
	}
}

void Turtle::partialEllipsoid(Point radius, 
										double slicesstart, double slicesend, int slices,
										double stacksstart, double stacksend, int stacks)
{
	double r=radius.Norm();
	if (r > 0)
	{	draw();
		Point old=*this;
		if (!screen->perspective 
		&& mode==wrapmode 
		&& !in(screen->Min, screen->Max))
			getInWindow();

		glPushMatrix();
		glTranslated(x,y,z);
		glRotated(roll(),0,1,0);
		glRotated(pitch()-90,1,0,0);
		glRotated(getHeading(),0,1,0);
	
		GLfloat rho, rho0, drho, theta, theta0, dtheta;
		GLfloat x, y, z;
		GLfloat s, t, ds, dt;
		GLint i, j, imin, imax;
		GLboolean normals;
		GLfloat nsign;
		
		normals = GL_TRUE;
		nsign = 1.0;
		
		rho0 = Float::deg2rad*stacksstart;
		theta0 = Float::deg2rad*slicesstart;
		drho = Float::deg2rad*(stacksend-stacksstart) / (GLfloat) stacks;
		dtheta = Float::deg2rad*(slicesend-slicesstart) / (GLfloat) slices;
		
		/* texturing: s goes from 0.0/0.25/0.5/0.75/1.0 at +y/+x/-y/-x/+y axis */
		/* t goes from -1.0/+1.0 at z = -radius/+radius (linear along longitudes) */
		/* cannot use triangle fan on texturing (s coord. at top/bottom tip varies) */
		
		if (!screen->textured) 
		{	
			if (stacksstart==0)
			{	/* draw +Z end as a triangle fan */
				glBegin(GL_TRIANGLE_FAN);
				glNormal3f(0.0, 0.0, 1.0);
				glVertex3f(0.0, 0.0, nsign*radius.z);
				for (j = 0; j <= slices; j++) 
				{	theta = j*dtheta+theta0;
					x = -sin(theta) * sin(drho);
					y = cos(theta) * sin(drho);
					z = nsign * cos(drho);
					if (normals)
						glNormal3f(x * nsign, y * nsign, z * nsign);
					glVertex3f(x*radius.x, y*radius.y, z*radius.z);
				}
				glEnd();
			}
		}
		ds = 1.0 / slices;
		dt = 1.0 / stacks;
		t = 1.0; 		/* because loop now runs from 0 */
		if (screen->textured) 
		{	imin = 0;
			imax = stacks;
		}
		else 
		{	imin = 1;
			imax = stacks - 1;
		}
		/* draw intermediate stacks as quad strips */
		for (i = imin; i < imax; i++) 
		{	rho = i*drho+rho0;
			glBegin(GL_QUAD_STRIP);
			s = 0.0;
			for (j = 0; j <= slices; j++) 
			{	theta = j*dtheta+theta0;
				x = -sin(theta) * sin(rho);
				y = cos(theta) * sin(rho);
				z = nsign * cos(rho);
				if (normals)
					glNormal3f(x * nsign, y * nsign, z * nsign);
				if (screen->textured) glTexCoord2f(s,t);
				glVertex3f(x*radius.x, y*radius.y, z*radius.z);
				x = -sin(theta) * sin(rho + drho);
				y = cos(theta) * sin(rho + drho);
				z = nsign * cos(rho + drho);
				if (normals)
					glNormal3f(x * nsign, y * nsign, z * nsign);
				if (screen->textured) glTexCoord2f(s,t-dt);
				s += ds;
				glVertex3f(x*radius.x, y*radius.y, z*radius.z);
			}
			glEnd();
			t -= dt;
		}
		if (!screen->textured) 
		{	if (stacksend==180)
			{
				/* draw -Z end as a triangle fan */
				glBegin(GL_TRIANGLE_FAN);
				glNormal3f(0.0, 0.0, -1.0);
				glVertex3f(0.0, 0.0, -nsign*radius.z);
				rho = Float::pi - drho;
				s = 1.0;
				t = dt;
				for (j = slices; j >= 0; j--) 
				{	theta = j * dtheta+theta0;
					x = -sin(theta) * sin(rho);
					y = cos(theta) * sin(rho);
					z = nsign * cos(rho);
					if (normals)
						glNormal3f(x * nsign, y * nsign, z * nsign);
					s -= ds;
					glVertex3f(x*radius.x, y*radius.y, z*radius.z);
				}
				glEnd();
			}
		}

		glPopMatrix();
		*this=old;
		draw();
	}
}

void Turtle::partialEllipsoidSh(Point radius, 
										  double slicesstart, double slicesend, int slices,
										  double stacksstart, double stacksend, int stacks)
{
	double r=radius.Norm();
	if (r > 0)
	{	draw();
		Point old=*this;
		if (!screen->perspective 
		&& mode==wrapmode 
		&& !in(screen->Min, screen->Max))
			getInWindow();

		GLfloat rho, rho0, drho, theta, theta0, dtheta;
		GLfloat s, t, ds, dt;
		GLint i, j, imin, imax;
		GLboolean normals;
		GLfloat nsign;
		
		glPushMatrix();
		glLoadIdentity();
		glRotated(roll(),0,1,0);
		glRotated(pitch()-90,1,0,0);
		glRotated(getHeading(),0,1,0);
		double mm[16];
		glGetDoublev(GL_MODELVIEW_MATRIX, mm);
		glPopMatrix();
		Matrix m;
		for (i=0;  i<3; i++)
		{	m.x.a[i]=mm[i];
			m.y.a[i]=mm[i+4];
			m.z.a[i]=mm[i+8];
		}
		glPushMatrix();
		glTranslated(x,y,z);
		glRotated(roll(),0,1,0);
		glRotated(pitch()-90,1,0,0);
		glRotated(getHeading(),0,1,0);
	
		normals = GL_TRUE;
		nsign = 1.0;
		
		rho0 = Float::deg2rad*stacksstart;
		theta0 = Float::deg2rad*slicesstart;
		drho = Float::deg2rad*(stacksend-stacksstart) / (GLfloat) stacks;
		dtheta = Float::deg2rad*(slicesend-slicesstart) / (GLfloat) slices;

		ShadowObject *so=graph->shadowObject;
		int si=so->nPoints,
			 sj=so->nPlanes;

		Point p, normal;

		// texturing: s goes from 0.0/0.25/0.5/0.75/1.0 at +y/+x/-y/-x/+y axis 
		// t goes from -1.0/+1.0 at z = -radius/+radius (linear along longitudes)
		// cannot use triangle fan on texturing (s coord. at top/bottom tip varies)
		
/*		if (!screen->textured) 
		{	
			if (stacksstart==0)
			{	// draw +Z end as a triangle fan
				glBegin(GL_TRIANGLE_FAN);
				glNormal3f(0.0, 0.0, 1.0);
				glVertex3f(0.0, 0.0, nsign*radius.z);
				for (j = 0; j <= slices; j++) 
				{	theta = j*dtheta+theta0;
					x = -sin(theta) * sin(drho);
					y = cos(theta) * sin(drho);
					z = nsign * cos(drho);
					if (normals)
						glNormal3f(x * nsign, y * nsign, z * nsign);
					glVertex3f(x*radius.x, y*radius.y, z*radius.z);
				}
				glEnd();
			}
		}
*/		ds = 1.0 / slices;
		dt = 1.0 / stacks;
		t = 1.0; 		// because loop now runs from 0
//		if (screen->textured) 
		{	imin = 0;
			imax = stacks;
		}
/*		else 
		{	imin = 1;
			imax = stacks - 1;
		}
*/		// draw intermediate stacks as quad strips
		for (i = imin; i < imax; i++) 
		{	rho = i*drho+rho0;
			glBegin(GL_QUAD_STRIP);
			s = 0.0;
			
			theta = theta0;
			p=Point( -sin(theta) * sin(rho),
						 cos(theta) * sin(rho),
								nsign * cos(rho));
			normal=p*nsign;
			if (normals)
				glNormal3dv((double*)&normal);
			if (screen->textured) glTexCoord2f(s,t);
			p=p*radius;
			glVertex3dv((double*)&p);
			so->checkPoints(si);
			so->points[si]=m*p+(*this);
			si++;
			
			p=Point(-sin(theta) * sin(rho + drho),
						cos(theta) * sin(rho + drho),
							  nsign * cos(rho + drho));
			normal=p*nsign;
			if (normals)
				glNormal3dv((double*)&normal);
			if (screen->textured) glTexCoord2f(s,t-dt);
			s += ds;
			p=p*radius;
			glVertex3dv((double*)&p);
			so->checkPoints(si);
			so->points[si]=m*p+(*this);
			si++;

			for (j = 1; j <= slices; j++) 
			{	theta = j*dtheta+theta0;
				p=Point(	-sin(theta) * sin(rho),
							 cos(theta) * sin(rho),
									nsign * cos(rho));
				normal=p*nsign;
				if (normals)
					glNormal3dv((double*)&normal);
				if (screen->textured) glTexCoord2f(s,t);
				p=p*radius;
				glVertex3dv((double*)&p);
				so->checkPoints(si);
				so->points[si]=m*p+(*this);
				si++;
		
				p=Point(	-sin(theta) * sin(rho + drho),
							 cos(theta) * sin(rho + drho),
							      nsign * cos(rho + drho));
				normal=p*nsign;
				if (normals)
					glNormal3dv((double*)&normal);
				if (screen->textured) glTexCoord2f(s,t-dt);
				s += ds;
				p=p*radius;
				glVertex3dv((double*)&p);
				so->checkPoints(si);
				so->points[si]=m*p+(*this);
				si++;

				so->checkPlanes(sj);
				so->planes[sj].p[2]=si-2;
				so->planes[sj].p[1]=si-3;
				so->planes[sj].p[0]=si-4;
//				so->planes[sj].neigh[0]=sj+2;
//				so->planes[sj].neigh[1]=sj+4;
				sj++;

				so->checkPlanes(sj);
				so->planes[sj].p[2]=si-3;
				so->planes[sj].p[1]=si-2;
				so->planes[sj].p[0]=si-1;
//				so->planes[sj].neigh[0]=sj;
				sj++;
			}
			glEnd();
			t -= dt;
		}
/*		if (!screen->textured) 
		{	if (stacksend==180)
			{
				// draw -Z end as a triangle fan
				glBegin(GL_TRIANGLE_FAN);
				glNormal3f(0.0, 0.0, -1.0);
				glVertex3f(0.0, 0.0, -nsign*radius.z);
				rho = Float::pi - drho;
				s = 1.0;
				t = dt;
				for (j = slices; j >= 0; j--) 
				{	theta = j * dtheta+theta0;
					x = -sin(theta) * sin(rho);
					y = cos(theta) * sin(rho);
					z = nsign * cos(rho);
					if (normals)
						glNormal3f(x * nsign, y * nsign, z * nsign);
					s -= ds;
					glVertex3f(x*radius.x, y*radius.y, z*radius.z);
				}
				glEnd();
			}
		}
*/		so->nPoints=si;
		so->nPlanes=sj;

		glPopMatrix();
		*this=old;
		draw();
	}
}

void Turtle::fillPie(double angle, Point radius)
{
	if (radius.x > 0)
	{	draw();
		Point old=*this;
		if (!screen->perspective 
		&& mode==wrapmode 
		&& !in(screen->Min, screen->Max))
			getInWindow();

		glPushMatrix();
			glColor4ubv((byte*)&pen.floodcolor);
			glTranslated(x,y,z);
			if (screen->perspective)
			{	glRotated(roll(),0,1,0);
				glRotated(pitch(),1,0,0);
				glRotated(-getHeading(),0,0,1);
			}else
				glRotated(-heading,0,0,1);

				GLUquadricObj *q=gluNewQuadric();
				if (screen->textured)
					gluQuadricTexture(q, GL_TRUE);
				gluPartialDisk(q, 0, radius.x, 
					int(3*radius.x), 1, 0, angle);
				gluDeleteQuadric(q);
			glColor4ubv((byte*)&pen.color);
		glPopMatrix();
		*this=old;
		draw();
	}
}

Color *pixels;

void Turtle::pix(int w,int h, Color color)
{
	pixels[w+h*screen->winrangex]=color; 

/*	static int i=0;
	if (!(i & 0xf))
	{	glDrawPixels(screen->winrangex, screen->winrangey, 
			GL_RGBA, GL_UNSIGNED_BYTE, (void*)pixels);
		mySwapBuffers();
	}
	i++;
*/
}

// floodFillX is taken from UCBLogo's xgraphics.c 
// & improved & translated to OpenGL

void Turtle::floodFillX(int x, int y, Color fill, Color rim)
{
	pixels=(Color*)GC_::malloc(screen->winrangex*screen->winrangey*sizeof(Color));
	if (!pixels)
		return;
	glReadPixels(int(graph->winorigin.x), int(graph->winorigin.y), 
		screen->winrangex, screen->winrangey, 
		GL_RGBA, GL_UNSIGNED_BYTE, (void*)pixels);

	floodFillXrecursiv(x,y, fill, rim);

	Point r;
	glGetDoublev(GL_CURRENT_RASTER_POSITION, (double*)&r);

	Point p=Point(screen->Min.x,screen->Min.y,0);
	glRasterPos3dv((double*)&p);
	glDrawPixels(screen->winrangex, screen->winrangey, 
		GL_RGBA, GL_UNSIGNED_BYTE, (void*)pixels);
	free(pixels);
}

void Turtle::floodFillXrecursiv(int x, int y, Color fill, Color rim)
{
	int i, lastBorder, leftLimit, rightLimit;
	Color c, *p;
	
	// Seek left
	leftLimit=-1;
	p=&pixels[x+y*screen->winrangex];
	for (i=x; i >= 0; i--) 
	{	c=*p;
		if (c==rim || c==fill) 
			break;
		*p-- = fill;
		leftLimit=i;
	}
	if (leftLimit==-1) 
		return;
	
	// Seek right
	rightLimit=x;
	p=&pixels[x+1+y*screen->winrangex];
	for (i=x+1; i < screen->winrangex; i++) 
	{	c=*p;
		if (c==rim || c==fill)
			break;
		*p++ = fill;
		rightLimit=i;
	}
	// Look at lines above and below and start paints
	// Above
	if (y > 0) 
	{	lastBorder=1;
		p=&pixels[leftLimit+(y-1)*screen->winrangex];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c = *p;
			if (lastBorder) 
			{	if (c!=rim && c!=fill) 
				{	floodFillXrecursiv(i, y-1, fill, rim);
					lastBorder=0;
				}
			}
			else if (c==rim) 
				lastBorder=1;
			p++;
		}
	}
	// Below
	if (y < screen->winrangey-1) 
	{	lastBorder=1;
		p=&pixels[leftLimit+(y+1)*screen->winrangex];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c = *p;
			if (lastBorder) 
			{	if (c!=rim && c!=fill) 
				{	floodFillXrecursiv(i, y+1, fill, rim);
					lastBorder=0;
				}
			} 
			else if (c==rim) 
				lastBorder=1;
			p++;
		}
	}
}

void Turtle::floodFillX2(int x, int y, Color newColor, Color oldColor)
{
	pixels=(Color*)GC_::malloc(screen->winrangex*screen->winrangey*sizeof(Color));
	if (!pixels)
		return;
	glReadPixels(int(graph->winorigin.x), int(graph->winorigin.y), 
		screen->winrangex, screen->winrangey, 
		GL_RGBA, GL_UNSIGNED_BYTE, (void*)pixels);

	floodFillX2recursiv(x,y, newColor,oldColor);

	Point p=Point(screen->Min.x,screen->Min.y,0);
	glRasterPos3dv((double*)&p);
	glDrawPixels(screen->winrangex, screen->winrangey, 
		GL_RGBA, GL_UNSIGNED_BYTE, (void*)pixels);
	free(pixels);
}

// surface fill does not work yet!
void Turtle::floodFillX2recursiv(int x, int y, Color newColor, Color oldColor)
{
	int i, lastBorder, leftLimit, rightLimit;
	if (oldColor == newColor)	// Nothing to be done
		return;
	
	Color c, *p;
	
	// Seek left
	leftLimit=-1;
	p=&pixels[x+y*screen->winrangex];
	for (i=x; i >= 0; i--) 
	{	c=*p;
		if (c != oldColor) 
			break;
		*p-- = newColor;
		leftLimit=i;
	}
	if (leftLimit==-1) 
		return;
	
	// Seek right
	rightLimit=x;
	p=&pixels[x+1+y*screen->winrangex];
	for (i=x+1; i < screen->winrangex; i++) 
	{	c=*p;
		if (c != oldColor) 
			break;
		*p++ = newColor;
		rightLimit=i;
	}
	// Look at lines above and below and start paints
	// Above
	if (y > 0) 
	{	lastBorder=1;
		p=&pixels[leftLimit+(y-1)*screen->winrangex];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c = *p;
			if (lastBorder) 
			{	if (c == oldColor) 
				{	floodFillX2recursiv(i, y-1, newColor, oldColor);
					lastBorder=0;
				}
			}
			else if (c != oldColor) 
				lastBorder=1;
			p++;
		}
	}
	// Below
	if (y < screen->winrangey-1) 
	{	lastBorder=1;
		p=&pixels[leftLimit+(y+1)*screen->winrangex];
		for (i=leftLimit; i <= rightLimit; i++) 
		{	int c;
			c = *p;
			if (lastBorder) 
			{	if (c == oldColor) 
				{	floodFillX2recursiv(i, y+1, newColor, oldColor);
					lastBorder=0;
				}
			} 
			else if (c != oldColor) 
				lastBorder=1;
			p++;
		}
	}
}

void Turtle::fill(bool surface) 
{
	if (direct)
		console->dispatchMessages();
	if (checkThrowing()) return;
	draw();
	Point old=*this;
	getInWindow();

	glDisable(GL_BLEND);
	glColor4ubv((byte*)&pen.floodcolor);
	Point wp=getWindowPos();
	wp.x-=graph->winorigin.x;
	wp.y-=graph->winorigin.y;
	if (surface)
	{	//floodFill2(screen->winrangex/2.0+x,screen->winrangey/2.0+y);
		//CQuickFill *qf=new CQuickFill(screen->winrangex, screen->winrangey);
		//qf->QuickFill(wp.x,wp.y, pen.floodcolor);
		//delete qf;
		floodFillX2(int(wp.x), int(wp.y), pen.floodcolor, pen.color);
	}else
	{	//floodFill(screen->winrangex/2.0+x,screen->winrangey/2.0+y);
		//CQuickFill *qf=new CQuickFill(screen->winrangex, screen->winrangey);
		//qf->QuickFill(wp.x,wp.y, pen.floodcolor, pen.color);
		//delete qf;
		floodFillX(int(wp.x), int(wp.y), pen.floodcolor, pen.color);
	}
	glColor4ubv((byte*)&pen.color);
	glEnable(GL_BLEND);
		
	*this=old;
	draw();
	if (direct)
		mySwapBuffers();
	checkThrowing();
}

// link with: freetype219MT.lib ftgl_static_MT.lib 

void Turtle::Label(wxString str)
{
	font.load();

	if (font.ftfont)
	{	draw();
		glPushMatrix();
			glTranslatef(x,y,z);
			if (screen->perspective)
				glScaled(1.5,1.5,1.5);

			double ad,as;
			ad=font.ftfont->Advance(str.c_str());
			as=font.ftfont->Ascender();
			if (screen->perspective)
			{	glRotated(roll(),0,1,0);
				glRotated(pitch(),1,0,0);
				glRotated(90-getHeading(),0,0,1);
			}else
				glRotated(90-heading,0,0,1);

			switch (font.alignx)
			{	case 1:	break;
				case 0:	glTranslatef(-ad/2,0,0); break;
				case -1:	glTranslatef(-ad,0,0); break;
			}
			switch (font.aligny)
			{	case 1: break;
				case 0: glTranslatef(0,-as/3,0); break;
				case -1: glTranslatef(0,-as,0); break;
			}
			font.ftfont->Render(str.c_str());
		glPopMatrix();
		draw();
	}
}

Point Turtle::LabelSize(wxString str)
{
	font.load();

	if (font.ftfont)
	{	float llx,lly,llz,urx,ury,urz;
		if (screen->perspective)
			glScaled(1.5,1.5,1.5);
		font.ftfont->BBox(str.c_str(),llx,lly,llz,urx,ury,urz);
		return Point(urx-llx,ury-lly,urz-llz);
	}
	return Point();
}

void Turtle::setLabelSize(double x, double y)
{	font.sizex=x;
	font.sizey=y;
}
wxString Turtle::getLabelFont()
{	return font.name;
}
void Turtle::setLabelFont(wxString face)
{	font.name=face;
}
void Turtle::setLabelWeight(int weight)
{	font.weight=weight;
}
void Turtle::setLabelAlign(int alignx, int aligny)
{	font.alignx=alignx;
	font.aligny=aligny;
}

void Turtle::setPixel(Point p, Color c)
{
	glBegin(GL_POINTS);
		glColor4ubv((byte*)(&c));
		glVertex3dv((double*)&p);
	glEnd();
	if (recorder->recording)
		*recorder << new R_SetPixel(p, c);
}

Point Turtle::getWindowPos()
{
	Point old=*this;
	if (!screen->perspective 
	&& mode==wrapmode 
	&& !in(screen->Min, screen->Max))
		getInWindow();

	Point wp;
	gluProject(x,y,z, 
		screen->modelview_matrix, 
		screen->projection_matrix, 
		screen->viewport, 
		&wp.x,&wp.y,&wp.z);
	*this=old;
	return wp;
}

int Turtle::Pixel()	// Can this be made faster & easier? I hope so!
{
	Point wp=getWindowPos();
	Color c;
	glReadPixels(int(wp.x),int(wp.y), 1,1, GL_RGBA, GL_UNSIGNED_BYTE, &c);
	return c;
}


BitmapP Turtle::bitCopy(double w, double h)
{
	Point old=*this;
	if (!screen->perspective 
	&& mode==wrapmode 
	&& !in(screen->Min, screen->Max))
		getInWindow();

	Point p1=getWindowPos().round();
	*this+=Point(w,h,0);
	Point win=(getWindowPos()-p1).abs().round();
	int winx=int(win.x), winy=int(win.y);
	*this=old;

	Color *pixels=(Color*)GC_::malloc(winx*winy*4);
	if (!pixels)
		return BitmapP(Unbound);
	draw();
	glReadPixels(int(p1.x),int(p1.y), winx,winy, 
		GL_RGBA, GL_UNSIGNED_BYTE, pixels);
	draw();
	BitmapP b=new Bitmap(int(w),int(h), winx,winy, pixels);
	return b;
}

void Turtle::bitPaste(BitmapP b)
{
	draw();
	Point old=*this;
	if (!screen->perspective 
	&& mode==wrapmode 
	&& !in(screen->Min, screen->Max))
		getInWindow();

//	double p[4];
//	glGetDoublev(GL_CURRENT_RASTER_POSITION, p);

	glRasterPos3dv((double*)this);
	glDrawPixels(b->winx,b->winy,
		GL_RGBA,GL_UNSIGNED_BYTE, b->pixels);

//	glRasterPos4dv(p);
//	Point p=Point(-screen->winrangex/2,-screen->winrangey/2,0);
	Point p=Point(screen->Min.x,screen->Min.y,0);
	glRasterPos3dv((double*)&p);
	
	Float::reset_FPU();// I dont know why this is neccessary, but it HELPS!

	*this=old;
	draw();
}

void Turtle::scroll(int width, int height, double sx, double sy)
{
	Point old=*this;
	if (!screen->perspective 
	&& mode==wrapmode 
	&& !in(screen->Min, screen->Max))
		getInWindow();
	glRasterPos3dv((double*)&graph->winorigin);
	*this=old;

	width=screen->winrangex;
	height=screen->winrangey;

	Color *pixels=(Color*)GC_::malloc(width*height*4);
	if (!pixels)
		return;
	draw();
	glReadPixels(int(sx),int(sy), int(width-sx),int(height-sy), 
		GL_RGBA, GL_UNSIGNED_BYTE, pixels);
	glDrawPixels(int(width-sx),int(height-sy),
		GL_RGBA,GL_UNSIGNED_BYTE, pixels);
	free(pixels);
//	glCopyPixels(sx,sy, width-sx,height-sy, GL_COLOR);
	if (direct)
		mySwapBuffers();
	draw();
}

Point Turtle::scrollCalibrate(double sx, double sy)
{
	prepareToDraw();
	return Point(
		(sx*screen->winrangex)/800,(sy*screen->winrangey)/600,0);
}
