#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/string.h>
#include <wx/utils.h>
#include <wx/textctrl.h>
#include <wx/html/helpctrl.h>
#include <wx/txtstrm.h>
#include <wx/wfstream.h>

#include <stdio.h>
#include <malloc.h>

#ifdef WIN32
	#include <process.h>
#else
	#include <unistd.h>
#endif

#include "Main.h"

#include "String.h"
#include "File.h"
#include "Pair.h"
#include "Procedur.h"
#include "Object.h"
#include "Table.h"
#include "List.h"
#include "Quote.h"
#include "Colon.h"
#include "Runparsd.h"
#include "Tree.h"
#include "Int.h"
#include "UInt8.h"

#include "Error.h"
#include "Eval.h"
#include "Stack.h"
#include "Init.h"

#include "Color.h"
#include "Console.h"
#include "EdDefsExt.h"
#include "Edit.h"
#include "EdFrame.h"
#include "LogoApp.h"

extern Function lTextScreen;

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

NodeP lPrimitives   (PairPP) {return root->contents(PROCS, PERMANENT);}
NodeP lLogoConstants(PairPP) {return root->contents(VALS , PERMANENT);}
NodeP lProcedures   (PairPP) {return root->contents(PROCS, NORMAL_OBJECT);}
NodeP lNames        (PairPP) 
{	return new Pair(0, 
			 new Pair(root->contents(VALS, NORMAL_OBJECT),0));
}
NodeP lPropertyLists(PairPP) 
{	return new List(0, 
						 0, 
						 root->contents(PLISTS, NORMAL_OBJECT), _ListEnd);
}

NodeP lContents (PairPP) {return root->contents( NORMAL_OBJECT );}

NodeP lburied (PairPP) {return root->contents( BURIED );}
NodeP ltraced (PairPP) {return root->contents( TRACED );}
NodeP lstepped(PairPP) {return root->contents( STEPPED);}

NodeP lbury (PairPP arg) {return arg[1]->  bury( PROC_BURIED );}
NodeP ltrace(PairPP arg) {return arg[1]->  bury( PROC_TRACED );}
NodeP lstep (PairPP arg) {return arg[1]->  bury( PROC_STEPPED);}

NodeP lunBury (PairPP arg) {return arg[1]->unBury( PROC_BURIED );}
NodeP lunTrace(PairPP arg) {return arg[1]->unBury( PROC_TRACED );}
NodeP lunStep (PairPP arg) {return arg[1]->unBury( PROC_STEPPED);}


NodeP lisdefined  (PairPP arg) {return arg[1]->procType(0);}
NodeP lisPrimitive(PairPP arg) {return arg[1]->procType(1);}
NodeP lisMacro    (PairPP arg) {return arg[1]->procType(2);}

NodeP lisProcedure(PairPP arg) 
{
	NodeP name = arg[1]->a_name(1);
	if (!throwing) 
		return Node::boolnode( name->intern()->proc_or_prim() != Unbound );
	return Unbound;
}

NodeP lisName(PairPP arg) 
{
	NodeP name = arg[1]->a_name(1);
	if (!throwing) 
		return Node::boolnode( name->intern()->var() != Unbound );
	return Unbound;
}

NodeP lArity(PairPP arg)
{
   NodeP name = arg[1]->a_name(1);

   if (!throwing) 
	{	ObjectP obj = name->intern();
		
		if (obj->prim()->is.primitive)
		{	int min = obj->prim()->min;
			if (min == Primitive::ok_if_no_arg) 
				min = 0;
			return new List(
				new Int(min),
			   new Int(obj->prim()->std),
			   new Int(obj->prim()->max), 
				new Int(obj->prim()->priority), _ListEnd);

		}
		else if (obj == Unbound) 
		{	error(DK_HOW_UNREC, name);
			return Unbound;
		}else 
		{	return new List(
				new Int(obj->proc()->min),
				new Int(obj->proc()->std),
				new Int(obj->proc()->max),
				new Int(priority_prefix), _ListEnd);
		}
	}
	return Unbound;
}

NodeP lThing(PairPP arg) 
{
	NodeP	name = arg[1]->a_name(1),
		 val=Unbound;

	if (!throwing) val = name->intern()->var();

	while (val==Unbound  &&  !throwing)
		val = error(NO_VALUE, arg[1]);


	return val;
}


NodeP lmake(PairPP arg) 
{
	NodeP name = arg[1]->a_name(1);

	if (!throwing) 
	{	ObjectP what = name->intern();

		what->setvar( arg[2] );

		if (what->bit.val_traced) 
		{
			NodeP tvar=0;	tvar = arg[2]->maybe_quote();
			writer << "Make " << QuoteP(new Quote(what)) << ' ' << tvar;

			if (ufun) 
				writer << " in " << ufun << endL << this_line;
			writer << endL;
		}
	}
	return Unbound;
}

NodeP llocal(PairPP arg) 
{
	if (Evaluator::tailcall==tailcall_tail) 
		return Unbound;


	PairP li = arg[1];

	while (li->islist()  &&  exist(2)  &&  !throwing)
		li = error(BAD_DATA, li);

	int vsp = Evaluator::varStack;

	if (li->islist())		// local [a b]
		while (li  &&  !throwing) 
		{
			while (! li->first()->is.word  &&  !throwing)
				li->setfirst( error(BAD_DATA, li->first()) ); // prevent crash in _apply
			
			if (!throwing) 
			{	li->setfirst( li->first()->intern() );	// local [a b] faster next time
				ObjectP a = ObjectP(li->first());
				if (a->not_local(vsp)) 
					Evaluator::varStack <<  a;
				a->setvar( Unbound );
				a->tell_shadow();
			}
			if (checkThrowing()) break;
			next(li);
		}
	else		// (local "a "b)
	{	int nr=1;
		while (arg!=argStack  &&  !throwing) 
		{
			while (arg!=argStack  &&  !(*++arg)->is.word  &&  !throwing)
				*arg = error(BAD_DATA, *arg, nr);

			ObjectP a = (*arg)->intern();
			if (!throwing) 
			{	if (a->not_local(vsp)) 

					Evaluator::varStack <<  a;
				a->setvar( Unbound );
				a->tell_shadow();
			}
			if (checkThrowing()) break;
			nr++;
		}
	}
	Evaluator::var = Evaluator::varStack;    // so eval won't undo our work
	return Unbound;
}


NodeP Pair::to_helper(bool macro_flag) 
{
	if (ufun  &&  loader.file()==stdin)
	{	error(NOT_INSIDE, 0);
		return Unbound;
	}
	if (!this) 									// verifiy this
	{	error(NOT_ENOUGH, 0);
		return Unbound;
	}
	ObjectP obj = ObjectP(first());				// verify object
	if (type_of(obj) != OBJECT)
		error(BAD_DATA_UNREC, obj);
	else
		if ((obj->proc_or_prim() != Unbound  &&  loader.file()==stdin)
		 ||  obj->proc_or_prim()->is.primitive )
			error(ALREADY_DEFINED, obj);
	
	File::deependProcName=obj->name();

	PairP args=tail(),  formals=0;
	int min=0, std=0, max=0, old_std=-1;
		
	if (!throwing)
	{	NodeP old_proc = obj->proc_or_prim();		// remember std of old proc
		if (old_proc != Unbound) 
			old_std =  old_proc->is.primitive ? 
				PrimitiveP(old_proc)->std  :  ProcedureP(old_proc)->std;

		PairP lastnode=0, tnode=0;

		while (args)										// scan formals 
		{	NodeP arg=args->first();
			next(args);
	
			if (arg->islist()  &&  max !=-1)	// scan default arguments
			{	PairP &pair= (PairP&) arg;
				pair->make_runparse();
				RunparsedP &run = (RunparsedP&) pair;
				arg = run->parsed();
				max++;

				if (type_of(pair->first()) != COLON) 
					pair->setfirst(ObjectP(pair->first()));
				else
					pair->setfirst( ColonP(pair->first())->colon() );

				if (!pair->tail())		// formal is rest
					max=-1;
			}else
				if (type_of(arg)==COLON  &&  min==max)	// usual formals
				{	arg = ColonP(arg)->colon();
					min++, std++, max++;
				}else
					if (type_of(arg)==INT_)				// set std
					{	Int &irg = *IntP(arg);
					
						if ((unsigned)irg.int_ <= (unsigned) max
						&&            irg.int_ >=            min) 
							std = irg.int_;
					}else 
						min++, std++, max++;	// colonless formal

			tnode = new Pair(arg, 0);

			if (formals) lastnode->settail( tnode );
			else formals=tnode;
			lastnode=tnode;
		}
	}
	if (!throwing) 
	{	PairP bodylist=0, bodywords=0;
		File::toPending++;    // for int or quit signal

		if (!throwing  &&  File::toPending  &&  !loader.eof())	// read procedure`s body
		{	//Line::nr=1;
			Line::nr=Line::readnr;
			StringP body = loader.read("");
//			c_out << "to_helper " << body << endL;
			bodylist = new Line( body->parser(true), 0);
//			c_out << "to_helper " << bodylist << endL;
			if (NodeP(body)==_Null_Word)
				bodywords = 
					new Pair(File::currentLine, 
					new Pair(_End, 0));
			else
				bodywords = 
					new Pair(File::currentLine, 
					new Pair(body, 
					new Pair(_End, 0)));
		}
		if (File::toPending  &&  !throwing) 
		{	obj->setproc( ProcedureP( new Procedure(
				formals, bodylist, bodywords, min, std, max) ) );

			obj->bit.proc_macro = macro_flag;

			if (std != old_std  &&  old_std >= 0) 
				Tree::the_generation = new Int(Tree::the_generation->int_+1);
			
			if (loader.file()==stdin 
			||  _LoadNoisily->var()->equal(_True))
				c_out << ";" << obj << " defined" << endL;
		}
		File::toPending=0;
	}
	File::deependProcName=0;
	return Unbound;
}

NodeP lto    (PairPP arg) {return arg[1]->to_helper(false);}
NodeP l_Macro(PairPP arg) {return arg[1]->to_helper(true);}


ProcedureP Pair::define_helper(bool macro_flag, PairP name) 
{
	PairP val=0;
	ObjectP obj=0;
	int old_std=-1;

	Line::fn=Line::findFile(StringP(new String("unnamed")));

	if (name) 								// verify name
	{	name = PairP(name->a_name(1));
		if (!throwing) 
		{	obj = name->intern();
			NodeP p = obj->proc_or_prim();

			int redef = _RedefP->var()->equal(_True);

			if (!redef && p->is.primitive)		// grumble if primitive and not redefp
			{	error(IS_PRIM, obj);
				return ProcedureP(Unbound);
			}else									// remember old std arg nr
				if (p != Unbound) 
					old_std=  p->is.primitive ?
						PrimitiveP(p)->std  :  ProcedureP(p)->std;
		}
		if (!throwing)		// verify this (should be a listlist)
		{	val=this;
			while ( (val->isnolist()  ||  !val->all_lists() )  &&  !throwing) 
				val = error(BAD_DATA, val);
		}
	}else			// lambda (proc with no name)
		val=this;

	if throwing return ProcedureP(Unbound);

	if (val)
		val=PairP(val->deepCopy());

	int min=0, std=0, max=0;
	PairP formals=0;
	if (val)
		formals = val->first();
	
	if (!throwing) 
	{	if (formals) 
		{	formals->make_runparse();
			formals = RunparsedP(formals)->parsed();
			val->setfirst( formals );
		}
		PairP formal=formals;

		while (formal) 
		{	NodeP arg = formal->first();

			if (arg->islist()  &&  max !=-1) 
			{
				PairP &list = (PairP&) arg;
				list->make_runparse();
				list = RunparsedP(list)->parsed();
				formal->setfirst( list );
				max++;
				if (! list->tail())		// formal is rest
					max=-1;
			}else
			{	if (type_of(arg)==INT_)
				{	IntP &i = (IntP&) arg;
					if (unsigned(i->int_) <= unsigned(max) 
					&&           i->int_  >=          min ) 
					std = i->int_;
				}else 
					if (max==min) 
						min++, max++, std++;
					else 
					{	error(BAD_DATA_UNREC, arg);
						break;
					}
			}
			next(formal);
			if (checkThrowing()) break;
		}
	}
	if (!throwing)
	{	PairP body=0;
		if (val)
			body = val->tail();

		ProcedureP proc=new Procedure(formals, body, NIL, min, std, max);
		if (!name)
			return proc;
		else 
		{	obj->setproc(ProcedureP(proc));
			if (macro_flag)
				obj->bit.proc_macro=true;
			else
				obj->bit.proc_macro=false;

			if (std != old_std  &&  old_std >= 0) 
				Tree::the_generation = new Int(Tree::the_generation->int_+1);
		}
	}
	return ProcedureP(Unbound);
}

NodeP ldefine   (PairPP arg) {return arg[2]->define_helper(false, arg[1]);}
NodeP l_defMacro(PairPP arg) {return arg[2]->define_helper( true, arg[1]);}


NodeP lcopyDef(PairPP arg) 
{
	int redef = _RedefP->var()->equal(_True);
	int new_std, old_std=-1;
	ObjectP
		arg1 = ObjectP( arg[1]->a_name(1) ),
		arg2 = ObjectP( arg[2]->a_name(2) );

	if (arg2->numberP()) error(BAD_DATA_UNREC, arg2, 2);
	if (arg1->numberP()) error(BAD_DATA_UNREC, arg1);

	if (!throwing) 
	{	arg1= arg1->intern();
		arg2= arg2->intern();
	}
	if (!throwing  &&  arg2->proc_or_prim()==Unbound)
		error(DK_HOW, arg2, 2);

	if (!throwing  &&  !redef  &&  arg1->proc_or_prim()->is.primitive)
		error(IS_PRIM, arg1);

	if (!throwing) 
	{	NodeP
			old_proc=arg1->proc_or_prim(),
			new_proc=arg2->proc_or_prim();

		if (old_proc != Unbound) 
			old_std = old_proc->is.primitive ?  PrimitiveP(old_proc)->std  
														:  ProcedureP(old_proc)->std;

		new_std= new_proc->is.primitive ?  PrimitiveP(new_proc)->std  
												  :  ProcedureP(new_proc)->std;

		if (old_std != new_std  &&  old_std >= 0) 
			Tree::the_generation = new Int(Tree::the_generation->int_+1);
		
		arg1->setproc_or_prim( new_proc );
		arg1->setflag( PROC_BURIED );
		
		if (arg2->bit.proc_macro) 
			arg1->setflag( PROC_MACRO );
		else
			arg1->clearflag( PROC_MACRO );
	}
	return Unbound;
}


NodeP lText(PairPP arg) 
{
	NodeP name = arg[1]->a_name(1);

	if (!throwing) 
	{	NodeP proc=0;	proc = name->intern()->proc_or_prim();

		if (proc==Unbound) 
		{	error(DK_HOW_UNREC, name); return Unbound;
		}
		else if (proc->is.primitive) 
		{	error(IS_PRIM, name); return Unbound;
		}
		else return new Pair(
			ProcedureP(proc)->formals(),
			ProcedureP(proc)->bodylist());
	}
	return Unbound;
}

NodeP lfullText(PairPP arg) 
{
	NodeP name = arg[1]->a_name(1);

	if (!throwing) 
	{	NodeP proc_or_prim = name->intern()->proc_or_prim();

		if (proc_or_prim==Unbound) 
		{	error(DK_HOW_UNREC, name); return Unbound;
		}
		else if (proc_or_prim->is.primitive) 
		{	error(IS_PRIM, name); return Unbound;
		}
		else return ProcedureP(proc_or_prim)->get_bodywords(name);
	}
	return Unbound;
}



NodeP Pair::printOut(int just_titles)	// >0 for _printOutTitles, <0 for _edit
{
	PairP proclist=0, varlist=0, plistlist=0;
	three_lists(proclist, varlist, plistlist);
	
	File::backslashes=true;
	
	while (proclist) 
	{	NodeP proc_or_prim;
		if (!proclist->first()  ||  proclist->first()->is.aggregate) 
		{	error(BAD_DATA_UNREC, proclist->first());
			break;
		}else
			proc_or_prim = proclist->first()->intern()->proc_or_prim();
		
		if (proc_or_prim==Unbound)
		{	if (just_titles < 0) 
				writer << "to " << proclist->first() << endL << "end" << endL << endL;
			else 
			{	error(DK_HOW_UNREC, proclist->first());
				break;
			}
		}else 
			if (proc_or_prim->is.primitive) 
			{	error(IS_PRIM, proclist->first());
				break;
			}else 
			{	PairP bodywords = ProcedureP(proc_or_prim)->
					get_bodywords(proclist->first());

				if (just_titles > 0) 
				{	if (bodywords->first()->is.list)
						File::printNoBracket(writer.file(), bodywords->first());
					else 
					{	charP str = bodywords->first()->string()->expand_slash();
						writer << str;
						::free(str);
					}
				}else 
					while (bodywords) 
					{	if (bodywords->first()->is.list)
							File::printNoBracket(writer.file(), bodywords->first());
						else 
						{	charP str = bodywords->first()->string()->expand_slash();
							writer << str;
							::free(str);
						}
						writer << endL;
						next(bodywords);
					}
				writer << endL;
			}
		next(proclist);
		if (checkThrowing()) break;
	}
	while (varlist && !throwing) 
	{	PairP var;
		if (varlist->first()->is.aggregate) 
		{	error(BAD_DATA_UNREC, varlist->first());
			break;
		}else
			var = varlist->first()->intern()->var()->maybe_quote();
		
		if (var==Unbound) 
		{	if (just_titles >= 0) 
			{	error(NO_VALUE, varlist->first());
				break;
			}
		}else 
			writer << "Make \"" << varlist->first() << ' ';
			File::print(writer.file(), var);
			writer << endL;

		next(varlist);
		if (checkThrowing()) break;
	}
	while (plistlist && !throwing) 
	{	if (plistlist->first()->is.aggregate) 
		{	error(BAD_DATA_UNREC, plistlist->first());
			break;
		}else 
		{	PairP plist = plistlist->first()->intern()->plist();

			if (plist && just_titles > 0) 
				writer << "PList " << plistlist->first()->maybe_quote() 
					<< '=' << plist << endL;
			else
				while (plist) 
				{	writer << "pProp "
						<< plistlist->first()->maybe_quote() << ' '
						<< plist   -> first()->maybe_quote() << ' ';
					File::print(writer.file(), plist->second()->maybe_quote());
					writer << endL;

					plist = plist->tail()->tail();
				}
		}
		next(plistlist);
		if (checkThrowing()) break;
	}
	File::backslashes=false;
	return Unbound;
}

NodeP lprintOut      (PairPP arg) {return arg[1]->printOut(0);}
NodeP lprintOutTitles(PairPP arg) {return arg[1]->printOut(1);}


char tmp_filename[500]="";

NodeP ledit(PairPP arg) 
{
	if (environment.editor != wxEmptyString)
	{	if (tmp_filename[0]=='\0'  ||  !noarg) 
			sprintf(tmp_filename, "%s/logo%d.lg",
				environment.tempdir.c_str(), getpid());

		if (!noarg) 
		{	FILEP old_file = writer.file();
			writer.setfile( fopen(tmp_filename, "w") );
			if (writer) 
			{	arg[1]->printOut(-1);
				writer.close();
				writer.setfile(old_file);
			}else 
			{	error(FILE_ERROR, new String("Could not create editor file"));
				writer.setfile(old_file);
				return Unbound;
			}
		}
		if throwing return Unbound;

	#ifdef WIN32
		if ( spawnlp(P_WAIT, 
				environment.editor, 
				environment.editorname, tmp_filename, 0) )
	#else
		wxString command=wxString(environment.editor)+' '+wxString(tmp_filename);
	//	wxEnableTopLevelWindows(false);
	//	if (wxExecute(command, wxEXEC_SYNC))
		system(command.c_str());
		if (false)
	#endif
		{	error(FILE_ERROR, new String("Could not launch the editor"));
			return Unbound;
		}
	#ifndef WIN32
	//	wxEnableTopLevelWindows(true);
	#endif
		StringP old_line = File::currentLine;
		FILEP   old_file = loader.file();

		loader.setfile( fopen(tmp_filename, "r") );
		if (loader) 
		{	ValStatus save_val_status = Evaluator::val_status;

			while (!loader.eof() && !throwing )
			{	File::currentLine = loader.read("");
				PairP exec_list = File::currentLine->parser(true);
				Evaluator::val_status = no_value_allowed;
				if (exec_list) 
					exec_list->eval_driver();
			}
			loader.close();
			Evaluator::val_status = save_val_status;
		}else
			error(FILE_ERROR, new String("Could not read editor file"));

		loader.setfile(old_file);
		File::currentLine=old_line;
	}
	else
		if (tmp_filename[0]=='\0' || !noarg)
		{
//			c_out << ";Hint: better save your procedures in a file and edit that in the editor!" << endL;

			sprintf(tmp_filename, "%s/logo%d.lg",
				environment.tempdir.c_str(), getpid());

			if (!noarg) 
			{	FILEP old_file = writer.file();
				writer.setfile( fopen(tmp_filename, "w") );
				if (writer) 
				{	arg[1]->printOut(-1);
					writer.close();
					writer.setfile(old_file);
				}else 
				{	error(FILE_ERROR, new String("Could not create editor file"));
					writer.setfile(old_file);
					return Unbound;
				}
			}
			if throwing return Unbound;
			theApp->edFrame->FileOpen(tmp_filename);
			c_out << ";editing... press any key in the console to load changes! ";
			console->charmodeOn();
			console->getch();
			console->charmodeOff();
			c_out << "OK" << endL;
			theApp->edFrame->FileSave(tmp_filename);
			theApp->edFrame->FileClose(tmp_filename);
			console->SetFocus();

			StringP old_line = File::currentLine;
			FILEP   old_file = loader.file();

			loader.setfile( fopen(tmp_filename, "r") );
			if (loader) 
			{	ValStatus save_val_status = Evaluator::val_status;

				while (!loader.eof() && !throwing )
				{	File::currentLine = loader.read("");
					PairP exec_list = File::currentLine->parser(true);
					Evaluator::val_status = no_value_allowed;
					if (exec_list) 
						exec_list->eval_driver();
				}
				loader.close();

				Evaluator::val_status = save_val_status;
			}else
				error(FILE_ERROR, new String("Could not read editor file"));

			loader.setfile(old_file);
			File::currentLine=old_line;
		}
		else
			theApp->edFrame->FileOpen(tmp_filename);
	return Unbound;
}

NodeP leditFile(PairPP arg) 
{
	StringP str = arg[1]->string();

	if (!throwing) 
	{	String::copy_noparity(tmp_filename, str->ptr(), str->length());
		return ledit(argStack);
	}else
		return Unbound;
}


NodeP lerase(PairPP arg) 
{
	PairP proclist=0, varlist=0, plistlist=0;
	arg[1]->three_lists(proclist, varlist, plistlist);
	
	if (proclist)
		Tree::the_generation = new Int(Tree::the_generation->int_+1);
	
	int redef = _RedefP->var()->equal(_True);

	while (proclist)
	{	if (proclist->first())
		{	if (proclist->first()->is.aggregate) 
			{	error(BAD_DATA_UNREC, proclist->first());
				break;
			}
			ObjectP obj = proclist->first()->intern();

			if ( !redef  &&  obj->proc_or_prim()->is.primitive )
			{	error(IS_PRIM, obj);
				break;
			}

			obj->setproc( ProcedureP(Unbound) );
		}
		next(proclist);
	}
	while (varlist && !throwing)
	{	if (varlist->first())
		{	if (varlist->first()->is.aggregate) 
			{	error(BAD_DATA_UNREC, varlist->first());
				break;
			}
			varlist->first()->intern()->setvar( Unbound );
		}
		next(varlist);
	}
	while (plistlist && !throwing)
	{	if (plistlist->first())
		{	if (plistlist->first()->is.aggregate) 
			{	error(BAD_DATA_UNREC, plistlist->first());
				break;
			}
			plistlist->first()->intern()->setplist(NIL);
		}
		next(plistlist);
	}
	return Unbound;
}

NodeP leraseAll          (PairPP) {return root->erase_all( true,  true,  true);}
NodeP leraseProcedures   (PairPP) {return root->erase_all( true, false, false);}
NodeP leraseNames        (PairPP) {return root->erase_all(false,  true, false);}
NodeP lerasePropertyLists(PairPP) {return root->erase_all(false, false,  true);}


NodeP lCalls(PairPP) // try to reconstruct the nonexisting stackframes
{
	PairP callslist=0;
	int 
		*index=Evaluator::varPStack.base,
		  *max=Evaluator::varPStack.index;
	NodeP *sp=Evaluator::stack.base;

	while (index < max)
	{	index++;
		callslist=new Pair(0, callslist);
		VarP var=&Evaluator::varStack.base[*index+1];
		VarP maxVar;
		if (index < max)
			maxVar=&Evaluator::varStack.base[*(index+1)];
		else
			maxVar=Evaluator::varStack.index+1;

		PairP varlist=0, tnode=0, lastnode=0;
		while (var < maxVar)
		{	if (var->object)
			{	VarP newvar=0; // search the real value of the local

				for (VarP i=var+1; i <= Evaluator::varStack.index; i++)
					if (i->object==var->object)	// search in varStack first
					{	newvar=i;
						break;
					}
				if (newvar) // if found again in varStack...
					tnode=new Pair(
						new Pair(newvar->object, 
						new Pair(newvar->val, 0)),0);
				else
				{	 // if not found again in varStack search in root
					ObjectP obj=root->find(var->object->name());
					if (obj)
					{	NodeP val=obj->var();
						tnode=new Pair(
							new Pair(obj, 
							new Pair(val, 0)),0);
					}
				}
				if (varlist)
					lastnode->settail(tnode);
				else
					varlist=tnode;
				lastnode=tnode;
			}
			var++;
		}
		varlist=new Pair(*sp,varlist);
		sp+=7;
		callslist->setfirst(varlist);
	}
	return callslist;
}

charP String::convertQMtoP(charP ptr, int len) 
{
	static char result[32];
	char c;
	charP p;

	for (p=result;  len--;  *p++= c)
	{	c = *ptr++;
		if (c=='?')
			c='P';
	}
	*p=0;
	return result;
}

StringP String::unspace()
{
	return make(_ptr, 0, _length, type, copy_unspace);
}

NodeP lhelp(PairPP arg) 
{
	int lines;
	wxString filename, line;
	StringP str=0;
	
	if noarg
		filename = wxString(environment.helpfiles)+"/index.txt";
	else
	{	str=arg[1]->a_string(1);
		if throwing
		{	error(BAD_DATA_UNREC, arg[1]);
			return Unbound;
		}
		if (str->equalP(_Null_Word))
			filename = wxString(environment.helpfiles)+"/index.txt";
		else
		{	str=StringP(str->lower())->unspace();
			filename = wxString(environment.helpfiles)+'/'
				+String::convertQMtoP(str->ptr(), str->length())+".txt";
		}
	}
	wxFileInputStream input(filename);
	if (!input.Ok())
	{	if (arg==argStack)
			writer << "No help available." << endL;
		else

			writer << "No help available on " << str << '.' << endL;
	}else 
	{	lTextScreen(argStack);
		wxTextInputStream istr(input);
		lines=0;
		
		while (!throwing && !istr.Eof())
		{	line=istr.ReadLine();
			line+='\n';
			writer << line.c_str();
			lines++;
		}
	}
	return Unbound;
}
