#ifdef WIN32
	#include <crtdbg.h>
#endif

#ifdef WIN32
	#include <windows.h>
#endif
#include <stdio.h>

#include <wx/string.h>

#include "Main.h"
#include "Node.h"
#include "Pair.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "String.h"
#include "Error.h"
#include "Init.h"
#include "io.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

PORTOUT PortOut;
PORTWORDOUT PortWordOut;
PORTDWORDOUT PortDWordOut;
PORTIN PortIn;
PORTWORDIN PortWordIn;
PORTDWORDIN PortDWordIn;
SETPORTBIT SetPortBit;
CLEARPORTBIT ClearPortBit;
NOTPORTBIT NotPortBit;
GETPORTBIT GetPortBit;
RIGHTPORTSHIFT RightPortShift;
LEFTPORTSHIFT LeftPortShift;
ISDRIVERINSTALLED IsDriverInstalled;

#ifdef WIN32
HMODULE ioDLL;
#else
void *ioDLL;
#endif

void __cdecl UnloadIODLL() 
{
#ifdef WIN32
	FreeLibrary(ioDLL);
#endif
}

bool LoadIODLL() 
{
#ifdef WIN32
	ioDLL = LoadLibrary("io");
	if (ioDLL == NULL) return false;

	PortOut       = (PORTOUT)      GetProcAddress(ioDLL, "PortOut");
	PortWordOut   = (PORTWORDOUT)  GetProcAddress(ioDLL, "PortWordOut");
	PortDWordOut  = (PORTDWORDOUT) GetProcAddress(ioDLL, "PortDWordOut");
	PortIn        = (PORTIN)       GetProcAddress(ioDLL, "PortIn");
	PortWordIn    = (PORTWORDIN)   GetProcAddress(ioDLL, "PortWordIn");
	PortDWordIn   = (PORTDWORDIN)  GetProcAddress(ioDLL, "PortDWordIn");
	SetPortBit    = (SETPORTBIT)   GetProcAddress(ioDLL, "SetPortBit");
	ClearPortBit  = (CLEARPORTBIT) GetProcAddress(ioDLL, "ClrPortBit");
	NotPortBit    = (NOTPORTBIT)   GetProcAddress(ioDLL, "NotPortBit");
	GetPortBit    = (GETPORTBIT)   GetProcAddress(ioDLL, "GetPortBit");
	RightPortShift=(RIGHTPORTSHIFT)GetProcAddress(ioDLL, "RightPortShift");
	LeftPortShift = (LEFTPORTSHIFT)GetProcAddress(ioDLL, "LeftPortShift");
	IsDriverInstalled = (ISDRIVERINSTALLED)GetProcAddress(ioDLL, "IsDriverInstalled");

	atexit(UnloadIODLL);
#endif
	return true;
}

NodeP lPortOut(PairPP arg)
{
	if (ioDLL == NULL) 
	{	error(FILE_ERROR, new String("Could not load io.dll"));
		return Unbound;
	}
	int 
		adr=arg[1]->a_int(1),
		val=arg[2]->a_int(2);
	if throwing return Unbound;
	if (adr > Int16::max*2+1 || adr <-Int16::max-1)
	{	error(BAD_DATA, arg[1]);
		return Unbound;
	}
	if (val > UInt8::max || val < 0)
	{	error(BAD_DATA, arg[2], 2);
		return Unbound;
	}
	PortOut(short(adr), (unsigned char)(val));
	return Unbound;
}

NodeP lPortIn(PairPP arg)
{
	if (ioDLL == NULL) 
	{	error(FILE_ERROR, new String("Could not load io.dll"));
		return Unbound;
	}
	int adr=arg[1]->a_int(1);
	if throwing return Unbound;
	if (adr > Int16::max*2+1 || adr <-Int16::max-1)
	{	error(BAD_DATA, arg[1]);
		return Unbound;
	}
	unsigned char val=PortIn(short(adr));
	return new UInt8(val);
}

NodeP lsetPortBit(PairPP arg)
{
	if (ioDLL == NULL) 
	{	error(FILE_ERROR, new String("Could not load io.dll"));
		return Unbound;
	}
	int adr=arg[1]->a_int(1),
		bit=arg[2]->a_int(2);
	if throwing return Unbound;
	if (adr > Int16::max*2+1 || adr <-Int16::max-1)
	{	error(BAD_DATA, arg[1]);
		return Unbound;
	}
	if (bit >= int(8*sizeof(unsigned char)) || bit < 0)
	{	error(BAD_DATA, arg[2], 2);
		return Unbound;
	}
	SetPortBit(adr, bit);
	return Unbound;
}

NodeP lclearPortBit(PairPP arg)
{
	if (ioDLL == NULL) 
	{	error(FILE_ERROR, new String("Could not load io.dll"));
		return Unbound;
	}
	int adr=arg[1]->a_int(1),
		bit=arg[2]->a_int(2);
	if throwing return Unbound;
	if (adr > Int16::max*2+1 || adr <-Int16::max-1)
	{	error(BAD_DATA, arg[1]);
		return Unbound;
	}
	if (bit >= int(8*sizeof(unsigned char)) || bit < 0)
	{	error(BAD_DATA, arg[2], 2);
		return Unbound;
	}
	ClearPortBit(adr, bit);
	return Unbound;
}

NodeP lnotPortBit(PairPP arg)
{
	if (ioDLL == NULL) 
	{	error(FILE_ERROR, new String("Could not load io.dll"));
		return Unbound;
	}
	int adr=arg[1]->a_int(1),
		bit=arg[2]->a_int(2);
	if throwing return Unbound;
	if (adr > Int16::max*2+1 || adr <-Int16::max-1)
	{	error(BAD_DATA, arg[1]);
		return Unbound;
	}
	if (bit >= int(8*sizeof(unsigned char)) || bit < 0)
	{	error(BAD_DATA, arg[2], 2);
		return Unbound;
	}
	NotPortBit(adr, bit);
	return Unbound;
}

NodeP lgetPortBit(PairPP arg)
{
	if (ioDLL == NULL) 
	{	error(FILE_ERROR, new String("Could not load io.dll"));
		return Unbound;
	}
	int adr=arg[1]->a_int(1),
		bit=arg[2]->a_int(2);
	if throwing return Unbound;
	if (adr > Int16::max*2+1 || adr <-Int16::max-1)
	{	error(BAD_DATA, arg[1]);
		return Unbound;
	}
	if (bit >= int(8*sizeof(unsigned char)) || bit < 0)
	{	error(BAD_DATA, arg[2], 2);
		return Unbound;
	}
	short val=GetPortBit(adr, bit);
	return val? NodeP(_True) : NodeP(_False);
}

NodeP lleftPortShift(PairPP arg)
{
	if (ioDLL == NULL) 
	{	error(FILE_ERROR, new String("Could not load io.dll"));
		return Unbound;
	}
	int adr=arg[1]->a_int(1),
		by=arg[2]->a_int(2);
	if throwing return Unbound;
	if (adr > Int16::max*2+1 || adr <-Int16::max-1)
	{	error(BAD_DATA, arg[1]);
		return Unbound;
	}
	if (by >= int(8*sizeof(unsigned char)) || by < -8*int(sizeof(unsigned char)))
	{	error(BAD_DATA, arg[2], 2);
		return Unbound;
	}
	LeftPortShift(adr, by);
	return Unbound;
}

NodeP lrightPortShift(PairPP arg)
{
	if (ioDLL == NULL) 
	{	error(FILE_ERROR, new String("Could not load io.dll"));
		return Unbound;
	}
	int adr=arg[1]->a_int(1),
		by=arg[2]->a_int(2);
	if throwing return Unbound;
	if (adr > Int16::max*2+1 || adr <-Int16::max-1)
	{	error(BAD_DATA, arg[1]);
		return Unbound;
	}
	if (by >= int(8*sizeof(unsigned char)) || by < -8*int(sizeof(unsigned char)))
	{	error(BAD_DATA, arg[2], 2);
		return Unbound;
	}
	RightPortShift(adr, by);
	return Unbound;
}
